/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor;

import java.io.File;
import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.*;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.commons.collections.map.LRUMap;
import org.xquark.extractor.common.Constants;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.metadata.*;
import org.xquark.extractor.runtime.ExtractorConfiguration;
import org.xquark.extractor.runtime.Selection;
import org.xquark.jdbc.datasource.JDBCDataSource;
import org.xquark.jdbc.datasource.JDBCDataSourceFactory;
import org.xquark.jdbc.datasource.SingleConnectionDataSource;
import org.xquark.jdbc.typing.DBMSInfoMap;
import org.xquark.jdbc.typing.JDBCProperties;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.ModuleManager;
import org.xquark.xquery.metadata.MetaDataImpl;

/**
 * Extractor XML/DBC DataSource implementation.
 */
public class Extractor implements XMLDataSource, Constants {

	private static final String RCSRevision = "$Revision: 1.33 $";
	private static final String RCSName = "$Name:  $";

	public static final String JNDI_SCHEME = "jndi:";
	
	static {
		MessageLibrary.setup("org.xquark.extractor.common.resources.MessageLibrary", new Locale("en", "US"));
	}

	private DataSource dataSource = null;
	private Properties configuration = null;
	private MetaDataManager metaDataManager = null;
	private MetaDataImpl metaData = null;
	private Map cachedStatements = Collections.synchronizedMap(new LRUMap(100));

	private String name = "Extractor";
	// TODO: check if this should not go somewhere else
	private boolean useScrollableResultSets = false;
	private String configURI = null;
	private String userName = null;
	private String databaseName = null;
	private String databaseVersion = null;
		
	/**
	 * Constructor to use with a configuration file URL or a JNDI name
	 * referencing a JDBC datasource instance. The configuration file is
	 * expected to contain JDBC connection parameters or a JNDI name. Choose
	 * this last solution if you both need to specify a configuration file and a
	 * JNDI name. See the <em>XQuery Language Reference Guide"</em> for more
	 * information about configuration files.
	 * 
	 * @param URI
	 *            either the configuration file URL, or a JNDI reference that
	 *            MUST be prefixed with the "jndi:" string to be recognized as
	 *            this.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public Extractor(String URI) throws XMLDBCException {
		this(URI, null);
	}
	
	/**
	 * Use this constructor when you need a specific classloader that will be
	 * used to load JDBC drivers.
	 * 
	 * @param URI
	 *            either a file URL, or a JNDI reference that MUST be prefixed
	 *            with the "jndi:" string to be recognized as this.
	 * @param loader
	 *            an application-provided class loader that will be used to load
	 *            JDBC drivers.
	 * @throws XMLDBCException
	 *             API exception
	 * @see #Extractor(String)
	 */
	public Extractor(String URI, ClassLoader loader) throws XMLDBCException {
		if (URI.startsWith(JNDI_SCHEME))
			initDataSource(getJNDIDataSource(URI.substring(JNDI_SCHEME.length())));
		else
			initDataSource(URI, loader);
	}
	
	/**
	 * Constructor to use with an extractor configuration file. See the
	 * <em>XQuery Language Reference Guide</em> for more information about
	 * configuration files.
	 * 
	 * @param configFile
	 *            a Java symbolic file.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public Extractor(File configFile) throws XMLDBCException {
		this(configFile, null);
	}
	
	/**
	 * Use this constructor when you need a specific classloader that will be
	 * used to load JDBC drivers.
	 * 
	 * @param configFile
	 *            a Java symbolic file
	 * @param loader
	 *            an application-provided class loader that will be used to load
	 *            JDBC drivers.
	 * @throws XMLDBCException
	 *             API exception
	 * @see #Extractor(File)
	 */
	public Extractor(File configFile, ClassLoader loader)
	throws XMLDBCException {
		try {
			String configURI = configFile.toURL().toString();
			initDataSource(configURI, loader);
		}
		catch (MalformedURLException e) {
			throw new XMLDBCException("Invalid configuration file name " + configFile, e); 
		}
	}
	
	/**
	 * Constructor to use with a JDBC data source.
	 * 
	 * @param ds
	 *            a JDBC data source.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public Extractor(DataSource ds) throws XMLDBCException {
		initDataSource(ds);
	}

	/**
	 * Constructor to use with a JDBC data source and an extractor configuration
	 * file URL. See the <em>XQuery Language Reference Guide</em> for more
	 * information about configuration files.
	 * 
	 * @param ds
	 *            a JDBC data source.
	 * @param configURL
	 *            the URL of the configuration file.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public Extractor(DataSource ds, String configURL) throws XMLDBCException {
		if (configURI != null)
			initDataSource(ds, configURI);
		else 
			initDataSource(ds);
	}

	/**
	 * Constructor to use with a JDBC data source and an extractor configuration
	 * file. See the <em>XQuery Language Reference Guide</em> for more
	 * information about configuration files.
	 * 
	 * @param ds
	 *            a JDBC data source.
	 * @param configFile
	 *            a Java symbolic file
	 * @throws XMLDBCException
	 *             API exception
	 */
	public Extractor(DataSource ds, File configFile) throws XMLDBCException {
		if (configFile != null) { 
			try {
				String configURI = configFile.toURL().toString();
				initDataSource(ds, configURI);
			}
			catch (MalformedURLException e) {
				throw new XMLDBCException("Invalid configuration file name " + configFile, e); 
			}
		}
		else 
			initDataSource(ds);
	}

	/**
	 * Construction to use when JDBC connection parameters are available.
	 * 
	 * @param jdbcURL
	 *            the JDBC URL of the JDBC data source.
	 * @param username
	 *            the login name for the JDBC data source.
	 * @param password
	 *            the password for the login name.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public Extractor(String jdbcURL, String username, String password) throws XMLDBCException {
		JDBCDataSourceFactory dsFactory = JDBCDataSourceFactory.newInstance();
		try {
			initDataSource(dsFactory.newDataSource(jdbcURL, username, password));
		}
		catch (SQLException ex) {
			throw new XMLDBCException("Error while accessing datasource "
					+ username+ '@' + jdbcURL, ex);
		}
	}
	
	/**
	 * Use this constructor when you need a specific classloader that will be
	 * used to load JDBC drivers.
	 * 
	 * @param jdbcURL
	 *            the JDBC URL of the JDBC data source.
	 * @param username
	 *            the login name for the JDBC data source.
	 * @param password
	 *            the password for the login name.
	 * @param loader
	 *            an application-provided class loader that will be used to load
	 *            JDBC drivers.
	 * @throws XMLDBCException
	 *             API exception
	 * @see #Extractor(String, String, String)
	 */
	public Extractor(String jdbcURL, String username, String password, ClassLoader loader)
	throws XMLDBCException {
		JDBCDataSourceFactory dsFactory = JDBCDataSourceFactory.newInstance(loader);
		try {
			initDataSource(dsFactory.newDataSource(jdbcURL, username, password));
		}
		catch (SQLException ex) {
			throw new XMLDBCException("Error while accessing datasource "
					+ username + '@' + jdbcURL, ex);
		}
	}
	
	/**
	 * Constructs an Extractor using a single JDBC connection. This connection
	 * should not be used while the XQBridge object is active.
	 * 
	 * @param jdbcConnection
	 *            a jdbc connection.
	 * @throws XMLDBCException
	 *             API exception
	 * @deprecated prevents the use of an external or internal data source.
	 */
	public Extractor(Connection jdbcConnection) throws XMLDBCException {
		this(jdbcConnection, null);
	}
	
	/**
	 * Constructs an Extractor using a single JDBC connection and a
	 * configuration file. The connection should not be used while the
	 * XQBridge object is active.
	 * 
	 * @param jdbcConnection
	 *            a jdbc connection.
	 * @param configURL
	 *            the URL of the configuration file.
	 * @throws XMLDBCException
	 *             API exception
	 * @deprecated prevents the use of an external or internal data source.
	 */
	public Extractor(Connection jdbcConnection, String configURL) throws XMLDBCException {
		this(jdbcConnection, configURL, null);
	}

	/**
	 * Use this constructor when you need a specific classloader that will be
	 * used to load JDBC drivers.
	 * 
	 * @param jdbcConnection
	 *            a jdbc connection.
	 * @param configURL
	 *            the URL of the configuration file.
	 * @param loader
	 *            an application-provided class loader that will be used to load
	 *            JDBC drivers.
	 * @throws XMLDBCException
	 *             API exception
	 * @deprecated prevents the use of an external or internal data source.
	 * @see #Extractor(Connection, String)
	 */
	public Extractor(Connection jdbcConnection, String configURL, ClassLoader loader)
	throws XMLDBCException {
		try {
			if (jdbcConnection == null)
				initDataSource(configURL, loader);
			else if (configURL != null)
				initDataSource(new SingleConnectionDataSource(jdbcConnection), configURL);
			else
				initDataSource(new SingleConnectionDataSource(jdbcConnection));
		}
		catch (SQLException ex) {
			throw new XMLDBCException("Error while accessing datasource: " + ex.getMessage(), ex);
		}
	}
	
	private void initDataSource(String configURI, ClassLoader loader) throws XMLDBCException {
		this.configURI = configURI;
		try {
			configuration = ExtractorConfiguration.loadConfiguration(configURI);
		}
		catch (MalformedURLException e) {
			throw new XMLDBCException("Invalid XML/DBC URI "
					+ ExtractorDriver.EXTRACTOR_URL_PREFIX + configURI, e); 
		}
		if (configuration.getProperty(LABEL_SPEC_NAME) != null)
			name = configuration.getProperty(LABEL_SPEC_NAME);
		if (configuration.getProperty(LABEL_SPEC_JNDI) != null) {
			dataSource = getJNDIDataSource(configuration.getProperty(LABEL_SPEC_JNDI));
		} else {
			String driver = configuration.getProperty(LABEL_SPEC_DRIVER);
			String url = configuration.getProperty(LABEL_SPEC_URL);
			String user = configuration.getProperty(LABEL_SPEC_USER);
			String password = configuration.getProperty(LABEL_SPEC_PASSWORD);
			Properties jdbcProps = new Properties();
			jdbcProps.setProperty("user", user);
			jdbcProps.setProperty("password", password);
			JDBCDataSourceFactory dsFactory = JDBCDataSourceFactory.newInstance(loader);
			try {
				dataSource = dsFactory.newDataSource(driver, url, jdbcProps);
			} catch (SQLException ex) {
				throw new XMLDBCException("Error while creating datasource: "+ex.getMessage(), ex);
			}
		}
		Connection jdbcConnection = null;
		try {
			jdbcConnection = dataSource.getConnection();
			DatabaseMetaData dm = jdbcConnection.getMetaData();
			databaseName = dm.getDatabaseProductName();
			databaseVersion = dm.getDatabaseProductVersion();
		} catch (SQLException ex) {
			throw new XMLDBCException("Error while accessing datasource: "+ex.getMessage(), ex);
		} finally {
			try {
				if (jdbcConnection != null)
					jdbcConnection.close();
			} catch (SQLException e) {
			}
		}
	}
	
	private void initDataSource(DataSource ds) throws XMLDBCException {
		dataSource = ds;
		String url = null;
		String user = null;
		Connection jdbcConnection = null;
		try {
			jdbcConnection = ds.getConnection();
			DatabaseMetaData dm = jdbcConnection.getMetaData();
			url = dm.getURL();
			user = dm.getUserName();
			databaseName = dm.getDatabaseProductName();
			databaseVersion = dm.getDatabaseProductVersion();
			configuration = ExtractorConfiguration.getDefaultProperties(url, user);
		} catch (SQLException ex) {
			throw new XMLDBCException("Error while accessing datasource: "+ex.getMessage(), ex);
		} finally {
			try {
				if (jdbcConnection != null)
					jdbcConnection.close();
			} catch (SQLException e) {
			}
		}
	}

	private void initDataSource(DataSource ds, String configURI) throws XMLDBCException {
		dataSource = ds;
		this.configURI = configURI;
		try {
			configuration = ExtractorConfiguration.loadConfiguration(configURI);
			if (configuration.getProperty(LABEL_SPEC_NAME) != null)
				name = configuration.getProperty(LABEL_SPEC_NAME);
		} catch (MalformedURLException e) {
			throw new XMLDBCException("Invalid XML/DBC URI "+ExtractorDriver.EXTRACTOR_URL_PREFIX+configURI, e); 
		}
		String url = null;
		String user = null;
		Connection jdbcConnection = null;
		try {
			jdbcConnection = ds.getConnection();
			DatabaseMetaData dm = jdbcConnection.getMetaData();
			url = dm.getURL();
			user = dm.getUserName();
			databaseName = dm.getDatabaseProductName();
			databaseVersion = dm.getDatabaseProductVersion();
		} catch (SQLException ex) {
			throw new XMLDBCException("Error while accessing datasource: "+ex.getMessage(), ex);
		} finally {
			try {
				if (jdbcConnection != null)
					jdbcConnection.close();
			} catch (SQLException e) {
			}
		}
		if (url != null && user != null) {
			String confUrl = configuration.getProperty(LABEL_SPEC_URL);
			if (confUrl == null) 
				configuration.setProperty(LABEL_SPEC_URL, url);
			else if (!url.equals(confUrl))
				throw new XMLDBCException("Datasource and configuration file URLs do not match");
			String confUser = configuration.getProperty(LABEL_SPEC_USER);
			if (confUser == null) 
				configuration.setProperty(LABEL_SPEC_USER, user);
			else if (!user.equals(confUser))
				throw new XMLDBCException("Datasource and configuration file user names do not match");
		}
	}
		
	private DataSource getJNDIDataSource(String ref) throws XMLDBCException {
		try {
			InitialContext ctx = new InitialContext();
			return (DataSource)ctx.lookup(ref);
		} catch (NamingException e) {
			throw new XMLDBCException("Error while locating JNDI object "+ref, e);
		}
	}
	
	/**
	 * Sets the internal JDBC datasource log writer.
	 * @param writer
	 */
	public void setLogWriter(PrintWriter writer) {
		try {
			dataSource.setLogWriter(writer);
		}
		catch (SQLException ex) {
		}
 	}

	/**
	 * Get the internal JDBC datasource log writer.
	 * @return java.io.PrintWriter
	 */
	public PrintWriter getLogWriter() {
		try {
            return dataSource.getLogWriter();
        }
		catch (SQLException e) {
			return null;
        }
	}

	/**
	 * Set the internal JDBC datasource login timeout.
	 * @param seconds the login timeout in seconds.
	 */
	public void setLoginTimeout(int seconds) {
		try {
			dataSource.setLoginTimeout(seconds);
		} catch (UnsupportedOperationException ex) {
		} catch (SQLException ex) {
		}
	}

	/**
	 * Get the internal JDBC datasource login timeout.
	 * @return the login timeout in seconds.
	 */
	public int getLoginTimeout() {
		try {
            return dataSource.getLoginTimeout();
		} catch (UnsupportedOperationException ex) {
			return -1;
		} catch (SQLException ex) {
            return -1;
        }
	}
	
	/**
	 * Closes the extractor and the underlying resources among which the internal
	 * JDBC data source except if it was passed at construction time.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public void close() throws XMLDBCException {
		if (dataSource instanceof JDBCDataSource) {
			try {
				((JDBCDataSource)dataSource).close();
				dataSource = null;
			}
			catch (SQLException ex) {
				throw new XMLDBCException("Error while closing extractor", ex);
			}
		}
	}

	/**
	 * Returns the internal JDBC data source.
	 * @return the internal JDBC source.
	 */
	public DataSource getJdbcDataSource() {
		return dataSource;
	}
	
	/**
	 * Creates a new XML/DBC connection instance.
	 * @return an XMLConnection to the XML/SQL datasource.
	 * @throws XMLDBCException
	 *             API exception
	 */
	public XMLConnection getConnection() throws XMLDBCException {
		return getExtractorConnection();
	}
	
	/**
	 * Creates a new XML/DBC connection instance for the extractor, avoiding
	 * a cast on {@link #getConnection()} to access Extractor specific features.
	 * @return an ExtractorConnection
	 * @throws XMLDBCException
	 *             API exception
	 */
	public ExtractorConnection getExtractorConnection() throws XMLDBCException {
		try {
            return new ExtractorConnection(this, dataSource.getConnection());
        }
		catch (SQLException e) {
			throw new XMLDBCException(e.getMessage(), e);
        }
	}
	
	/**
	 * Not implemented by Extractor.
	 * @see org.xquark.xml.xdbc.XMLDataSource#getConnection(java.lang.String, java.lang.String)
	 */
	public XMLConnection getConnection(String user, String password)
	throws XMLDBCException {
		throw new XMLDBCException("Unsupported operation getConnection(user, password). Use getConnection() instead.");
	}

	/**
	 * Accessor to the "use scrollable result set" option.
	 * @return true, if the extractor must use scrollable JDBC result sets.
	 */
	public synchronized boolean useScrollableResultSets() {
		return useScrollableResultSets;
	}

	/**
	 * To indicate the Extractor that scrollable JDBC result sets must be used.
	 * @param use a boolean.
	 */
	public synchronized void setUseScrollableResultSets(boolean use) {
		useScrollableResultSets = use;
	}

	/**
	 * Gets the user name used to log to the Extractor.
	 * @return a string
	 */
	public String getUserName() {
		return configuration.getProperty(LABEL_SPEC_USER);
	}
	
	/**
	 * Gets the XML/DBC URL of the Extractor (XML data source).
	 * @return generally the JDBC URL prefixed with "xdbc:xquark:extractor:".
	 */
	public String getURL() {
		return ExtractorDriver.EXTRACTOR_URL_PREFIX + configuration.getProperty(LABEL_SPEC_URL);
	}
	
	String getDatabaseName() {
		return databaseName;
	}
	
	String getDatabaseVersion() {
		return databaseVersion;
	}
	
	synchronized MetaDataManager getMetaDataManager(Connection jdbcConnection, boolean refresh) throws XMLDBCException {
		if (refresh || metaDataManager == null) {
			initMetaData(jdbcConnection);
		}
		return metaDataManager;
	}
	
	synchronized MetaDataImpl getDataSourceMetaData(Connection jdbcConnection, boolean refresh) throws XMLDBCException {
		if (refresh || metaData == null) {
			initMetaData(jdbcConnection);
		}
		return metaData;
	}
	
	Map getCachedStatements() {
		return cachedStatements;
	}
	
	private void initMetaData(Connection jdbcConnection) throws XMLDBCException {
		/* create a CustomizedEncoder*/
		CustomizedEncoder encoder = new CustomizedEncoder();
		Properties substitutionMap = (Properties) configuration.get(LABEL_SPEC_SUBST);
		encoder.setSubstitutionMap(substitutionMap);
		Integer nameCase = (Integer) configuration.get(LABEL_SPEC_SUBST_NAMECASE);
		if (null == nameCase) {
			nameCase = new Integer(CASE_MIXED);
		}
		encoder.setNameCase(nameCase.intValue());
		Selection selection = (Selection) configuration.get("selections");
		if (selection != null && selection.noSubElementDef())
			selection = null;

		try { 
			metaDataManager = new MetaDataManager(DBMSInfoMap.getDBMSInfo(jdbcConnection));
			String dbName = databaseName;
			if (dbName.startsWith("Oracle")) {
				loadMetaDataForOracle(jdbcConnection, selection, encoder);
			} else if (dbName.startsWith("Microsoft SQL Server")) {
				loadMetaDataForMicrosoftSqlServer(jdbcConnection, selection, encoder);
			} else if (dbName.startsWith("Adaptive Server Enterprise")) {
				loadMetaDataForSybase(jdbcConnection, selection, encoder);
			} else if (dbName.startsWith("PROGRESS")) {
				loadMetaDataForProgress(jdbcConnection, selection, encoder);
			} else if (dbName.startsWith("MySQL")) {
				loadMetaDataForMySql(jdbcConnection, selection, encoder);
			} else {
				throw new XMLDBCNotSupportedException(MessageLibrary.getMessage("D_N_SUP_DB_VENDOR", dbName), null);
			}
		} catch (SQLException ex) {
			throw new XMLDBCException("Error while loading metadata", ex);
		}
		ModuleManager moduleManager = null;
		if (metaData != null)
			moduleManager = metaData.getModuleManager();
		metaData = metaDataManager.buildMetaDataForParser(name);
		if (moduleManager != null) {
			for (Iterator it = moduleManager.getModules().iterator();it.hasNext();) {
			}
		}
		cachedStatements.clear();
	}

	private void loadMetaDataForMicrosoftSqlServer(Connection jdbcConnection, Selection selection, QNameEncoder encoder) 
	throws SQLException
	{
		/* Load metadata in MetadataManager */
		Loader metadataLoader = new org.xquark.extractor.microsoft.Loader(jdbcConnection, metaDataManager.getDbmsInfo(), encoder);
		metaDataManager.setMetaData(metadataLoader.loadSite(selection, new JDBCProperties() {
			public boolean useResultSet4ColumnMetadata() {
				return false;
			}
			public boolean useStaticMappingFromNativeType() {
				return true;
			}
		}));
		// metaDataManager.setTableMap(metadataLoader._tableMap);	// QUESTION : why removed ?
	}

	private void loadMetaDataForSybase(Connection jdbcConnection, Selection selection, QNameEncoder encoder) 
	throws SQLException
	{
		/* Load metadata in MetadataManager */
		Loader metadataLoader = new org.xquark.extractor.sybase.Loader(jdbcConnection, metaDataManager.getDbmsInfo(), encoder);
		metaDataManager.setMetaData(metadataLoader.loadSite(selection, new JDBCProperties() {
			public boolean useResultSet4ColumnMetadata() {
				return false;
			}
			public boolean useStaticMappingFromNativeType() {
				return true;
			}
		}));
		// metaDataManager.setTableMap(metadataLoader._tableMap);	// QUESTION : why removed ?
	}

	private void loadMetaDataForOracle(Connection jdbcConnection, Selection selection, QNameEncoder encoder) 
	throws SQLException 
	{
		/* Load metadata in MetadataManager */
		Loader metadataLoader = new org.xquark.extractor.oracle.Loader(jdbcConnection, metaDataManager.getDbmsInfo(), encoder);
		metaDataManager.setMetaData(metadataLoader.loadSite(selection, new JDBCProperties() {
			public boolean useResultSet4ColumnMetadata() {
				return false;
			}
			public boolean useStaticMappingFromNativeType() {
				return true;
			}
		}));
		metaDataManager.setTableMap(metadataLoader._tableMap);
	}

	private void loadMetaDataForProgress(Connection jdbcConnection, Selection selection, QNameEncoder encoder) 
	throws SQLException
	{
		/* Load metadata in MetadataManager */
		Loader metadataLoader = new org.xquark.extractor.progress.Loader(jdbcConnection, metaDataManager.getDbmsInfo(), encoder);
		metaDataManager.setMetaData(metadataLoader.loadSite(selection, new JDBCProperties() {
			public boolean useResultSet4ColumnMetadata() {
				return false;
			}
			public boolean useStaticMappingFromNativeType() {
				return true;
			}
		}));
		metaDataManager.setTableMap(metadataLoader._tableMap);
	}

	private void loadMetaDataForMySql(Connection jdbcConnection, Selection selection, QNameEncoder encoder) 
	throws SQLException
	{
		/* Load metadata in MetadataManager */
		Loader metadataLoader = new org.xquark.extractor.mysql.Loader(jdbcConnection, metaDataManager.getDbmsInfo(), encoder);
		metaDataManager.setMetaData(metadataLoader.loadSite(selection, new JDBCProperties() {
			public boolean useResultSet4ColumnMetadata() {
				return false;
			}
			public boolean useStaticMappingFromNativeType() {
				return true;
			}
		}));
		metaDataManager.setTableMap(metadataLoader._tableMap);
	}

	/**
	 * Show the configuration of the wrapper
	 */
	void display() {
		if (configuration == null)
			return;
		for (Enumeration e = configuration.propertyNames(); e.hasMoreElements();) {
			String key = (String) e.nextElement();
			System.out.println("[" + key + "] : [" + configuration.getProperty(key) + "]");
		}
	}
	
}
