/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.algebra;

import java.sql.Types;
import java.util.ArrayList;
import java.util.List;

import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.xfunctions.*;
import org.xquark.jdbc.typing.DbType;
import org.xquark.jdbc.typing.JavaTypeInfo;

/**
 * Browse an expression tree to set the sql type field.
 * 1. SqlTypeAtom are built for attributes basing on database metadata and
 * on XML type for literals.
 * 2. SqlTypeStructure are built for simple relations.
 * 3. and SqlTypeCombinedStructure for joins.
 * 4. Predefined types are used for other algebra nodes, e.g., logical expressions.
 */
public final class AlgebraTypeVisitor extends DefaultAlgebraVisitor
implements Constants
{
	private static final String RCSRevision = "$Revision: 1.16 $";
	private static final String RCSName = "$Name:  $";

	private DbType ptBit = new DbType(Types.BIT);
	private DbType ptBoolean = new DbType(DbType.BOOLEAN);
	private DbType ptDate = new DbType(Types.DATE);
	private DbType ptDecimal = new DbType(Types.DECIMAL);
	private DbType ptDouble = new DbType(Types.DOUBLE);
	private DbType ptFloat = new DbType(Types.FLOAT);
	private DbType ptInteger = new DbType(Types.INTEGER);
	private DbType ptNull = new DbType(Types.NULL);
	private DbType ptTime = new DbType(Types.TIME);
	private DbType ptTimeStamp = new DbType(Types.TIMESTAMP);
	private DbType ptVarChar = new DbType(Types.VARCHAR);

	public AlgebraTypeVisitor() {
	}

	public void reinit() {
	}

	public void visit(Expression arg) throws SqlWrapperException {
		Debug.assertTrue(false, "NYI!!" + arg.toString());
	}

	public void visit(Attribute arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(
		        		arg.getAttribute().getMappingInfo().getDBType(),
		        		arg.getAttribute().getName(),
		        		SqlType.ONE
		        		));
	}

	public void visit(Rowid arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(new DbType(DbType.ORACLE_ROWID), arg.getName()));
	}

	public void visit(AttributeExpression arg) throws SqlWrapperException {
		SqlTypeAtom type = null;

		Expression underlyinExpr = arg.getUnderlyinExpr();
		if (underlyinExpr != null && !arg.getName().equals("*")) {
			// AttributeExpression may be created during rewritting without XQuery expression...
			underlyinExpr.accept(this);
			type = (SqlTypeAtom) underlyinExpr.getType().clone();
			type.setName(arg.getAttribute());
		}

		arg.setType(type);
	}

	public void visit(BinOpArithmetic arg) throws SqlWrapperException {
		arg.getLeftOperand().accept(this);
		arg.getRightOperand().accept(this);
		arg.setType(new SqlTypeAtom(ptDouble, null, SqlType.ONE));
	}

	public void visit(BinOpBoolean arg) throws SqlWrapperException {
		arg.getLeftOperand().accept(this);
		arg.getRightOperand().accept(this);
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(BinOpCompare arg) throws SqlWrapperException {
		arg.getLeftOperand().accept(this);
		arg.getRightOperand().accept(this);
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(BinOpCompareAny arg) throws SqlWrapperException {
		arg.getLeftOperand().accept(this);
		arg.getRightOperand().accept(this);
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(BinOpDifference arg) throws SqlWrapperException {
		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		arg.setType((SqlType) lOprnd.getType().clone());
	}

	public void visit(BinOpIntersect arg) throws SqlWrapperException {
		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		arg.setType((SqlType) lOprnd.getType().clone());
	}

	public void visit(BinOpMerge arg) throws SqlWrapperException {
		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		arg.setType((SqlType) lOprnd.getType().clone());
	}

	public void visit(BinOpOuterJoin arg) throws SqlWrapperException {
		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType leftType = (SqlType) lOprnd.getType().clone();
		SqlType rightType = (SqlType) rOprnd.getType().clone();
		SqlTypeCombinedStructure type = new SqlTypeCombinedStructure();
		type.addStructure(leftType);
		type.addStructure(rightType);
		arg.setType(type);

		List predicateList = arg.getPredicateList();
		if (null != predicateList) {
			Expression predicate = null;
			for (int i = 0; i < predicateList.size(); i++) {
				predicate = (Expression) predicateList.get(i);
				predicate.accept(this);
			}
		}
	}

	public void visit(BinOpUnion arg) throws SqlWrapperException {
		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = (SqlType) lOprnd.getType().clone();
		arg.setType(type);
	}

	public void visit(DummyTable arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(null, null, SqlType.MANY));
	}

	public void visit(FunAggregate arg) throws SqlWrapperException {
		Expression oprnd = arg.getOperand();
		oprnd.accept(this);

		SqlType type = null;
		int oprtr = arg.getOperator();
		switch (oprtr) {
			case AVG :
				type = new SqlTypeAtom(ptDouble, null, SqlType.ONE); // TODO: This typing may not be consistent with XQuery one. Is it used ?
				break;
			case COUNT :
				type = new SqlTypeAtom(ptInteger, null, SqlType.ONE);
				break;
			case MIN :
				type = (SqlType) oprnd.getType().clone();
				break;
			case MAX :
				type = (SqlType) oprnd.getType().clone();
				break;
			case SUM :
				type = (SqlType) oprnd.getType().clone();
				break;
			default :
				Debug.assertTrue(false, "internal error");
				break;
		}
		arg.setType(type);
	}

	public void visit(IfThenElse arg) throws SqlWrapperException {
		Expression iif = arg.getIf();
		iif.accept(this);

		Expression tthen = arg.getThen();
		tthen.accept(this);

		Expression eelse = arg.getElse();
		eelse.accept(this);
		/** @todo check type : iff : boolean, tthen = eelse */

		int tthenCode = ((SqlTypeAtom) tthen.getType()).getType().getPrimitiveTypeCode();
		int eelseCode = ((SqlTypeAtom) eelse.getType()).getType().getPrimitiveTypeCode();
		Debug.assertTrue(tthenCode == eelseCode, "tthenCode == eelseCode");

		arg.setType((SqlTypeAtom) tthen.getType().clone());
	}

	public void visit(Join arg) throws SqlWrapperException {
		List operandList = arg.getOperandList();
		SqlType type = null;
		Expression operand = null;
		if (1 < operandList.size()) {
			type = new SqlTypeCombinedStructure();
			for (int i = 0; i < operandList.size(); i++) {
				operand = (Expression) operandList.get(i);
				operand.accept(this);
				SqlType operandType = (SqlType) operand.getType().clone();
				((SqlTypeCombinedStructure) type).addStructure(operandType);
			}
		}
		else {
			operand = (Expression) operandList.get(0);
			operand.accept(this);
			type = (SqlType) operand.getType().clone();
		}
		arg.setType(type);

		List predicateList = arg.getPredicateList();
		if (null != predicateList) {
			Expression predicate = null;
			for (int i = 0; i < predicateList.size(); i++) {
				predicate = (Expression) predicateList.get(i);
				predicate.accept(this);
			}
		}
	}

    public void visit(ExternalVariable arg) throws SqlWrapperException {
        SqlType sType = null;
        
        switch (arg.getMappingInfo().getJavaType()) { // TODO : share with MappingInfo.setParameter()
            case JavaTypeInfo.JAVA_STRING :
            case JavaTypeInfo.SCHEMA_URI :
            case JavaTypeInfo.SCHEMA_QNAME :
         	case JavaTypeInfo.SCHEMA_DURATION :
                sType = new SqlTypeAtom(ptVarChar, null, SqlType.ONE);
                break;
            case JavaTypeInfo.JAVA_BOOLEAN :
                sType = new SqlTypeAtom(ptBit, null, SqlType.ONE);
                break;
            case JavaTypeInfo.JAVA_BIG_DECIMAL :
                sType = new SqlTypeAtom(ptDecimal, null, SqlType.ONE);
                break;
            case JavaTypeInfo.JAVA_FLOAT :
                sType = new SqlTypeAtom(ptFloat, null, SqlType.ONE);
                break;
            case JavaTypeInfo.JAVA_DOUBLE :
                sType = new SqlTypeAtom(ptDouble, null, SqlType.ONE);
                break;
            case JavaTypeInfo.JAVA_LONG :
                sType = new SqlTypeAtom(ptInteger, null, SqlType.ONE);
                break;
        	case JavaTypeInfo.SCHEMA_DATETIME :
                sType = new SqlTypeAtom(ptDate, null, SqlType.ONE);
                break;
            default :
        }
        arg.setType(sType);
    }

	public void visit(LitBoolean arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(LitDate arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptDate, null, SqlType.ONE));
	}

	public void visit(LitDecimal arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptDecimal, null, SqlType.ONE));
	}

	public void visit(LitDouble arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptDouble, null, SqlType.ONE));
	}

	public void visit(LitFloat arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptFloat, null, SqlType.ONE));
	}

	public void visit(LitInteger arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptInteger, null, SqlType.ONE));
	}

	public void visit(LitList arg) throws SqlWrapperException {
		SqlTypeAtom type = null;
		List list = arg.getLitList();
		if (null != list && !list.isEmpty()) {
			Literal element = (Literal) list.get(0);
			element.accept(this);
			type = (SqlTypeAtom) element.getType();
			for (int i = 1; i < list.size(); i++) {
				element = (Literal) list.get(i);
				element.setType(type);
			}
			type = new SqlTypeAtom(type.getType(), null, SqlType.MANY);
		}
		else
			type = new SqlTypeAtom(ptNull, null, SqlType.MANY);

		arg.setType(type);
	}

	public void visit(LitString arg) throws SqlWrapperException {
		// Consider it as a varchar to avoid useless length comparisons
		arg.setType(new SqlTypeAtom(new DbType(Types.VARCHAR,
									        arg.getValue().length()),
									        null,
									        SqlType.ONE
									        ));
	}

	public void visit(LitNull arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptNull, null, SqlType.ONE));
	}

	public void visit(RenameItem arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		SqlTypeAtom type = (SqlTypeAtom) operand.getType();
		type.setName(arg.getName());
		arg.setType(type);
	}

	public void visit(RenameRelation arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		SqlType oprndType = operand.getType();
		SqlType type = null;
		if (oprndType instanceof SqlTypeAtom) {
			type = (SqlTypeAtom) oprndType.clone();
			((SqlTypeAtom) type).setRelationName(arg.getName());
		} else if (oprndType instanceof SqlTypeStructure) {
			type = (SqlTypeStructure) oprndType.clone();
			((SqlTypeStructure) type).setName(arg.getName());
		} else if (oprndType instanceof SqlTypeCombinedStructure) {
			List attrList = ((SqlTypeCombinedStructure) oprndType).getAttributes();
			type = new SqlTypeStructure(arg.getName(), attrList, SqlType.MANY);
		}

		arg.setType(type);
	}

	public void visit(SortSpecification arg) throws SqlWrapperException {
		Expression expr = arg.getSortExpression();
		expr.accept(this);
		arg.setType(expr.getType());
	}

	public void visit(Table arg) throws SqlWrapperException {
		List attrList = arg.getAttributes();
		List attrTypeList = new ArrayList();
		SqlTypeStructure type = new SqlTypeStructure();
		type.setMultiplicity(SqlType.MANY);

		for (int i = 0; i < attrList.size(); i++) {
			Expression attr = (Expression) attrList.get(i);
			attr.accept(this);
			SqlTypeAtom attrType = (SqlTypeAtom) attr.getType();
			type.addAttribute(attrType);
		}
		arg.setType(type);
	}

	public void visit(TempValue arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType(operand.getType());
	}

	public void visit(UnOpAggregate arg) throws SqlWrapperException {
		visit((UnOpProject) arg);
		SqlType type = arg.getType();
		type.setMultiplicity(SqlType.ONE);
		/* arg.setType(type); */
	}

	public void visit(UnOpExists arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(UnOpIsNull arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(UnOpGroup arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType((SqlType) operand.getType().clone());
	}

	public void visit(UnOpMinus arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType(operand.getType());
	}

	public void visit(UnOpNot arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType(new SqlTypeAtom(ptBit, null, SqlType.ONE));
	}

	public void visit(UnOpProject arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		List selectList = arg.getItemList();
		List itemTypeList = new ArrayList();

		for (int i = 0; i < selectList.size(); i++) {
			Expression item = (Expression) selectList.get(i);
			item.accept(this);
			itemTypeList.add((SqlTypeAtom) item.getType().clone());
		}

		SqlType type = null;
		if (1 < itemTypeList.size()) {
			type = new SqlTypeStructure(itemTypeList, SqlType.MANY);
		} else {
			type = (SqlTypeAtom) itemTypeList.get(0);
			type.setMultiplicity(SqlType.MANY);
		}

		arg.setType(type);
	}

	public void visit(UnOpRestrict arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType((SqlType) operand.getType().clone());

		List predicateList = arg.getPredicateList();
		if (null != predicateList) {
			Expression predicate = null;
			for (int i = 0; i < predicateList.size(); i++) {
				predicate = (Expression) predicateList.get(i);
				predicate.accept(this);
			}
		}
	}

	public void visit(UnOpSort arg) throws SqlWrapperException {
		Expression operand = arg.getOperand();
		operand.accept(this);
		arg.setType((SqlType) operand.getType().clone());

		List list = arg.getSortSpecificationList();
		SortSpecification sortSpec = null;
		Expression sortExpr = null;
		Expression predicate = null;
		for (int i = 0; i < list.size(); i++) {
			sortSpec = (SortSpecification) list.get(i);
			sortExpr = sortSpec.getSortExpression();
			sortExpr.accept(this);
		}
	}

	public void visit(AfTrim arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		// don't know the resulting length : consider it as a varchar to avoid
		// the length comparison problems
		arg.setType(new SqlTypeAtom(ptVarChar, null, SqlType.ONE));
	}

	public void visit(AfCast arg) throws SqlWrapperException {
		((Expression) arg.getExpression()).accept(this);
		SqlTypeAtom type = new SqlTypeAtom(arg.getTargetType(), arg.getName());
		type.setName(arg.getName());
		arg.setType(type);
	}

	public void visit(XfCast arg) throws SqlWrapperException {
		((Expression) arg.getExpression()).accept(this);
		arg.setType(new SqlTypeAtom(arg.getTargetType(), null));
	}

	public void visit(XfAbs arg) throws SqlWrapperException {
		Expression para = (Expression) arg.getArgument(0);
		para.accept(this);
		arg.setType((SqlType) para.getType().clone());
	}

	public void visit(XfCeiling arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptInteger, null, SqlType.ONE));
	}

	public void visit(XfConcat arg) throws SqlWrapperException {
		Expression parameter;
		int size = 0;
		for (int i = 0; i < arg.getArgumentNumber(); i++) {
			parameter = (Expression) arg.getArgument(i);
			parameter.accept(this);
			size += ((SqlTypeAtom) parameter.getType()).getType().getLength();
		}

		// Consider it as a varchar to avoid useless length comparisons
		arg.setType(new SqlTypeAtom(new DbType(Types.VARCHAR, size), null, SqlType.ONE));
	}

	public void visit(XfContains arg) throws SqlWrapperException {
		Expression para = (Expression) arg.getArgument(0);
		para.accept(this);
		para = (Expression) arg.getArgument(1);
		para.accept(this);
		arg.setType(new SqlTypeAtom(ptBoolean, null, SqlType.ONE));
	}

	public void visit(XfCurrentDateTime arg) throws SqlWrapperException {
		arg.setType(new SqlTypeAtom(ptTimeStamp, null, SqlType.ONE));
	}

	public void visit(XfDate arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptDate, null, SqlType.ONE));
	}

	public void visit(XfDateTime arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptTimeStamp, null, SqlType.ONE));
	}

	public void visit(XfEndsWith arg) throws SqlWrapperException {
		Expression para = (Expression) arg.getArgument(0);
		para.accept(this);
		para = (Expression) arg.getArgument(1);
		para.accept(this);
		arg.setType(new SqlTypeAtom(ptBoolean, null, SqlType.ONE));
	}

	public void visit(XfFloor arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptInteger, null, SqlType.ONE));
	}

	public void visit(XfLowerCase arg) throws SqlWrapperException {
		Expression para = (Expression) arg.getArgument(0);
		para.accept(this);
		arg.setType(para.getType());
	}

	public void visit(XfRound arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptInteger, null, SqlType.ONE));
	}

	public void visit(XfStartsWith arg) throws SqlWrapperException {
		Expression para = (Expression) arg.getArgument(0);
		para.accept(this);
		para = (Expression) arg.getArgument(1);
		para.accept(this);
		arg.setType(new SqlTypeAtom(ptBoolean, null, SqlType.ONE));
	}

	public void visit(XfStringLength arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptInteger, null, SqlType.ONE));
	}

	public void visit(XfSubString arg) throws SqlWrapperException {
		Expression string = (Expression) arg.getArgument(0);
		string.accept(this);
		Expression startIndex = (Expression) arg.getArgument(1);
		startIndex.accept(this);
		Expression length = (Expression) arg.getArgument(2);
		int size = 0;
		if (length == null && startIndex instanceof LitInteger) {
			size = (int) ((SqlTypeAtom) string.getType()).getType().getLength() - ((LitInteger) startIndex).getValue().intValue() + 1;
		}
		else {
			length.accept(this);
			if (length instanceof LitInteger)
				size = ((LitInteger) length).getValue().intValue();
		}

		if (size == 0) // do not matter with length problems and consider it as a varchar...
			arg.setType(new SqlTypeAtom(ptVarChar, null, SqlType.ONE));
		else {
			SqlTypeAtom type = (SqlTypeAtom) string.getType().clone();
			type.getType().setLength(size);
			arg.setType(type);
		}
	}

	public void visit(XfTime arg) throws SqlWrapperException {
		((Expression) arg.getArgument(0)).accept(this);
		arg.setType(new SqlTypeAtom(ptTime, null, SqlType.ONE));
	}

	public void visit(XfUpperCase arg) throws SqlWrapperException {
		Expression para = (Expression) arg.getArgument(0);
		para.accept(this);
		arg.setType(para.getType());
	}

}
