/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.algebra;

import java.sql.Types;
import java.util.*;

import org.xquark.extractor.common.*;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.runtime.IDProvider;
import org.xquark.extractor.xfunctions.*;
import org.xquark.jdbc.typing.DbType;
import org.xquark.schema.ComplexType;
import org.xquark.schema.ElementDeclaration;
import org.xquark.xpath.Axis;
import org.xquark.xpath.NodeKind;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.*;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.typing.*;

/**
 * Builds the algebra tree from a database XQueryExpression (QDB).
 */
public final class GenAlgebraVisitor extends DefaultParserVisitor implements Constants {
	private static final String RCSRevision = "$Revision: 1.52 $";
	private static final String RCSName = "$Name:  $";

	/* the "return value" of each visit function is pushed in this stack */
	private Stack _returnStack = new Stack();

	/* stores the context for related Path Expression and sortby */
	private Stack _typeContextStack = new Stack();

	/* temporary sqlType builder for guiding generation */ // TODO: check if basing on query flags is not enough
	private TypeInterpreter _typeInterpreter;
	private AlgebraTypeVisitor _typeVistor = new AlgebraTypeVisitor();
	private MetaDataManager _mdm;

	/* variable table that stores the translation of the variables */
	private VariableTable _variableTable = null;

	// NOTE: Attributes here, are "columns" in the relational terminology
	/* Locally used here for predicate exploration in FLWOR translation */
	private BuildReferredAttributes _brav = new BuildReferredAttributes();

	/* Locally used in FLWOR translation */
	private AddKeys _akv = new AddKeys();

	/* the final algebra tree generated by a visite */
	private Expression _algebraTree = null;

	/** This object maintains a correpondence between path waited by reconstruction
	 * and attributes of the top relation.
	 */
	private Mapper _mapper = null;

	/** if there is a SortedOuterUnion in the query ,
	 * this list contains all the arguments of the SortedOuterUnion.
	 * This help adding keys for these arguments
	 */
	private List _flwrsInSortedOuterUnion = new ArrayList();

	static final String NUMBER_COLUMN_PREFIX = "SqlWrapperNumber_";

	/* counters for generate unique name */
	protected IDProvider _attIDProvider = null;
	protected IDProvider _relIDProvider = null;

	/* a flag to treat view in nested xquery */ // QUESTION : not used ? Only stored from AddKeyVisitors and published
	private boolean _heavyEliminationNeeded = false;

	public GenAlgebraVisitor(TypeInterpreter ti, MetaDataManager mdm, IDProvider attIDProvider, IDProvider relIDProvider) {
		setTypeInterpreter(ti);
		setMetaDataManager(mdm);
		_attIDProvider = attIDProvider;
		_relIDProvider = relIDProvider;
	}

	public void setTypeInterpreter(TypeInterpreter ti) {
		_typeInterpreter = ti;
	}

	public void setMetaDataManager(MetaDataManager mdm) {
		_mdm = mdm;
	}

	public boolean isHeavyEliminationNeeded() {
		return _heavyEliminationNeeded;
	}

	public void noticSortedOuterUnion(List argumentsOfSortedOuterUnion) {
		XQueryExpression xExpr = null;
		Variable var = null;
		FLWRExpression flwr = null;
		for (int i = 0; i < argumentsOfSortedOuterUnion.size(); i++) {
			xExpr = (XQueryExpression) argumentsOfSortedOuterUnion.get(i);
			if (xExpr instanceof Variable) { // TODO: Check if a variable can be produced by front-end 
				_flwrsInSortedOuterUnion.add(((Variable) xExpr).getExpression());
			} else if (xExpr instanceof FLWRExpression) {
				_flwrsInSortedOuterUnion.add(xExpr);
			} else {
				Debug.assertTrue(false, "Internal logic error.");
			}
		}
	}

	public void reinit() {
		_returnStack.clear();
		_typeContextStack.clear();
		_variableTable = null;
		_flwrsInSortedOuterUnion.clear();
		_heavyEliminationNeeded = false;
	}

	public Expression getAlgebraTree() {
		//Trace.enter(this, "getAlgebraTree()");

		Expression retVal = (Expression) _returnStack.peek();
		//Trace.exit(this, "getAlgebraTree()");
		return retVal;
	}

	public void visit(BinOpANDExpression arg) throws XQueryException {
		Expression retVal = null;

		XQueryExpression xLOprnd, xROprnd;
		Expression lOprnd, rOprnd;
		SqlType lType, rType;

		// check the signature
		xLOprnd = arg.getLeftOperand();
		xROprnd = arg.getRightOperand();

		lType = _typeInterpreter.interprete(xLOprnd.getQType());
		rType = _typeInterpreter.interprete(xROprnd.getQType());
		if (!lType.isBoolean() || !rType.isBoolean()) {
			throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_SIG", arg), null);
		}

		if (xLOprnd instanceof ValueBoolean) {
			boolean lValue = ((ValueBoolean) xLOprnd).getBooleanValue();
			if (xROprnd instanceof ValueBoolean) {

				/* both left Operand and right operand are boolean constants
				 *  so the return value is a boolean constant
				 */
				boolean rValue = ((ValueBoolean) xROprnd).getBooleanValue();
				retVal = new LitBoolean(lValue && rValue);
				_returnStack.push(retVal); // a return value
			} else {

				/* left Operand is boolean constant but  right operand is not  */
				if (lValue) {
					xROprnd.accept(this); // a return value
					retVal = (Expression) _returnStack.peek();
				} else {
					retVal = new LitBoolean(false);
					_returnStack.push(retVal); // a return value
				}
			}
		} else {
			if (xROprnd instanceof ValueBoolean) {

				/* left Operand is not boolean constant but  right operand is. */
				boolean rValue = ((ValueBoolean) xROprnd).getBooleanValue();
				if (rValue) {
					xLOprnd.accept(this); // a return value
					retVal = (Expression) _returnStack.peek();
				} else {
					retVal = new LitBoolean(false);
					_returnStack.push(retVal); // a return value
				}
			} else {

				/* neither left Operand nor right operand are boolean constants */
				xLOprnd.accept(this);
				lOprnd = (Expression) _returnStack.pop();
				xROprnd.accept(this);
				rOprnd = (Expression) _returnStack.pop();

				retVal = new BinOpBoolean(Constants.AND, lOprnd, rOprnd);
				_returnStack.push(retVal); // a return value
			}
		}

		retVal.setOrginalXExpr(arg);
	}

	public void visit(BinOpORExpression arg) throws XQueryException {
		Expression retVal = null;

		XQueryExpression xLOprnd, xROprnd;
		Expression lOprnd, rOprnd;
		SqlType lType, rType;

		// check the signature
		xLOprnd = arg.getLeftOperand();
		xROprnd = arg.getRightOperand();
		lType = _typeInterpreter.interprete(xLOprnd.getQType());
		rType = _typeInterpreter.interprete(xROprnd.getQType());
		if (!lType.isBoolean() || !rType.isBoolean()) {
			throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_SIG", arg), null);
		}

		if (xLOprnd instanceof ValueBoolean) {
			boolean lValue = ((ValueBoolean) xLOprnd).getBooleanValue();
			if (xROprnd instanceof ValueBoolean) {

				/* both left Operand and right operand are boolean constants
				 *  so the return value is a boolean constant
				 */
				boolean rValue = ((ValueBoolean) xROprnd).getBooleanValue();
				retVal = new LitBoolean(lValue || rValue);
				_returnStack.push(retVal); // a return value
			} else {

				/* left Operand is boolean constant but  right operand is not  */
				if (lValue) {
					retVal = new LitBoolean(true);
					_returnStack.push(retVal); // a return value
				} else {
					xROprnd.accept(this); // a return value
					retVal = (Expression) _returnStack.peek();
				}
			}
		} else {
			if (xROprnd instanceof ValueBoolean) {

				/* left Operand is not boolean constant but  right operand is. */
				boolean rValue = ((ValueBoolean) xROprnd).getBooleanValue();
				if (rValue) {
					retVal = new LitBoolean(true);
					_returnStack.push(retVal); // a return value
				} else {
					xLOprnd.accept(this); // a return value
					retVal = (Expression) _returnStack.peek();
				}
			} else {

				/* neither left Operand nor right operand are boolean constants */
				xLOprnd.accept(this);
				lOprnd = (Expression) _returnStack.pop();
				xROprnd.accept(this);
				rOprnd = (Expression) _returnStack.pop();

				retVal = new BinOpBoolean(OR, lOprnd, rOprnd);
				_returnStack.push(retVal); // a return value
			}
		}

		retVal.setOrginalXExpr(arg);
	}

	public void visit(CastTreatExpression arg) throws XQueryException {
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_OPTR", arg), null);
	}

	public void visit(FLWRExpression arg) throws XQueryException {
		/* create a new VariableTable for the new scope*/
		VariableTable newVT = new VariableTable();
		newVT.setDefautVariableTable(_variableTable);
		_variableTable = newVT;

		Expression retVal = null;
		Mapper mapper = new Mapper(); // TODO: check the generation of mapper is correct (2nd Qdb of UCR11 is strange...)
		// at least items generated for key could be mapped on already present ?

		/* generate algebra for FOR,LET clause */
		List variables = arg.getVariables();
		Iterator iter = variables.iterator();
		Variable variable;
		int bindingType;

		List tableInstanceList = new ArrayList();
		List bindingTypeList = new ArrayList();

		List pseudoIdPathList = new ArrayList();

		/* for the case of a artificial for clause */
		String pseudoIdPath = null;
		List idOrder = new ArrayList();
		VarMappingInfo varInfo;

		List predicateListAddedByForBinding = new ArrayList();
		while (iter.hasNext()) {
			variable = ((Variable) iter.next());
			bindingType = variable.getBindingType();
			switch (bindingType) {
				case LET_BINDINGTYPE :
					visitLetBinding(variable);
					break;
				case FOR_BINDINGTYPE :
					varInfo = visitForBinding(variable);
					if (null != varInfo) {
						tableInstanceList.add(varInfo.relation);
						idOrder.add(varInfo);
					} else {
						visit(variable);
						Expression atomicExpr = (Expression) _returnStack.pop();

						if (!(atomicExpr instanceof Literal))
							predicateListAddedByForBinding.add(new UnOpIsNull(atomicExpr, true));

						if (variable.getExpression() instanceof ValueInteger) {
							pseudoIdPath = "$" + variable.getVarName().getName();
							idOrder.add(pseudoIdPath);
							pseudoIdPathList.add(pseudoIdPath);
						}
					}
					break;
				default :
					throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_XEXPR", variable.toString()), null);
			}
		}

		/* translate predicate in XQuery into Expression list */
		List aPredicates = new ArrayList();
		aPredicates.addAll(predicateListAddedByForBinding);
		XQueryExpression whereClause = arg.getWhereClause();
		if (null != whereClause) {
			List xPredicates = new ArrayList();
			xPredicates = decomposePredicate(whereClause, xPredicates);
			iter = xPredicates.iterator();
			XQueryExpression xPredicate;
			Expression aPredicate;
			while (iter.hasNext()) {
				xPredicate = (XQueryExpression) iter.next();
				xPredicate.accept(this);
				aPredicate = (Expression) _returnStack.pop();
				aPredicate.accept(_brav);
				aPredicates.add(aPredicate);
			}
		}

		if (1 == tableInstanceList.size())
			retVal = new UnOpRestrict((Expression) tableInstanceList.get(0), aPredicates);
		else if (1 < tableInstanceList.size())
			retVal = new Join(tableInstanceList, aPredicates);

		/* generate algebra for RETURN clause */
		XQueryExpression returnClause = arg.getReturnClause();
		List subPathList = null;
		XQueryExpression subPath = null;

		List selectList = new ArrayList();
		Variable xExpr;
		Expression aExpr;

		if (returnClause instanceof XQueryExpressionSequence) {
			/* normal case */
			List xExprs = ((XQueryExpressionSequence) returnClause).getSubExpressions();

			Debug.assertTrue(null != xExprs, "null != xExprs");
			SqlType xExprSqlType;

			for (int j = 0; j < xExprs.size(); j++) {
				xExpr = (Variable) xExprs.get(j);
				subPathList = xExpr.getUniqueSubPathList();
				if (null != subPathList && !subPathList.isEmpty()) {
					for (int i = 0; i < subPathList.size(); i++) {
						subPath = (XQueryExpression) subPathList.get(i);
						// NOTE: Front-end may generate let that have to be ignored
						// 		(these lets may be useful for other backends that do
						// 		not want to expand variable for perf reasons)
						if (!subPath.getQType().isMultiple()) {
							subPath.accept(this);
							aExpr = (Expression) _returnStack.pop();
							selectList.add(aExpr);
						}
					}
				}
				// TODO:  the following branch is to be removed ? Used "for $x in 1" ?
				else if (xExpr instanceof Variable && ((Variable) xExpr).getExpression() instanceof ValueInteger) {
					xExpr.accept(this);
					aExpr = (Expression) _returnStack.pop();
					selectList.add(aExpr);
				}
			}
		} else { // for variable by example
			returnClause.accept(this);
			aExpr = (Expression) _returnStack.pop();
			selectList.add(aExpr);
		}

		if (null == retVal) /* nothing is producted by JoinBuilder*/ {
			/* check SortedOuterUnion */
			boolean isSortedOuterUnion = false;

			// the way for detecting the existance of SOU is simplified
			XQueryExpression tmpExpr = returnClause;
			if (returnClause instanceof XQueryExpressionSequence && ((XQueryExpressionSequence) returnClause).getSubExpressions().size() == 1)
				tmpExpr = (XQueryExpression) ((XQueryExpressionSequence) returnClause).getSubExpressions().get(0);
			else if (tmpExpr instanceof Variable)
				tmpExpr = ((Variable) tmpExpr).getExpression();
			else if (tmpExpr instanceof InternalFunctionCall)
				isSortedOuterUnion = true;

			if (isSortedOuterUnion) {
				/* sorted-outer-union */
				retVal = (Expression) selectList.get(0);
			} else {
				/* case : a special case where this is not For Clause in the query
				 * and return clause contains several  values
				 */

				if (!aPredicates.isEmpty()) {
					/* generate UnOpRestrict for aPredicates whick was not treaded by joinbuilder */
					retVal = new UnOpRestrict(new DummyTable(), aPredicates);
					retVal = new UnOpProject(retVal, selectList, false, _attIDProvider);
				} else
					retVal = new UnOpProject(new DummyTable(), selectList, false, _attIDProvider);

				/* second pass of making map , for catch the AttributesExpressions which
				 * are genreated from  TupleExpressions
				 */
				mapper.createMap(((UnOpProject) retVal).getItemList());

				/* add pseudoId(s) to the result */
				if (pseudoIdPathList.size() > 0) {
					for (int i = 0; i < pseudoIdPathList.size(); i++) {
						pseudoIdPath = (String) pseudoIdPathList.get(i);
						MapItem pseudoId = mapper.map(pseudoIdPath);
						//Debug.assertTrue(null!= pseudoId, "null!= pseudoId");
						if (pseudoId != null)
							mapper.addIdentifier(pseudoIdPath, pseudoId.getRelatedExpression());
					}
				}

				retVal.setMapper(mapper);
			}
		} else {
			/* first pass of making map , for catch the TupleExpressions that
			 * disappear after adding generation of selectList to UnOp project
			 */
			List tupleExpressionList = new ArrayList();
			for (int i = 0; i < selectList.size(); i++) {
				Expression selectItem = (Expression) selectList.get(i);
				if (selectItem instanceof TupleExpression) {
					tupleExpressionList.add(selectItem);
				}
			}
			if (0 < tupleExpressionList.size()) {
				mapper.createMap(tupleExpressionList);
			}

			retVal = new UnOpProject(retVal, selectList, false, _attIDProvider);

			/* second pass of making map , for catch the AttributesExpressions which
			 * are generated from TupleExpressions
			 */
			mapper.createMap(((UnOpProject) retVal).getItemList());

			if (_flwrsInSortedOuterUnion.contains(arg)) {
				/* add key(s) to the result */
				List keyItemList = new ArrayList();

				/*add key(s) to all table instances*/
				List allKeys = new ArrayList();

				String pathForkey = null;
				for (int i = 0; i < idOrder.size(); i++) {
					Object obj = idOrder.get(i);
					if (obj instanceof VarMappingInfo) {
						VarMappingInfo item = (VarMappingInfo) obj;
						pathForkey = "key/$" + item.varName;
						_akv.reinit();
						item.relation.accept(_akv);
						allKeys = _akv.getKeys();

						/* added for support view in nested xquery */
						if (_akv.isHeavyEliminationNeeded())
							_heavyEliminationNeeded = true;
						/* end of -- added for support view in nested xquery */

						Debug.assertTrue(0 < allKeys.size(), "0 < allKeys.size()");

						for (int j = 0; j < allKeys.size(); j++) {
							Expression keyAttribute = (Expression) allKeys.get(j);
							// TODO verify if the  keyAttribute is already in the retVal*/
							keyAttribute = ((UnOpProject) retVal).addItem(keyAttribute);
							allKeys.set(j, keyAttribute);
						}
						mapper.addIdentifier(pathForkey, allKeys);
					} else { // String
						/* generate a pseudoId*/
						pseudoIdPath = (String) obj;
						MapItem pseudoId = mapper.map(pseudoIdPath);
						Debug.assertTrue(null != pseudoId, "null!= pseudoId");
						mapper.addIdentifier(pseudoIdPath, pseudoId.getRelatedExpression());
					}

				}
			}
			retVal.setMapper(mapper);
		}

		/* translate sortby clause */
		List xOrderbySpecList = arg.getOrderBy();
		if (null != xOrderbySpecList && (!xOrderbySpecList.isEmpty())) {

			List aSortSpecList = new ArrayList();
			XQueryExpression xOrderbyExpr = null;
			SortSpecification aSortSpec = null;
			Expression projectItem = null;
			AttributeExpression aSortExpr = null;
			int order = 0;
			String strOrderbytExpr = null;

			PlaceHolder placeHolder = null;
			Object obj = null;
			MapItem sortItemMap = null;
			for (int i = 0; i < xOrderbySpecList.size(); i++) {
				obj = xOrderbySpecList.get(i);

				if (obj instanceof PlaceHolder) {
					List sortSpecList = generateSortSpecList((PlaceHolder) obj, mapper);
					aSortSpecList.addAll(sortSpecList);
				} else {
					xOrderbyExpr = (XQueryExpression) obj;

					if (xOrderbyExpr.getOrder(0) == Constants.PLACE_ORDER) {
						xOrderbyExpr.accept(this);
						Expression relation = (Expression) _returnStack.pop();
						if (relation instanceof TupleExpression) {
							relation = ((TupleExpression) relation).getTableInstance();
						}
						if (relation instanceof RenameRelation) {
							RenameRelation item = (RenameRelation) relation;
							_akv.reinit();
							item.accept(_akv);
							List allKeys = _akv.getKeys();
							for (int j = 0; j < allKeys.size(); j++) {
								Expression key = (Expression) allKeys.get(j);
								aSortSpecList.add(new SortSpecification(key, ASCENDING_ORDER));
							}
						}
					} else {
						/* find out corresponding AttributeExpression */
						strOrderbytExpr = xOrderbyExpr.getStringValue();

						sortItemMap = mapper.map(strOrderbytExpr);
						if (null == sortItemMap) {
							int index = strOrderbytExpr.indexOf("/child::text()");
							if (-1 == index) {
								strOrderbytExpr += "/child::text()";
							} else {
								strOrderbytExpr = strOrderbytExpr.substring(0, index);
							}
							sortItemMap = mapper.map(strOrderbytExpr);
						}

						if (null == sortItemMap) {
							if (!xOrderbyExpr.getQType().isMultiple()) {
								xOrderbyExpr.accept(this);
								aExpr = (Expression) _returnStack.pop();
								if (aExpr instanceof AttributeExpression) {
									aSortExpr = (AttributeExpression) aExpr;
								}
							}
						} else {
							projectItem = sortItemMap.getRelatedExpression();
							aSortExpr = new AttributeExpression(projectItem instanceof AttributeExpression ? ((AttributeExpression) projectItem).getTableInstance() : null, projectItem.getName());
							aSortExpr.setOrginalXExpr(xOrderbyExpr);
							aSortExpr.setUnderlyingExpr(projectItem);
						}

						Debug.assertTrue(null != aSortExpr, "failed in finding order by expression in return clause");

						order = xOrderbyExpr.getOrder(0);
						if (order == -1)
							order = 0;
						aSortSpec = new SortSpecification(aSortExpr, order);
						aSortSpecList.add(aSortSpec);
					}
				}
			}
			/* added for sybase*/
			removeDuplicatedExpr(aSortSpecList);
			retVal = new UnOpSort(retVal, aSortSpecList);
			/* the following line is added so that the RemoveProjectReplaceVisitor can reach the UnOpSort*/
			 ((UnOpProject) ((UnOpSort) retVal).getOperand()).setAssociatedSort((UnOpSort) retVal);
			retVal.setMapper(mapper);
		}

		retVal.setOrginalXExpr(arg);

		/* remove the VariableTable from VariableTable chaine*/
		_variableTable = _variableTable.getDefautVariableTable();

		_returnStack.push(retVal);
	}

	private List generateSortSpecList(PlaceHolder placeHolder, Mapper mapper) {
		List varNameList = new ArrayList();
		List varList = placeHolder.getList();
		Variable var = null;
		for (int i = 0; i < varList.size(); i++) {
			var = (Variable) varList.get(i);
			varNameList.add(var.toString());
		}
		return generateSortSpecList(varNameList, mapper);
	}

	private List generateSortSpecList(List varNameList, Mapper mapper) {
		List retVal = new ArrayList();
		List idMapItemList = mapper.getIdentifierList();
		MapItem mapItem = null;
		MapItem childMapItem = null;
		Expression sortExpr = null;
		String varName = null;
		if (idMapItemList != null) {
			for (int i = 0; i < idMapItemList.size(); i++) {
				mapItem = (MapItem) idMapItemList.get(i);
				varName = mapItem.getPath().substring(4);
				if (varNameList.contains(varName)) {
					/*mapItem is of current level */
					if (mapItem.isLeafPath()) {
						sortExpr = mapItem.getRelatedExpression();
						retVal.add(createSortSpecification(sortExpr, ASCENDING_ORDER));
					} else {
						List children = mapItem.getChildren();
						for (int j = 0; j < children.size(); j++) {
							childMapItem = (MapItem) mapper.map((String) children.get(j));
							sortExpr = childMapItem.getRelatedExpression();
							retVal.add(createSortSpecification(sortExpr, ASCENDING_ORDER));
						}
					}
				}
			}
		}

		return retVal;
	}

	static public void removeDuplicatedExpr(List sortSpecList) {
		List exprList = new ArrayList();

		SortSpecification sortSpec = null;
		AttributeExpression expr = null;

		for (int i = 0; i < sortSpecList.size(); i++) {
			sortSpec = (SortSpecification) sortSpecList.get(i);
			exprList.add(sortSpec.getSortExpression());
		}

		boolean duplicated = false;
		for (int i = exprList.size() - 1; i > -1; i--) {
			expr = (AttributeExpression) exprList.get(i);
			for (int j = 0; j < i; j++) {
				duplicated = UnOpProject.isIdentical((AttributeExpression) exprList.get(j), expr);
				if (duplicated) {
					sortSpecList.remove(i);
					break;
				}
			}
		}
	}

	private SortSpecification createSortSpecification(Expression expr, int order) {
		String attrName = expr.getName();
		Debug.assertTrue(null != attrName, "null != attrName");

		AttributeExpression sortExpr = new AttributeExpression(expr instanceof AttributeExpression ? ((AttributeExpression) expr).getTableInstance() : null, expr.getName());
		sortExpr.setUnderlyingExpr(expr);

		return new SortSpecification(sortExpr, order);
	}

	public void visit(InternalFunctionCall arg) throws XQueryException {
		String functionName;
		functionName = arg.getFuncName().getLocalName();
		if (functionName.equalsIgnoreCase("SOU")) {
			visitSortedOuterUnion(arg);
		} else
			throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_XEXPR", arg), null);
	}

	public void visitSortedOuterUnion(InternalFunctionCall arg) throws XQueryException {
		Expression retVal = null;
		List args = arg.getArguments();
		List aExprs = new ArrayList();
		List sortSpecListList = new ArrayList();

		/* pick out UnOpprojects and UnOpSorts*/
		Mapper mapper = null;
		UnOpProject project = null;
		for (int i = 0; i < args.size(); i++) {
			((XQueryExpression) args.get(i)).accept(this);
			Expression aExpr = (Expression) _returnStack.pop();
			if (aExpr instanceof UnOpProject) {
				aExprs.add(aExpr);
				sortSpecListList.add(null);
			} else if (aExpr instanceof UnOpSort) {
				UnOpSort sort = (UnOpSort) aExpr;
				sortSpecListList.add(sort.getSortSpecificationList());
				project = (UnOpProject) sort.getOperand();
				//                mapper = project.getMapper();
				//                mapper.addSortItemList(sort.getSortSpecificationList());
				aExprs.add(project);
			} else {
				Debug.assertTrue(false, "Internal error.");
			}
		}

		/* build a globle mapper */
		mapper = ((Expression) aExprs.get(0)).getMapper();

		try {
			mapper = (Mapper) mapper.clone();
		} catch (CloneNotSupportedException e) {
			throw new XQueryException(e.getMessage());
		}

		int sqlWrapperPartBitMap = 1;
		mapper.setPart(sqlWrapperPartBitMap);
		for (int i = 1; i < aExprs.size(); i++) {
			Mapper tmpMapper = ((Expression) aExprs.get(i)).getMapper();
			sqlWrapperPartBitMap = sqlWrapperPartBitMap << 1;
			tmpMapper.setPart(sqlWrapperPartBitMap);
			mapper.merge(tmpMapper);
		}

		Numbering num = new Numbering(mapper.getIdentifierList(), args.size());
		num.number();
		mapper.setNumbering(num);

		List mapItemList = mapper.getMapItemList();
		for (int i = 0; i < mapItemList.size(); i++) {
			MapItem mapItem = (MapItem) mapItemList.get(i);
			if (mapItem.isLeafPath()) {
				Expression relatedExpr = mapItem.getRelatedExpression();
				relatedExpr.accept(_typeVistor);
			}
		}

		/* rebuild select list of each participating relation */
		sqlWrapperPartBitMap = 1;
		RenameItem partAttr = null;
		RenameItem numberAttr = null;
		int numbers[];
		int number;
		List sortSpecList = null;
		SortSpecification sortSpec = null;
		AttributeExpression sortExpr = null;
		XQueryExpression sortXExpr = null;
		Expression projectItem = null;
		for (int i = 0; i < aExprs.size(); i++) {
			//            Debug.assertTrue(aExprs.get(i) instanceof UnOpProject,"aExprs.get(i) instanceof UnOpProject");
			project = (UnOpProject) aExprs.get(i);

			reBuildSelectList(project, ((Expression) aExprs.get(i)).getMapper(), mapper, 0 == i);
			if (0 == i) {
				rebuildSortSpecListList(mapper, sortSpecListList);
			}

			project.setMapper(mapper);

			LitInteger part = new LitInteger(sqlWrapperPartBitMap);
			partAttr = new RenameItem(part, "SQLWrapper_Part");
			project.addItem(partAttr);

			numbers = num.getNumbersForQuery(i + 1);
			for (int j = 1; j < numbers.length; j++) {
				number = numbers[j];
				numberAttr = new RenameItem(new LitInteger(number), NUMBER_COLUMN_PREFIX + Integer.toString(j));
				project.addItem(numberAttr);
			}

			sqlWrapperPartBitMap = sqlWrapperPartBitMap << 1;
		}
		MapItem sqlWrapperPart = new MapItem(partAttr.getName(), null, partAttr.getName(), null);

		mapper.setPartAttribute(partAttr);

		retVal = new BinOpMerge((Expression) aExprs.get(0), (Expression) aExprs.get(1));
		for (int i = 2; i < aExprs.size(); i++) {
			retVal = new BinOpMerge(retVal, (Expression) aExprs.get(i));
		}
		retVal.setMapper(mapper);

		/* generate UnOpSort */
		num._mapper = mapper;

		num.setSortSpecListList(sortSpecListList);
		sortSpecList = num.getSortList();
		retVal = new UnOpSort(retVal, sortSpecList);

		retVal.setMapper(mapper);
		_returnStack.push(retVal);
	}

	private List rebuildSortSpecListList(Mapper mapper, List sortSpecListList) {
		SortSpecification sortSpec = null;
		AttributeExpression sortExpr = null;
		XQueryExpression xExpr = null;
		List sortSpecList = null;

		List mapItemList = new ArrayList();
		MapItem mapItem = null;
		String attrName = null;

		for (int i = 0; i < sortSpecListList.size(); i++) {
			sortSpecList = (List) sortSpecListList.get(i);
			if (null == sortSpecList) {
				continue;
			}
			for (int j = sortSpecList.size() - 1; j >= 0; j--) {
				sortSpec = (SortSpecification) sortSpecList.get(j);
				sortExpr = (AttributeExpression) sortSpec.getSortExpression();
				mapItem = findMapItem(mapper, sortExpr);
				Debug.assertTrue(null != mapItem && mapItem.isLeafPath(), "null !=mapItem && mapItem.isLeafPath()");
				if (mapItemList.contains(mapItem)) {
					/* */
					sortSpecList.remove(j);
				} else {
					mapItemList.add(mapItem);
					attrName = mapItem.getItemName();
					sortExpr.setAttribute(attrName);
				}
			}
		}
		return sortSpecListList;
	}

	private MapItem findMapItem(Mapper mapper, Expression expr) {
		MapItem retVal = null;

		String strXExpr = expr.getOrginalXExpr().toString();

		retVal = mapper.map(strXExpr);
		if (null == retVal) {
			int index = strXExpr.indexOf("/child::text()");
			if (-1 == index) {
				strXExpr += "/child::text()";
			} else {
				strXExpr = strXExpr.substring(0, index);
			}
			retVal = mapper.map(strXExpr);
		}
		return retVal;
	}

	protected String insertNullItem(UnOpProject project, int index, MapItem mapItem) {
		Expression nullItem = new LitNull();
		SqlTypeAtom itemType = (SqlTypeAtom) mapItem.getRelatedExpression().getType();
		DbType primitiveType = (DbType) itemType.getType();
		nullItem = new AfCast(nullItem, primitiveType);

		nullItem = new RenameItem(nullItem, mapItem.getItemName());
		Expression newExpr = project.insertItem(index, nullItem);

		return newExpr.getName();
	}

	protected void reBuildSelectList(UnOpProject project, Mapper local, Mapper global, boolean updateMap) {
		List globalItems = global.getMapItemList();
		List localItems = local.getMapItemList();
		List selectItems = project.getItemList();

		int iG = 0; /* a pointer to globalItems */
		int leafPathCounter = 0; /* count the number of leaf path */
		int iS = 0; /* a pointer to selectlItems */

		MapItem localItem;
		MapItem globalItem;
		int j = 0;
		for (iG = 0; iG < globalItems.size(); iG++) {
			globalItem = (MapItem) globalItems.get(iG);
			if (globalItem.isLeafPath()) {
				leafPathCounter++;
				if (null == local.map(globalItem.getPath())) {
					String newName = insertNullItem(project, leafPathCounter - 1, globalItem);
					if (updateMap) {
						globalItem.setItemName(newName);
					}
				}
			}
		}
	}

	public void visit(ListOpArithExpression arg) throws XQueryException {
		Expression retVal = null;

		/**2.5.1 Arithmetic operators
		 *  Syntax
		 *  AdditiveExpr    ::=    Expr ("+" | "-") Expr
		 *  MultiplicativeExpr    ::=    Expr ("*" | "div" | "mod") Expr
		 *  UnaryExpr    ::=    ("-" | "+") Expr
		 *
		 *  XQuery provides the usual arithmetic operators for addition, subtraction,
		 *  multiplication, division, and modulus, in the usual binary and unary forms
		 *  and with the usual meanings.
		 *
		 *  When both operands of an arithmetic operator are numeric, the result is
		 *  straightforward. When one or more operands is a node, the content of the
		 *  node is extracted by an implicit call to the data function and converted
		 *  to a number before the operation is performed; if this conversion is not
		 *  possible, an error results. When one operand is a number and the other
		 *  is a sequence of numbers, the result of the operator is a sequence that
		 *  is created by applying the operator pairwise to the numeric operand and
		 *  each individual member of the sequence operand, preserving the original
		 *  order.
		 */

		XQueryExpression xLOprnd = arg.getLeftOperand();
		XQueryExpression xROprnd = arg.getRightOperand();
		int xOprtr = arg.getOperator();
		SqlType lType = _typeInterpreter.interprete(xLOprnd.getQType());
		SqlType rType = _typeInterpreter.interprete(xROprnd.getQType());

		xLOprnd.accept(this);
		Expression aLOprnd = (Expression) _returnStack.pop();

		xROprnd.accept(this);
		Expression aROprnd = (Expression) _returnStack.pop();

		/* verify operand types*/
		if (!lType.isNumeric() || !rType.isNumeric()) {
			throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_SIG", arg), null);
		}

		/* translate */
		if (lType.isAtom() && rType.isAtom()) {
			/* signature 1 Numeric_Atom <op> Numeric_Atom */
			retVal = new BinOpArithmetic(xOprtr, aLOprnd, aROprnd);
		} else {
			throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_SIG", arg), null);
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	protected void VisitColumeArithmeticOperatorAtom(Expression relation, int operator, Expression leftOperand, Expression rightOperand) {
		/* find out the underlying UnOpProject node */
		Expression cursor = relation;
		while (!(cursor instanceof UnOpProject)) {
			/** @todo  binary algebra and join is to treated */
			cursor = ((UnaryAlgebra) cursor).getOperand();
		}
		if (!(cursor instanceof UnOpProject)) {
			//Trace.exit(this, "VisitColumeArithmeticOperatorAtom");
			throw new SqlWrapperException(MessageLibrary.getMessage("INT_LOG_ERR"), null);
		}
		UnOpProject project = (UnOpProject) cursor;

		/* push the operation into the UnOpProject node */
		project.arithmeticOperation(operator, leftOperand, rightOperand);
	}

	public void visit(ListOpCompExpression arg) throws XQueryException {
		Expression retVal = null;

		int xOprtr = arg.getOperator();
		// add LARS 06/04/04 to support valuecomp
		// TODO better!!!
		if (xOprtr > 5)
			xOprtr = xOprtr - 6;

		XQueryExpression lXOprnd = arg.getLeftOperand();

		SqlType lOprndType = _typeInterpreter.interprete(lXOprnd.getQType());
		lXOprnd.accept(this);
		Expression lOprnd = (Expression) _returnStack.pop();

		XQueryExpression rXOprnd = arg.getRightOperand();

		SqlType rOprndType = _typeInterpreter.interprete(rXOprnd.getQType());
		rXOprnd.accept(this);
		Expression rOprnd = (Expression) _returnStack.pop();

		if (lOprndType.isAtom()) {
			if (rOprndType.isAtom()) {
				/* signature :  compare ( atomic  , atomic ) : boolean */
				retVal = new BinOpCompare(xOprtr, lOprnd, rOprnd);
			} else if (rOprndType.isColumn()) {
				/* signature :  compare ( atomic  , column ) : boolean */
				rOprnd = removeUselessSort(rOprnd);
				retVal = new BinOpCompareAny(xOprtr, lOprnd, rOprnd);
			} else if (rOprndType.isTuple()) {
				/* signature :  compare ( atomic, tuple ) : boolean */
				Debug.assertTrue(false, "NYI!");
			} else
				throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_SIG", arg), null);
		} else if (rOprndType.isAtom()) {
			if (lOprndType.isColumn()) {
				/* signature :  compare ( column , atomic ) : boolean */
				lOprnd = removeUselessSort(lOprnd);
				retVal = new BinOpCompareAny(inverseCompareOperator(xOprtr), rOprnd, lOprnd);
			} else if (lOprndType.isTuple()) {
				/* signature :  compare ( tuple , atomic ) : boolean */
				Debug.assertTrue(false, "NYI!");
			} else
				throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_SIG", arg), null);
		} else if (lOprndType.isColumn() && rOprndType.isColumn()) {
			/* signature :  compare ( column , column ) : boolean */
			retVal = visitColumnCompareColumn(xOprtr, lOprnd, rOprnd);
		} else
			throw new InternalErrorException(MessageLibrary.getMessage("IE_INV_SIG", arg), null);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	protected Expression visitColumnCompareColumn(int operator, Expression leftOperand, Expression rightOperand) {
		/* make out AttributeExpression for leftOperand */
		Expression column = (Expression) ((Relation) leftOperand).nameTest("*").get(0);
		String attrName = column.getName();
		AttributeExpression leftAttribute = new AttributeExpression(null, attrName);
		leftAttribute.setUnderlyingExpr(column);

		column = (Expression) ((Relation) rightOperand).nameTest("*").get(0);
		attrName = column.getName();
		AttributeExpression rightAttribute = new AttributeExpression(null, attrName);
		rightAttribute.setUnderlyingExpr(column);

		/* translation for =,!= */
		// TODO  implement more performent translation for >,>=,<,<= */
		Join join = new Join();

		RenameRelation left = new RenameRelation(leftOperand, "left");
		join.addOperand(left);
		RenameRelation right = new RenameRelation(rightOperand, "right");
		join.addOperand(right);

		leftAttribute.setTableInstance(left);
		rightAttribute.setTableInstance(right);
		BinOpCompare predicate = new BinOpCompare(operator, leftAttribute, rightAttribute);
		join.addPredicate(predicate);

		return new UnOpExists(join);
	}

	protected int inverseCompareOperator(int comparisonOperator) {
		int retVal = -1;
		switch (comparisonOperator) {
			case LT_COMPOP :
				retVal = GT_COMPOP;
				break;
			case GT_COMPOP :
				retVal = LT_COMPOP;
				break;
			case GEQ_COMPOP :
				retVal = LEQ_COMPOP;
				break;
			case LEQ_COMPOP :
				retVal = GEQ_COMPOP;
				break;
			case EQ_COMPOP :
				retVal = EQ_COMPOP;
				break;
			case NEQ_COMPOP :
				retVal = NEQ_COMPOP;
			default :

				break;
		}

		return retVal;
	}

	public void visit(ListOpEXCEPTExpression arg) throws XQueryException {
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
	}

	public void visit(ListOpINTERSECTExpression arg) throws XQueryException {
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
	}

	public void visit(ListOpUNIONExpression arg) throws XQueryException {
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
	}

	public void visit(LocatedExpression arg) throws XQueryException {
		/** treat ablolute locatedExpression
		 * ablolute locatedExpression
		 */
		List steps = arg.getSteps();

		int numSteps = steps.size();
		Step step = (Step) steps.get(numSteps - 1);
		step.setQType(arg.getQType());

		Expression prefixAlgebraExpr = null;
		/*  the algebra translation of a prefix path expression */
		QType prefixQType = null;

		/* for inner LocatedExpression , convert the context to prefixQType and prefixAlgebraExpr */
		if (arg.getContext()) {
			/* case : in predicate of LocatedExpression  or in Sort expression , there are often
			* LocatedExpressions that base on a context
			* eg.   offered_by/data()
			* Collection("SQLWrapper1:items")/items[offered_by/text() = $U/userid]
			*/
			Context context = (Context) _typeContextStack.peek();

			/* make out prefixAlgebraExpr */
			prefixQType = context.qType;
			SqlType xExprType = _typeInterpreter.interprete(prefixQType);

			/* find out underlying table instance */
			Set tableInstanceSet = ((Relation) context.aExpr).providedTableInstances();
			Object[] tableInstances = tableInstanceSet.toArray();
			Debug.assertTrue(1 == tableInstances.length, "1 == tableInstances.length");
			//            RenameRelation tableInstance =  (RenameRelation)tableInstances[0];
			RenameRelation tableInstance;
			if (tableInstances[0] instanceof RenameRelation)
				tableInstance = (RenameRelation) tableInstances[0];
			else
				tableInstance = null;

			if (xExprType.isColumn()) {
				// case 1_2 For $I In Collection("SqlWrapper1:items")/items/description [ ]
				List attributes = ((Relation) context.aExpr).nameTest("*");
				Debug.assertTrue(1 == attributes.size(), "1== attributes.size()");
				Expression attribute = (Expression) attributes.get(0);

				prefixAlgebraExpr = new AttributeExpression(tableInstance, attribute.getName());
				((AttributeExpression) prefixAlgebraExpr).setUnderlyingExpr(attribute);

			} else if (xExprType.isRelation()) {
				// case 1_1 For $I In Collection("SqlWrapper1:items")/items[]
				prefixAlgebraExpr = new AttributeExpression(tableInstance, "*");
			}
			/* make out prefixQType */
			//            prefixQType = ((QTypeCollection) prefixQType).getCollectionType();
		}

		_returnStack.push(prefixAlgebraExpr);
		Iterator iter = steps.iterator();
		int axis;

		while (iter.hasNext()) {
			step = (Step) iter.next();
			prefixAlgebraExpr = (Expression) _returnStack.pop();
			axis = step.getAxis();
			switch (axis) {
				case Axis.CHILD :
				case Axis.NONE :
					visit(step, prefixQType, prefixAlgebraExpr);
					break;
				case Axis.SELF :
					XQueryExpression stepXExpr = step.getExpression();
					if (stepXExpr instanceof NodeTest && NodeKind.NODE == ((NodeTest) stepXExpr).getKind()) {
						Expression retVal = prefixAlgebraExpr;
						retVal.setOrginalXExpr(step);
						_returnStack.push(retVal);
						break;
					}
				default :
					throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", step), null);
			}
			prefixQType = step.getQType();
		}

		((Expression) _returnStack.peek()).setOrginalXExpr(arg);
	}

	public void visit(QName arg) throws XQueryException {
		/* this fonction should only be called when a QName acts as a implicte LocatedExpression
		* eg. offered_by, name
		* (FOR $U IN Collection("SQLWrapper1:users")/users
		*    FOR $I IN Collection("SQLWrapper1:items")/items[offered_by = $U/userid]
		*        RETURN  <tag> {$U/name ,$I/itemno , $B/itemno} <tag> ) SORT BY name
		*/
		Context context = (Context) _typeContextStack.peek();
		Expression retVal = null;

		SqlType conType = null == context.qType ? null : _typeInterpreter.interprete(context.qType);
		// conType : SqlType of context
		Debug.assertTrue(null != conType, "null!=conType");

		if (conType.isRelation()) {

			/* case A, eg.  Collection("SQLWrapper1:items")/items[offered_by = "U03"] */

			/* find out underlying table instance */
			Set tableInstanceSet = ((Relation) context.aExpr).providedTableInstances();
			Object[] tableInstances = tableInstanceSet.toArray();

			Debug.assertTrue(1 == tableInstances.length, "1 == tableInstances.length");
			//            RenameRelation tableInstance =  (RenameRelation)tableInstances[0];
			RenameRelation tableInstance;
			if (tableInstances[0] instanceof RenameRelation)
				tableInstance = (RenameRelation) tableInstances[0];
			else
				tableInstance = null;

			String attrName = arg.getLocalName();
			if (attrName.equals("*"))
				throw new NotSupport("WildCard(*) in localName is not supported", null);

			/* find out underlying expression. */
			List underlyingExprs = ((Relation) context.aExpr).nameTest(attrName);
			Debug.assertTrue(1 == underlyingExprs.size(), "1==underlyingExprs.size()");
			Expression underlyingExpr = (Expression) underlyingExprs.get(0);
			AttributeExpression attrExpr = new AttributeExpression(tableInstance, underlyingExpr.getName());
			attrExpr.setUnderlyingExpr(underlyingExpr);

			retVal = attrExpr;
		} else {
			throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(QuantifiedExpression arg) throws XQueryException {
		/* create a new VariableTable for the new scope*/
		VariableTable newVT = new VariableTable();
		newVT.setDefautVariableTable(_variableTable);
		_variableTable = newVT;
		Expression retVal = null;

		/* generate algebra for FOR,LET clause */
		List variables = arg.getVariables();
		Iterator iter = variables.iterator();

		List tableInstanceList = new ArrayList();
		List bindingTypeList = new ArrayList();
		List pseudoIdPathList = new ArrayList();

		/* for the case of a artificial for clause */
		String pseudoIdPath = null;
		List predicateListAddedByForBinding = null;

		while (iter.hasNext()) {
			Variable variable = ((Variable) iter.next());
			VarMappingInfo varInfo = visitForBinding(variable);
			if (varInfo != null) {
				tableInstanceList.add(varInfo.relation);
			} else {
				visit(variable);
				Expression atomicExpr = (Expression) _returnStack.pop();
				if (!(atomicExpr instanceof Literal)) {
					if (predicateListAddedByForBinding == null)
						predicateListAddedByForBinding = new ArrayList();
					predicateListAddedByForBinding.add(new UnOpIsNull(atomicExpr, true));
				}
				//   dit que je n'aurai plus de for $v in 1 , 2002-03-07 // SR 06/08/03 c'est faux cf UCR11
				if (variable.getExpression() instanceof ValueInteger) {
					pseudoIdPath = "$" + variable.getVarName().getName();
					pseudoIdPathList.add(pseudoIdPath);
				}
			}
		}

		XQueryExpression xPredicate = arg.getConstraintExpresson();
		xPredicate.accept(this);
		Expression aPredicate = (Expression) _returnStack.pop();
		if (Constants.EVERY_QUANTIFIER == arg.getKind())
			aPredicate = new UnOpNot(aPredicate);

		if (1 == tableInstanceList.size()) {
			retVal = new UnOpRestrict((Expression) tableInstanceList.get(0), aPredicate);
			if (Constants.SOME_QUANTIFIER == arg.getKind()) {
				retVal = new UnOpExists(retVal, false);
			} else // Constants.EVERY_QUANTIFIER == arg.getKind()) 
				retVal = new UnOpExists(retVal, true);
		} else if (1 < tableInstanceList.size()) {
			ArrayList aPredicates = new ArrayList(1);
			aPredicates.add(aPredicate);
			retVal = new Join(tableInstanceList, aPredicates);

			Value val = new ValueInteger("1", null);
			val.accept(this);
			//((Variable)variables.get(0)).accept(this);
			Expression aExpr = (Expression) _returnStack.pop();
			ArrayList selectList = new ArrayList(1);
			selectList.add(aExpr);

			retVal = new UnOpProject(retVal, selectList, false, _attIDProvider);
			Mapper mapper = new Mapper();
			mapper.createMap(((UnOpProject) retVal).getItemList());
			retVal.setMapper(mapper);

			if (Constants.SOME_QUANTIFIER == arg.getKind())
				retVal = new UnOpExists(retVal, false);
			else // Constants.EVERY_QUANTIFIER == arg.getKind())
				retVal = new UnOpExists(retVal, true);

			//remove the VariableTable from VariableTable chaine
			_variableTable = _variableTable.getDefautVariableTable();
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(SortedExpression arg) throws XQueryException {
		Expression retVal = null;

		/* check type */
		XQueryExpression xOprnd = arg.getExpression();
		SqlType oprndType = _typeInterpreter.interprete(xOprnd.getQType());
		if (!oprndType.isRelation() && !oprndType.isColumn()) {
			throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
		}

		xOprnd.accept(this);
		retVal = (Expression) _returnStack.pop();

		Context context = new Context();
		context.xExpr = xOprnd;
		context.aExpr = retVal;
		context.qType = xOprnd.getQType();
		_typeContextStack.push(context);

		ArrayList aSortSpecList = new ArrayList();
		XQueryExpression xSortSpec;

		Expression aSortExpr;
		int aDirection;
		SortSpecification aSortSpec;

		Iterator iter = arg.getSortClauses().iterator();
		while (iter.hasNext()) {
			xSortSpec = (XQueryExpression) iter.next();
			xSortSpec.accept(this);

			aSortExpr = (Expression) _returnStack.pop();
			// OLIVIER - W3C WD 16 August 2002
			//aDirection = xSortSpec.getOrder();
			aDirection = xSortSpec.getOrder(0);

			aSortSpec = new SortSpecification(aSortExpr, aDirection);
			aSortSpecList.add(aSortSpec);
		}
		retVal = new UnOpSort(retVal, aSortSpecList);

		_typeContextStack.pop();

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(Step arg) throws XQueryException {
		Debug.assertTrue(false, "Function \"visit(Step arg)\" should not be called.");
	}

	protected void visit(Step arg, QType prefixQType, Expression prefixAlgebraExpr) throws XQueryException {
		Expression retVal = null;
		Expression algebraExpr;
		XQueryExpression stepXExpr = arg.getExpression();
		// the XQueryExpression contained in the Step

		SqlType pType = null == prefixQType ? null : _typeInterpreter.interprete(prefixQType);
		// pType : SqlType of prefix path
		SqlType tType = _typeInterpreter.interprete(arg.getQType());
		// tType : SqlType of this step

		if (null == prefixQType && null == prefixAlgebraExpr) {
			/* case 0 eg. Collection("SQLWrapper1:items") */
			/* case 0 eg. $I    */
			arg.getExpression().accept(this);

			retVal = (Expression) _returnStack.pop();
		} else if (prefixAlgebraExpr instanceof PseudoExpression) {
			if (!(stepXExpr instanceof QName)) {
				throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg), null);
			}

			String tableName = ((QName) stepXExpr).getLocalName();
			CollectionName colName = ((PseudoExpression) prefixAlgebraExpr).getCollectionName();
			colName.setTable(tableName);

			List tableList = _mdm.findTables(colName);

			org.xquark.extractor.metadata.Table mTable = null;

			if (null != tableList && 1 == tableList.size()) {
				mTable = (org.xquark.extractor.metadata.Table) tableList.get(0);
			} else if (null == tableList || 1 > tableList.size()) {
				throw new SqlWrapperException(MessageLibrary.getMessage("IE_TAB_NOT_FND", arg));
			} else { /* tableList.size() > 1 */
				throw new SqlWrapperException(MessageLibrary.getMessage("E_AMBG", arg));
			}
			retVal = new Table(mTable);
			retVal = new RenameRelation(retVal, _relIDProvider);
		} else if (pType.isRelation() && tType.isRelation()) {
			Debug.assertTrue(stepXExpr instanceof QName, "stepXExpr instanceof QName");

			String localName = ((QName) stepXExpr).getLocalName();
			if (localName.equals("*")) {
				throw new NotSupport("WildCard(*) in localName is not supported", null);
			}

			int prefixTypeItemNumber = ((SqlTypeStructure) pType).getAttributes().size();
			int thisTypeItemNumber = ((SqlTypeStructure) tType).getAttributes().size();
			if (prefixTypeItemNumber == thisTypeItemNumber) {
				/* case 1_a eg. Collection("SQLWrapper1:items")/items[iteno>"1002"][end_date>"01/05/2000"]  */
				/* there is nothing to do for the NameTest (items)  */
				retVal = prefixAlgebraExpr;
			} else {
				/* case 1_b eg.
				 *   Let $J = (  FOR $U IN Collection("SQLWrapper1:users")/users
				 *               for $I IN Collection("SQLWrapper1:items")/items[offered_by = $U/userid]
				 *                FOR $B IN Collection("SQLWrapper1:bids") /bids
				 *                RETURN  <tag> {$U/name ,$I/itemno , $B/itemno} <tag>
				 *            )
				 *   FOR $R IN $J/itemno[...]                 // !!: $J/itemno
				 *   RETURN ( $R )
				 * In "$J/itemno" , the type of $J and $J/itemno are both relation ,
				 * but we can distinguish this case with case 1_a by the the number
				 * of attributes in the relation.
				 */
				Debug.assertTrue(stepXExpr instanceof QName, "stepXExpr instanceof QName");

				String name = ((QName) stepXExpr).getLocalName();
				if (name.equals("*")) {
					throw new NotSupport("WildCard(*) in localName is not supported", null);
				}

				List attrList = ((Relation) prefixAlgebraExpr).nameTest(name);
				Debug.assertTrue(attrList.size() > 1, "attrList.size()>1");

				List selectList = new ArrayList();
				Iterator iter = attrList.iterator();
				AttributeExpression attr;
				while (iter.hasNext()) {
					Object item = iter.next();
					attr = new AttributeExpression(new RenameRelation(prefixAlgebraExpr, _relIDProvider), ((Expression) item).getName());
					selectList.add(attr);
				}
				retVal = new UnOpProject(prefixAlgebraExpr, selectList, false, _attIDProvider);
			}
		} else if (pType.isRelation() && tType.isColumn()) {
			// case 2 eg. Collection("SQLWrapper1:items")/items[iteno>"1002"][end_date>"01/05/2000"]/description

			Debug.assertTrue(stepXExpr instanceof QName, "stepXExpr instanceof QName");
			String attrName = ((QName) stepXExpr).getLocalName();
			if (attrName.equals("*")) {
				throw new NotSupport("WildCard(*) in localName is not supported", null);
			}

			if (!(prefixAlgebraExpr instanceof UnOpProject)) {

				/*find out underlying table instance */
				Set tableInstanceSet = ((Relation) prefixAlgebraExpr).providedTableInstances();
				Object[] tableInstances = tableInstanceSet.toArray();
				Debug.assertTrue(1 == tableInstances.length, "1 == tableInstances.length");
				//                RenameRelation tableInstance =  (RenameRelation)tableInstances[0];
				RenameRelation tableInstance;
				if (tableInstances[0] instanceof RenameRelation) {
					tableInstance = (RenameRelation) tableInstances[0];
				} else {
					tableInstance = null;
				}

				/* create AttributeExpression*/
				List underlyingExprs = ((Relation) prefixAlgebraExpr).nameTest(attrName);
				Debug.assertTrue(1 == underlyingExprs.size(), "1==underlyingExprs.size()");
				Expression underlyingExpr = (Expression) underlyingExprs.get(0);
				AttributeExpression attrExpr = new AttributeExpression(tableInstance, underlyingExpr.getName());
				attrExpr.setUnderlyingExpr(underlyingExpr);

				retVal = new UnOpProject(prefixAlgebraExpr, _attIDProvider);
				((UnOpProject) retVal).addItem(attrExpr);
			} else {
				/* find out the selected expression. */
				List itemList = ((Relation) prefixAlgebraExpr).nameTest(attrName);
				Debug.assertTrue(1 == itemList.size(), "1==itemList.size()");
				Expression item = (Expression) itemList.get(0);
				/* replace old UnOpPreject*/
				retVal = new UnOpProject(((UnOpProject) prefixAlgebraExpr).getOperand(), item, _attIDProvider);
			}
		} else if (pType.isColumn()) {
			// case 3 eg. Collection("SQLWrapper1:items")/items/description/text()

			Debug.assertTrue((stepXExpr instanceof NodeTest), "(stepXExpr instanceof NodeType)");
			NodeTest nodeType = (NodeTest) stepXExpr;
			Debug.assertTrue((nodeType.getKind() == NodeKind.TEXT), "nodeType.getKind() == NodeKind.TEXT");

			if (prefixAlgebraExpr instanceof UnOpProject) {
				UnOpProject project = (UnOpProject) prefixAlgebraExpr;
				Expression item = (Expression) project.getItemList().get(0);
				DbType string = new DbType(java.sql.Types.VARCHAR);
				item = new XfCast(item, string);
				project.setItem(0, item);
			} else {
				Debug.assertTrue(prefixAlgebraExpr instanceof Table, "prefixAlgebraExpr instanceof Table");
				Debug.nyi("table with only one column");
			}
			retVal = prefixAlgebraExpr;
		} else if (pType.isTuple()) {
			/* case A, eg. $I/itemno ;$I/description
			 *   FOR $I IN Collection("SQLWrapper1:items")/items[offered_by = $U/userid]
			 *   RETURN  <tag> {$I/itemno , $I/description } <tag>
			 */

			/* find out tableInstance*/
			RenameRelation ti = null;

			if (prefixAlgebraExpr instanceof TupleExpression) {
				ti = ((TupleExpression) prefixAlgebraExpr).getTableInstance();
			} else if (prefixAlgebraExpr instanceof AttributeExpression) {
				ti = ((AttributeExpression) prefixAlgebraExpr).getTableInstance();
			} else {
				throw new SqlWrapperException("internal logic error", null);
			}

			/* find items (in the tuple) whoes name is QName */
			Debug.assertTrue(stepXExpr instanceof QName, "stepXExpr instanceof QName");
			QName qName = (QName) stepXExpr;
			String name = qName.getLocalName();
			if (name.equals("*")) {
				throw new NotSupport("WildCard(*) in localName is not supported", null);
			}

			/* find out underlying expression */
			List attributes = ti.nameTest(name);
			Debug.assertTrue(1 <= attributes.size(), "1<=attributes.size()");
			/** @todo the case where there are multiplue expressions in "attributes" */
			Expression underlyingExpr = (Expression) attributes.get(0);
			AttributeExpression attrExpr = new AttributeExpression(ti, underlyingExpr.getName());
			attrExpr.setUnderlyingExpr(underlyingExpr);
			retVal = attrExpr;
		} else if (pType.isAtom()) {

			// this branch will be called by relative LocatedExpression
			// (predicate or sort sepcification)
			if (stepXExpr instanceof NodeTest) {
				int kind = ((NodeTest) stepXExpr).getKind();
				if (kind == NodeKind.TEXT) {
					DbType dbString = new DbType(java.sql.Types.VARCHAR);
					retVal = new XfCast(prefixAlgebraExpr, dbString);
				} else {
					throw new SqlWrapperException("Invalid node type in Located Expression :" + arg);
				}
			} else {
				throw new SqlWrapperException("Invalid step in Located Expression :" + arg);
			}
		} else {
			throw new SqlWrapperException(MessageLibrary.getMessage("G_INT_LOG_ERR", "Type Analysis Error"), null);
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);

		// There must not be predicates left in the located expression
		// NOTE: the code processing it, is available in CVS history
		Debug.assertTrue(arg.getPredicates() == null, "XQuery frontend should not produce predicates in located expression.");
	}

	public void visit(UnOpMinusExpression arg) throws XQueryException {
		Expression retVal = null;

		XQueryExpression xExpr = arg.getExpression();
		xExpr.accept(this);
		Expression aOprnd = (Expression) _returnStack.pop();

		boolean minus = arg.getMinus();

		if (minus) {
			retVal = new UnOpMinus(aOprnd);
		} else {
			retVal = aOprnd;
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueBoolean arg) throws XQueryException {
		LitBoolean retVal = new LitBoolean(arg.getBooleanValue());
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueDecimal arg) throws XQueryException {
		LitDecimal retVal = new LitDecimal(arg.getDecimalValue());
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueDouble arg) throws XQueryException {
		LitDouble retVal = new LitDouble(arg.getDoubleValue());
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueFloat arg) throws XQueryException {
		LitFloat retVal = new LitFloat(arg.getFloatValue());
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueInteger arg) throws XQueryException {
		LitInteger retVal = new LitInteger(arg.getIntValue());
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueString arg) throws XQueryException {
		LitString retVal = new LitString(arg.getStringValue());
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(ValueText arg) throws XQueryException {
		throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
	}

	public void visit(Variable arg) throws XQueryException {
		Expression translation = (Expression) _variableTable.get(arg.getName());
		if (translation != null) {
			try {
				_returnStack.push(translation.clone());
			} catch (CloneNotSupportedException e) {
				throw new XQueryException(e.getMessage());
			}
		} else
			throw new SqlWrapperException("Variable " + arg.toString() + " is not defined.");
	}

	public void visit(org.xquark.xquery.parser.ExternalVariable arg) throws XQueryException {
		QName varQName = ((Variable) arg.getObjectValue()).getVarName();
		_returnStack.push(new ExternalVariable(varQName.getNameSpace(), varQName.getLocalName(), arg.getQType().getSimpleType()));
	}

	protected void visitLetBinding(Variable arg) throws XQueryException {
		XQueryExpression xExpr = arg.getExpression();
		xExpr.accept(this);

		Expression retVal = (Expression) _returnStack.pop();
		retVal.setOrginalXExpr(arg);
		_variableTable.put(arg.getName(), retVal);
	};

	protected VarMappingInfo visitForBinding(Variable arg) throws XQueryException {
		XQueryExpression varXExpr = arg.getExpression();
		SqlType xExprType = _typeInterpreter.interprete(varXExpr.getQType());
		VarMappingInfo retVal = null;

		if ((varXExpr instanceof InternalFunctionCall) && ((InternalFunctionCall) varXExpr).getFuncName().getLocalName().equalsIgnoreCase("SOU")) {
			visitLetBinding(arg);
		} else if (null == xExprType) {
			throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_EXPR", varXExpr.toString()));
		} else if (xExprType.isColumn() || xExprType.isRelation()) {
			// case 1_1 For $I In Collection("SqlWrapper1:items")/items
			// case 1_2 For $I In Collection("SqlWrapper1:items")/items/description
			Expression tmpExpr = null;

			varXExpr.accept(this);
			Expression relation = (Expression) _returnStack.pop();

			String varName = arg.getName();
			if (relation instanceof RenameRelation)
				retVal = new VarMappingInfo(varName, (RenameRelation) relation);
			else
				retVal = new VarMappingInfo(varName, new RenameRelation(relation, _relIDProvider));

			Expression varTrans = null;
			if (xExprType.isColumn()) {
				// case 1_2 For $I In Collection("SqlWrapper1:items")/items/description
				List attributes = ((Relation) relation).nameTest("*");
				Debug.assertTrue(1 == attributes.size(), "1== attributes.size()");

				Expression attribute = (Expression) attributes.get(0);
				varTrans = new AttributeExpression(retVal.relation, attribute.getName());
				((AttributeExpression) varTrans).setUnderlyingExpr(attribute);
				varTrans.setOrginalXExpr(arg);
			} else if (xExprType.isRelation()) {
				varTrans = createTupleExpression(arg, retVal.relation);
			}
			_variableTable.put(varName, varTrans);
		} else
			visitLetBinding(arg);

		return retVal;
	}

	protected TupleExpression createTupleExpression(Variable arg, RenameRelation relation) throws XQueryException {
		// case 1_1 For $I In Collection("SqlWrapper1:items")/items

		SqlTypeStructure stVar = (SqlTypeStructure) _typeInterpreter.interprete(arg.getQType());

		List elementNameList = ((SqlTypeStructure) stVar).getNames();
		boolean varBindedOnSeq = false;
		if (null == elementNameList.get(0)) {
			/* Variable arg is binded to a squence */
			relation.accept(_typeVistor);
			elementNameList = ((SqlTypeStructure) relation.getType()).getNames();
			varBindedOnSeq = true;
		}

		String elementName;

		List attrExprs = new ArrayList(); // Attributes in the tuple (varTrans)
		Expression attrExpr;

		List attrExprsFromNameTest = null; //
		Expression exprFromNameTest = null;

		Iterator iter = elementNameList.iterator();

		XQueryExpression xVirtualExpr = null;
		while (iter.hasNext()) {
			elementName = (String) iter.next();

			attrExprsFromNameTest = ((Relation) relation).nameTest(elementName);
			Debug.assertTrue(1 == attrExprsFromNameTest.size(), "1== attributes.size()");
			exprFromNameTest = (Expression) attrExprsFromNameTest.get(0);

			attrExpr = new AttributeExpression(relation, exprFromNameTest.getName());
			((AttributeExpression) attrExpr).setUnderlyingExpr(exprFromNameTest);

			if (!varBindedOnSeq) {
				xVirtualExpr = makeVirtualLocatedeExpression(arg, elementName);
			}

			attrExpr.setOrginalXExpr(xVirtualExpr);
			attrExprs.add(attrExpr);
		}

		return new TupleExpression(relation, attrExprs, arg);
	}

	private LocatedExpression makeVirtualLocatedeExpression(Variable variable, String elementName) {
		try {
			/* create a located expression */
			ArrayList steps = new ArrayList();
			Step step = null;

			step = new Step(false, Axis.NONE, variable, null, null);
			steps.add(step);

			step = new Step(true, Axis.CHILD, new QName(elementName, null), null, null);
			steps.add(step);

			LocatedExpression retVal = new LocatedExpression(steps, null);

			/* find out the QType of this retVal */
			QType varQtype = variable.getQType();
			if (varQtype instanceof QTypeElement) {
				/* find out the sub element whoes name == elementName */
				ComplexType varType = (ComplexType) varQtype.getType();
				QName varQName = (QName) varQtype.getName();
				ElementDeclaration elementDeclaration = varType.getElementDeclaration(varQName.getNameSpace(), elementName);
				if (null == elementDeclaration) {
					elementDeclaration = varType.getElementDeclaration(null, elementName);
				}

				/* create a QTypeElement */
				QTypeElement qType = null;
				if (null != elementDeclaration) {

					qType = new QTypeElement(new QName(elementDeclaration.getNamespace(), null, elementName, null, null), elementDeclaration, elementDeclaration.getType());
				}

				Debug.assertTrue(null != qType, "null != qType");
				retVal.setQType(qType);
			} else if (varQtype instanceof QTypeSequence) {
				List qteList = ((QTypeSequence) varQtype).getList();
				QTypeElement qType = null;
				for (int i = 0; i < qteList.size(); i++) {
					QTypeElement item = (QTypeElement) qteList.get(i);
					if (((QName) item.getName()).getLocalName().equals(elementName)) {
						qType = item;
						break;
					}
				}
				Debug.assertTrue(null != qType, "null != qType");
				retVal.setQType(qType);
			}

			return retVal;
		} catch (XQueryException ex) {
			throw new SqlWrapperException(MessageLibrary.getMessage("G_INT_LOG_ERR", "make virtual located expression failed"), ex);
		}
	}

	public void visit(XQueryExpression arg) throws XQueryException {
		throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_EXPR", arg.toString() + " : " + arg.getClass()), null);
	}

	public void visit(XQueryExpressionSequence arg) throws XQueryException {
		Expression retVal = null;
		List xExprs = arg.getSubExpressions();

		/* create structure for the virtual table */
		QType qtElementOfCollection = arg.getQType();

		// CHANGE LARS 03/05/04 delete obsolete code
		if (qtElementOfCollection instanceof QTypeAtom) {
			retVal = createLitList(arg);
		} else {
			throw new XQueryException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg), null);
		}
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	// CHANGE LARS 03/05/04 delete obsolete code

	protected LitList createLitList(XQueryExpressionSequence arg) throws XQueryException {
		QTypeAtom qta = (QTypeAtom) arg.getQType();

		/* add values to retVal */
		XQueryExpression xExpr = null;
		Expression aExpr = null;
		List xExprList = arg.getSubExpressions();
		List aExprList = new ArrayList(xExprList.size());

		for (int i = 0; i < xExprList.size(); i++) {
			xExpr = (Value) xExprList.get(i);
			xExpr.accept(this);
			aExpr = (Expression) _returnStack.pop();
			aExprList.add(aExpr);
		}

		return new LitList(aExprList);
	}

	public void visit(FunctionABS arg) throws XQueryException {
		Expression retVal = null;
		arg.getArgument(0).accept(this);
		retVal = (Expression) _returnStack.pop();
		retVal = new XfAbs(retVal);
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionCEILING arg) throws XQueryException {
		Expression retVal = null;
		arg.getArgument(0).accept(this);
		retVal = (Expression) _returnStack.pop();
		retVal = new XfCeiling(retVal);
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionFLOOR arg) throws XQueryException {
		Expression retVal = null;
		arg.getArgument(0).accept(this);
		retVal = (Expression) _returnStack.pop();
		retVal = new XfFloor(retVal);
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionROUND arg) throws XQueryException {
		Expression retVal = null;
		arg.getArgument(0).accept(this);
		retVal = (Expression) _returnStack.pop();
		retVal = new XfRound(retVal);
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionAVG arg) throws XQueryException {
		_returnStack.push(visitAggregationFunction(arg, AVG));
	}

	public void visit(FunctionCOLLECTION arg) throws XQueryException {
		Expression retVal = null;
		CollectionName colName = new CollectionName(arg.getSourceName(), arg.getCollectionName());

		if (null != colName.getTable() && !colName.tableIsWildCard()) {
			/* table name is designated, we should be able to find one and only one table in meta-data */

			/* case : collection("Extractor:use-case.items") or collection("Extractor:items")
			 * or collection("Extractor:use-case.items") */
			List tableList = _mdm.findTables(colName);
			org.xquark.extractor.metadata.Table mTable = null;

			if (null != tableList && 1 == tableList.size()) {
				mTable = (org.xquark.extractor.metadata.Table) tableList.get(0);
			} else if (null == tableList || 1 > tableList.size()) {
				throw new SqlWrapperException(MessageLibrary.getMessage("IE_TAB_NOT_FND", arg));
			} else { /* tableList.size() > 1 */
				throw new SqlWrapperException(MessageLibrary.getMessage("E_AMBG", arg));
			}
			retVal = new Table(mTable);
			retVal = new RenameRelation(retVal, _relIDProvider);
		} else {
			PseudoExpression ps = new PseudoExpression();
			ps.setCollectionName(colName);
			retVal = ps;
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionCONCAT arg) throws XQueryException {
		Expression retVal = null;

		List xArgumentList = arg.getArguments();
		int argNumber = xArgumentList.size();

		XQueryExpression xArgument = null;
		Expression aArgument = null;

		if (0 == argNumber) {
			retVal = new LitString("");
		} else if (1 == argNumber) {
			xArgument = (XQueryExpression) xArgumentList.get(0);
			xArgument.accept(this);
			retVal = (Expression) _returnStack.pop();
		} else {
			retVal = new XfConcat();
			for (int i = 0; i < xArgumentList.size(); i++) {
				xArgument = (XQueryExpression) xArgumentList.get(i);
				xArgument.accept(this);
				aArgument = (Expression) _returnStack.pop();
				((XfConcat) retVal).setArgument(i, aArgument);
			}
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionCONTAINS arg) throws XQueryException {
		arg.getArgument(0).accept(this);
		Expression arg0 = (Expression) _returnStack.pop();

		arg.getArgument(1).accept(this);
		Expression arg1 = (Expression) _returnStack.pop();

		XfContains retVal = new XfContains();
		retVal.setArgument(0, arg0);
		retVal.setArgument(1, arg1);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionCURRENT_DATETIME arg) throws XQueryException {
		XfCurrentDateTime retVal = new XfCurrentDateTime();
		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionCOUNT arg) throws XQueryException {
		_returnStack.push(visitAggregationFunction(arg, COUNT));
	}

	public void visit(FunctionDATA arg) throws XQueryException {
		Expression retVal = null;

		XQueryExpression xArg = arg.getArgument(0);
		xArg.accept(this);
		retVal = (Expression) _returnStack.pop();
		retVal.setOrginalXExpr(arg);

		_returnStack.push(retVal);
	}

	public void visit(FunctionEXISTS arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;
		XQueryExpression xArg = (XQueryExpression) arg.getArgument(0);

		xArg.accept(this);
		Expression aArg = (Expression) _returnStack.pop();

		aArg = removeUselessSort(aArg);

		QType xArgQtype = xArg.getQType();
		SqlType xArgSqlType = _typeInterpreter.interprete(xArgQtype);

		if (xArgSqlType.isRelation() || xArgSqlType.isColumn()) {
			retVal = new UnOpExists(aArg, false); /* exists (aArg)*/
		} else if (xArgSqlType.isAtom()) {
			retVal = new UnOpIsNull(aArg, true); /* aArg IS NOT NULL */
		} else if (xArgSqlType.isTuple()) {
			TupleExpression tuple = (TupleExpression) aArg;
			List itemList = tuple.getItemList();
			Expression item = null;
			UnOpIsNull isNull = null;

			for (int i = 0; i < itemList.size(); i++) {
				item = (Expression) itemList.get(i);
				isNull = new UnOpIsNull(item, true);
				if (null != retVal) {
					retVal = new BinOpBoolean(OR, retVal, isNull);
				} else {
					retVal = isNull;
				}
			}
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionEMPTY arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;
		XQueryExpression xArg = (XQueryExpression) arg.getArgument(0);

		xArg.accept(this);
		Expression aArg = (Expression) _returnStack.pop();

		aArg = removeUselessSort(aArg);

		QType xArgQtype = xArg.getQType();
		SqlType xArgSqlType = _typeInterpreter.interprete(xArgQtype);

		if (xArgSqlType.isRelation() || xArgSqlType.isColumn()) {
			retVal = new UnOpExists(aArg, true); /* not exists (aArg)*/
		} else if (xArgSqlType.isAtom()) {
			retVal = new UnOpIsNull(aArg, false); /* aArg IS NULL */
		} else if (xArgSqlType.isTuple()) {
			TupleExpression tuple = (TupleExpression) aArg;
			List itemList = tuple.getItemList();
			Expression item = null;
			UnOpIsNull isNull = null;

			for (int i = 0; i < itemList.size(); i++) {
				item = (Expression) itemList.get(i);
				isNull = new UnOpIsNull(item);
				if (null != retVal) {
					retVal = new BinOpBoolean(AND, retVal, isNull);
				} else {
					retVal = isNull;
				}
			}
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionDATE arg) throws XQueryException {
		XfDate retVal = new XfDate();

		XQueryExpression xDate = (XQueryExpression) arg.getArgument(0);
		xDate.accept(this);
		Expression aArg = (Expression) _returnStack.pop();
		retVal.setArgument(0, aArg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionDATETIME arg) throws XQueryException {
		XfDateTime retVal = new XfDateTime();

		XQueryExpression xDate = (XQueryExpression) arg.getArgument(0);
		xDate.accept(this);
		Expression aArg = (Expression) _returnStack.pop();
		retVal.setArgument(0, aArg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionDISTINCT_VALUES arg) throws XQueryException {
		Expression retVal = null;

		/* check type */
		XQueryExpression xOprnd = arg.getArgument(0);
		SqlType oprndType = _typeInterpreter.interprete(xOprnd.getQType());

		xOprnd.accept(this);
		retVal = (Expression) _returnStack.pop();

		if (oprndType.isRelation() || oprndType.isColumn()) {

			if (retVal instanceof UnOpProject) {
				((UnOpProject) retVal).setDistinct(true);
			} else if (retVal instanceof Table || retVal instanceof RenameRelation || retVal instanceof UnOpRestrict) {
				Set tis = ((Relation) retVal).providedTableInstances();
				Debug.assertTrue(1 == tis.size(), "1 ==tis.size())");
				RenameRelation ti = (RenameRelation) tis.iterator().next();
				List attrExprs = new ArrayList();
				// Attributes in the tuple (varTrans)
				Expression attrExpr;

				List attrExprsFromNameTest = ((Relation) retVal).nameTest("*");
				//
				Expression exprFromNameTest = null;
				String name = null;

				for (int i = 0; i < attrExprsFromNameTest.size(); i++) {
					exprFromNameTest = (Expression) attrExprsFromNameTest.get(i);
					name = exprFromNameTest.getName();
					attrExpr = new AttributeExpression(ti, name);
					((AttributeExpression) attrExpr).setUnderlyingExpr(exprFromNameTest);

					attrExprs.add(attrExpr);
				}
				retVal = new UnOpProject(retVal, _attIDProvider);
				((UnOpProject) retVal).addItemList(attrExprs);
				((UnOpProject) retVal).setDistinct(true);
			} else if (retVal instanceof UnOpSort) {
				retVal = ((UnOpSort) retVal).getOperand();
				((UnOpProject) retVal).setDistinct(true);
			} else {
				// TODO:  implementation for distinct_values */
				Debug.nyi("cas imprevu");
			}
		} else if (oprndType.isAtom()) {
		} else if (oprndType.isTuple() && (xOprnd.getQType() instanceof QTypeElement)) {
		} else {
			throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_EXPR", arg + " : " + arg.getClass()), null);
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionENDS_WITH arg) throws XQueryException {
		arg.getArgument(0).accept(this);
		Expression arg0 = (Expression) _returnStack.pop();

		arg.getArgument(1).accept(this);
		Expression arg1 = (Expression) _returnStack.pop();

		XfEndsWith retVal = new XfEndsWith();
		retVal.setArgument(0, arg0);
		retVal.setArgument(1, arg1);

		_returnStack.push(retVal);
	}

	public void visit(FunctionLOWER_CASE arg) throws XQueryException {
		arg.getArgument(0).accept(this);
		Expression arg0 = (Expression) _returnStack.pop();

		XfLowerCase retVal = new XfLowerCase();
		retVal.setArgument(0, arg0);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionMAX arg) throws XQueryException {
		_returnStack.push(visitAggregationFunction(arg, MAX));
	}

	protected Expression visitAggregationFunction(PrimitiveFunctionCall arg, int oprtr) throws XQueryException {
		Expression retVal = null;

		XQueryExpression xOprd = arg.getArgument(0);
		SqlType argType = _typeInterpreter.interprete(xOprd.getQType());

		xOprd.accept(this);
		Expression relation = (Expression) _returnStack.pop();

		/* NOTE:
		 * Removes Sorts in aggregates we don't know how to process at now.
		 * Sorts might be necessary when XPath expressions like $x[1 to N] will be
		 * supported. Indeed, ordering may change result of functions like SUM.
		 */
		relation = removeUselessSort(relation);

		if (argType.isColumn() || argType.isRelation()) {

			/* create UnOpAggregate */
			AttributeExpression expr;
			Expression tmpExpr = null;
			if (argType.isColumn()) {
				tmpExpr = (Expression) ((Relation) relation).nameTest("*").get(0);
				expr = new AttributeExpression(null, tmpExpr.getName());
				expr.setUnderlyingExpr(tmpExpr);
			} else { /* argType.isRelation(). This is only possible for count(*) */
				Object[] vti = ((Relation) relation).providedTableInstances().toArray();
				Debug.assertTrue(vti.length == 1, "vti.length == 1");
				if (vti[0] instanceof RenameRelation) {
					expr = new AttributeExpression((RenameRelation) vti[0], "*");
				} else {
					relation = new RenameRelation(relation, _relIDProvider);
					expr = new AttributeExpression((RenameRelation) relation, "*");
				}
			}

			FunAggregate aggregationFunction = new FunAggregate(oprtr, expr, false);
			aggregationFunction.setOrginalXExpr(arg);
			retVal = new UnOpAggregate(_attIDProvider);

			((UnOpAggregate) retVal).setOperand(relation);
			((UnOpAggregate) retVal).insertItemWithoutRename(0, aggregationFunction);

			Mapper mapper = new Mapper();
			mapper.createMap(((UnOpAggregate) retVal).getItemList());
			retVal.setMapper(mapper);

			retVal.setOrginalXExpr(arg);
		} else if (argType.isAtom()) {
			switch (oprtr) {
				case AVG :
				case MAX :
				case MIN :
				case SUM :
					retVal = relation;
					retVal.setOrginalXExpr(arg);
					break;
				case COUNT :
					retVal = new UnOpIsNull(relation);
					retVal = new IfThenElse(retVal, new LitInteger(0), new LitInteger(1));
					break;
				default :
					Debug.assertTrue(false, "invalid aggregation function ( " + arg + ").");
			}
		} else if (argType.isTuple()) {
			/** @todo  */
			throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_SIG", arg), null);
		} else {
			throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_SIG", arg), null);
		}

		retVal.setOrginalXExpr(arg);
		return retVal;
	}

	/**
	 * Removes Sorts in aggregates or set functions.
	 * @param relation
	 * @return
	 */
	private Expression removeUselessSort(Expression relation) {
		if (relation instanceof UnOpSort) {
			XQueryExpression xqExpr = relation.getOrginalXExpr();
			relation = ((UnOpSort) relation).getOperand();
			relation.setOrginalXExpr(xqExpr);
		}
		return relation;
	}

	public void visit(FunctionMIN arg) throws XQueryException {
		_returnStack.push(visitAggregationFunction(arg, MIN));
	}

	public void visit(FunctionNOT arg) throws XQueryException {
		XQueryExpression xOprnd;
		Expression aOprnd;
		SqlType type;

		// check the signature
		xOprnd = arg.getArgument(0);
		type = _typeInterpreter.interprete(xOprnd.getQType());
		if (!type.isBoolean()) {
			throw new NotSupport(MessageLibrary.getMessage("T_N_SUP_SIG", "NOT"), null);
		}

		xOprnd.accept(this);
		aOprnd = (Expression) _returnStack.pop();

		UnOpNot retVal = new UnOpNot(aOprnd);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionNUMBER arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;
		XQueryExpression xArg = arg.getArgument(0);

		SqlTypeAtom sqlType = (SqlTypeAtom) _typeInterpreter.interprete(xArg.getQType());
		xArg.accept(this);
		retVal = (Expression) _returnStack.pop();

		if (sqlType.isString()) {
			retVal = new XfCast(retVal, new DbType(java.sql.Types.DOUBLE));
		} else if (sqlType.isNumeric()) {
			/* return retVal */
		} else {
			throw new InternalErrorException("IE_INV_XEXPR" + arg);
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionDECIMAL arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;
		XQueryExpression xArg = arg.getArgument(0);

		SqlTypeAtom sqlType = (SqlTypeAtom) _typeInterpreter.interprete(xArg.getQType());
		xArg.accept(this);
		retVal = (Expression) _returnStack.pop();

		if (sqlType.isString()) {
			retVal = new XfCast(retVal, new DbType(java.sql.Types.DECIMAL));
		} else if (!sqlType.isNumeric())
			throw new InternalErrorException("IE_INV_XEXPR" + arg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionDOUBLE arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;

		XQueryExpression xArg = arg.getArgument(0);

		SqlTypeAtom sqlType = (SqlTypeAtom) _typeInterpreter.interprete(xArg.getQType());
		xArg.accept(this);
		retVal = (Expression) _returnStack.pop();

		if (sqlType.isString()) {
			retVal = new XfCast(retVal, new DbType(java.sql.Types.DOUBLE));
		} else if (!sqlType.isNumeric())
			throw new InternalErrorException("IE_INV_XEXPR" + arg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionINTEGER arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;
		XQueryExpression xArg = arg.getArgument(0);

		SqlTypeAtom sqlType = (SqlTypeAtom) _typeInterpreter.interprete(xArg.getQType());
		xArg.accept(this);
		retVal = (Expression) _returnStack.pop();

		if (sqlType.isString()) {
			retVal = new XfCast(retVal, new DbType(java.sql.Types.INTEGER));
		} else if (!sqlType.isNumeric())
			throw new InternalErrorException("IE_INV_XEXPR" + arg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionFLOAT arg) throws XQueryException {
		Debug.assertTrue(1 == arg.getArguments().size(), "1 == arg.getArguments().size()");

		Expression retVal = null;
		XQueryExpression xArg = arg.getArgument(0);

		SqlTypeAtom sqlType = (SqlTypeAtom) _typeInterpreter.interprete(xArg.getQType());
		xArg.accept(this);
		retVal = (Expression) _returnStack.pop();

		if (sqlType.isString()) {
			retVal = new XfCast(retVal, new DbType(java.sql.Types.FLOAT));
		} else if (!sqlType.isNumeric())
			throw new InternalErrorException("IE_INV_XEXPR" + arg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionSTARTS_WITH arg) throws XQueryException {
		arg.getArgument(0).accept(this);
		Expression arg0 = (Expression) _returnStack.pop();

		arg.getArgument(1).accept(this);
		Expression arg1 = (Expression) _returnStack.pop();

		XfStartsWith retVal = new XfStartsWith();
		retVal.setArgument(0, arg0);
		retVal.setArgument(1, arg1);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionSTRING arg) throws XQueryException {
		Expression retVal = null;

		arg.getArgument(0).accept(this);
		retVal = (Expression) _returnStack.pop();
		retVal = new XfCast(retVal, new DbType(Types.VARCHAR));
		retVal.setOrginalXExpr(arg);

		_returnStack.push(retVal);
	}

	public void visit(FunctionSTRING_LENGTH arg) throws XQueryException {
		arg.getArgument(0).accept(this);
		Expression arg0 = (Expression) _returnStack.pop();

		XfStringLength retVal = new XfStringLength();
		retVal.setArgument(0, arg0);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionSUBSTRING arg) throws XQueryException {
		XfSubString retVal = new XfSubString();
		arg.getArgument(0).accept(this);
		Expression aArg = (Expression) _returnStack.pop();
		retVal.setArgument(0, aArg);

		arg.getArgument(1).accept(this);
		aArg = (Expression) _returnStack.pop();
		retVal.setArgument(1, aArg);

		if (2 < arg.getArguments().size()) {
			arg.getArgument(2).accept(this);
			aArg = (Expression) _returnStack.pop();
			retVal.setArgument(2, aArg);
		}

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionUPPER_CASE arg) throws XQueryException {
		arg.getArgument(0).accept(this);
		Expression arg0 = (Expression) _returnStack.pop();

		XfUpperCase retVal = new XfUpperCase();
		retVal.setArgument(0, arg0);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionSUM arg) throws XQueryException {
		_returnStack.push(visitAggregationFunction(arg, SUM));
	}

	public void visit(FunctionTIME arg) throws XQueryException {
		XfTime retVal = new XfTime();

		XQueryExpression xTime = (XQueryExpression) arg.getArgument(0);
		xTime.accept(this);
		Expression aArg = (Expression) _returnStack.pop();
		retVal.setArgument(0, aArg);

		retVal.setOrginalXExpr(arg);
		_returnStack.push(retVal);
	}

	public void visit(FunctionTRUE arg) throws XQueryException {
		_returnStack.push(new LitBoolean(true));
	}

	public void visit(FunctionFALSE arg) throws XQueryException {
		_returnStack.push(new LitBoolean(false));
	}

	protected List decomposePredicate(XQueryExpression predicate, List predicateList) {
		if (null == predicateList)
			predicateList = new ArrayList();

		XQueryExpression xP1 = predicate;
		BinOpANDExpression and;

		if (xP1 instanceof BinOpANDExpression) {
			and = (BinOpANDExpression) xP1;
			decomposePredicate(and.getLeftOperand(), predicateList);
			decomposePredicate(and.getRightOperand(), predicateList);
		} else
			predicateList.add(predicate);

		return predicateList;
	}

	protected class Context {
		private static final String RCSRevision = "$Revision: 1.52 $";
		private static final String RCSName = "$Name:  $";

		public XQueryExpression xExpr = null;
		public QType qType = null;
		public Expression aExpr = null;
		Context() {
		}
	}

	private class VarMappingInfo {
		public String varName = null;
		public RenameRelation relation = null;
		public VarMappingInfo(String varName, RenameRelation relation) {
			this.varName = varName;
			this.relation = relation;
		}
	}
}
