/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.algebra;

import java.util.List;
import java.util.Map;

import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.sql.SqlExpression;

/**
 * Used in the the following kind of queries:
 * 
 * for $i in collection("TABLE")/TABLE[count(COL2)=1]
 * return
 *	<a>
 *		{$i/COL1}
 *	</a>
 *
 */
public final class IfThenElse extends Expression {

    private static final String RCSRevision = "$Revision: 1.4 $";
    private static final String RCSName = "$Name:  $";

    protected Expression _if;
    protected Expression _then;
    protected Expression _else;

    public IfThenElse() {}

    public IfThenElse(Expression iif, Expression tthen, Expression eelse) {
        setIf(iif);
        setThen(tthen);
        setElse(eelse);
    }

    synchronized Object clone(Map clonedExprs)
        throws CloneNotSupportedException {
        //Trace.enter(this, "clone(Map clonedExprs)");
        IfThenElse retVal = (IfThenElse) super.clone(clonedExprs);
        retVal.setIf(
            (getIf() == null) ? null : (Expression) getIf().clone(clonedExprs));
        retVal.setThen(
            (getThen() == null)
                ? null
                : (Expression) getThen().clone(clonedExprs));
        retVal.setElse(
            (getElse() == null)
                ? null
                : (Expression) getElse().clone(clonedExprs));
        clonedExprs.put(this, retVal);
        //Trace.exit(this, "clone(Map clonedExprs)");
        return retVal;
    }

    public Expression getIf() {
        return _if;
    }

    public void setIf(Expression iif) {
        _if = iif;
    }

    public Expression getThen() {
        return _then;
    }

    public void setThen(Expression tthen) {
        _then = tthen;
    }

    public Expression getElse() {
        return _else;
    }

    public void setElse(Expression eelse) {
        _else = eelse;
    }

    public List getOperands() {
        return null;
    }

    public boolean replaceChild(Expression oldChild, Expression newChild) {
        //Trace.enter(this,"replaceChild(Expression oldChild, Expression newChild)");
        boolean retVal = false;

        if (getIf().equals(oldChild)) {
            setIf(newChild);
            retVal = true;
        }
        else if (getThen().equals(oldChild)) {
            setThen(newChild);
            retVal = true;
        }
        else if (getElse().equals(oldChild)) {
            setElse(newChild);
            retVal = true;
        }

        //Trace.exit(this,"replaceChild(Expression oldChild, Expression newChild)");
        return retVal;
    }

    public SqlExpression accept(GenSqlVisitor visitor)
        throws SqlWrapperException {
        return visitor.visit(this);
    }

    public void accept(AlgebraVisitor visitor) throws SqlWrapperException {
        visitor.visit(this);
    }

    /**
     * The semantics chosen for equals of algebra nodes is "the same algebra operator
     * hierarchy referring to the same database objects".
     * @param o the expression to be compared with this.
     */
    public boolean deepEquals(Object o) {
        if (o instanceof IfThenElse) {
            IfThenElse cast = (IfThenElse) o;
            return _if.deepEquals(cast.getIf())
                && _then.deepEquals(cast.getThen())
                && _else.deepEquals(cast.getElse());
        }
        return false;
    }
}
