/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.algebra;

import java.util.*;

import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.sql.SqlExpression;
import org.apache.commons.collections.set.ListOrderedSet;;

public final class Join extends Expression implements Relation, PredicateHolder {
	private static final String RCSRevision = "$Revision: 1.6 $";
	private static final String RCSName = "$Name:  $";
	
	protected List _predicateList;
	protected List _operandList;
	protected List _keys;
	
	private Set _visibleTableInstance = ListOrderedSet.decorate(new HashSet());


	public Join(List operandList, List predicateList) {
		setOperandList(operandList);
		setPredicateList(predicateList);
	}

	public Join() {
		_operandList = new ArrayList();
	}

	synchronized Object clone(Map clonedExprs) throws CloneNotSupportedException {
		Join retVal = (Join) super.clone(clonedExprs);
		retVal.setOperandList(AlgebraTools.clone(getOperandList(), clonedExprs));
		retVal.setPredicateList(AlgebraTools.clone(getPredicateList(), clonedExprs));

		clonedExprs.put(this, retVal);
		return retVal;
	}

	public List getPredicateList() {
		return _predicateList;
	}

	public void setPredicateList(List predicateList) {
		_predicateList = predicateList;

		if (null != _predicateList) {
			Iterator itr;
			itr = _predicateList.iterator();
			while (itr.hasNext()) {
				((Expression) itr.next()).setFather(this);
			}
		}
	}

	public void addPredicate(Expression predicate) {
		if (null == predicate) {
			return;
		}

		if (null == _predicateList) {
			_predicateList = new ArrayList();
		}

		_predicateList.add(predicate);
		predicate.setFather(this);
	}

	public void addPredicateList(List predicateList) {
		if (null == predicateList)
			return;

		if (null == _predicateList) {
			_predicateList = new ArrayList();
		}

		_predicateList.addAll(predicateList);

		Expression predicate = null;
		for (int i = 0; i < predicateList.size(); i++) {
			predicate = (Expression) predicateList.get(i);
			predicate.setFather(this);
		}

	}

	public void removePredicate(Expression predicate) {
		_predicateList.remove(predicate);
	}

	/**
	 * Access method for the _operandList property.
	 * @return the current value of the _operandList property
	 */
	public List getOperandList() {
		return _operandList;
	}

	/**
	 * Sets the value of the _operandList property.
	 * 
	 * @param aOperandList
	 *            the new value of the _operandList property
	 */
	public void setOperandList(List aOperandList) {
		_operandList = aOperandList;
		if (null != _operandList) {
			Iterator itr;
			itr = _operandList.iterator();
			while (itr.hasNext()) {
				((Expression) itr.next()).setFather(this);
			}
		}
		_visibleTableInstance.clear();
	}

	public void addOperandList(List aOperandList) {
		if (null != aOperandList) {
			if (null == _operandList) {
				_operandList = new ArrayList();
			}
			_operandList.addAll(aOperandList);

			Iterator itr;
			itr = _operandList.iterator();
			while (itr.hasNext()) {
				((Expression) itr.next()).setFather(this);
			}
		}
		_visibleTableInstance.clear();
	}

	public List getOperands() {
		return _operandList;
	}

	public void addOperand(Expression relation) {
		if (null == relation)
			return;
		if (null == _operandList)
			_operandList = new ArrayList();
		_operandList.add(relation);
		relation.setFather(this);
		_visibleTableInstance.clear();
	}

	public boolean predicateUp(Expression predicate) {
		return false;
	}

	public Set providedTableInstances() {
		return visibleTableInstances();
	};

	public Set visibleTableInstances() {
		if (_visibleTableInstance.isEmpty()) {
			Relation operand = null;
			for (int i = 0; i < _operandList.size(); i++) {
				operand = (Relation) _operandList.get(i);
				_visibleTableInstance.addAll(operand.providedTableInstances());
			}
		}
		return _visibleTableInstance;
	}

	public AttributeExpression findNonNullAttribute() {
		Iterator it = _operandList.iterator();
		AttributeExpression ret = null;

		while (it.hasNext() && ret == null) {
			ret = ((Relation) it.next()).findNonNullAttribute();
		}
		return ret;
	}

	public List getItems() {
		return null;
	}

	public void setKeys(List keys) {
		_keys = keys;
	}

	public List getKeys() {
		return _keys;
	}

	public List nameTest(String name) { return null;}

	public boolean replaceChild(Expression oldChild, Expression newChild) {
		boolean retVal = false;

		List list = getPredicateList();
		Expression expr = null;

		if (null != list) {
			for (int i = 0; i < list.size(); i++) {
				expr = (Expression) list.get(i);
				if (expr.equals(oldChild)) {
					list.set(i, newChild);
					newChild.setFather(this);
					retVal = true;
					break;
				}
			}
		}

		list = getOperandList();

		expr = null;
		for (int i = 0; i < list.size(); i++) {
			expr = (Expression) list.get(i);
			if (expr.equals(oldChild)) {
				list.set(i, newChild);
				newChild.setFather(this);
				retVal = true;
				break;
			}
		}

		return retVal;
	}

	public SqlExpression accept(GenSqlVisitor visitor) throws SqlWrapperException {
		return visitor.visit(this);
	}

	public void accept(AlgebraVisitor visitor) throws SqlWrapperException {
		visitor.visit(this);
	}

	/**
	 * @see Expression#deepEquals(Object)
	 */
	public boolean deepEquals(Object o) {
		if (o instanceof Join) {
			Join cast = (Join) o;
			return AlgebraTools.areExprListEquivalent(_operandList, cast.getOperandList()) && AlgebraTools.areExprListEquivalent(_predicateList, cast.getPredicateList());
		}
		return false;
	}
}
