/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.metadata;

import java.io.*;
import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.ParseException;

import org.xquark.extractor.common.MetaDataException;
import org.xquark.jdbc.typing.*;
import org.xquark.schema.SchemaException;
import org.xquark.schema.SimpleType;
import org.xquark.schema.datatypes.*;
import org.xquark.schema.datatypes.URI.MalformedURIException;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.typing.QAtomicSerializer;

/**
 * Processes XML/SQL mapping issues (type checking, JDBC interface, mapping...)
 * for extractor. 
 * 
 * NOTE: At the present time there is no particular reason with the metadata
 * and the reconstruction should use the same class except for future use of flags
 * or if the junction between the reconstruction an metadata could be done.
 */
public class ExtractorMappingInfo extends MappingInfo
{
	private static int BUF_SIZE = 256;
	
    private boolean isBuiltInXMLType = true;
    private ColumnMetaData columnMetaData = null;
    
	/** Whether this column must be retrieved using JDBC getString */
	private boolean useGetString = false;

    
	/**
	 * Used at the present time, by GenAlgebraVisitor for temporary table.
	 * @param cmeta
	 * @throws MetaDataException
	 */
	public ExtractorMappingInfo(DbType cmeta, TypeMap typeMap)
	throws MetaDataException
	{
		setXMLType(generateSimpleType(cmeta));
		setDBType(cmeta);
		useGetString = typeMap.useGetString(dbType.getJDBCType());
	}

	/**
	 * Used at the present time, by metadata for constructing an XML type basing
	 * on SQL type and set some flags (not used yet)
	 * @param cmeta
	 * @throws MetaDataException
	 */
	public ExtractorMappingInfo(ColumnMetaData cmeta, TypeMap typeMap)
	throws MetaDataException
	{
	    this(cmeta.getType(), typeMap);
		setColumnMetaData(cmeta);
	}
	
	/**
	 * This constructor is to be used only at reconstruction time when no simple
	 * type string generation and MappingInfo flags are needed. At reconstruction
	 * time, no reliable DBType is available. This should be corrected if flags
	 * like @link MappingInfo#useStreams were needed.
	 * (see @link org.xquark.extractor.algebra.GenAlgebraVisitor)
	 * @param XMLType
	 */
	public ExtractorMappingInfo(SimpleType XMLType, TypeMap typeMap)
	{
		super(XMLType);
		useGetString = typeMap.useGetString(
				ABSTRACT_SCHEMA_JDBC_MAPPING[sType.getPrimitive().getType()]);
	}
	
	/**
	 * Performed on demand, if data is fetched from database.
	 */
	public void checkForExtraction() throws XQueryException {
		if (columnMetaData != null && !columnMetaData.getSqlType().isExtraction())
			throw new XQueryException("Extraction is not supported for the " + columnMetaData.getTypeName() + " data type.");
	}

	private String generateSimpleType(DbType dbType)
	throws MetaDataException
	{
		StringBuffer buffer = new StringBuffer();
		isBuiltInXMLType = true;
		switch (dbType.getJDBCType()) {
			case java.sql.Types.ARRAY :
				throw new MetaDataException("The JDBC ARRAY type is not currently supported for extraction.");
			case java.sql.Types.BIGINT :
				buffer.append("long");
			    break;
			case java.sql.Types.BINARY :
				isBuiltInXMLType = false;
				buffer.append("<simpleType><restriction base=\"base64Binary\"><length value=\"");
				buffer.append(dbType.getLength());
				buffer.append("\"/></restriction></simpleType>");
				break;
			case java.sql.Types.BIT : case DbType.BOOLEAN :
			    buffer.append("boolean");
			    break;
			case java.sql.Types.CHAR :
				isBuiltInXMLType = false;
				buffer.append("<simpleType><restriction base=\"string\"><length value=\"");
				buffer.append(dbType.getLength());
				buffer.append("\"/></restriction></simpleType>");
				break;
			case java.sql.Types.DATE :
				buffer.append("date");
				break;
			case java.sql.Types.DECIMAL : case java.sql.Types.NUMERIC :
				isBuiltInXMLType = false;
				buffer.append("<simpleType><restriction base=\"decimal\"><totalDigits value=\"");
				buffer.append(dbType.getLength());
				buffer.append("\"/><fractionDigits value=\"");
				buffer.append(dbType.getScale());
				buffer.append("\"/></restriction></simpleType>");
				break;
			case java.sql.Types.DISTINCT :
			    throw new MetaDataException("The JDBC CLOB type is not currently supported for extraction.");
			case java.sql.Types.DOUBLE : case java.sql.Types.FLOAT :
				buffer.append("double");
				break;
			case DbType.ORACLE_ROWID:
				buffer.append("string");
				break;
			case java.sql.Types.INTEGER :
				buffer.append("integer");
				break;
			case java.sql.Types.JAVA_OBJECT :
			    throw new MetaDataException("The JDBC JAVA_OBJECT type is not currently supported for extraction.");
			case java.sql.Types.NULL :
			    throw new MetaDataException("The JDBC NULL type is not currently supported for extraction.");
			case java.sql.Types.OTHER :
			    throw new MetaDataException("The JDBC OTHER type is not currently supported for extraction.");
			case java.sql.Types.REAL :
				buffer.append("float");
				break;
			case java.sql.Types.REF :
			    throw new MetaDataException("The JDBC REF type is not currently supported for extraction.");
			case java.sql.Types.SMALLINT :
				buffer.append("short");
				break;
			case java.sql.Types.STRUCT :
			    throw new MetaDataException("The JDBC STRUCT type is not currently supported for extraction.");
			case java.sql.Types.TIME :
				buffer.append("time");
				break;
			case java.sql.Types.TIMESTAMP :
				buffer.append("dateTime");
				break;
			case java.sql.Types.TINYINT :
				buffer.append("byte");
				break;
			case java.sql.Types.BLOB :
			case java.sql.Types.LONGVARBINARY :	case java.sql.Types.VARBINARY :
				isBuiltInXMLType = false;
				buffer.append("<simpleType><restriction base=\"base64Binary\"><maxLength value=\"");
				buffer.append(dbType.getLength());
				buffer.append("\"/></restriction></simpleType>");
				break;
			case java.sql.Types.CLOB :
			case java.sql.Types.LONGVARCHAR : case java.sql.Types.VARCHAR :
				isBuiltInXMLType = false;
				buffer.append("<simpleType><restriction base=\"string\"><maxLength value=\"");
				buffer.append(dbType.getLength());
				buffer.append("\"/></restriction></simpleType>");
				break;
		}
		return buffer.toString();
	}
	
    public boolean isBuiltInXMLType() {
        return isBuiltInXMLType;
    }

	public String getParameter(ResultSet rs, int i, QAtomicSerializer qSerializer, ValidationContextProvider nsContext) throws XQueryException {
		String value = null;
	    try {
	        if (useGetString) {
				try {
					// Try a getString + convert/toXMLString if JDBC cannot provide a java type compatible with schema
					value = rs.getString(i);
					if (value == null)
					    return null;
					switch(javaType) {
					    // perf shortcut for numbers 
						case JAVA_FLOAT : case JAVA_DOUBLE :
						case JAVA_LONG : case JAVA_BIG_DECIMAL :
							return qSerializer.removeNotSignificantDigits(value);

						case JAVA_STRING :
							return qSerializer.serialize(sType, value);
						case SCHEMA_QNAME :
							return qSerializer.serialize(new QName(value));
						case SCHEMA_URI :
							return qSerializer.serialize(new URI(value));
						case SCHEMA_DATETIME :
							return qSerializer.serialize(sType, (DateTime)sType.convert(value, true, nsContext));
						case SCHEMA_DURATION :
							return qSerializer.serialize(sType, new Duration(value));
						case JAVA_BOOLEAN :
							return qSerializer.serialize(sType, (Boolean)sType.convert(value, true, nsContext));
						case SCHEMA_BYTE_ARRAY :
							return qSerializer.serialize(sType, (ByteArray)sType.convert(value, true, nsContext));
						default:
							return null;
					}
				}
				catch (SchemaException e1) {
					throw new XQueryException("Schema exception raised when converting \""
							+ value + "\" database value to XML. Please check underlying exception.", e1);
				}
				catch (SQLException e1) {
					throw new XQueryException("JDBC exception raised when retrieving data from resultset using backup getString() method. Please check underlying exception.", e1);
				}
	        }
	        else {
	            switch(javaType) {
	            	case JAVA_STRING :
	            		if (useStreams){
	            			Reader reader = rs.getCharacterStream(i);
	            			char[] cb = new char[BUF_SIZE];
	            			CharArrayWriter caw = new CharArrayWriter();
	            			int len = -1;
	            			try {
								while ((len = reader.read(cb)) > 0)
									caw.write(cb, 0, len);
							} catch (IOException e1) {
								throw new XQueryException(
										"IOexception raised when retrieving data from resultset using backup getCharacterStream() method. Please check underlying exception.",
										e1);
							}
							if (caw.size() == 0 && rs.wasNull())
								value = null;
							else
								value = caw.toString();
	            		}
	            		else
	            			value = rs.getString(i);
	            	    if (value == null)
	            	        return null;
	            	    return qSerializer.serialize(sType, value);
	            	case SCHEMA_QNAME :
						value = rs.getString(i);
						if (value == null)
							return null;
	            	    return qSerializer.serialize(new QName(value));
	            	case SCHEMA_URI :
						value = rs.getString(i);
						if (value == null)
							return null;
	            	    return qSerializer.serialize(new URI(value));
	            	case SCHEMA_DATETIME :
	            	    Timestamp ts = rs.getTimestamp(i);
						if (ts == null)
							return null;
	            	    return qSerializer.serialize(sType, new DateTime(ts.getTime()));
	            	case SCHEMA_DURATION :
						value = rs.getString(i);
						if (value == null)
							return null;
	            	    return qSerializer.serialize(sType, new Duration(value));
	            	case JAVA_LONG :
	            	    long l = rs.getLong(i);
	            	    if (rs.wasNull())
	            	        return null;
	            	    return qSerializer.serialize(new Long(l));
	            	case JAVA_BIG_DECIMAL :
	            	    BigDecimal bd = rs.getBigDecimal(i);
						if (bd == null)
							return null;
	            	    return qSerializer.serialize(bd);
	            	case JAVA_FLOAT :
						float f = rs.getFloat(i);
						if (rs.wasNull())
							return null;
	            	    return qSerializer.serialize(new Float(f));
	            	case JAVA_DOUBLE :
						double d = rs.getDouble(i);
						if (rs.wasNull())
							return null;
	            	    return qSerializer.serialize(new Double(d));
	            	case JAVA_BOOLEAN :
						boolean b = rs.getBoolean(i);
						if (rs.wasNull())
							return null;
	            	    return qSerializer.serialize(sType, new Boolean(b));
	            	case SCHEMA_BYTE_ARRAY :
	            		byte[] ba = null;
	            		if (useStreams){
	            			InputStream is = rs.getBinaryStream(i);
	            			ba = new byte[BUF_SIZE];
	            			ByteArrayOutputStream baos = new ByteArrayOutputStream();
	            			int len = -1;
	            			try {
								while ((len = is.read(ba)) > 0)
									baos.write(ba, 0, len);
							} catch (IOException e1) {
								throw new XQueryException(
										"IOexception raised when retrieving data from resultset using backup getCharacterStream() method. Please check underlying exception.",
										e1);
							}
							if (baos.size() == 0 && rs.wasNull())
								ba = null;
							else
								ba = baos.toByteArray();
	            		}
	            		else
	            			ba = rs.getBytes(i);
						if (ba == null)
							return null;
	            	    return qSerializer.serialize(sType, ba);
	            	default:
	            	    return null;
	            }
	        }
        }
        catch (SQLException e) {
			try {
				// Try a getString + convert/toXMLString if JDBC cannot provide a java type compatible with schema
			    // For instance MySQL does not support getTimeStamp on a literal since no TO_DATE exist
			    // TODO: try to detect there is no column to avoid catching exception
				value = rs.getString(i);
				return sType.toXMLString(sType.convert(value, true, nsContext), nsContext);
			}
			catch (SQLException e1) {
				throw new XQueryException("JDBC exception raised when retrieving data from resultset using backup getString() method. Please check underlying exception.", e1);
			}
			catch (SchemaException e1) {
				throw new XQueryException("Schema exception raised when converting \""
						+ value + "\" database value to XML. Please check underlying exception.", e1);
			}
        }
		catch (MalformedURIException e) {
		    throw new XQueryException("Data retrieved from resultset is not an URI. Please check underlying exception.", e);
		}
        catch (ParseException e) {
			throw new XQueryException("Data retrieved from resultset is not a Duration. Please check underlying exception.", e);
        }
	}
    
    public ColumnMetaData getColumnMetaData() {
        return columnMetaData;
    }

    public void setColumnMetaData(ColumnMetaData columnMetaData) {
        this.columnMetaData = columnMetaData;
    }
}
