/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.metadata;

import java.sql.*;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.common.MetaDataException;
import org.xquark.extractor.runtime.Selection;
import org.xquark.jdbc.typing.ColumnMetaData;
import org.xquark.jdbc.typing.DBMSInfo;
import org.xquark.jdbc.typing.JDBCProperties;

/**
 * Base class for metadata path tree building from JDBC metadata applying
 * filters defined in the configuration file. This class will be derived for
 * every RDBMS supported.
 */
public abstract class Loader {

	private static final String RCSRevision = "$Revision: 1.24 $";
	private static final String RCSName = "$Name:  $";

	private static Log log = LogFactory.getLog(Loader.class);

	/* Maintains a map from XML names to DBMS name for table and columns */
	public Properties _tableMap = new Properties();

	protected Connection _connection;
	protected DatabaseMetaData _databaseMetaData;
	protected DBMSInfo _dbmsInfo;
	protected HashSet _nameSpaces = new HashSet();
	protected QNameEncoder _encoder = null;

	public Loader(Connection connection, DBMSInfo dbmsInfo, QNameEncoder encoder) throws SQLException {
		_connection = connection;
		_databaseMetaData = connection.getMetaData();
		_dbmsInfo = dbmsInfo;
		_encoder = encoder;
	}

	/**
	 * Main entry point performing path tree construction. Browse JDBC catalogs
	 * calling sub-procedure for schemas if matching the selection.
	 * @param selection A user-defined filter used for catalogs.
	 * @return The root metadata node created for the source (Site).
	 * @throws MetaDataException
	 * @see org.xquark.extractor.metadata.NamedNode
	 */
	public Site loadSite(Selection selection, JDBCProperties JDBCprop) throws MetaDataException {

		Site retVal = new Site(null);
		List children = new ArrayList();
		List subSelections = new ArrayList();
		ResultSet resultSet = null;
		String name = null;

		_nameSpaces.clear();

		try {
			String defCatalog = getDefaultCatalog();

			if (null == selection) {
				selection = new Selection(null, null, Selection.INCLUDES, new HashMap());
			}
			
			 if (selection.noSubElementDef()) {
				Selection catalogSel = new Selection(defCatalog, null, Selection.INCLUDES, new HashMap());
				selection.addToMap(defCatalog, catalogSel);
			}

			/* JDBC catalog browsing */
			resultSet = _databaseMetaData.getCatalogs();
			Selection subSelection = null;
			while (resultSet.next()) {
				name = resultSet.getString("TABLE_CAT");
				subSelection = selection.includesSubElement(name);
				if (subSelection != null) {
					// TODO : No indications left by YS on this...
					//                        if (CASE_TOLOWER == _nameCase ) name = name.toLowerCase();
					//                        else if (CASE_TOUPPER== _nameCase) name = name.toUpperCase();
					children.add(name);
					subSelections.add(subSelection);
				}
			}
			resultSet.close();

			/* Case where the catalog name is unspecified */
			subSelection = selection.includesSubElement(null);
			if (subSelection != null) {
				if (defCatalog == null) {			
					children.add(null);
					subSelections.add(subSelection);
				} else {
					children.add(defCatalog);
					subSelection.setNameForDefault(defCatalog);
					subSelections.add(subSelection);
				}
			}

			if (children.isEmpty()) {
				throw new MetaDataException("Wrong catalog configuration specification. Check your configuration file.");
			}

			/* Browse filtered catalogs for building */
			for (int i = 0; i < children.size(); i++) {
				String child = (String) children.get(i);
				subSelection = (Selection) subSelections.get(i);
				retVal.addChild(loadCatalog(child, subSelection, JDBCprop));
			}

		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}
		return retVal;
	}

	/**
	 * Create the Catalog metadata node for a specific catalog. Browse JDBC
	 * schemas calling sub-procedure for schemas if matching the selection.
	 * @param catalogName The catalog browsed.
	 * @param selection A user-defined filter used for schemas.
	 * @return the {@link org.xquark.extractor.metadata.Catalog Catalog} node
	 * built.
	 * @throws MetaDataException
	 */
	protected Catalog loadCatalog(String catalogName, Selection selection, JDBCProperties JDBCprop) throws MetaDataException {
		//Trace.enter(this, "loadCatalog(String name, Selection selection)", catalogName);

		List children = new ArrayList();
		List subSelections = new ArrayList();
		ResultSet resultSet = null;
		String name = null;

		/* Catalog node creation */
		Catalog retVal = null;
		if (catalogName == null || catalogName.length() == 0)
			retVal = new Catalog(null);
		else
			retVal = new Catalog(catalogName);

		/* JDBC schema browsing */
		try {
			String defSchema = getDefaultSchema();

			if (selection.noSubElementDef()) {
				Selection schemaSel = new Selection(defSchema, null, Selection.INCLUDES, new HashMap());
				selection.addToMap(defSchema, schemaSel);
			}

			resultSet = _databaseMetaData.getSchemas();
			Selection subSelection = null;
			while (resultSet.next()) {
				name = resultSet.getString("TABLE_SCHEM");
				subSelection = selection.includesSubElement(name);
				if (subSelection != null) {
					String targetNameSpace = subSelection.getTargetNamespace();
					//                        if (targetNameSpace == null) targetNameSpace = "";
					if (_nameSpaces.contains(targetNameSpace))
						throw new MetaDataException("Same target namespace : [" + targetNameSpace + "] used twice");
					_nameSpaces.add(targetNameSpace);
					children.add(name);
					subSelections.add(subSelection);
				}
			}
			resultSet.close();

			/* Case where the catalog name is unspecified */
			subSelection = selection.includesSubElement(null);
			if (subSelection != null) {
				if (defSchema == null) {			
					children.add(null);
					subSelections.add(subSelection);
				} else {
					children.add(defSchema);
					subSelection.setNameForDefault(defSchema);
					subSelections.add(subSelection);
				}
			}

			if (children.isEmpty()) {
				throw new MetaDataException("Wrong schema configuration specification. Check your configuration file.");
			}

			/* Browse filtered schema for building */
			for (int i = 0; i < children.size(); i++) {
				String child = (String) children.get(i);
				subSelection = (Selection) subSelections.get(i);
				retVal.addChild(loadSchema(catalogName, child, subSelection, JDBCprop));
			}
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		//Trace.exit(this, "loadCatalog(String name, Selection selection)", catalogName);
		return retVal;

	}

	/**
	 * Create the Schema metadata node for a specific schema. Browse JDBC
	 * tables encoding names if demanded by user and calling sub-procedures if 
	 * table is matching the selection for both tables and keys.
	 * @param catalogName The catalog that contains the schema.
	 * @param schemaName The schema browsed.
	 * @param selection A user-defined filter used for tables.
	 * @return the {@link org.xquark.extractor.metadata.Schema Schema} node
	 * built.
	 * @throws MetaDataException
	 */
	protected Schema loadSchema(String catalogName, String schemaName, Selection selection, JDBCProperties JDBCprop) throws MetaDataException {
		//Trace.enter(this, "loadSchema ( String catalogName, String schemaName, Selection selection )", schemaName);

		List children = new ArrayList();
		List subSelections = new ArrayList();
		NamedNode catalog = null;
		ResultSet resultSet = null;
		String name = null;
		_nameSpaces.clear();
		String alias = null;

		/* Schema node creation */
		Schema retVal = new Schema(schemaName);
		if (selection != null) {
			retVal.setTargetNamespace(selection.getTargetNamespace());
			retVal.setElementFormDefault(selection.getElementFormDefault());
		}

		/* JDBC tables browsing */
		try {
			String[] tableTypes = { "TABLE", "VIEW", "SYNONYM" };
			String tableType = null;
			resultSet = _databaseMetaData.getTables(catalogName, schemaName, null, tableTypes);

			if (selection == null || selection.noSubElementDef()) {
				/* not indications of which tables to be loaded, so all tables will be loaded */
				while (resultSet.next()) {
					tableType = resultSet.getString("TABLE_TYPE");
					name = resultSet.getString("TABLE_NAME");
					alias = _encoder.encode(name);
					if (!Selection.verifyName(alias)) {
						log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
						log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
						continue; // out of selection
					}

					if (tableType.equalsIgnoreCase("TABLE")) {
						children.add(new Relation(name, alias, Table.TABLE));
						subSelections.add(null);
					} else if (tableType.equalsIgnoreCase("VIEW")) {
						children.add(new Relation(name, alias, Table.VIEW));
						subSelections.add(null);
					} else if (tableType.equalsIgnoreCase("SYNONYM")) {
						children.add(new Relation(name, alias, Table.SYNONYM));
						subSelections.add(null);
					} else {
						log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
					}
				}
				resultSet.close();
			} else {
				/* */
				Selection subSelection = null;
				while (resultSet.next()) {
					name = resultSet.getString("TABLE_NAME");
					subSelection = selection.includesSubElement(name);
					if (subSelection != null) {
						tableType = resultSet.getString("TABLE_TYPE");
						alias = subSelection.getAlias();
						if (alias == null) {
							alias = _encoder.encode(name);
							if (!Selection.verifyName(alias)) {
								log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
								log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
								continue; // out of selection
							}
						}

						if (tableType.equalsIgnoreCase("TABLE")) {
							children.add(new Relation(name, alias, Table.TABLE));
							subSelections.add(subSelection);
						} else if (tableType.equalsIgnoreCase("VIEW")) {
							children.add(new Relation(name, alias, Table.VIEW));
							subSelections.add(subSelection);
						} else if (tableType.equalsIgnoreCase("SYNONYM")) {
							children.add(new Relation(name, alias, Table.SYNONYM));
							subSelections.add(subSelection);
						} else {
							log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
						}
					}
				}
				resultSet.close();
			}

			// TODO : temporary code (written by YS)
			String path = null; /* the prefix of key in _tableMap Properties */
			if (null != catalogName)
				path = catalogName;
			if (schemaName != null)
				if (path == null)
					path = schemaName;
				else
					path += '.' + schemaName;

			String encodedName = null;

			/* browse "relation" objects for table and key loading */
			for (int i = 0; i < children.size(); i++) {
				Relation child = ((Relation) children.get(i));
				Selection subSelection = (Selection) subSelections.get(i);

				putTable(path, child.alias, child.name); // register table

				NamedNode childNode = null;
				if (Table.TABLE == child.type) {
					childNode = loadTable(catalogName, schemaName, child.name, child.alias, subSelection, JDBCprop);
					if (childNode != null) {
						loadKeys(catalogName, schemaName, (Table) childNode, subSelection);
					}
				} else {
					childNode = loadTable(catalogName, schemaName, child.name, child.alias, subSelection, JDBCprop);
				}
				if (childNode != null) {
					retVal.addChild(childNode);
				}
			}
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		//Trace.exit(this, "loadSchema ( String catalogName, String schemaName, Selection selection )", schemaName);
		return retVal;
	}

	/**
	 * Create the table and columns metadata nodes for a specific table. Browse
	 * JDBC {@link java.sql.ResultSetMetaData ResultSetMetaData} for column
	 * information encoding names if demanded by user.
	 * @param catalogName
	 * @param schemaName
	 * @param origTableName
	 * @param tableName
	 * @param selection
	 * @return
	 * @throws MetaDataException
	 */
	protected Table loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection, JDBCProperties JDBCprop) throws MetaDataException {
		//Trace.enter(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);

		/* Table node creation */
		Table retTable = new Table(origTableName, tableName);
		retTable.setType(Table.TABLE);

		/* Get dummy SQL query string to explore resultset metadata (columns) */
		List columnsMeta = null;
		try {
			columnsMeta = _dbmsInfo.getTableMetadata(catalogName, schemaName, origTableName, _connection, JDBCprop);
		} catch (SQLException ex) {
			log.error(ex);
			log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", "Table : " + schemaName + "." + origTableName));
			//Trace.exit(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);
			return null;
		}

		/* Browse result set metadata */
		Attribute attribute = null;
		ColumnMetaData column = null;
		Selection subSelection = null;
		String alias = null;
		List pkAttributes = new ArrayList();
		// list of column nodes forming PK
		boolean currentColumnIsInPK = false;

		//Trace.message(this, "loadTable(...)", "Loading columns...");

		String path = null; /* the prefix of key in _tableMap Properties */
		if (null != catalogName)
			path = catalogName + '.' + schemaName + '.' + tableName;
		else
			path = schemaName + '.' + tableName;

		for (int i = 0; i < columnsMeta.size(); i++) {
			column = (ColumnMetaData) columnsMeta.get(i);

			currentColumnIsInPK = false;
			alias = null;

			if (selection == null || selection.noSubElementDef()) {
				alias = _encoder.encode(column.getColumnName());
				if (!Selection.verifyName(alias)) {
					log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
					log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column.getColumnName()));
					continue; // out of selection
				}
			} else {
				subSelection = selection.includesSubElement(column.getColumnName());
				if (subSelection == null) {
					continue; // out of selection
				} else {
					alias = subSelection.getAlias();
					if (alias == null) {
						alias = _encoder.encode(column.getColumnName());
						if (!Selection.verifyName(alias)) {
							log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
							log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column.getColumnName()));
							continue;
						}
					}
					if (subSelection.getPKColumn()) {
						currentColumnIsInPK = true;
					}
				}
			}
			putAttribute(path, alias, column.getColumnName()); // register column

			/* Attribute (column) node creation */
			ExtractorMappingInfo mappingInfo = null;
			try {
				mappingInfo = new ExtractorMappingInfo(column, _dbmsInfo.getTypeMap());
			} catch (MetaDataException e) {
				MessageLibrary.getMessage("M_SUP_DATATYPE", new String[] { column.getType().getNativeType(), column.getTableName(), column.getColumnName()});
				continue;
			}

			attribute = new Attribute(column.getColumnName(), alias);
			attribute.setMappingInfo(mappingInfo);
			attribute.setNullable(column.isNullable());

			/* Add column to metadata tree */
			retTable.addChild(attribute);

			/*
			 * Add column metadata node to Table node PK list as defined by
			 * user in configuration file.
			 */
			// QUESTION : the user PK is the first so used in priority
			if (currentColumnIsInPK) {
				pkAttributes.add(attribute);
			}
		}

		/* Add PK column metadata node list to Table node */
		if (!pkAttributes.isEmpty()) {
			retTable.addKey(pkAttributes);
		}

		//Trace.exit(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);
		return retTable;
	}

	/**
	 * Add the real Primary Key to Table metadata node basing on JDBC metadata.
	 * @param catalogName
	 * @param schemaName
	 * @param table
	 * @param selection
	 * @throws MetaDataException
	 */
	protected void loadKeys(String catalogName, String schemaName, Table table, Selection selection) throws MetaDataException {
		//Trace.enter(this, "loadKeys(String catalogName, String schemaName, Table table, Selection selection)", table.getName());

		try {
			//		    /* load Primary Keys for this table */
			//            Trace.message( this,"loadTable(...)", "creating pseudoAttribute column...");
			String column_name = null;
			String alias = null;
			Selection subSelection = null;
			Attribute attribute = null;
			List key = new ArrayList();

			ResultSet resultSet = _databaseMetaData.getPrimaryKeys(catalogName, schemaName, table.getName());

			String path = null; /* the prefix of key in _tableMap Properties */
			if (null != catalogName) {
				path = catalogName + '.' + schemaName + '.' + table.getName();
			} else {
				path = schemaName + '.' + table.getName();
			}

			/* Loop on columns of the PK for the specified table */
			while (resultSet.next()) {
				column_name = resultSet.getString("COLUMN_NAME");

				/* generate and verify alias*/
				alias = null;
				if (selection == null || selection.noSubElementDef()) {
					alias = _encoder.encode(column_name);
				} else {
					subSelection = selection.includesSubElement(column_name);
					if (subSelection == null) {
						alias = _encoder.encode(column_name);
					} else {
						alias = subSelection.getAlias();
						if (null == alias) {
							alias = _encoder.encode(column_name);
						}
					}
				}

				//Trace.message(this, "loadKeys(String catalogName, String schemaName, Table table, Selection selection)", column_name + "--" + alias);

				attribute = (Attribute) table.findChild(column_name);
				if (attribute == null) {
					putAttribute(path, alias, column_name);
					attribute = new Attribute(column_name, alias);
				}
				key.add(attribute);
			}
			resultSet.close();

			/* Add PK to Table metadata node */
			if (!key.isEmpty()) {
				table.addKey(key);
			}
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage());
		}
		//Trace.exit(this, "loadKeys(String catalogName, String schemaName, Table table, Selection selection)", table.getName());
	}

	/**
	 * Add the table to map between XML names and native DBMS names.
	 * @param path
	 * @param encodedTableName
	 * @param originalName
	 * @throws MetaDataException
	 */
	protected void putTable(String path, String encodedTableName, String originalName) throws MetaDataException {
		String key = null;
		key = path + '.' + encodedTableName;

		Object value = null;
		value = _tableMap.getProperty(key);
		if (null == value) {
			_tableMap.setProperty(key, originalName);
		} else {
			throw new MetaDataException(MessageLibrary.getMessage("M_NAME_CONF", key, originalName));
		}
	}

	/**
	 * Add the column to map between XML names and native DBMS names.
	 * @param path
	 * @param encodedAttributeName
	 * @param originalName
	 * @throws MetaDataException
	 */
	protected void putAttribute(String path, String encodedAttributeName, String originalName) throws MetaDataException {
		String key = null;
		key = path + '.' + encodedAttributeName;

		Object value = null;
		value = _tableMap.getProperty(key);
		if (null == value) {
			_tableMap.setProperty(key, originalName);
		} else {
			throw new MetaDataException(MessageLibrary.getMessage("M_NAME_CONF", key, originalName));
		}
	}
	
	/**
	 * Returns the current catalog for a connection
	 * @return the current catalog name, or null if there is none
	 * @throws SQLException
	 */
	protected String getDefaultCatalog() throws SQLException {
		return _connection.getCatalog();
	}
	
	/**
	 * Returns the current schema for a connection
	 * @return the current schema name, or null if there is none
	 * @throws SQLException
	 */
	protected String getDefaultSchema() throws SQLException {
		return _databaseMetaData.getUserName();
	}
	
	/**
	 * Struct-like data class.
	 */
	class Relation {

		public String name;
		public String alias;
		public int type;

		public Relation(String name, String alias, int type) {
			this.name = name;
			this.alias = alias;
			this.type = type;
		}

		public Relation(String name, int type) {
			this.name = name;
			this.type = type;
			this.alias = name;
		}
	}
}
