/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.metadata;

import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.*;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.extractor.algebra.CollectionName;
import org.xquark.extractor.runtime.Selection;
import org.xquark.jdbc.typing.DBMSInfo;
import org.xquark.schema.SchemaReader;
import org.xquark.schema.Type;
import org.xquark.serialize.XMLSerializer;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.metadata.*;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.typing.QType;
import org.xquark.xquery.typing.QTypeElement;

/**
 * Object that encapsulates relational metadata information for a data source (
 * a relational database. This object is also used, basing on this information,
 * to build medatata used by XQuery parser and the XDBC API
 * (@see org.xquark.xquery.metadata.MetaDataImpl}).
 */
public class MetaDataManager {
	private static final String RCSRevision = "$Revision: 1.17 $";
	private static final String RCSName = "$Name:  $";

	/** root NamedNode of the tree repressenting the relational metadata (site,
	 * catalogs, schema, tables, views. A site is another term for data source.
	 * It is loaded by @link Loader basing on @link Selection. 
	 */
	private Site _site = null;
	private Properties _tableMap = null;
	private Map _viewMap = null;
	private DBMSInfo dbmsInfo = null;
	
	public MetaDataManager(DBMSInfo dbmsInfo) {
	    this.dbmsInfo = dbmsInfo;
	}

	public void setMetaData(Site site) {
		_site = site;
	}

	public List findTables(CollectionName colName) {
		//Trace.enter(this, "findTables(Collection_Name colName)", colName.pprint());
		List retVal = findTables(colName.getCatalog(), colName.getSchema(), colName.getTable());
		//        if (null == retVal || retVal.isEmpty()) {
		//            retVal = findTables(colName.getCatalog(), colName.getSchema(), _encoder.decode(colName.getTable()));
		//        }
		//Trace.enter(this, "findTables(Collection_Name colName)", colName.pprint());
		return retVal;
	}

	private List findTables(String catalogName, String schemaName, String tableName) {
		//Trace.enter(this, "findTables(String catalog String schema,String table)");

		List retVal = new ArrayList();

		List nodeList = null;
		Catalog node = null;

		if (null == catalogName || catalogName.equals("*")) {
			nodeList = _site.getChildren();
			for (int i = 0; i < nodeList.size(); i++) {
				node = (Catalog) nodeList.get(i);
				retVal.addAll(findTables(node, schemaName, tableName));
			}
		} else {
			node = (Catalog) _site.findChild(catalogName);
			if (null != node) {
				retVal.addAll(findTables(node, schemaName, tableName));
			}
		}

		//Trace.exit(this, "findTables(String catalog String schema,String table)");
		return retVal;
	}

	private List findTables(Catalog catalog, String schemaName, String tableName) {
		//Trace.enter(this, "findTables(Catalog catalog, String schemaName, String tableName)", catalog.getName());

		List retVal = new ArrayList();

		List nodeList = null;
		Schema node = null;

		if (null == schemaName || schemaName.equals("*")) {
			nodeList = catalog.getChildren();
			for (int i = 0; i < nodeList.size(); i++) {
				node = (Schema) nodeList.get(i);
				List tables = findTables(node, tableName);
				if (null != tables) {
					retVal.addAll(findTables(node, tableName));
				}
			}
		} else {
			node = (Schema) catalog.findChild(schemaName);
			if (null != node) {
				retVal.addAll(findTables(node, tableName));
			}
		}

		//Trace.exit(this, "findTables(Catalog catalog, String schemaName, String tableName)", catalog.getName());
		return retVal;
	}

	private List findTables(Schema schema, String tableName) {
		//Trace.enter(this, "findTables(Schema schema, String tableName)", schema.getName());

		List retVal = null;
		Table node = null;

		if (null == tableName || tableName.equals("*")) {
			retVal = schema.getChildren();
		} else {
			node = (Table) schema.findChildByAlias(tableName);
			if (null != node) {
				retVal = new ArrayList();
				retVal.add(node);
			}
		}

		//Trace.exit(this, "findTables(Schema schema, String tableName)", schema.getName());
		return retVal;
	}

	public void setTableMap(Properties tableMap) {
		_tableMap = tableMap;
	}

	public void setViewMap(Map viewMap) {
		_viewMap = viewMap;
	}

	public String originalNameOf(String fullEncodedName) {
		return _tableMap.getProperty(fullEncodedName);
	}

	public String pprint() {
		String retVal = null;
		return retVal;
	}

	/**
	 * Builds metadata used by XQuery parser and for XDBC API.
     * @param name datasource symbolic name read in the configuration file or a
     * default one.
     * @return
     * @throws XMLDBCException
     */
    public MetaDataImpl buildMetaDataForParser(String name) throws XMLDBCException {

		if (_site == null)
			return null;
		StringBuffer bufSchema = new StringBuffer();
		StringBuffer bufCollections = new StringBuffer();

		List catalogs = this._site.getChildren();
		//org.xquark.xml.misc.Debug.print("Catalog list = " + catalogs);
		String nameSpace = null;
		if (catalogs != null)
			for (int cat = 0; cat < catalogs.size(); cat++) {
				Catalog catalog = (Catalog) catalogs.get(cat);

				List schemas = catalog.getChildren();
				//org.xquark.xml.misc.Debug.print("schemas = " + schemas);
				
				if (schemas != null) {
					for (int i = 0; i < schemas.size(); i++) {
						Schema schema = (Schema) schemas.get(i);
						//org.xquark.xml.misc.Debug.print("schema = " + schema);
						
						String targetNamespace = schema.getTargetNamespace();
						//org.xquark.xml.misc.Debug.print("targetNamespace = " + targetNamespace);

						int elementFormDefault = schema.getElementFormDefault();
						//org.xquark.xml.misc.Debug.print("elementFormDefault = " + elementFormDefault);

						bufSchema.append("\t\t<schema");
						bufSchema.append("\n\t\t\txmlns=\"http://www.w3.org/2001/XMLSchema\"");
						bufSchema.append("\n\t\t\telementFormDefault=\"" + Selection.ELEMENT_FORM_DEFAULT[elementFormDefault] + "\"");
						if (targetNamespace == null) {
							bufSchema.append(">");
							/*
							bufSchema.append("targetNamespace=\"");
							bufSchema.append("http://");
							if (catalog.getName() != null) bufSchema.append(catalog.getName());
							bufSchema.append(schema.getName());
							bufSchema.append("\">");
							 */
						} else {
							/*
							bufSchema.append("\n\t\t\t\txmlns:");
							bufSchema.append(localns);
							bufSchema.append("=\"http://www.xquark.org/");
							bufSchema.append(name);
							bufSchema.append("Schema\"");
							 */
							bufSchema.append(" targetNamespace=\"");
							bufSchema.append(targetNamespace);
							bufSchema.append("\">");
						}

						List tables = schema.getChildren();
						//org.xquark.xml.misc.Debug.print("tables = " + tables);
						if (tables != null) {
							for (int j = 0; j < tables.size(); j++) {
								Table table = (Table) tables.get(j);

								List attributes = table.getChildren();
								if (attributes == null)
									continue;

								String collectionName = table.getAlias();
								String fullCollectionName = table.getAlias();
								if (schema.getName() != null && schema.getName().length() != 0)
									fullCollectionName = schema.getName() + "." + fullCollectionName;
								if (catalog.getName() != null && catalog.getName().length() != 0)
									fullCollectionName = catalog.getName() + "." + fullCollectionName;

								bufSchema.append("\n\t\t\t<element name=\"");
								//bufTable.append(fullCollectionName);
								bufSchema.append(collectionName);
								bufSchema.append("\">");
								bufSchema.append("\n\t\t\t\t<complexType>");

								bufSchema.append("\n\t\t\t\t\t<sequence>");
								bufCollections.append("\n\t\t<collection name=\"");
								bufCollections.append(fullCollectionName);
								//bufCollections.append(collectionName);
								bufCollections.append("\">");
								if (targetNamespace == null) {
									bufCollections.append("\n\t\t\t<step name=\"");
									bufCollections.append(collectionName);
									bufCollections.append("\">");
								} else {
									bufCollections.append("\n\t\t\t<step ns=\"");
									bufCollections.append(targetNamespace);
									bufCollections.append("\" name=\"");
									bufCollections.append(collectionName);
									bufCollections.append("\">");
								}

								for (int k = 0; k < attributes.size(); k++) {
									Attribute attribute = (Attribute) attributes.get(k);
									ExtractorMappingInfo mappingInfo = attribute.getMappingInfo();

									bufSchema.append("\n\t\t\t\t\t\t<element name=\"");
									bufSchema.append(attribute.getAlias());
									if (mappingInfo.getColumnMetaData().isNullable()) {
										bufSchema.append("\" minOccurs=\"0");
									}
									if (mappingInfo.isBuiltInXMLType()) {
										bufSchema.append("\" type=\"");
										bufSchema.append(mappingInfo.getStringXMLType());
										bufSchema.append("\"/>");
									}
									else {
										bufSchema.append("\">");
										bufSchema.append("\n\t\t\t\t\t\t\t");
										bufSchema.append(mappingInfo.getStringXMLType());
										bufSchema.append("\n\t\t\t\t\t\t</element>");
									}

									if (Selection.QUALIFIED != elementFormDefault || targetNamespace == null) {
										bufCollections.append("\n\t\t\t\t<step name=\"");
										bufCollections.append(attribute.getAlias());
										bufCollections.append("\"/>");
									} else {
										bufCollections.append("\n\t\t\t\t<step ns=\"");
										bufCollections.append(targetNamespace);
										bufCollections.append("\"  name=\"");
										bufCollections.append(attribute.getAlias());
										bufCollections.append("\"/>");
									}
								}

								bufSchema.append("\n\t\t\t\t\t</sequence>");
								bufSchema.append("\n\t\t\t\t</complexType>");
								bufSchema.append("\n\t\t\t</element>");

								bufCollections.append("\n\t\t\t</step>");
								bufCollections.append("\n\t\t</collection>");
							}
						}
						bufSchema.append("\n\t\t</schema>");
					}
				}
			}

		MetaDataImpl metadata = new MetaDataImpl(null);

		StringBuffer buf = new StringBuffer();
		buf.append("<?xml version='1.0'?>");
		buf.append("\n<xm:metadata");
		buf.append("\n\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
		buf.append("\n\txmlns:xm=\"http://www.xquark.org/XMLDBC/Metadata\"");
		buf.append("\n\tsource=\"");
		buf.append(name);
		buf.append("\">");
		buf.append("\n\t<schemas>\n");
		buf.append(bufSchema.toString());
		buf.append("\n\t</schemas>");
		buf.append("\n\t<collections>");
		buf.append(bufCollections.toString());
		buf.append("\n\t</collections>");
		buf.append("\n</xm:metadata>");

		//Trace.log(this, "buildMetaDataForParser(String name)", "metadata schema :");
		//Trace.log(this, "buildMetaDataForParser(String name)", buf.toString());
		
		//System.out.println(buf.toString());

		try {
			MetaParser metaparser = new MetaParser(metadata);
			MetaWrapper metawrapper = metaparser.createWrapper(new InputSource(new StringReader(buf.toString())), true);
			metadata.addMetaWrapper(metawrapper);
		} catch (MetadataException e) {
			throw new XMLDBCException("build metadata failed", e);
		}

		if (_viewMap == null || _viewMap.isEmpty()) return metadata;

		// metadata of views
		// add schemas and collections
		XQueryParser parser = null;
		StringWriter typestring = new StringWriter();
		XMLSerializer serializer = new XMLSerializer(typestring);
		serializer.setIndent(true);
		SchemaReader schemareader = new SchemaReader(metadata.getSchemaManager());
		schemareader.setContentHandler(serializer);
		Collection keys = _viewMap.keySet();
		for (Iterator it = keys.iterator(); it.hasNext();) {
			String key = (String) it.next();
			Object value = _viewMap.get(key);
			if (value instanceof String) {
				try {
					String viewDefinition = (String) value;
					if (parser == null)
						parser = new XQueryParser(new StringReader("dummy"));
					parser.setFactory("org.xquark.xml.xqueryparser.common.ParserFactory");
					parser.ReInit(new StringReader(viewDefinition));
					XQueryModule module = parser.Start(metadata,metadata.getSchemaManager(),metadata.getModuleManager(),new VarCounter());
					XQueryExpression expr = null;
					QType qtype = null;
					Type type = null;
					QName elementname = null;
					String targetNamespace = null;
					expr = (XQueryExpression)module.getExpressions().get(0);
					qtype = expr.getQType();
					type = qtype.getType();
					elementname = (QName)((QTypeElement)qtype).getName();
					targetNamespace = elementname.getNameSpace();					
					if (type == null || !(qtype instanceof QTypeElement)) {
						this._viewMap.put(key, null);
						continue;
					}
					this._viewMap.put(key, module);
					
					// add string to corresponding strings
					// add to schemas
					if (targetNamespace != null) {
						bufSchema.append("\n\t\t<schema");
						bufSchema.append("\n\t\t\txmlns=\"http://www.w3.org/2001/XMLSchema\"");
						bufSchema.append("\n\t\t\telementFormDefault=\"unqualified\"");
						if (targetNamespace == null) {
							bufSchema.append(">");
						} else {
							bufSchema.append(" targetNamespace=\"");
							bufSchema.append(targetNamespace);
							bufSchema.append("\">");
						}
						bufSchema.append("\n\t\t\t<element name=\"");
						bufSchema.append(elementname.getName());
						bufSchema.append("\">");						
						
						typestring.flush();
						schemareader.parse(type);
						bufSchema.append(typestring.toString());
						
						bufSchema.append("\n\t\t\t</element>");
						bufSchema.append("\n\t\t</schema>");
					}
					// add to collections
					bufCollections.append("\n\t\t<collection name=\"");
					bufCollections.append(key);
					bufCollections.append("\">");					
					bufCollections.append("\n\t\t\t<step name=\"");
					bufCollections.append(elementname.getName());
					bufCollections.append("\"/>");
					bufCollections.append("\n\t\t</collection>");
					
				} catch (XQueryException e) {
					throw new XMLDBCException("Could not parse view definition", e);
				} catch (SAXException e) {
					throw new XMLDBCException("Could not parse view definition", e);
				} catch (IOException e) {
					throw new XMLDBCException("Could not parse view definition", e);
				} catch (ParseException e) {
					throw new XMLDBCException("Could not parse view definition", e);
				}
			}
		}
		// create metadata string
		buf.setLength(0);
		buf.append("<?xml version='1.0'?>");
		buf.append("\n<xm:metadata");
		buf.append("\n\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
		buf.append("\n\txmlns:xm=\"http://www.xquark.org/XMLDBC/Metadata\"");
		buf.append("\n\tsource=\"");
		buf.append(name);
		buf.append("\">");
		buf.append("\n\t<schemas>\n");
		buf.append(bufSchema.toString());
		buf.append("\n\t</schemas>");
		buf.append("\n\t<collections>");
		buf.append(bufCollections.toString());
		buf.append("\n\t</collections>");
		buf.append("\n</xm:metadata>");

		//Trace.log(this, "buildMetaDataForParser(String name)", "metadata schema :");
		//Trace.log(this, "buildMetaDataForParser(String name)", buf.toString());
		//System.out.println(buf.toString());

		metadata = new MetaDataImpl(null);
		try {
			MetaParser metaparser = new MetaParser(metadata);
			MetaWrapper metawrapper = metaparser.createWrapper(new InputSource(new StringReader(buf.toString())), true);
			metadata.addMetaWrapper(metawrapper);
		} catch (MetadataException e) {
			throw new XMLDBCException("build metadata failed", e);
		}
		return metadata;
	}
    public DBMSInfo getDbmsInfo() {
        return dbmsInfo;
    }

}
