/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.progress;

import java.util.*;

import org.xquark.extractor.algebra.*;
import org.xquark.extractor.runtime.IDProvider;

/**
 * This visitor browse both operands of an outer join and if more than one table
 * involved in the join condition come from the operand (the operand is a join
 * subquery) the operand is renamed and references to original tables are changed.
 * Used by Progress extractor implementation. 
 */
// TODO (YS): this implementation of smiplifyOuterJoin will not work for the case
// where outer join predicate is "like i.a = b.b(+) OR i.c=b.d(+)"
public class SimplifyOuterJoinVisitor extends DefaultCompleteVisitor {

	private static final String RCSRevision = "$Revision: 1.5 $";
	private static final String RCSName = "$Name:  $";

	protected IDProvider _attIDProvider = null;
	protected IDProvider _relIDProvider = null;

	public SimplifyOuterJoinVisitor(IDProvider attIDProvider, IDProvider relIDProvider) {
		_attIDProvider = attIDProvider;
		_relIDProvider = relIDProvider;
	}

	public void visit(BinOpOuterJoin arg) {
		//Trace.enter(this, "visit(BinOpOuterJoin arg)", arg);

		super.visit(arg);
		Set participants = arg.participants();

		Expression lOprnd = arg.getLeftOperand();
		Set lTis = ((Relation) lOprnd).providedTableInstances();
		boolean lOprndTobeSimplyfied = containsManyElementsof(participants, lTis);

		Expression rOprnd = arg.getRightOperand();
		Set rTis = ((Relation) rOprnd).providedTableInstances();
		boolean rOprndTobeSimplyfied = containsManyElementsof(participants, rTis);

		if (lOprndTobeSimplyfied || rOprndTobeSimplyfied) {
			/* create RenameRelation and replace map */
			Map replaceMap = new Hashtable();
			if (lOprndTobeSimplyfied) {
				RenameRelation renameRelation = createRenameRelationProject(lOprnd, replaceMap);
				arg.replaceChild(lOprnd, renameRelation);
			}

			if (rOprndTobeSimplyfied) {
				RenameRelation renameRelation = createRenameRelationProject(rOprnd, replaceMap);
				arg.replaceChild(rOprnd, renameRelation);
			}

			TopDownReplacer replacer = new TopDownReplacer(replaceMap);
			BottomUpTractor tractor = new BottomUpTractor((AlgebraVisitor) replacer);
			arg.accept(tractor);
		} else {
			/* this node is a simple outer join . nothing to do */
		}

		//Trace.exit(this, "visit(BinOpOuterJoin arg)", arg);
	}

	private RenameRelation createRenameRelationProject(Expression relation, Map replaceMap) {
		//Trace.enter(this, "createRenameRelationProject(Expression relation)");

		UnOpProject project = new UnOpProject(relation, _attIDProvider);
		RenameRelation retVal = new RenameRelation(project, _relIDProvider);

		List attributeList = new ArrayList();
		AttributesVisitor visitor = new AttributesVisitor(attributeList);
		relation.accept(visitor);

		AttributeExpression attrExpr = null;
		Expression newExpr = null;
		AttributeExpression newAttrExpr = null;
		for (int i = 0; i < attributeList.size(); i++) {
			attrExpr = (AttributeExpression) attributeList.get(i);
			newExpr = project.addItem(attrExpr); // newAttrExpr maybe a RenameItem

			newAttrExpr = new AttributeExpression(retVal, newExpr.getName());
			newAttrExpr.setUnderlyingExpr(newExpr);

			replaceMap.put(attrExpr, newAttrExpr);
		}

		//Trace.exit(this, "createRenameRelationProject(Expression relation)");
		return retVal;
	}

	//	private RenameRelation searchRenameRelation(Set relations, String name){
	//		RenameRelation retVal = null;
	//		return retVal;
	//	}

	/** 
	 * @return true if colA contains more than one element of colB.
	 */
	private boolean containsManyElementsof(Collection colA, Collection colB) {
		boolean retVal = false;
		int counter = 0;

		Iterator iter = colB.iterator();
		while (iter.hasNext()) {
			Object item = iter.next();
			if (colA.contains(item)) {
				counter++;
			}
		}
		if (1 < counter) {
			retVal = true;
		}

		return retVal;
	}

	class AttributesVisitor extends DefaultSimpleVisitor {

		/* a list of AttributeExpression */
		private List _attributeList = null;

		private RenameRelation _curTI = null;

		public AttributesVisitor(List list) {
			_attributeList = list;
		}

		public void visit(UnOpProject arg) {
			//Trace.enter(this, "visit(UnOpProject arg)");

			List itemList = arg.getItemList();
			Iterator iter = itemList.iterator();
			Expression attribute = null;
			AttributeExpression attrExpr = null;
			while (iter.hasNext()) {
				attribute = (Expression) iter.next();
				attrExpr = new AttributeExpression(_curTI, attribute.getName());
				attrExpr.setUnderlyingExpr(attribute);
				_attributeList.add(attrExpr);
			}

			//Trace.exit(this, "visit(UnOpProject arg)");
		}

		public void visit(Table arg) {
			//Trace.enter(this, "visit(Table arg)");

			/* union of attributes and used key */
			Set attributeSet = new HashSet();
			attributeSet.addAll(arg.getAttributes());
			if (null != arg.getUsedKeyAttributeList()) {
				attributeSet.addAll(arg.getUsedKeyAttributeList());
			}

			/* create AttributeExpressions and add them to _attributeList.*/
			Iterator iter = attributeSet.iterator();
			Attribute attribute = null;
			AttributeExpression attrExpr = null;
			while (iter.hasNext()) {
				attribute = (Attribute) iter.next();
				if (attribute instanceof PseudoAttribute) {
					attrExpr = new Rowid(_curTI, attribute.getName());
				} else {
					attrExpr = new AttributeExpression(_curTI, attribute.getName());
				}
				attrExpr.setUnderlyingExpr(attribute);
				_attributeList.add(attrExpr);
			}

			//Trace.exit(this, "visit(Table arg)");
		}

		public void visit(RenameRelation arg) {
			//Trace.enter(this, "visit(RenameRelation arg)");
			_curTI = arg;
			super.visit((UnaryOperator) arg);
			_curTI = null;
			//Trace.exit(this, "visit(RenameRelation arg)");
		}
	}
}
