/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.runtime;

import java.io.InputStream;
import java.math.BigDecimal;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.extractor.ExtractorConnection;
import org.xquark.extractor.algebra.ExternalVariable;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.sql.PlaceHolderInfo;
import org.xquark.extractor.sql.RuntimeVarInfo;
import org.xquark.extractor.sql.StatementInfo;
import org.xquark.schema.Type;
import org.xquark.xml.xdbc.PreparedXMLStatement;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xquery.parser.*;

/**
 * Extractor XML/DBC prepared statement implementation. Based on the design
 * of split compilation and execution processes in the superclass. 
 */
public class PreparedExtractorStatement extends ExtractorStatement implements PreparedXMLStatement {

	private static Log log = LogFactory.getLog(PreparedExtractorStatement.class);

	private static final byte PARAM_OBJECT = 0;
	private static final byte PARAM_XML = 1;
	private static final byte PARAM_DECIMAL = 2;
	private static final byte PARAM_INPUTSTREAM = 3;
	private static final byte PARAM_BYTES = 4;
	private static final byte PARAM_BOOLEAN = 5;
	private static final byte PARAM_DATE = 6;
	private static final byte PARAM_DOUBLE = 7;
	private static final byte PARAM_FLOAT = 8;
	private static final byte PARAM_LONG = 9;
	private static final byte PARAM_STRING = 10;
    private static final byte PARAM_SEQUENCE = 11;

	/* cacheable structures */
	protected Query _query = null;

	/* execution structures */
	/**
	 * Key : expanded name of the external variable. Value a List RuntimeVarInfo
	 * used to set parameters in the JDBC statements prepared statements
	 */
	protected Map _externalVarMap = new HashMap();

	/**
	 * Key: an SQLExecution node. Value: a List of NodeStatementInfo to execute
	 * at every execution of the XDBC prepared statements
	 */
	protected Map _queryMap = new HashMap();


	/**
	 * Constructor
	 * @param connection
	 * @param stmtId
	 * @param query
	 * @throws XMLDBCException
	 */
	public PreparedExtractorStatement(ExtractorConnection connection, int stmtId, QueryFactory queryFactory, MetaDataManager metadataManager, String query, Map cachedStatements) throws XMLDBCException {
		super(connection, stmtId, queryFactory, metadataManager);
		if (cachedStatements != null) {
			_query = (Query) cachedStatements.get(query);
		}
		if (_query == null) {
			_query = queryFactory.createCompiledQuery(query);
			if (cachedStatements != null) {
				cachedStatements.put(query, _query);
			}
		}
		// create execution structures
		try {
			prepareJDBCStatements();
		} catch (SQLException e) {
			throw new XMLDBCException(MessageLibrary.getMessage("E_ERR_SQL", e.getMessage()));
		}
	}

	public void close() throws XMLDBCException {
		if (!_isClosed) {
			// browse compiled query to delete temp tables, views and close prepared statements
			CompiledQueryTreeExecutor executor = new CompiledQueryTreeExecutor() {

				public void execute(Query.SQLExecutionInfo sInfo) throws SQLException {
						// delete temp tables and views
					if (_stmt == null)
						_stmt = createJdbcStatement();

					if (sInfo.sql != null)
						sInfo.sql.cleanAll(_stmt);

					// close prepared statements
					NodeStatementInfo nsi = getPreparedStatements(sInfo);
					nsi.mainQueryStmt.close();
					Iterator it = null;
					if (nsi.firstExecStmts != null) {
						it = nsi.firstExecStmts.iterator();
						while (it.hasNext())
							 ((PreparedStatement) it.next()).close();
					}
					if (nsi.nextExecStmts != null) {
						it = nsi.nextExecStmts.iterator();
						while (it.hasNext())
							 ((PreparedStatement) it.next()).close();
					}
				}
			};
			try {
				browseCompiledQuery(executor);
			} catch (SQLException e) { /* no op */
			}

			_externalVarMap.clear();
			_queryMap.clear();

			super.close();
		}
	}

	private NodeStatementInfo getPreparedStatements(Query.SQLExecutionInfo sInfo) {
		return (NodeStatementInfo) _queryMap.get(sInfo);
	}

	public PreparedStatement createPreparedJdbcStatement(String query) throws SQLException {
		PreparedStatement retVal = null;
		if (_connection.getExtractor().useScrollableResultSets())
			retVal = _connection.getJdbcConnection().prepareStatement(query, ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_READ_ONLY);
		else
			retVal = _connection.getJdbcConnection().prepareStatement(query, ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);

		return retVal;
	}

	/*********************************************************************/
	/*                           PARAMETER SETTERS                       */
	/*********************************************************************/
	private void setVariable(String ns, String varName, Object value, byte valueType) throws XMLDBCException {
		HashMap varMap = _query.getModule().getVariables();
		if (varMap == null)
			throw new XMLDBCException("Variable $" + ((ns==null)?"":"{" + ns + "}") + varName + " is not defined.");
		varMap = (HashMap)varMap.get(ns);
		if (varMap == null)
			throw new XMLDBCException("Variable $" + ((ns==null)?"":"{" + ns + "}") + varName + " is not defined.");
		Variable var = (Variable)varMap.get(varName);
		if (var == null)
			throw new XMLDBCException("Variable $" + ((ns==null)?"":"{" + ns + "}") + varName + " is not defined.");

		List placeHolders = (List) _externalVarMap.get(ExternalVariable.getExpandedName(ns, varName));
		if (placeHolders == null)
			return;

		RuntimeVarInfo varInfo = null;
		for (int i = 0; i < placeHolders.size(); i++) {
			varInfo = (RuntimeVarInfo) placeHolders.get(i);
			try {
				switch (valueType) {
					case PARAM_OBJECT :
						varInfo.setObject(value);
						break;
					case PARAM_XML :
						varInfo.setParameter((String) value);
						break;
					case PARAM_DECIMAL :
						varInfo.setBigDecimal((BigDecimal) value);
						break;
					case PARAM_INPUTSTREAM :
						StreamInfo sInfo = (StreamInfo) value;
						varInfo.setInputStream(sInfo.stream, sInfo.length);
						break;
					case PARAM_BYTES :
						varInfo.setBytes((byte[]) value);
						break;
					case PARAM_BOOLEAN :
						varInfo.setBoolean(((Boolean) value).booleanValue());
						break;
					case PARAM_DATE :
						varInfo.setDate((Date) value);
						break;
					case PARAM_DOUBLE :
						varInfo.setDouble(((Double) value).doubleValue());
						break;
					case PARAM_FLOAT :
						varInfo.setFloat(((Float) value).floatValue());
						break;
					case PARAM_LONG :
						varInfo.setLong(((Long) value).longValue());
						break;
					case PARAM_STRING :
						varInfo.setString((String) value);
						break;
//					case PARAM_SEQUENCE :
//						varInfo.setSequence((ArrayList) value);
//						break;
					default :
						}

			} catch (Exception e) {
				throw new XMLDBCException(MessageLibrary.getMessage("ST_VAR_SETTING"), e);
			}
		}
	}

	public void setExternalVariable(String ns, String varName, String value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_XML);
	}

	public void setExternalVariable(String varName, String value) throws XMLDBCException {
		setExternalVariable(null, varName, value);
	}

	public void setObject(String ns, String varName, Object value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_OBJECT);
	}

	public void setObject(String varName, Object value) throws XMLDBCException {
		setObject(null, varName, value);
	}

	public void setBigDecimal(String ns, String varName, BigDecimal value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_DECIMAL);
	}

	public void setBigDecimal(String varName, BigDecimal value) throws XMLDBCException {
		setBigDecimal(null, varName, value);
	}

	protected class StreamInfo {
		InputStream stream;
		int length;

		public StreamInfo(InputStream stream, int length) {
			this.stream = stream;
			this.length = length;
		}
	}

	public void setInputStream(String ns, String varName, InputStream value, int length) throws XMLDBCException {
		setVariable(ns, varName, new StreamInfo(value, length), PARAM_INPUTSTREAM);
	}

	public void setInputStream(String varName, InputStream value, int length) throws XMLDBCException {
		setInputStream(null, varName, value, length);
	}

	public void setBytes(String ns, String varName, byte[] value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_BYTES);
	}

	public void setBytes(String varName, byte[] value) throws XMLDBCException {
		setBytes(null, varName, value);
	}

	public void setBoolean(String ns, String varName, boolean value) throws XMLDBCException {
		setVariable(ns, varName, new Boolean(value), PARAM_BOOLEAN);
	}

	public void setBoolean(String varName, boolean value) throws XMLDBCException {
		setBoolean(null, varName, value);
	}

	public void setDate(String ns, String varName, Date value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_DATE);
	}

	public void setDate(String varName, Date value) throws XMLDBCException {
		setDate(null, varName, value);
	}

	public void setDouble(String ns, String varName, double value) throws XMLDBCException {
		setVariable(ns, varName, new Double(value), PARAM_DOUBLE);
	}

	public void setDouble(String varName, double value) throws XMLDBCException {
		setDouble(null, varName, value);
	}

	public void setFloat(String ns, String varName, float value) throws XMLDBCException {
		setVariable(ns, varName, new Float(value), PARAM_FLOAT);
	}

	public void setFloat(String varName, float value) throws XMLDBCException {
		setFloat(null, varName, value);
	}

	public void setLong(String ns, String varName, long value) throws XMLDBCException {
		setVariable(ns, varName, new Long(value), PARAM_LONG);
	}

	public void setLong(String varName, long value) throws XMLDBCException {
		setLong(null, varName, value);

	}

	public void setString(String ns, String varName, String value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_STRING);
	}

	public void setString(String varName, String value) throws XMLDBCException {
		setString(null, varName, value);
	}

    public void setSequence(String ns, String varName, List value) throws XMLDBCException {
		setVariable(ns, varName, value, PARAM_SEQUENCE);
    }

    public void setSequence(String varName, List value) throws XMLDBCException {
        setSequence(null, varName, value);
    }

    public void clearVariables() throws XMLDBCException {
		_externalVarMap.clear();
	}

	/*********************************************************************/
	/*                             EXECUTION                             */
	/*********************************************************************/
	public boolean execute() throws XMLDBCException {
		if (_currentResultSet != null)
			_currentResultSet.close();
		_currentResultSet = runQuery(_query.getCompiledQuery());
		return true;
	}

	public XMLResultSet executeQuery() throws XMLDBCException {
		execute();
		return _currentResultSet;
	}

	protected SynchronizedResultSet executeSQL(Query.SQLExecutionInfo sqlInfo) throws XQueryException, XMLDBCException {
		SynchronizedResultSet retVal = null;
		NodeStatementInfo pStmts = (NodeStatementInfo) _queryMap.get(sqlInfo);

		try {
			// TODO: check that all param were set ?

			// Execute create or reset statements (empty temp tables)
			if (pStmts.firstExecStmts != null) {
				if (pStmts.firstPerformed) { // resets
					for (int i = 0; i < pStmts.nextExecStmts.size(); i++) {
						((PreparedStatement) pStmts.nextExecStmts.get(i)).executeUpdate();
					}
				} else { // create
					for (int i = 0; i < pStmts.firstExecStmts.size(); i++) {
						((PreparedStatement) pStmts.firstExecStmts.get(i)).executeUpdate();
					}
					pStmts.firstPerformed = true;
				}
			}

			// create result set
			retVal = new SynchronizedResultSet(pStmts.mainQueryStmt, sqlInfo.mapper, _metadataManager.getDbmsInfo());
		} catch (XMLDBCException ex) {
			throw ex;
		} catch (SQLException ex) {
			throw new XMLDBCException(ex.getMessage(), ex);
		}

		return retVal;
	}

	private void prepareJDBCStatements() throws SQLException {
		CompiledQueryTreeExecutor executor = new CompiledQueryTreeExecutor() {

			public void execute(Query.SQLExecutionInfo sInfo) throws SQLException {
				List requestList = sInfo.sql.getRequestList();

				if (log.isDebugEnabled()) {
					log.debug("=================< SQL >=================");
					for (int i = 0; i < requestList.size(); i++)
						log.debug(((StatementInfo) requestList.get(i)).sStmt);
				}

				//Trace.log("================< Mapper >================");
				//Trace.log(mapper.pprint());

				// Add an item in query map
				NodeStatementInfo pStmts = new NodeStatementInfo();
				_queryMap.put(sInfo, pStmts);

				// create "init" prepared statements
				StatementInfo stmtInfo = null;
				PreparedStatement pStmt = null;
				if (requestList.size() > 1)
					pStmts.firstExecStmts = new ArrayList();
				for (int i = 0; i < requestList.size() - 1; i++) {
					stmtInfo = (StatementInfo) requestList.get(i);
					pStmt = createPreparedJdbcStatement(stmtInfo.sStmt);
					pStmts.firstExecStmts.add(pStmt);
					registerPlaceHolders(stmtInfo.parameterList, pStmt);
				}

				// create "main" prepared statements
				stmtInfo = (StatementInfo) requestList.get(requestList.size() - 1);
				pStmt = createPreparedJdbcStatement(stmtInfo.sStmt);
				pStmts.mainQueryStmt = pStmt;
				registerPlaceHolders(stmtInfo.parameterList, pStmt);

				// create "reset" prepared statements
				requestList = sInfo.sql.getResetStatementList();

				if (log.isDebugEnabled()) {
					log.debug("==============< SQL Reset >==============");
					for (int i = 0; i < requestList.size(); i++)
						log.debug(((StatementInfo) requestList.get(i)).sStmt);
				}

				if (!requestList.isEmpty())
					pStmts.nextExecStmts = new ArrayList();
				for (int i = 0; i < requestList.size(); i++) {
					stmtInfo = (StatementInfo) requestList.get(i);
					pStmt = createPreparedJdbcStatement(stmtInfo.sStmt);
					pStmts.nextExecStmts.add(pStmt);
					registerPlaceHolders(stmtInfo.parameterList, pStmt);
				}
			}
		};
		browseCompiledQuery(executor);
	}

	private void registerPlaceHolders(List placeholders, PreparedStatement pStmt) {
		for (int i = 0; i < placeholders.size(); i++) {
			registerPlaceHolder((PlaceHolderInfo) placeholders.get(i), pStmt);
		}
	}

	private void registerPlaceHolder(PlaceHolderInfo placeholder, PreparedStatement pStmt) {
		List runTimeVarInfos = (List) _externalVarMap.get(placeholder.varId);
		if (runTimeVarInfos == null) {
			runTimeVarInfos = new ArrayList();
			_externalVarMap.put(placeholder.varId, runTimeVarInfos);
		}
		runTimeVarInfos.add(new RuntimeVarInfo(placeholder, pStmt));
	}

	/*********************************************************************/
	/*                             UTILITIES                             */
	/*********************************************************************/
	protected class NodeStatementInfo {
		boolean firstPerformed = false;
		List firstExecStmts = null;
		List nextExecStmts = null;
		PreparedStatement mainQueryStmt = null;
	}

	protected void browseCompiledQuery(CompiledQueryTreeExecutor executor) throws SQLException {
		Iterator it = _query.getCompiledQuery().iterator();
		while (it.hasNext())
			browseQueryTree(executor, ((Query.CompiledExpression) it.next()).sqlInfoTree);
	}

	protected void browseQueryTree(CompiledQueryTreeExecutor executor, Query.SQLExecutionInfo sInfo) throws SQLException {
		// in the case there is no Qdb
		if (sInfo == null)
			return;
			
		if (sInfo.sql != null)
			executor.execute(sInfo);

		// recursive call
		if (sInfo.childrenQueries != null) {
			Iterator it = sInfo.childrenQueries.iterator();
			while (it.hasNext())
				browseQueryTree(executor, (Query.SQLExecutionInfo) it.next());
		}
	}

	protected interface CompiledQueryTreeExecutor {
		void execute(Query.SQLExecutionInfo sInfo) throws SQLException;
	}

	// add LARS 24/05/04
	/**
	 * Gets the array of parameter names of the statement
	 */
	public javax.xml.namespace.QName[] getParameterNames() {
		ArrayList names = new ArrayList();
		XQueryModule module = _query.getModule();
		if (module.getDeclList() != null) {
			for (int i = 0; i < module.getDeclList().size(); i++) {
				QName name = (QName) module.getDeclList().get(i);
				HashMap map = null;
				if (module.getVariables() != null) {
					map = (HashMap) module.getVariables().get(name.getNameSpace());
					if (map != null) {
						Variable vari = (Variable) map.get(name.getLocalName());
						if (vari != null && vari.getExpression() == null) {
							names.add(new javax.xml.namespace.QName(vari.getVarName().getNameSpace(),vari.getVarName().getLocalName()));
						}
					}
				}
			}
		}
		javax.xml.namespace.QName qnames[] = new javax.xml.namespace.QName[names.size()];
		return (javax.xml.namespace.QName[]) names.toArray(qnames);
	}
	/**
	 * Gets the base schema type of a particular parameter given it's name
	 */
	public javax.xml.namespace.QName getParameterType(javax.xml.namespace.QName varName) {
		XQueryModule module = _query.getModule();
		org.xquark.xquery.parser.ExternalVariable extVar = (org.xquark.xquery.parser.ExternalVariable) module.getExternalVariables().get(varName);
		Type type = extVar.getQType().getSimpleType().getBaseType();
		return new javax.xml.namespace.QName(type.getNamespace(), type.getName());
	}

}
