/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.fusion;

import java.io.*;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.parsers.SAXParserFactory;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.xml.sax.SAXException;
import org.xquark.mediator.Mediator;
import org.xquark.serialize.XMLSerializer;
import org.xquark.xml.xdbc.*;

/**
 * This class is the command-line launcher for Fusion. It is mostly useful
 * for test and demo purposes, as the JVM and connection setup overhead
 * are prohibitive in most cases.
 */
public class Main {
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name:  $";

	private static final String HELP_OPT1 = "-help";
	private static final String HELP_OPT2 = "-?";
	private static final String CONF_OPT = "-conf";
	private static final String FILE_OPT1 = "-file";
	private static final String FILE_OPT2 = "-f";
	private static final String DEBUG_OPT = "-debug";

	private static SAXParserFactory parserFactory;

	static {
		parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(true);
	}

	private String configurationFile;
	private Mediator mediator = null;

	/** The constructor for the XQFusion object. 
	 */
	public Main() {
	}
	
	/** The constructor for the XQFusion object. 
	 * @param confFile the XQFusion configuration file
	 */
	public Main(String confFile) {
		if (confFile == null)
			throw new NullPointerException("Illegal null value for configuration file");
		configurationFile = confFile;
	}
	
	private Mediator getMediator() throws XMLDBCException {
		if (mediator == null) {
			if (configurationFile != null) {
				mediator = new Mediator(configurationFile);
			} else {
				mediator = new Mediator();
			}
		}
		return mediator;
	}
	
	public XMLConnection getConnection() throws XMLDBCException {
		return getMediator().getConnection();
	}

	/* This part of the file deals with command-line execution */

	/**
	 * The method called from the command line to run Fusion as a standalone program.
	 * Use the -help option to display a list of available options.
	 * @param args the command line parameters
	 */
	public static void main(String[] args) {
		String configFile = null;
		String batchFile = null;
		ArrayList parameters = new ArrayList();

		for (int i = 0; i < args.length; i++) {
			if (!args[i].startsWith("-")) {
				// not an option
				parameters.add(args[i]);
				
			} else {
				// Check options are in first position
				if (parameters.size() > 0) {
					usage();
					System.err.println("Options must appear before the command parameters.");
					System.exit(-1);
				}
				// one-parameter option
				if (HELP_OPT1.equals(args[i]) || HELP_OPT2.equals(args[i])) {
					usage();
					System.exit(0);
				} else {
					// two-parameter option
					if (i + 1 == args.length) {
						System.err.println("Missing value for option " + args[i]);
						usage();
						System.exit(-1);
					} else if (DEBUG_OPT.equals(args[i])) {
						Logger.getRootLogger().setLevel(Level.DEBUG);
					} else if (CONF_OPT.equals(args[i])) {
						configFile = args[++i];
					} else if (FILE_OPT1.equals(args[i]) || FILE_OPT2.equals(args[i])) {
						batchFile = args[++i];
					} else {
						// Unknown option
						System.err.println("Invalid option " + args[i]);
						usage();
						System.exit(-1);
					}
				}
			}
		}

		List parametersList = null;
		if (batchFile != null) {
			parametersList = parseBatchFile(batchFile);
		} else {
			checkParameters(parameters);
			parametersList = new ArrayList();
			parametersList.add(parameters);
		}

		Main fusion = null;
		if (configFile != null) {
			try {
				configFile = new File(configFile).toURL().toString();
				fusion = new Main(configFile);
			} catch (MalformedURLException ex) {
				System.err.println("Invalid configuration file specification.");
				printError(ex);
				System.exit(-1);
			}
		} else {
			fusion = new Main();
		}
		
//		try {
//			PrintWriter writer = new PrintWriter(System.out, true);
//			Iterator it = fusion.getMediator().getWrappers().values().iterator();
//			while (it.hasNext()) {
//				XMLDataSource s = ((XDBCWrapper) it.next()).getDataSource();
//				s.setLogWriter(writer);
//			}
//		} catch (Exception e) {
//			System.err.println("Error in trace.");
//			printError(e);
//			System.exit(-1);
//		}

		int result = fusion.executeQuery(parametersList);
		System.exit(result);
	}

	private static void checkParameters(List params) {
		int size = params.size();
		if (size < 1 || size > 2) {
			System.err.println("Wrong number of parameters.");
			System.exit(-1);
		}
	}

	private static List parseBatchFile(String batchFile) {
		ArrayList result = new ArrayList();
		BufferedReader reader = null;
		try {
			File f = new File(batchFile);
			reader = new BufferedReader(new FileReader(f));
		} catch (IOException ex) {
			System.err.println("Could not find specified batch file");
			System.exit(-1);
		}
		try {
			while (true) {
				String line = reader.readLine();
				if (line == null)
					break;
				if (line.startsWith("#") || line.trim().length() == 0)
					continue;
				java.util.StringTokenizer it = new java.util.StringTokenizer(line);
				ArrayList params = new ArrayList();
				while (it.hasMoreTokens()) {
					params.add(it.nextToken());
				}
				checkParameters(params);
				result.add(params);
			}
		} catch (IOException e) {
			System.err.println("Error while reading specified batch file");
			System.exit(-1);
		}
		return result;
	}

	private static String readQueryFile(String fileName) {
		try {
			StringBuffer retVal = new StringBuffer();
			String tmpStr;
			BufferedReader in = new BufferedReader(new FileReader(fileName));
			while ((tmpStr = in.readLine()) != null) {
				if (retVal.length() > 0)
					retVal.append('\n');
				retVal.append(tmpStr);
			}
			in.close();
			return retVal.toString();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not find query file " + fileName);
		} catch (IOException ex) {
			System.err.println("Error while reading query file " + fileName);
		}
		return null;
	}

	private int executeQuery(List parametersList) {
		XMLConnection connection = null;		
		XMLStatement stat = null;
		int result = 0;
		try {
            connection = getConnection();
			stat = connection.createStatement();
			Iterator it = parametersList.iterator();
			while (it.hasNext()) {
				List params = (List) it.next();
				String queryFile = (String) params.get(0);
				String resultFile = null;
				if (params.size() > 1)
					resultFile = (String) params.get(1);
				String query = readQueryFile(queryFile);
				if (query != null) {
					OutputStream out = System.out;
					try {
						if (resultFile != null) {
							out = new BufferedOutputStream(new FileOutputStream(new File(resultFile)));
						}
					} catch (IOException ex) {
						out = null;
						System.err.println("Could not open result file " + resultFile);
						result = -1;
					}
					if (out != null) {
						try {
							stat.setBaseURI(new File(queryFile).toURL().toString());
							XMLResultSet rs = stat.executeQuery(query);
							XMLSerializer handler = new XMLSerializer(out);
							rs.setContentHandler(handler);
							handler.startDocument();
							while (rs.hasNext())
								rs.nextAsSAX();
							handler.endDocument();
							rs.close();
							if (out != System.out)
								out.close();
						} catch (SAXException ex) {
							System.err.println("Error while processing query result " + queryFile);
							printError(ex);
							result = -1;
						} catch (IOException ex) {
							System.err.println("Error while creating query result " + queryFile);
							printError(ex);
							result = -1;
						} catch (XMLDBCException ex) {
							System.err.println("Error while executing query " + queryFile);
							printError(ex);
							result = -1;
						}
					}
				}
			}
        } catch (XMLDBCException e) {
			System.err.println("Could not create connection and statement");
			printError(e);
			System.exit(-1);
        } finally {
			try {
				if (stat != null) stat.close();
				if (connection != null) connection.close();
			} catch (XMLDBCException e) {
				System.err.println("Could not close connection and statement");
				printError(e);
				System.exit(-1);
			}
        }
        return result;
	}

	/**
	 * Print an error message to the standard error output.
	 */
	private static void printError(Throwable e) {
		XMLDBCException.printError(e, Logger.getRootLogger().isDebugEnabled());
	}

	private static void usage() {
		System.out.println("Usage: java org.xquark.fusion.Main <options> [<file1> [<file2>]]");
		System.out.println();
		System.out.println("file1:           XQuery file, required unless the -file option is used");
		System.out.println("file2:           optional output file, output to stdout by default");
		System.out.println();
		System.out.println("Options:");
		System.out.println("-?, -help:       this message");
		System.out.println("-conf conffile:  configuration file, required when accessing data sources.");
		System.out.println("                 It can be omitted when querying only documents.");
		System.out.println("-f file:         file containing a list of files to be processed.");
		System.out.println("                 Each line of the file has one query file and an optional output file.");
		System.out.println("                 Lines starting with # are comments");
	}

}
