/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2004 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.DOMUtils;

import java.util.Iterator;

import org.w3c.dom.*;
import org.xml.sax.SAXException;
import org.xquark.schema.*;
import org.xquark.schema.validation.PSVInfoSetProvider;
import org.xquark.util.DOMReader;
import org.xquark.xpath.datamodel.TypedAttribute;
import org.xquark.xpath.datamodel.TypedElement;
import org.xquark.xpath.datamodel.TypedNode;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.typing.QAtomicSerializer;


/**
 * Define a typed DOM-tree walker that produces SAX events.
 *
 * <p><B> WARNING : DOM must support namespaces !!!</p>
 * <p>Special Feature:</p>
 * <code>http://wwww.xquark.org/sax/features/fragment</code> disables SAX
 * startDocument() and endDocument() events triggering
 *
 * @see org.xml.sax.XMLReader, org.w3c.dom.Element
 */
public class TypedDOMReader extends DOMReader {
	private static final String RCSRevision = "$Revision: 1.7 $";
	private static final String RCSName     = "$Name:  $";

	protected PSVInfoSetProvider psvisp = null;
	protected QAtomicSerializer qAtomicSerializer = null;

	public TypedDOMReader() {
		super();
	}
    
	public TypedDOMReader(Document doc) {
		super(doc);
	}
    
	public TypedDOMReader(Element root) {
		super(root);
	}

	public void setPSVInfoSetProvider(PSVInfoSetProvider psvisp) {
		this.psvisp = psvisp;
	}

	public void setQAtomicSerializer(QAtomicSerializer qAtomicSerializer) {
		this.qAtomicSerializer = qAtomicSerializer;
	}
    
	protected void generateText(Node node) throws SAXException {
		if (psvisp != null && node.getNodeValue() != null && node instanceof TypedNode) {
			psvisp.setActualValue(-1,((TypedNode)node).getTypedValue());
			psvisp.setNormalizedValue(-1,((TypedNode)node).getNormalizedStringValue());
		}
		//super.generateText(node);
		String value = null;
		if (qAtomicSerializer != null) {
			Object typedValue = ((TypedNode)node).getTypedValue();
			try {
				value = qAtomicSerializer.serialize((SimpleType)((TypedNode)node).getType(), typedValue);
			} catch (XQueryException e) {
				throw new SAXException(e);
			}
		}
		if (value == null) {
			value = node.getNodeValue();
		}
		if (value != null) {
			contentHandler.characters(value.toCharArray(), 0, value.length());
		}
	}


	protected void generateStartElement(Node node) throws SAXException {
		
		// TODO do it using inheritance
		String uri = node.getNamespaceURI();
		if (uri == null) uri = "";
		String local = node.getLocalName();
		if (local == null) local = "";
		String prefix = node.getPrefix();
		String qname = prefix == null ? local : prefix + ":" +local;
		
		wAtts.setDOMAttributes(node.getAttributes());
		Iterator it = wAtts.getPrefixIterator(); // OPTIM SR 23/05/2001
		if (it != null)
			while (it.hasNext()) {
				Attr a = (Attr)it.next();
				contentHandler.startPrefixMapping(
				a.getLocalName().equals(XMLNS_PREFIX) ? "": a.getLocalName(), // default prefix
				a.getValue());
			}
			
		if (psvisp != null) {
			psvisp.pushElementInfoset(uri, local);
			if (node instanceof TypedElement) {
				TypedElement elt = (TypedElement)node;
				psvisp.initElementValidationInfo((ElementDeclaration)elt.getDeclaration(),elt.getType(),false,SchemaConstants.LAX);
				for (int i = 0; i < wAtts.getLength(); i++) {
					String namespace = wAtts.getURI(i);
					String localName = wAtts.getLocalName(i);
					String attrValue = wAtts.getValue(i);
					TypedAttribute attr = (TypedAttribute)elt.getAttributeNodeNS(namespace, localName);
					int index = psvisp.addAttributePSVI(namespace, localName, (AttributeDeclaration)attr.getDeclaration(), true);
					psvisp.setActualValue(index, attr.getTypedValue());
					psvisp.setNormalizedValue(index, attr.getNormalizedStringValue());
				}
			}
		}
		
		contentHandler.startElement(uri, local, qname, wAtts);
 
//		super.generateStartElement(node);
	}
    
	protected void generateEndElement(Node node) throws SAXException {
		super.generateEndElement(node);
		if (psvisp != null) {
			psvisp.popElementInfoset();
		}
	}

}


