/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator;

import java.io.PrintWriter;
import java.net.MalformedURLException;
import java.util.*;

import org.apache.commons.collections.map.LRUMap;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.mediator.runtime.MediatorConfiguration;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.*;

/** 
 * This class represent the mediator as a source that can be accessed by
 * xdbc.
 * 
 */
public class Mediator implements XMLDataSource {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.18 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************
	/** 
	 * A hashmap for wrappers.
	 * Key (type : String) : xdbc String
	 * Value (type : SubAccessor) : a class encapsuling informations and 
	 * 				connection to a wrapper
	 */
	private HashMap wrappers = new HashMap();
	private MetaDataImpl metadata = null ;
	private Map cachedStatements = Collections.synchronizedMap(new LRUMap(100));
	private String name = null;
	private String configURI = null;
	private PrintWriter logWriter = null;
	private ClassLoader classLoader = null;

	// **********************************************************************
	// * CONSTRUCTOR
	// **********************************************************************

	public Mediator() {
	}

	/**
	 * Initialize the mediator with a configuration file URI where wrappers
	 * location are defined.
	 */
	public Mediator(String configURI) throws XMLDBCException {
		this.configURI = configURI;
		Properties configuration = null;
		try {
			configuration = MediatorConfiguration.loadConfiguration(configURI);
		} catch (MalformedURLException e) {
			throw new XMLDBCException("Invalid XML/DBC URI "+MediatorDriver.MEDIATOR_URL_PREFIX+configURI, e); 
		}
		name = configuration.getProperty(MediatorConfiguration.KEY_NAME);
		List datasources = (List)configuration.get(MediatorConfiguration.KEY_SUBACCESSORS);
		for (int i = 0; i < datasources.size(); i++) {
			addXDBCWrapper((XDBCWrapper)datasources.get(i));
		}
	}

	public String getName() {
		return name;
	}
	
	public String getURL() {
		if (configURI == null) return null;
 		else return MediatorDriver.MEDIATOR_URL_PREFIX+configURI;
	}
	
	/**
	 * Add a new Subaccessor in the mediator hashmap.
	 * @param name the name identifing the subaccessor
	 * @param driver the driver used for the subaccessor
	 * @param connection the connection string given to the accessor
	 */
	public synchronized void addXDBCWrapper(XDBCWrapper wrapper) throws XMLDBCException {
		wrapper.setMediator(this);
		wrappers.put(wrapper.getName(), wrapper);
	}

	public void addXDBCWrapper(String name, XMLDataSource ds) throws XMLDBCException {
		addXDBCWrapper(new XDBCWrapper(name, ds));
	}

	public void addXDBCWrapper(String name, String driver, String uri) throws XMLDBCException {
		addXDBCWrapper(new XDBCWrapper(name, driver, uri));
	}

	public void addXDBCWrapper(String name, String driver, String uri, String login, String password) throws XMLDBCException {
		addXDBCWrapper(new XDBCWrapper(name, driver, uri, login, password));
	}

	/** 
	 *
	 */
	public synchronized Map getWrappers() {
		return Collections.unmodifiableMap(wrappers);
	}
	
	public synchronized XDBCWrapper getWrapper(String name) {
		return (XDBCWrapper) wrappers.get(name);
	}

	protected synchronized MetaDataImpl getDataSourceMetaData(boolean refresh) throws XMLDBCException {
		if (metadata == null || refresh) {
			metadata = new MetaDataImpl(null);
	 		if (wrappers != null) {
				MetaParser parser = new MetaParser(metadata);
				XDBCWrapper subi = null;
				XMLConnection coni = null;
				for (Iterator it = wrappers.values().iterator(); it.hasNext();) {
					try {
						subi = (XDBCWrapper) it.next();
						coni = null;
						XMLDataSourceMetaData metai = null;
						coni = subi.getConnection();
						metai = coni.getMetaData(true);
						MetaWrapper metawrapper = parser.createWrapper(true);
						XMLDocument doc = metai.getMetaData();
						doc.setContentHandler(parser);
						doc.getAsSAX();
						metadata.addMetaWrapper(metawrapper, subi.getName());
					} catch (XMLDBCException ex) {
						throw new XMLDBCException("Error while accessing metadata for "+subi.getName(), ex);						
					} catch (MetadataException e) {
						throw new XMLDBCException("Error while accessing metadata for "+subi.getName(), e);						
                    } catch (SAXException e) {
						throw new XMLDBCException("Error while parsing metadata for "+subi.getName(), e);						
                    } finally {
                    	if (coni != null) {
                    		coni.close();
                    	}
                    }
				}
	 		}
	 		cachedStatements.clear();
		}
		return metadata ;
	}

	// **********************************************************************
	// * XMLDATASOURCE IMPLEMENTATION
	// **********************************************************************
	/** 
	 * Attempts to establish a connection with the data source.
	 * @throws XMLDBCException If a data source access error occurs
	 * @return an XMLConnection object that represents a connection to the data source
	 */
	public XMLConnection getConnection() throws XMLDBCException {
		return new MediatorConnection(this);
	}

	/** Attempts to establish a connection with the data source.
	 * @param user the data source registered user
	 * @param password the user's password
	 * @throws XMLDBCException If a data source access error occurs
	 * @return an XMLConnection object that represents a connection to the data source
	 */
	public XMLConnection getConnection(String user, String password) throws XMLDBCException {
		throw new XMLDBCException("Unsupported operation getConnection(user, password). Use getConnection() instead.");
	}

	/** Sets the logging/tracing java.io.PrintWriter object that is used by the XDataSource object.
	 * @param writer the new logging/tracing java.io.PrintWriter object, or null to disable logging and tracing
	 */
	public void setLogWriter(PrintWriter writer) {
		logWriter = writer;
	}

	/** Retrieves the log writer currently used by the XDataSource object to print messages.
	 * @return the java.io.PrintWriter object used to log data source messages, or null if none is defined
	 */
	public PrintWriter getLogWriter() {
		return logWriter;
	}

	/** Sets the maximum time in seconds that the object will wait when attempting to log in to a data source. Setting the timeout to 0 will disable it.
	 * @param seconds the login time limit in seconds
	 */
	public void setLoginTimeout(int seconds) {}

	/** Gets the maximum time in seconds that the object can wait when attempting to log in to a data source
	 * @return the data source login timeout in seconds, 0 if no timeout is defined
	 */
	public int getLoginTimeout() {
		return -1;
	}
	
	Map getCachedStatements() {
		return cachedStatements;
	}
	
	/**
	 * Set the class loader used for loading drivers, mappings and user generators.
	 * @param loader ClassLoader 
	 */
	public void setClassLoader(ClassLoader loader) {
		classLoader = loader;
	}

	/**
	 * Returns the class loader used for loading drivers, mappings and user generators.
	 */
	public ClassLoader getClassLoader() {
		return classLoader != null ? classLoader : this.getClass().getClassLoader();
	}

	/**
	 * Loads an XML schema in memory. This is useful for bypassing the automatic schema loading based
	 * upon the schemaLocation attribute (e.g. if there is no such attribute).
	 */
	public void loadSchema(InputSource source) throws SAXException, XMLDBCException {
		if (metadata == null)
			getDataSourceMetaData(false);
		metadata.getSchemaManager().loadSchema(source);
	}

}
