/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator;

import java.io.IOException;
import java.net.URL;
import java.util.*;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.mediator.runtime.*;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.MetaDataImpl;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.xdbc.XMLDataSourceMetaDataConnectionWrapper;

/**
 * This class implements the interface that represents a connection with
 * a specific XML data source.  Within the context of an XMLConnection,
 * XQuery or XPath statements are executed and XML results are returned.
 * <BR>
 * This interface is used by both read-only connections and storage
 * connections, which require access to a larger set of interfaces, such
 * as interfaces representing XML collections and documents.
 * Classes providing read-only access to XML data sources can avoid
 * implementing the optional interfaces for XML storage management,
 * by deriving from the abstract class DefautReadOnlyXMLConnection.
 *
 */
public class MediatorConnection extends DefaultReadOnlyXMLConnection implements XMLConnection {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.27 $";
	private static final String RCSName = "$Name:  $";	

	private Mediator mediator = null;
	private XMLDataSourceMetaData metaData = null;
	private Map wrapperConnections = Collections.synchronizedMap(new HashMap());
	private boolean isClosed = false ;
	
	private String sourceName = null;
	private String baseURI = null;
	
	/* a list all active XMLStatement that this XMLConnection has created */
	private List activeStatementList = new ArrayList();
	
	// ************************************************************************
	// * INITIALISATION
	// ************************************************************************
	MediatorConnection(Mediator mediator) {
		this.mediator = mediator;
	}
	
	String getSourceName() {
		return mediator.getName();
	}
	
	// ************************************************************************
	// * XMLCONNECTION IMPLEMENTATION
	// ************************************************************************

	public Mediator getMediator() {
		return mediator;	
	}
	
	/*********************************************************************/
	/*************************** Connection management *******************/
	/*********************************************************************/
	/**
	 * To close the current data source connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() throws XMLDBCException {
		if (!isClosed) {
			isClosed = true ;
			XMLStatement statement;
			for (int i = activeStatementList.size() - 1; i > -1; i--) {
				statement = (XMLStatement) activeStatementList.get(i);
				statement.close();
			}
			Iterator it = wrapperConnections.values().iterator();
			while (it.hasNext()) {
				((XMLConnection)it.next()).close();
			}
		}
	}
	
	/**
	 * To know the connection state (opened or closed).
	 * @return true if the current connection is closed, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public boolean isClosed() throws XMLDBCException {
		return isClosed ;
	}
	
	/**
	 * Gets the metadata regarding this connection's data source. A
	 * connection's data source is able to provide information
	 * describing itself.
	 * @return an XMLDataSourceMetaData object for this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDataSourceMetaData getMetaData() throws XMLDBCException {
		return getMetaData(false);
	}

	/**
	 * Gets the metadata regarding this connection's data source. A
	 * connection's data source is able to provide information
	 * describing itself.
	 * @param refresh when it is true, reload metadata.
	 * @return an XMLDataSourceMetaData object for this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDataSourceMetaData getMetaData(boolean refresh) throws XMLDBCException {
		if (metaData == null || refresh) {
			metaData = new XMLDataSourceMetaDataConnectionWrapper(this, mediator.getDataSourceMetaData(refresh));
		}
		return metaData;
	}

	MetaDataImpl getMetadataAccess() throws XMLDBCException {
		return mediator.getDataSourceMetaData(false);
	}

	/**
	 * To get the current user name.
	 * @return the current user name.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public String getUserName() throws XMLDBCException {
		return null;
	}
	
	/**
	 * To get the current URL of this connection.
	 * @return the current URL of this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public String getURL() throws XMLDBCException {
		return mediator.getURL();
	}
	
	Map getWrappers() throws XMLDBCException {
		return mediator.getWrappers();
	}
	
	/*********************************************************************/
	/*************************** Create/prepare statements ***************/
	/*********************************************************************/
	
	/**
	 * Creates a statement object for sending XQuery (query type by default)
	 * statements to the data source.
	 * @return a new statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLStatement createStatement() throws XMLDBCException {
		XMLStatement retVal = createStatement(XQUERY_STRING_TYPE);
		activeStatementList.add(retVal);
		return retVal;
	}
	
	
	/**
	 * Creates a statement object for sending XQuery or XPath (according to the
	 * specified type) statements to the data source.
	 * @param queryType the type of query for this statement (see constants in this interface).
	 * @return a new statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLStatement createStatement(short queryType) throws XMLDBCException {
		return new MediatorStatement(new MediatorConnectionProxy());
	}
	
	/**
	 * Creates a prepared statement object for sending parameterized XQuery
	 * (query type by default) statements to the data source.
	 * The advantage of prepared statements over statements is that the request
	 * may be precompiled once and executed several times.
	 * Note: this interface is optional.
	 * @param query a parameterized XQuery.
	 * @return a new prepared statement containing the pre-compiled statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 * @throws XMLDBCNotSupportedException if the method is not supported.
	 */
	public PreparedXMLStatement prepareStatement(String query) throws XMLDBCException, XMLDBCNotSupportedException {
		PreparedXMLStatement retVal = new PreparedMediatorStatement(new MediatorConnectionProxy(), baseURI, query, mediator.getCachedStatements());
		retVal.setBaseURI(baseURI);
		activeStatementList.add(retVal);
		return retVal;
	}
	
	// specific methods
	XDBCWrapper getSubAccessorByName(String name) {
		Map wrappers = mediator.getWrappers();
		if (wrappers != null)
			for (Iterator it = wrappers.values().iterator(); it.hasNext() ;) {
				XDBCWrapper sub = (XDBCWrapper) it.next();
				if (sub.getName().equalsIgnoreCase(name))
					return sub;
			}
		return null;
	}
	
	// gets a NEW connection
	XMLConnection getSubConnectionByName(String name) throws MediatorException {
		XMLConnection result = (XMLConnection)wrapperConnections.get(name);
		if (result == null) {
			XDBCWrapper sub = getSubAccessorByName(name);
			if (sub == null) throw new MediatorException("Unknown wrapper "+name);
			try {
				result = sub.getConnection();
				wrapperConnections.put(name, result);
			} catch (XMLDBCException e) {
				throw new MediatorException("Could not create a connection for "+name, e);
			}
		}
		return result;
	}
	
	// add LARS 18/06/04
	public void loadModule(String moduleURL) throws XMLDBCException {
		try {
			URL url = new URL(moduleURL);
			InputSource source = new InputSource(url.openStream());
			source.setSystemId(url.toString());
			getMetadataAccess().getModuleManager().loadModule(source, getMetadataAccess(), getMetadataAccess().getSchemaManager());
		} catch (IOException xqe) {
			throw new XMLDBCException("Could not read module file", xqe);
		} catch (XQueryException xqe) {
			throw new XMLDBCException("Could not read module file", xqe);
		}
	}
	
	public void loadSchema(InputSource source) throws SAXException, XMLDBCException {
		getMetadataAccess().getSchemaManager().loadSchema(source);
	}

	public void setBaseURI(String baseURI) {
		this.baseURI = baseURI;
	}

	private class MediatorConnectionProxy implements _MediatorConnection {
				
		public XMLConnection getConnection() {
			return MediatorConnection.this;
		}
		
		public String getSourceName() {
			return MediatorConnection.this.getSourceName();
		}

		public XDBCWrapper getSubAccessorByName(String name) {
			return MediatorConnection.this.getSubAccessorByName(name);
		}

		public XMLConnection getSubConnectionByName(String name) throws MediatorException {
			return MediatorConnection.this.getSubConnectionByName(name);
		}

		public MetaDataImpl getMetadataAccess() throws XMLDBCException {
			return MediatorConnection.this.getMetadataAccess();
		}
		public void statementClosed(XMLStatement closedStatement) throws XMLDBCException {
			activeStatementList.remove(closedStatement);
		}
	}
}

