/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.algebra;

import java.util.*;

import org.xquark.mediator.decomposer.*;
import org.xquark.mediator.plan.*;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.xquery.normalize.GetBoundVariablesVisitor;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.hinttree.HintTree;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionCOLLECTION;
import org.xquark.xquery.parser.util.Constants;

/**
 * Represent a node where nodes or leaves can fix to it. Before atomization a DepMem binds only DepSource and DepVar. After atomization, a DepMem binds only DepVar.
 *  
 */
public class AlgFLWR extends Algebra {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.27 $";

    private static final String RCSName = "$Name:  $";

    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************
    private ArrayList subDepFLWRs = null;

    private AlgFLWR parentDepFLWR = null;

    //private ArrayList variables = null;
    private boolean monoSource = false;

    protected Set sources = null;

    protected HintTree hinttree = null;

    // telling if information has already been computed
    boolean hasInformation = false;

    // telling whether the subelements are done with nested loop
    int subhandling = -1;

    // relation between variables in where expression
    VariableDependencies vardependencies = null;

    //	private HashMap varVarWhereMap = null;
    //	private ArrayList allExprWhereList = null;
    //	private ArrayList complexExprWhereList = null;
    // maps for ordering purposes
    private HashMap varOrderMap = null;

    private HashMap orderVarMap = null;

    private ArrayList complexExprOrderList = null;

    private ArrayList varOrderList = null;

    private ArrayList varReOrderList = null;

    private ArrayList varHintList = null;

    // for compulsery hint
    private boolean hasCompulseryMerge = false;

    private boolean hasCompulseryNested = false;

    private IsolateWhereVisitor isolateVisitor = null;

    // ******************************************************************
    // * INITIALIZATION
    // ******************************************************************
    /**
     * 
     * @param typevisitor
     * @param expression
     * @param depmanager
     */
    public AlgFLWR(XQueryExpression expression, Variable var, AlgebraManager depmanager, boolean islet) throws MediatorException {
        super(expression, var, depmanager, islet);
        // set HintTree
        hinttree = ((FLWRExpression) expression).getHintTree();
        // fill var lists
        fillVarLists();
    }

    // ******************************************************************
    // * GET/SET METHODS
    // ******************************************************************
    public void setDependency(AlgFLWR parentDepFLWR) {
        if (this.parentDepFLWR != null)
            this.parentDepFLWR.getDependencyList().remove(this);
        this.parentDepFLWR = parentDepFLWR;
        ArrayList tmpList = parentDepFLWR.getDependencyList();
        if (tmpList == null) {
            tmpList = new ArrayList();
            tmpList.add(this);
            parentDepFLWR.setDependencyList(tmpList);
        } else if (!tmpList.contains(this))
            tmpList.add(this);
    }

    public ArrayList getDependencyList() {
        return subDepFLWRs;
    }

    public void setDependencyList(ArrayList subDepFLWRs) {
        this.subDepFLWRs = subDepFLWRs;
    }

    public boolean isRootQDB() {
        return (parentDepFLWR == null && parent == null);
    }

    public boolean isQDB() {
        return (isRootQDB() || parent == null);
    }

    public boolean isMonoSource() {
        return monoSource;
    }

    public void setMonoSource(boolean monoSource) {
        this.monoSource = monoSource;
    }

    public HintTree getHintTree() {
        return hinttree;
    }

    public void setHintTree(HintTree hinttree) {
        this.hinttree = hinttree;
    }

    public int getSubHandling() {
        return subhandling;
    }

    public void setSubHandling(int subhandling) {
        this.subhandling = subhandling;
    }

    public boolean hasCompulseryMerge() {
        return hasCompulseryMerge;
    }

    public boolean hasCompulseryNested() {
        return hasCompulseryNested;
    }

    public void updateVarDependencyList(ArrayList dependVars) {
        if (parentDepFLWR == null || dependVars == null || dependVars.isEmpty())
            return;
        for (int i = 0; i < dependVars.size(); i++) {
            Variable tmpvar = (Variable) dependVars.get(i);
            if (!((FLWRExpression) parentDepFLWR.getExpression()).getVariables().contains(tmpvar))
                getParentDepFLWR().addDependSourceVariable(tmpvar);
            else
                dependVars.remove(tmpvar);
        }
        if (!dependVars.isEmpty())
            getParentDepFLWR().updateVarDependencyList(dependVars);
    }

    private AtomizeWhereVisitor awvisitor = null;

    private void computeVarOrderList() throws MediatorException {
        ArrayList orders = ((FLWRExpression) expression).getOrderBy();
        if (orders != null) {
            varOrderList = new ArrayList(orders.size());
            complexExprOrderList = new ArrayList(1);
            orderVarMap = new HashMap();
            varOrderMap = new HashMap();
            GetBoundVariablesVisitor gbvv = new GetBoundVariablesVisitor(true, false);
            for (int i = 0; i < orders.size(); i++) {
                XQueryExpression orderExpr = (XQueryExpression) orders.get(i);
                //				if (orderExpr.getOrder(0) == Constants.PLACE_ORDER)
                //					continue;
                gbvv.reset();
                try {
                    orderExpr.accept(gbvv);
                } catch (XQueryException e) {
                    throw new MediatorException("Could not extract variables from expression : " + orderExpr, e);
                }
                ArrayList tmpList = gbvv.getAllVariables();
                orderVarMap.put(orderExpr, tmpList);
                if (tmpList.size() > 1) {
                    complexExprOrderList.add(orderExpr);
                    if (varOrderList != null && !varOrderList.containsAll(tmpList))
                        varOrderList = null;
                    ArrayList tmpOrders = (ArrayList) varOrderMap.get(tmpList);
                    if (tmpOrders == null) {
                        tmpOrders = new ArrayList(1);
                        varOrderMap.put(tmpList, tmpOrders);
                    }
                    tmpOrders.add(orderExpr);
                } else {
                    Variable tmpVar = (Variable) tmpList.get(0);
                    if (((FLWRExpression) expression).getVariables().contains(tmpVar)) {
                        int pos = varOrderList.indexOf(tmpVar);
                        if (pos >= 0 && pos != varOrderList.size() - 1)
                            varOrderList = null;
                        else if (pos == -1)
                            varOrderList.add(tmpVar);
                        ArrayList tmpOrders = (ArrayList) varOrderMap.get(tmpVar);
                        if (tmpOrders == null) {
                            tmpOrders = new ArrayList(1);
                            varOrderMap.put(tmpVar, tmpOrders);
                        }
                        tmpOrders.add(orderExpr);
                    }
                }
            }
            if (complexExprOrderList.isEmpty())
                complexExprOrderList = null;
        }
    }

    /*
     * private void updateVarListWithDependencies(ArrayList varlist) { updateVarListWithDependencies(varlist, null, false); }
     */
    private void updateVarListWithDependencies(ArrayList varlist, ArrayList rangelist, boolean reorder) {
        if (varlist == null || varlist.isEmpty())
            return;
        int i = 0;
        while (i < varlist.size()) {
            Variable tmpVari = (Variable) varlist.get(i);
            Algebra tmpNode = algManager.getMapNode(tmpVari);
            if (tmpNode != null) {
                ArrayList tmpList = tmpNode.getVarsDependingOn();
                if (tmpList != null) {
                    for (int j = 0; j < tmpList.size(); j++) {
                        Variable tmpVarj = (Variable) tmpList.get(j);
                        if (rangelist != null && !rangelist.contains(tmpVarj))
                            continue;
                        int posj = varlist.indexOf(tmpVarj);
                        if (posj > i) {
                            varlist.remove(tmpVarj);
                            varlist.add(i, tmpVarj);
                            if (reorder) {
                                if (!varReOrderList.contains(tmpVari))
                                    varReOrderList.add(tmpVari);
                                if (!varReOrderList.contains(tmpVarj))
                                    varReOrderList.add(tmpVarj);
                            }
                            i--;
                            break;
                        } else if (posj == -1) {
                            varlist.add(i, tmpVarj);
                            i--;
                            break;
                        }
                    }
                }
            }
            i++;
        }
    }

    private void verifyVarOrderList() {
        if (varOrderList == null) {
            varOrderList = ((FLWRExpression) expression).getVariables();
            if (((FLWRExpression) expression).getOrderBy() != null)
                varReOrderList = ((FLWRExpression) expression).getVariables();
        } else {
            varReOrderList = new ArrayList();
            // manage depending on variables
            int i = 0;
            while (i < varOrderList.size()) {
                Variable tmpVari = (Variable) varOrderList.get(i);
                Algebra tmpNode = algManager.getMapNode(tmpVari);
                ArrayList tmpList = tmpNode.getVarsDependingOn();
                if (tmpList != null) {
                    for (int j = 0; j < tmpList.size(); j++) {
                        Variable tmpVarj = (Variable) tmpList.get(j);
                        int posj = varOrderList.indexOf(tmpVarj);
                        if (posj > i) {
                            varOrderList.remove(tmpVarj);
                            varOrderList.add(i, tmpVarj);
                            if (!varReOrderList.contains(tmpVari))
                                varReOrderList.add(tmpVari);
                            if (!varReOrderList.contains(tmpVarj))
                                varReOrderList.add(tmpVarj);
                            i--;
                            break;
                        } else if (posj == -1) {
                            varOrderList.add(i, tmpVarj);
                            i--;
                            break;
                        }
                    }
                }
                i++;
            }
            if (varReOrderList.isEmpty())
                varReOrderList = null;
            // manage other variables
            ArrayList varlist = ((FLWRExpression) expression).getVariables();
            if (varOrderList.size() < varlist.size()) {
                for (i = 0; i < varlist.size(); i++) {
                    if (!varOrderList.contains(varlist.get(i))) {
                        varOrderList.add(varlist.get(i));
                    }
                }
            }
        }
    }

    private void fillList(ArrayList varlist, ArrayList allvars) {
        if (varlist.containsAll(allvars))
            return;
        for (int i = 0; i < allvars.size(); i++) {
            if (!varlist.contains(allvars.get(i))) {
                varlist.add(allvars.get(i));
            }
        }
    }

    /*
     * private boolean verifyHintTree(HintTree hinttree) { return verifyHintTree(hinttree, new ArrayList()); } private boolean verifyHintTree(HintTree hinttree, ArrayList hintVarList) { if (hinttree == null) return false; if (hinttree.getType() == HintTree.NESTED_LOOP_TYPE) { if (hinttree.getLeftHint() instanceof HintTree) { if (!verifyHintTree((HintTree) hinttree.getLeftHint(), hintVarList)) return false; } if (hinttree.getLeftHint() instanceof Variable) { hintVarList.add(hinttree.getLeftHint()); } if (hinttree.getRightHint() instanceof HintTree) { if (!verifyHintTree((HintTree) hinttree.getRightHint(), hintVarList)) return false; } if (hinttree.getRightHint() instanceof Variable) { hintVarList.add(hinttree.getRightHint()); } } else if (hinttree.getType() == HintTree.MERGE_TYPE) { ArrayList tmplistleft = (ArrayList) hintVarList.clone(); if (hinttree.getLeftHint() instanceof HintTree) { if (!verifyHintTree((HintTree) hinttree.getLeftHint(), tmplistleft)) return false;
     * tmplistleft.removeAll(hintVarList); } ArrayList tmplistright = (ArrayList) hintVarList.clone(); if (hinttree.getRightHint() instanceof HintTree) { if (!verifyHintTree((HintTree) hinttree.getRightHint(), tmplistright)) return false; tmplistright.removeAll(hintVarList); } hintVarList.addAll(tmplistleft); hintVarList.addAll(tmplistright); if (hinttree.getLeftHint() instanceof Variable) { hintVarList.add(hinttree.getLeftHint()); } if (hinttree.getRightHint() instanceof Variable) { hintVarList.add(hinttree.getRightHint()); } } return true; }
     */

    private boolean verifyVarHintList() {
        if (varHintList == null)
            return false;
        ArrayList vars = ((FLWRExpression) expression).getVariables();
        // verifying all variables of the hint tree are declared in flwr
        if (!vars.containsAll(varHintList))
            return false;
        // verifying that all variables are used only once
        for (int i = 0; i < varHintList.size(); i++) {
            if (varHintList.lastIndexOf(varHintList.get(i)) != i)
                return false;
        }
        // get all variables in vars but not in varlist
        ArrayList remainvars = (ArrayList) vars.clone();
        remainvars.removeAll(varHintList);
        if (!remainvars.isEmpty()) {
            updateVarListWithDependencies(remainvars, remainvars, false);
        }
        // verifying that variables in hint are bound to DepNodeSource on collections
        for (int i = 0; i < varHintList.size(); i++) {
            Variable vari = (Variable) vars.get(i);
            if (vari.getBindingType() != Constants.FOR_BINDINGTYPE)
                return false;
            XQueryExpression expri = vari.getExpression();
            if (expri instanceof LocatedExpression)
                expri = ((LocatedExpression) expri).getExpression();
            if (!(expri instanceof FunctionCOLLECTION))
                return false;
        }

        /*
         * // verifying that all variables of the flwr are used in the hint tree for (int i = 0; i < vars.size(); i++) { Variable vari = (Variable) vars.get(i); if (vari.getBindingType() == Constants.FOR_BINDINGTYPE && !varHintList.contains(vari)) return false; // for now no let if (vari.getBindingType() == Constants.LET_BINDINGTYPE) return false; // if (vari.getBindingType() == Constants.LET_BINDINGTYPE && varHintList.contains(vari)) // return false; }
         */
        return true;
    }

    private boolean updateVarHintList(HintTree hinttree) {
        if (hinttree == null)
            return false;
        ArrayList vars = ((FLWRExpression) expression).getVariables();
        /*
         * Object obj = hinttree.getLeftHint(); if (obj instanceof HintTree) updateVarListWithDependencies(((HintTree)obj).getVarList(),vars,false); obj = hinttree.getRightHint(); if (obj instanceof HintTree) updateVarListWithDependencies(((HintTree)obj).getVarList(),vars,false);
         */
        updateVarListWithDependencies(hinttree.getVarList(), vars, false);
        return true;
    }

    private void compeReOrdering(boolean hasMerge, ArrayList varList) {
        if (hasMerge)
            //varReOrderList = (ArrayList) hint.getVarList().clone();
            varReOrderList = (ArrayList) ((FLWRExpression) expression).getVariables().clone();
        else {
            if (varOrderList == null || varOrderList.isEmpty())
                return;
            ArrayList tmplist = (ArrayList) ((FLWRExpression) expression).getVariables().clone();
            tmplist.removeAll(varList);
            tmplist.addAll(0, varList);
            if (varReOrderList == null)
                varReOrderList = new ArrayList();
            else
                varReOrderList.clear();
            int pos = -1;
            for (int j = 0; j < tmplist.size(); j++) {
                Variable tmpVarj = (Variable) tmplist.get(j);
                int posj = varOrderList.indexOf(tmpVarj);
                if (posj > pos)
                    pos = posj;
                else
                    for (int i = 0; i < tmplist.size(); i++) {
                        Variable tmpVari = (Variable) tmplist.get(i);
                        if (!varReOrderList.contains(tmpVari))
                            varReOrderList.add(tmpVari);
                    }
            }
            if (varReOrderList.isEmpty())
                varReOrderList = null;
        }
    }

    /*
     * private boolean computeVarHintList(HintTree hinttree) { if (hinttree == null) return false; Object obj = hinttree.getLeftHint(); if (obj instanceof Variable) { if (varHintList.contains(obj)) return false; else varHintList.add(obj); } else computeVarHintList((HintTree) obj); obj = hinttree.getRightHint(); if (obj instanceof Variable) { if (varHintList.contains(obj)) return false; else varHintList.add(obj); } else computeVarHintList((HintTree) obj); return true; }
     */

    public void makeInformation() throws MediatorException {
        if (hasInformation)
            return;
        hasInformation = true;
        // verify that all children have only one variable
        if (!isMonoSource()) {
            if (algChildren != null) {
                for (int i = 0; i < algChildren.size(); i++)
                    if (((Algebra) algChildren.get(i)).getVariables().size() != 1)
                        throw new MediatorException("Found DepNode of more than one variable as child of DepFLWR");
            }
        } else {
            if (hinttree != null) {
                if (hinttree.getOuterType() != HintTree.NO_TYPE)
                    hinttree.setType(HintTree.NO_TYPE);
            }
        }
        // gather information on joins if any (where clause)

        XQueryExpression wexpr = ((FLWRExpression) expression).getWhereClause();
        if (wexpr != null) {
            vardependencies = new VariableDependencies(wexpr);
            //			if (awvisitor == null)
            //				awvisitor = new AtomizeWhereVisitor();
            //			try {
            //				wexpr.accept(awvisitor);
            //			} catch (XQueryException e) {
            //				throw new MediatorException("Could not atomize where clause : " + wexpr, e);
            //			}
            //			varVarWhereMap = awvisitor.getVarVarMap();
            //			allExprWhereList = awvisitor.getAllExprList();
            //			complexExprWhereList = awvisitor.getComplexExprList();
        }
        // gather inforamtion on hint tree if any
        if (hinttree != null && hinttree.getType() != HintTree.NO_TYPE) {
            varHintList = hinttree.getVarList();
            if (updateVarHintList(hinttree)) {
                if (verifyVarHintList()) {
                } else {
                    hinttree = null;
                }
            } else {
                hinttree = null;
            }
        }
        computeVarOrderList();
        if (hinttree != null)
            compeReOrdering(hinttree.hasMerge(), hinttree.getVarList());
        else
            compeReOrdering(algManager.getPlan().getDefaultSubHandling() == HintTree.MERGE_TYPE, ((FLWRExpression) expression).getVariables());
    }

    public XQueryExpression getWhereClause() {
        return ((FLWRExpression) expression).getWhereClause();
    }

    public VariableDependencies getVarDependencies() {
        return vardependencies;
    }

    // ******************************************************************
    // * CREATE EXECUTION PLAN
    // ******************************************************************
    // some documentation
    /*
     * for now hint tree only applies on FLWR expressions (not nested for now)
     * 
     * if (hint tree is correct) for now correct hint tree means that all (for) variables present in the FLWR expression are also present in the hint tree (let) variables are simply merged
     * 
     * if (hint tree is not present or hint tree is incorrect) variables are ordered in the following order first get the order implied by the order by clauses if any and if possible insert all depending on variables add all variables not in this list make list of variables to be reordered (because of dependencies)
     */

    //##################################################################################################################################
    //##################################################################################################################################
    //##################################################################################################################################
    private Operator createVarsOperator(ExecutionPlan plan, Operator operator, List varlist, List vars) throws MediatorException {
        //##################################################################################################################################
        //##################################################################################################################################
        //##################################################################################################################################
        for (int i = 0; i < vars.size(); i++) {
            operator = createVarOperator(plan, operator, varlist, (Variable) vars.get(i));
        }
        return operator;
    }

    //##################################################################################################################################
    //##################################################################################################################################
    //##################################################################################################################################
    private Operator createVarOperator(ExecutionPlan plan, Operator operator, List varlist, Variable var) throws MediatorException {
        //##################################################################################################################################
        //##################################################################################################################################
        //##################################################################################################################################
        // getting child
        Algebra depchildi = algManager.getMapNode(var);
        XQueryExpression depchildiexpr = depchildi.getExpression();
        //ArrayList varidependingon = (ArrayList) depchildi.getVarsDependingOn().clone();
        Variable vari = (Variable) depchildi.getVariables().get(0);
        if (!isMonoSource())
            algManager.addVarToScope(vari);
        else
            algManager.addVarsToScope(depchildi.getVariables());
        boolean isleti = depchildi.isLet();
        varlist.add(vari);
        //		if (log.isDebugEnabled()) {
        //			log.debug("*********** DEPFLWR CHILD STUFF *************************");
        //			log.debug("ClassName = " + depchildi.getClass().getName());
        //			log.debug("is let = " + depchildi.isLet());
        //			log.debug("depchildiexpr = " + depchildiexpr);
        //			log.debugt("vari = " + vari);
        //			log.debug("allVars = " + algManager.getScopeVariables());
        //			log.debug("definition vars depending on = " + depchildi.getVarsDependingOn());
        //			log.debug("where vars depending on = " + depchildi.getVarsWhereOn());
        //			log.debug("vars referenced by = " + depchildi.getVarsReferencedBy());
        //			log.debugt("children = " + depchildi.getChildren());
        //			log.debug("var dependencies = " + vardependencies);
        //			log.debug("---------------------------------------------------------");
        //		}
        // add where dependencies
        if (vardependencies != null) {
            ArrayList tmplist = vardependencies.getAllAssociatedVariables(vari);
            if (tmplist != null) {
                for (int i = 0; i < tmplist.size(); i++) {
                    Variable tmpvari = (Variable) tmplist.get(i);
                    if (algManager.getScopeVariables().contains(tmpvari)) {
                        if (depchildiexpr instanceof FLWRExpression && ((FLWRExpression) depchildiexpr).getVariables().contains(tmpvari)) {
                            if (!depchildi.getVarsWhereOn().contains(tmpvari))
                                depchildi.getVarsWhereOn().add(tmpvari);
                        } else {
                            if (!depchildi.getVarsDependingOn().contains(tmpvari))
                                depchildi.getVarsDependingOn().add(tmpvari);
                        }
                    }

                }
            }
        }
        boolean doSort = false;
        if (operator == null) {
            if (depchildi instanceof AlgLocated || depchildi instanceof AlgVar || depchildi instanceof AlgArithFunction)
                //algebra = new AlgNotSource(plan,depchildi.getExpression(),var,depchildi.isLet(),depchildi.hasIdentifier());
                operator = new OpNotSource(plan, depchildi);
            else
                operator = (Operator) depchildi.createOperator(plan);
            doSort = true;
        } else {
            if (!depchildi.getDepends()) {
                Operator tmpOp = depchildi.createOperator(plan);
                // test if sort clause and not source -> then sort in place !!!
                if (!(depchildi instanceof AlgSource) && varOrderMap != null) {
                    ArrayList orders = (ArrayList) varOrderMap.get(vari);
                    if (orders != null)
                        tmpOp = new OpSort(plan, tmpOp, orders);
                }
                if (isleti) {
                    operator = new OpMerge(plan, depchildiexpr, operator, tmpOp);
                } else {
                    operator = new OpCartesian(plan, depchildiexpr, operator, tmpOp);
                }
            } else { // varidependingon is not empty
                if (depchildi instanceof AlgLocated || depchildi instanceof AlgVar || depchildi instanceof AlgArithFunction) {
                    //algebra = new AlgSimpleMerge(plan, depchildiexpr, vari, algebra, depchildi.isLet());
                    operator = new OpSimpleMerge(plan, depchildi, operator);
                    doSort = true;
                } else if (depchildi instanceof AlgFunction && depchildi.getChildren() == null) {
                    operator = new OpSimpleMerge(plan, depchildiexpr, vari, operator, depchildi.isLet());
                    doSort = true;
                } else {
                    Operator[] tmplist = new Operator[2];
                    tmplist[0] = operator;
                    Operator tmpOp = depchildi.createOperator(plan);
                    // test if sort clause and not source -> then sort in place !!!
                    if (!(depchildi instanceof AlgSource) && varOrderMap != null) {
                        ArrayList orders = (ArrayList) varOrderMap.get(vari);
                        if (orders != null)
                            tmpOp = new OpSort(plan, tmpOp, orders);
                    }
                    tmpOp.setDepends(depchildi);
                    tmplist[1] = tmpOp;
                    operator = new OpSubQuery(plan, depchildiexpr, tmplist, true, false);
                }

            }
        }
        // handle restriction if any

        XQueryExpression tmpExpr = ((FLWRExpression) expression).getWhereClause();
        if (tmpExpr != null) {
            if (operator instanceof OpSource || operator instanceof OpSubQuery) {
                if (isolateVisitor == null)
                    isolateVisitor = new IsolateWhereVisitor(algManager.getScopeVariables());
                else
                    isolateVisitor.reset(algManager.getScopeVariables());
                isolateVisitor.putVar(true);
                try {
                    tmpExpr.accept(isolateVisitor);
                } catch (XQueryException e) {
                    throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
                }

                tmpExpr = isolateVisitor.getIsolatedExpression();
                if (tmpExpr != null) {
                    operator = operator.addCondition(plan, tmpExpr);
                    //operator = new OpRestrict(plan, tmpExpr, operator);
                }
                ((FLWRExpression) expression).setWhereClause(isolateVisitor.getRemainingExpression());
            } else {

                if (isolateVisitor == null)
                    isolateVisitor = new IsolateWhereVisitor(algManager.getScopeVariables());
                else
                    isolateVisitor.reset(algManager.getScopeVariables());
                isolateVisitor.putVar(true);
                try {
                    tmpExpr.accept(isolateVisitor);
                } catch (XQueryException e) {
                    throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
                }

                tmpExpr = isolateVisitor.getIsolatedExpression();
                if (tmpExpr != null) {
                    operator = operator.addCondition(plan, tmpExpr);
                    //algebra = new OpRestrict(plan, tmpExpr, algebra);
                }
                ((FLWRExpression) expression).setWhereClause(isolateVisitor.getRemainingExpression());

            }
        }
        // test if sort clause and not source -> then sort in place !!!
        if (doSort && !(depchildi instanceof AlgSource) && varOrderMap != null) {
            ArrayList orders = (ArrayList) varOrderMap.get(vari);
            if (orders != null)
                operator = new OpSort(plan, operator, orders);
        }
        return operator;
    }

    // utility
    private boolean extractConditionsfromMap(VariableDependencies dep, List varListLeft, List varListRight, ArrayList leftExpressions, ArrayList rightExpressions, ArrayList mergeRestrictions, ArrayList otherRestrictions) {
        boolean hasnoteq = false;
        for (int il = 0; il < varListLeft.size(); il++) {
            Variable leftvar = (Variable) varListLeft.get(il);
            if (vardependencies != null) {
                HashMap tmpmap = vardependencies.getJoinMap(leftvar);
                if (tmpmap != null) {
                    // there is at least something because of hint tree verification
                    for (int ir = 0; ir < varListRight.size(); ir++) {
                        Variable rightvar = (Variable) varListRight.get(ir);
                        ArrayList tmplist = (ArrayList) tmpmap.get(rightvar);
                        if (tmplist != null) {
                            for (int i = 0; i < tmplist.size(); i++) {
                                XQueryExpression tmpexpr = (XQueryExpression) tmplist.get(i);
                                if (tmpexpr instanceof ListOpCompExpression) {
                                    ListOpCompExpression listopcomp = (ListOpCompExpression) tmpexpr;
                                    XQueryExpression expr1 = listopcomp.getExpression1();
                                    XQueryExpression expr2 = listopcomp.getExpression2();
                                    if ((expr1 instanceof LocatedExpression || expr1 instanceof Variable) && (expr2 instanceof LocatedExpression || expr2 instanceof Variable)) {
                                        if (listopcomp.getOperator() == Constants.EQ_COMPOP) {
                                            //if (((LocatedExpression) expr1).startsWith((Variable) rightvar) && ((LocatedExpression) expr2).startsWith((Variable) leftvar)) {
                                            // change LARS 11/03/04
                                            if (((expr1 instanceof LocatedExpression && ((LocatedExpression) expr1).startsWith(rightvar)) || expr1 == rightvar) && ((expr2 instanceof LocatedExpression && ((LocatedExpression) expr2).startsWith(leftvar)) || expr2 == leftvar)) {
                                                XQueryExpression bufexpr = expr2;
                                                try {
                                                    listopcomp.setExpression2(expr1);
                                                    listopcomp.setExpression1(bufexpr);
                                                } catch (XQueryException e) {
                                                }
                                            }
                                            if (hasnoteq) {
                                                leftExpressions.add(mergeRestrictions.size() - 1, listopcomp.getExpression1());
                                                rightExpressions.add(mergeRestrictions.size() - 1, listopcomp.getExpression2());
                                                mergeRestrictions.add(mergeRestrictions.size() - 1, listopcomp);
                                            } else {
                                                leftExpressions.add(listopcomp.getExpression1());
                                                rightExpressions.add(listopcomp.getExpression2());
                                                mergeRestrictions.add(listopcomp);
                                            }
                                        } else if (!hasnoteq && (listopcomp.getOperator() == Constants.GEQ_COMPOP || listopcomp.getOperator() == Constants.GT_COMPOP || listopcomp.getOperator() == Constants.LEQ_COMPOP || listopcomp.getOperator() == Constants.LT_COMPOP)) {
                                            hasnoteq = true;
                                            //if (((LocatedExpression) expr1).startsWith((Variable) rightvar) && ((LocatedExpression) expr2).startsWith((Variable) leftvar)) {
                                            // change LARS 11/03/04
                                            if (((expr1 instanceof LocatedExpression && ((LocatedExpression) expr1).startsWith(rightvar)) || expr1 == rightvar) && ((expr2 instanceof LocatedExpression && ((LocatedExpression) expr2).startsWith(leftvar)) || expr2 == leftvar)) {
                                                XQueryExpression bufexpr = expr2;
                                                try {
                                                    listopcomp.setExpression2(expr1);
                                                    listopcomp.setExpression1(bufexpr);
                                                    if (listopcomp.getOperator() == Constants.GEQ_COMPOP)
                                                        listopcomp.setOperator(Constants.LEQ_COMPOP);
                                                    else if (listopcomp.getOperator() == Constants.GT_COMPOP)
                                                        listopcomp.setOperator(Constants.LT_COMPOP);
                                                    else if (listopcomp.getOperator() == Constants.LEQ_COMPOP)
                                                        listopcomp.setOperator(Constants.GEQ_COMPOP);
                                                    else if (listopcomp.getOperator() == Constants.LT_COMPOP)
                                                        listopcomp.setOperator(Constants.GT_COMPOP);
                                                } catch (XQueryException e) {
                                                }
                                            }
                                            leftExpressions.add(listopcomp.getExpression1());
                                            rightExpressions.add(listopcomp.getExpression2());
                                            mergeRestrictions.add(listopcomp);
                                        } else if (otherRestrictions != null)
                                            otherRestrictions.add(listopcomp);

                                    } else if (otherRestrictions != null)
                                        otherRestrictions.add(listopcomp);
                                } else if (otherRestrictions != null)
                                    otherRestrictions.add(tmpexpr);
                            }
                        }
                    }
                }
                // change LARS 11/03/04 all conditions -> complex conditions
                ArrayList conds = vardependencies.getAllComplexConditionsWithVar(leftvar);
                if (conds != null)
                    for (int i = 0; i < conds.size(); i++)
                        if (!otherRestrictions.contains(conds.get(i)))
                            otherRestrictions.add(conds.get(i));
            }
        }
        return hasnoteq;
    }

    //##################################################################################################################################
    //##################################################################################################################################
    //##################################################################################################################################
    private ArrayList getMainDependVars(Variable var) {
        ArrayList reslist = null;
        ArrayList tmplist = algManager.getMapNode(var).getVarsDependingOn();
        if (tmplist != null) {
            reslist = new ArrayList(1);
            for (int i = 0; i < tmplist.size(); i++) {
                ArrayList list = getMainDependVars((Variable) tmplist.get(i));
                if (list != null)
                    reslist.addAll(list);
                else
                    reslist.add(tmplist.get(i));
            }
            if (reslist.isEmpty())
                reslist = null;
        }
        return reslist;
    }

    private Operator createOperator(ExecutionPlan plan, HintTree hint, List vars) throws MediatorException {
        return createOperator(plan, hint, plan.getDefaultSubHandling(), vars);
    }

    private Operator createOperator(ExecutionPlan plan, HintTree hint, byte defaultHintType, List vars) throws MediatorException {
        //##################################################################################################################################
        //##################################################################################################################################
        //##################################################################################################################################

        //int typ = getAlgebraManager().getPlan().getDefaultSubHandling();
        byte hintType = defaultHintType;
        Object left = null;
        Object right = null;
        List varListLeft = null;
        List varListRight = null;
        boolean hasLeftloop = false;
        boolean hasRightLoop = false;
        if (hint != null && hint.getType() != HintTree.NO_TYPE) {
            hintType = hint.getType();
            left = hint.getLeftHint();
            right = hint.getRightHint();

            if (left instanceof Variable) {
                varListLeft = new ArrayList(1);
                varListLeft.add(left);
            } else
                varListLeft = ((HintTree) left).getVarList();

            if (right instanceof Variable) {
                varListRight = new ArrayList(1);
                varListRight.add(right);
            } else
                varListRight = ((HintTree) right).getVarList();
        } else if (vars != null) {
            if (vars.size() < 2)
                return null;
            else {
                // TODO make better division whenever possible
                // 1. find source variables
                // 2. associate all depending variables
                // find restriction
                // find joins

                boolean hasJoinMap = (vardependencies != null && vardependencies.getJoinsMap() != null && !vardependencies.getJoinsMap().isEmpty());
                
                varListLeft = new ArrayList();
                for (int i = 0; i < vars.size(); i++) {
                    Variable vari = (Variable) vars.get(i);
                    if (varListRight == null) {
                        varListLeft.add(vari);
                        if (!hasLeftloop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                            hasLeftloop = true;
                        if (hasJoinMap) {
                            Map tmpmap = vardependencies.getJoinMap(vari);
                            if (tmpmap != null) {
                                boolean found = false;
                                for (Iterator it = tmpmap.keySet().iterator(); it.hasNext();) {
                                    if (vars.contains(it.next())) {
                                        found = true;
                                        break;
                                    }
                                }
                                if (found)
                                    varListRight = new ArrayList();
                            }
                        }
                    } else {
                        ArrayList depends = getMainDependVars(vari);
                        if (depends != null) {
                            if (varListLeft.containsAll(depends)) {
                                varListLeft.add(vari);
                                if (!hasLeftloop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                                    hasLeftloop = true;
                            } else if (varListRight.containsAll(depends)) {
                                varListRight.add(vari);
                                if (!hasRightLoop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                                    hasRightLoop = true;
                            } else
                                return null;
                        } else {
                            varListRight.add(vari);
                            if (!hasRightLoop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                                hasRightLoop = true;
                        }
                    }
                }
                if (varListRight == null || varListRight.isEmpty())
                    return null;                
                if (hintType == HintTree.MERGE_TYPE) {
                    if (!hasLeftloop || !hasRightLoop) {
                        hintType = HintTree.NESTED_LOOP_TYPE;
                    }
                } 
                // exchange left and right if left has loop and right has not
                if (hintType == HintTree.NESTED_LOOP_TYPE) {
                    if (hasLeftloop || !hasRightLoop) {
                        List tmpVarList = varListRight;
                        varListRight = varListLeft;
                        varListLeft = tmpVarList;
                    }
                }
                //                if (hintType == HintTree.MERGE_TYPE) {
                //                    // test if simple joins exist
                //                    if (vardependencies == null || vardependencies.getJoinsMap() == null || vardependencies.getJoinsMap().isEmpty())
                //                        return null;
                //                    varListLeft = new ArrayList();
                //                    boolean hasLeftloop = false;
                //                    boolean hasRightLoop = false;
                //                    for (int i = 0; i < vars.size(); i++) {
                //                        Variable vari = (Variable) vars.get(i);
                //                        if (varListRight == null) {
                //                            varListLeft.add(vari);
                //                            if (!hasLeftloop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                //                                hasLeftloop = true;
                //                            Map tmpmap = vardependencies.getJoinMap(vari);
                //                            if (tmpmap != null) {
                //                                boolean found = false;
                //                                for (Iterator it = tmpmap.keySet().iterator(); it.hasNext();) {
                //                                    if (vars.contains(it.next())) {
                //                                        found = true;
                //                                        break;
                //                                    }
                //                                }
                //                                if (found)
                //                                    varListRight = new ArrayList();
                //                            }
                //                        } else {
                //                            ArrayList depends = getMainDependVars(vari);
                //                            if (depends != null) {
                //                                if (varListLeft.containsAll(depends)) {
                //                                    varListLeft.add(vari);
                //                                    if (!hasLeftloop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                //                                        hasLeftloop = true;
                //                                } else if (varListRight.containsAll(depends)) {
                //                                    varListRight.add(vari);
                //                                    if (!hasRightLoop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                //                                        hasRightLoop = true;
                //                                } else
                //                                    return null;
                //                            } else {
                //                                varListRight.add(vari);
                //                                if (!hasRightLoop && vari.getBindingType() == Constants.FOR_BINDINGTYPE)
                //                                    hasRightLoop = true;
                //                            }
                //                        }
                //                    }
                //                    if (!hasLeftloop || !hasRightLoop)
                //                        return null;
                //                    if (varListRight == null || varListRight.isEmpty())
                //                        return null;
                //                } else if (hintType == HintTree.NESTED_TYPE) {
                //                    return null;
                //                }
            }
        } else
            return null;

        if (hintType == HintTree.MERGE_TYPE) {
            ArrayList mergeRestrictions = new ArrayList(1);
            ArrayList otherRestrictions = new ArrayList(1);
            ArrayList leftExpressions = new ArrayList(1);
            ArrayList rightExpressions = new ArrayList(1);
            boolean hasnoteq = false;
            if (hint != null || vars.size() >= 2)
                hasnoteq = extractConditionsfromMap(vardependencies, varListLeft, varListRight, leftExpressions, rightExpressions, mergeRestrictions, otherRestrictions);

            if (mergeRestrictions.isEmpty() && otherRestrictions.isEmpty()) {
                // change 23/06/2003
                ArrayList varlist = new ArrayList();
                Operator leftAlg = null;
                if (varListLeft.size() == 1)
                    leftAlg = createVarOperator(plan, null, varlist, (Variable) varListLeft.get(0));
                else {
                    leftAlg = createOperator(plan, (HintTree) hint.getLeftHint(), null);
                    if (leftAlg == null)
                        leftAlg = createVarsOperator(plan, null, varlist, varListLeft);
                }
                varlist.removeAll(varListLeft);
                algManager.removeVarsFromScope(varListLeft);
                Operator rightAlg = null;
                if (varListRight.size() == 1)
                    rightAlg = createVarOperator(plan, null, varlist, (Variable) varListRight.get(0));
                else {
                    rightAlg = createOperator(plan, null, varListRight);
                    if (rightAlg == null)
                        rightAlg = createVarsOperator(plan, null, varlist, varListRight);
                }
                algManager.addVarsToScope(varListLeft);
                algManager.addVarsToScope(varListRight);
                return new OpCartesian(plan, expression, leftAlg, rightAlg);
            }

            // ATTENTION erase these conditions from dependencies to avoid having them used several times
            for (int i = 0; i < otherRestrictions.size(); i++) {
                XQueryExpression otheri = (XQueryExpression) otherRestrictions.get(i);
                ArrayList tmplist = vardependencies.getVarsOfCond(otheri);
                for (int j = 0; j < tmplist.size(); j++) {
                    Variable varj = (Variable) tmplist.get(j);
                    // change LARS 11/03/04 all condition -> complex condition
                    ArrayList varjlist = vardependencies.getAllComplexConditionsWithVar(varj);
                    if (varjlist != null)
                        varjlist.remove(otheri);
                }
            }

            if (mergeRestrictions.isEmpty()) {
            }
            if (hasnoteq) {
                ListOpCompExpression listopcomp = (ListOpCompExpression) mergeRestrictions.get(mergeRestrictions.size() - 1);
                if (listopcomp.getOperator() == Constants.LT_COMPOP || listopcomp.getOperator() == Constants.LEQ_COMPOP) {
                    ((XQueryExpression) leftExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
                    ((XQueryExpression) rightExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
                }
            }

            ArrayList varlist = new ArrayList();
            Operator leftAlg = null;
            if (hint == null) {
                //				left = vars.get(0);
                //				leftAlg = createVarOperator(plan, null, varlist, (Variable) left);
                //				varlist.remove(left);
                //				algManager.removeVarFromScope((Variable) left);
                // change LARS 11/03/04
                leftAlg = createOperator(plan, null, defaultHintType, varListLeft);
                if (leftAlg == null) {
                    leftAlg = createVarsOperator(plan, null, varlist, varListLeft);
                    varlist.removeAll(varListLeft);
                }
                algManager.removeVarsFromScope(varListLeft);
            } else if (left instanceof Variable) {
                leftAlg = createVarOperator(plan, null, varlist, (Variable) left);
                varlist.remove(left);
                algManager.removeVarFromScope((Variable) left);
            } else {
                leftAlg = createOperator(plan, (HintTree) left, null);
                if (leftAlg == null)
                    leftAlg = createVarsOperator(plan, null, varlist, varListLeft);
                algManager.removeVarsFromScope(((HintTree) left).getVarList());
                //leftAlg = new AlgSort(plan, leftAlg, leftExpressions);
            }
            if (!leftExpressions.isEmpty()) {
                if (!(leftAlg instanceof OpSort))
                    leftAlg = new OpSort(plan, leftAlg, leftExpressions);
                leftAlg.addIndexSpec(leftExpressions);
            }
            Operator rightAlg = null;
            if (hint == null) {
                //				if (vars.size() == 2) {
                //					rightAlg = createVarOperator(plan, null, varlist, (Variable) vars.get(1));
                //				} else {
                //					rightAlg = createOperator(plan, null, vars.subList(1, vars.size()));
                //					algManager.removeVarsFromScope(vars.subList(1, vars.size()));
                //				}
                // change LARS 11/03/04
                rightAlg = createOperator(plan, null, hintType, varListRight);
                if (rightAlg == null) {
                    rightAlg = createVarsOperator(plan, null, varlist, varListRight);
                    varlist.removeAll(varListRight);
                }
                algManager.removeVarsFromScope(varListRight);
            } else if (right instanceof Variable) {
                rightAlg = createVarOperator(plan, null, varlist, (Variable) right);
                //varlist.clear();
                //depmanager.removeVarFromScope((Variable) right);
            } else {
                rightAlg = createOperator(plan, (HintTree) right, null);
                if (rightAlg == null)
                    rightAlg = createVarsOperator(plan, null, varlist, varListRight);
                algManager.removeVarsFromScope(((HintTree) right).getVarList());
                //rightAlg = new AlgSort(plan, rightAlg, rightExpressions);
            }
            if (!rightExpressions.isEmpty()) {
                if (!(rightAlg instanceof OpSort))
                    rightAlg = new OpSort(plan, rightAlg, rightExpressions);
                rightAlg.addIndexSpec(rightExpressions);
            }
            Operator operator = null;
            if (mergeRestrictions.isEmpty())
                operator = new OpCartesian(plan, expression, leftAlg, rightAlg);
            else {
                operator = new OpSortMerge(plan, leftAlg, rightAlg, mergeRestrictions);
            }
            // add restriction if necessary
            if (!otherRestrictions.isEmpty()) {
                XQueryExpression tmpexpr = null;
                if (otherRestrictions.size() == 1)
                    tmpexpr = (XQueryExpression) otherRestrictions.get(0);
                else {
                    try {
                        tmpexpr = new BinOpANDExpression((XQueryExpression) otherRestrictions.get(0), (XQueryExpression) otherRestrictions.get(1), expression.getParentModule());
                        for (int i = 2; i < otherRestrictions.size(); i++)
                            tmpexpr = new BinOpANDExpression(tmpexpr, (XQueryExpression) otherRestrictions.get(i), expression.getParentModule());
                    } catch (XQueryException e) {
                        throw new MediatorException("Could not build where clause", e);
                    }
                }
                operator = operator.addCondition(plan, tmpexpr);
                //operator = new OpRestrict(plan, tmpexpr, operator);
            }
            // reset scope
            algManager.addVarsToScope((hint != null) ? hint.getVarList() : vars);
            // reset order list
            //((FLWRExpression) expression).getOrderBy().removeAll(rightExpressions);
            //((FLWRExpression) expression).getOrderBy().removeAll(leftExpressions);
            return operator;
        }
        if (hintType == HintTree.NESTED_LOOP_TYPE) {
            // create algebra
            Operator leftOp = null;
            ArrayList varlist = new ArrayList();
            if (hint == null) {
                leftOp = createOperator(plan, null, defaultHintType, varListLeft);
                if (leftOp == null)
                    leftOp = createVarsOperator(plan, null, varlist, varListLeft);
                varlist.addAll(varListLeft);
                algManager.addVarsToScope(varListLeft);
            } else if (left instanceof Variable) {
                //updateVarListWithDependencies(varListLeft, ((FLWRExpression) expression).getVariables());
                leftOp = createOperator(plan, null, hintType, varListLeft);
                if (leftOp == null)
                    leftOp = createVarsOperator(plan, null, varlist, varListLeft);
                varlist.addAll(varListLeft);
                algManager.addVarToScope((Variable) left);
            } else {
                leftOp = createOperator(plan, (HintTree) left, null);
                if (leftOp == null)
                    leftOp = createVarsOperator(plan, null, varlist, varListLeft);
                algManager.addVarsToScope(((HintTree) left).getVarList());
                varlist.addAll(((HintTree) left).getVarList());
            }
            Operator rightOp = null;
            if (hint == null) {
                rightOp = createOperator(plan, null, defaultHintType, varListRight);
                if (rightOp == null)
                    rightOp = createVarsOperator(plan, null, varlist, varListRight);
                varlist.addAll(varListRight);
                algManager.addVarsToScope(varListRight);
            } else if (right instanceof Variable) {
                //updateVarListWithDependencies(varListRight, ((FLWRExpression) expression).getVariables());
                rightOp = createOperator(plan, null, hintType, varListRight);
                if (rightOp == null)
                    rightOp = createVarsOperator(plan, null, varlist, varListRight);
                varlist.addAll(varListRight);
                algManager.addVarToScope((Variable) right);
            } else {
                rightOp = createOperator(plan, (HintTree) right, null);
                if (rightOp == null)
                    rightOp = createVarsOperator(plan, null, varlist, varListRight);
                algManager.addVarsToScope(((HintTree) right).getVarList());
            }
            //if (hint != null)
                rightOp.setValueDepends(true);
            Operator[] tmplistalg = new Operator[2];
            tmplistalg[0] = leftOp;
            tmplistalg[1] = rightOp;
            return new OpSubQuery(plan, expression, tmplistalg, true, false);
        } else {
            throw new MediatorException("Unsupported type of hint : " + hintType);
        }
    }

    private void verifySubElements() throws MediatorException {
        // rules of sub queries
        // if there is no outer hint -> take the default hint for all
        // if there is a single nested -> nested for all
        // if there is a merge and no nested -> merge for all
        if (subDepFLWRs == null || subDepFLWRs.isEmpty())
            return;
        boolean hasMerge = false;
        boolean hasNested = false;
        subhandling = HintTree.NO_TYPE; //getDepManager().getPlan().defaultsubqueryhandling;
        for (int i = 0; i < subDepFLWRs.size(); i++) {
            ((AlgFLWR) subDepFLWRs.get(i)).verifySubElements();
        }
        if (subhandling != HintTree.NESTED_TYPE) {
            for (int i = 0; i < subDepFLWRs.size(); i++) {
                AlgFLWR subdepi = (AlgFLWR) subDepFLWRs.get(i);
                HintTree ht = ((FLWRExpression) subdepi.getExpression()).getHintTree();
                if (ht != null) {
                    if (!hasNested && ht.getOuterType() == HintTree.NESTED_TYPE) {
                        subhandling = HintTree.NESTED_TYPE;
                        hasNested = true;
                        break;
                    }
                    if (!hasMerge && ht.getOuterType() == HintTree.MERGE_TYPE)
                        hasMerge = true;
                }
            }
        } else
            hasNested = true;

        // check that all sub elements have the same sort items for parent element (none is ok too)
        if (hasNested)
            subhandling = HintTree.NESTED_TYPE;
        else if (hasMerge)
            subhandling = HintTree.MERGE_TYPE;
        else
            subhandling = getAlgebraManager().getPlan().getDefaultSubHandling();

        algManager.addVarsToScope(getVariables());
        for (int i = 0; i < subDepFLWRs.size(); i++) {
            AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
            depi.makeInformation();
        }
        // verify that all sub elements can do merge (no condition outside)
        IsolateWhereVisitor visitor = new IsolateWhereVisitor();
        ArrayList varListLeft = ((FLWRExpression) expression).getVariables();
        for (int i = 0; i < subDepFLWRs.size(); i++) {
            AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
            varListLeft.addAll(((FLWRExpression) depi.getExpression()).getVariables());
            visitor.reset(varListLeft);
            if (((FLWRExpression) depi.getExpression()).getWhereClause() != null) {
                try {
                    ((FLWRExpression) depi.getExpression()).getWhereClause().accept(visitor);
                } catch (XQueryException e) {
                    throw new MediatorException("Could not isolated expression : " + ((FLWRExpression) depi.getExpression()).getWhereClause());
                }
            }
            varListLeft.removeAll(((FLWRExpression) depi.getExpression()).getVariables());

            if (visitor.getRemainingExpression() != null) {
                if (getParentDepFLWR() != null)
                    getParentDepFLWR().setSubHandling(HintTree.NESTED_TYPE);
                break;
            }
        }

        algManager.removeVarsFromScope(getVariables());
    }

    //##################################################################################################################################
    //##################################################################################################################################
    //##################################################################################################################################
    private Operator createOperator(ExecutionPlan plan, ArrayList indexSpecs, ArrayList subRestrictions, ArrayList outsideRestrictions) throws MediatorException {
        //##################################################################################################################################
        //##################################################################################################################################
        //##################################################################################################################################
        Operator retval = null;
        if (dependDefinitionVars.isEmpty())
            retval = createOperator(plan);
        else if (((FLWRExpression) expression).getWhereClause() == null)
            retval = createOperator(plan);
        else if (vardependencies.hasJoins())
            retval = createOperator(plan);
        else {
            ArrayList varListLeft = ((FLWRExpression) getParentDepFLWR().getExpression()).getVariables();
            ArrayList varListRight = ((FLWRExpression) getExpression()).getVariables();
            ArrayList mergeRestrictions = new ArrayList();
            ArrayList otherRestrictions = new ArrayList();
            ArrayList leftExpressions = new ArrayList();
            ArrayList rightExpressions = new ArrayList();

            boolean hasnoteq = extractConditionsfromMap(vardependencies, varListLeft, varListRight, leftExpressions, rightExpressions, mergeRestrictions, otherRestrictions);
            if (mergeRestrictions.isEmpty()) {
                retval = createOperator(plan);
            } else {
                if (hasnoteq) {
                    ListOpCompExpression listopcomp = (ListOpCompExpression) mergeRestrictions.get(mergeRestrictions.size() - 1);
                    if (listopcomp.getOperator() == Constants.LT_COMPOP || listopcomp.getOperator() == Constants.LEQ_COMPOP) {
                        ((XQueryExpression) leftExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
                        ((XQueryExpression) rightExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
                    }
                }
                // get all conditions only on subquery
                for (int i = 0; i < varListLeft.size(); i++) {
                    Variable vari = (Variable) varListLeft.get(i);
                    HashMap tmpmap = (HashMap) vardependencies.getJoinsMap().get(vari);
                    for (Iterator it = tmpmap.values().iterator(); it.hasNext();) {
                        ArrayList tmplist = (ArrayList) it.next();
                        for (int j = 0; j < tmplist.size(); j++) {
                            if (!mergeRestrictions.contains(tmplist.get(j)) && !otherRestrictions.contains(tmplist.get(j)))
                                otherRestrictions.add(tmplist.get(j));
                        }
                    }
                }

                // add index to parent algebra
                indexSpecs.add(leftExpressions);

                // modify where expression to contains only conditions on subquery
                // and add to ouside restrictions conditions either merge not allowed or all others
                XQueryExpression whereExpr = ((FLWRExpression) expression).getWhereClause();
                try {
                    for (int i = 0; i < mergeRestrictions.size(); i++) {
                        whereExpr = whereExpr.erasePart((XQueryExpression) mergeRestrictions.get(i));
                    }
                    for (int i = 0; i < otherRestrictions.size(); i++) {
                        whereExpr = whereExpr.erasePart((XQueryExpression) otherRestrictions.get(i));
                    }
                } catch (XQueryException e) {
                    throw new MediatorException("Could not erase part from expression : " + whereExpr);
                }
                ((FLWRExpression) expression).setWhereClause(whereExpr);
                //outsideRestrictions.add(otherRestrictions);
                if (otherRestrictions.isEmpty()) {
                    outsideRestrictions.add(null);
                } else {
                    if (otherRestrictions.size() == 1)
                        outsideRestrictions.add(otherRestrictions.get(0));
                    else {
                        try {
                            BinOpANDExpression tmpBinOp = new BinOpANDExpression((XQueryExpression) otherRestrictions.get(0), (XQueryExpression) otherRestrictions.get(1), expression.getParentModule());
                            for (int i = 2; i < otherRestrictions.size(); i++)
                                tmpBinOp = new BinOpANDExpression(tmpBinOp, (XQueryExpression) otherRestrictions.get(i), expression.getParentModule());
                            outsideRestrictions.add(tmpBinOp);
                        } catch (XQueryException e) {
                            throw new MediatorException("Could not aggregate list of expressions : " + otherRestrictions);
                        }
                    }
                }

                //				if (localRestrictions.isEmpty()) {
                //					((FLWRExpression)getExpression()).setWhereClause(null);
                //				}
                //				else {
                //					if (localRestrictions.size() == 1)
                //						((FLWRExpression)getExpression()).setWhereClause((XQueryExpression)localRestrictions.get(0));
                //					else {
                //						try {
                //							BinOpANDExpression tmpBinOp = new BinOpANDExpression((XQueryExpression) localRestrictions.get(0), (XQueryExpression) localRestrictions.get(1), depmanager.getTypeVisitor());
                //							for (int i = 2; i < localRestrictions.size(); i++)
                //								tmpBinOp = new BinOpANDExpression(tmpBinOp, (XQueryExpression) localRestrictions.get(i), depmanager.getTypeVisitor());
                //							((FLWRExpression)getExpression()).setWhereClause(tmpBinOp);
                //						} catch (XQueryException e) {
                //							throw new MediatorException("Could not aggregate list of expressions : " + localRestrictions);
                //						}
                //					}
                //				}

                retval = createOperator(plan);
                retval.addIndexSpec(rightExpressions);
                subRestrictions.add(mergeRestrictions);
            }
        }
        return retval;
    }

    public Operator createOperator(ExecutionPlan plan) throws MediatorException {
        // definition of return value
        Operator operator = null;
        // gather information concerning where expression if any
        makeInformation();
        // gather information on subelements if any
        if (getParentDepFLWR() == null)
            verifySubElements();

        Operator[] subOperators = null;
        ArrayList subRestrictions = null;
        ArrayList outsideExpressions = null;
        ArrayList indexSpecs = null;

        if (subDepFLWRs != null && !subDepFLWRs.isEmpty()) {

            algManager.addVarsToScope(((FLWRExpression) expression).getVariables());

            if (subhandling == HintTree.NESTED_TYPE) {
                subOperators = new Operator[subDepFLWRs.size() + 1];
                //subElements.add(algebra);
                //ArrayList subWhereExpressions = new ArrayList();
                for (int i = 0; i < subDepFLWRs.size(); i++) {
                    AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
                    ArrayList varsi = ((FLWRExpression) depi.getExpression()).getVariables();
                    //					if (log.isDebugEnabled()) {
                    //						log.debug("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
                    //						log.debug("ClassName = " + depi.getClass().getName());
                    //						log.debug("is let = " + depi.isLet());
                    //						log.debug("depiexpr = " + depi.getExpression());
                    //						log.debug("vari = " + depi.getVariables());
                    //						log.debug("definition vars depending on = " + depi.getVarsDependingOn());
                    //						log.debug("where vars depending on = " + depi.getVarsWhereOn());
                    //						log.debug("vars referenced by = " + depi.getVarsReferencedBy());
                    //						log.debug("children = " + depi.getChildren());
                    //					}
                    Operator subAlg = depi.createOperator(plan);
                    if (!depi.getVarsDependingOn().isEmpty()) // add 10/02/2003
                        subAlg.setValueDepends(true);
                    if (!depi.getVarsWhereOn().isEmpty()) // add 10/02/2003
                        subAlg.setWhereDepends(true);
                    subOperators[i + 1] = subAlg;
                    //subWhereExpressions.add(depi.getWhereClause());
                }
            }
            if (subhandling == HintTree.MERGE_TYPE) {
                subOperators = new Operator[subDepFLWRs.size() + 1];
                //subElements.add(algebra);
                if (outsideExpressions == null)
                    outsideExpressions = new ArrayList();
                if (indexSpecs == null)
                    indexSpecs = new ArrayList();
                subRestrictions = new ArrayList();
                for (int i = 0; i < subDepFLWRs.size(); i++) {
                    AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
                    ArrayList varsi = ((FLWRExpression) depi.getExpression()).getVariables();
                    //					if (log.isDebugEnabled()) {
                    //						log.debug("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
                    //						log.debug("ClassName = " + depi.getClass().getName());
                    //						log.debug("is let = " + depi.isLet());
                    //						log.debug("depiexpr = " + depi.getExpression());
                    //						log.debug("vari = " + depi.getVariables());
                    //						log.debug("definition vars depending on = " + depi.getVarsDependingOn());
                    //						log.debug("where vars depending on = " + depi.getVarsWhereOn());
                    //						log.debug("vars referenced by = " + depi.getVarsReferencedBy());
                    //						log.debug("children = " + depi.getChildren());
                    //					}
                    Operator subAlg = depi.createOperator(plan, indexSpecs, subRestrictions, outsideExpressions);
                    if (!depi.getVarsDependingOn().isEmpty()) // add 10/02/2003
                        subAlg.setValueDepends(true);
                    if (!depi.getVarsWhereOn().isEmpty()) // add 10/02/2003
                        subAlg.setWhereDepends(true);
                    subOperators[i + 1] = subAlg;
                }

            }
            algManager.removeVarsFromScope(((FLWRExpression) expression).getVariables());
        }

        //if (hinttree != null && hinttree.getType() != HintTree.NO_TYPE) {
        operator = createOperator(plan, hinttree, ((FLWRExpression) expression).getVariables());
        if (operator != null) {
            // if flwr variables are not in modified hint tree create standard algebra
            if (hinttree != null) {
                ArrayList othervars = (ArrayList) ((FLWRExpression) expression).getVariables();
                if (!varHintList.containsAll(othervars)) {
                    othervars.removeAll(varHintList);
                    operator = createVarsOperator(plan, operator, varHintList, othervars);
                }
            }
        }
        if (operator == null) {
            // create order of variable depending on sort list, dependencies and declaration
            if (!isMonoSource()) {
                computeVarOrderList();
                verifyVarOrderList();
            } else {
                varOrderList = new ArrayList();
                varOrderList.add(((FLWRExpression) expression).getVariables().get(0));
            }
            //			if (log.isDebugEnabled()) {
            //				log.debug("********** WHERE STUFF **************************");
            //				log.debug("where = " + ((FLWRExpression) expression).getWhereClause());
            //				log.debug("vardependencies = " + vardependencies);
            //				log.debug("varOrderMap = " + varOrderMap);
            //				log.debug("orderVarMap = " + orderVarMap);
            //				log.debug("varOrderList = " + varOrderList);
            //				log.debug("varReOrderList = " + varReOrderList);
            //				log.debug("children = " + algChildren);
            //			}
            ArrayList varlist = new ArrayList(varOrderList.size());
            for (int i = 0; i < varOrderList.size(); i++) {
                operator = createVarOperator(plan, operator, varlist, (Variable) varOrderList.get(i));
                algManager.getScopeVariables().add(varOrderList.get(i));
            }
            algManager.getScopeVariables().removeAll(varOrderList);
        }

        if (subOperators != null) {
            subOperators[0] = operator;
            if (subhandling == HintTree.NESTED_TYPE)
                operator = new OpSubQuery(plan, expression, subOperators, true, true);
            if (subhandling == HintTree.MERGE_TYPE) {
                if (subRestrictions.isEmpty())
                    operator = new OpSubQuery(plan, expression, subOperators, true, true);
                else {
                    operator.addIndexSpecs(indexSpecs);
                    operator = new OpSubQuery(plan, expression, subOperators, true, true, subRestrictions, outsideExpressions);
                }
            }
        }

        // if it is a root node, than we must add final projection
        if (parentDepFLWR == null) {
            if (parent != null && (variables != null && !variables.isEmpty())) {
                // start add 30/05/2003
                boolean retvalue = false;
                XQueryExpression retexpr = ((FLWRExpression) expression).getReturnClause();
                while (!retvalue) {
                    if (retexpr instanceof LocatedExpression || retexpr instanceof Variable)
                        retvalue = true;
                    else {
                        if (retexpr instanceof XQueryExpressionSequence) {
                            XQueryExpressionSequence retseq = (XQueryExpressionSequence) retexpr;
                            if (retseq.getSubExpressions().size() == 1)
                                retexpr = (XQueryExpression) retseq.getSubExpressions().get(0);
                            else
                                break;
                        } else
                            break;
                    }
                }
                if (retvalue)
                    operator = new OpNotSource(plan, retexpr, (Variable) variables.get(0), operator, islet);
                else
                    // end add 30/05/2003
                    operator = new OpNotSource(plan, null, (Variable) variables.get(0), operator, islet);
            }
        }
        if (varReOrderList != null && !varReOrderList.isEmpty()) {
            ArrayList neworders = new ArrayList();
            ArrayList orders = ((FLWRExpression) expression).getOrderBy();
            boolean doit = false;
            if (orders != null)
                for (int i = 0; i < orders.size(); i++) {
                    Variable tmpvar = null;
                    XQueryExpression tmpexpr = null;
                    if (orders.get(i) instanceof Variable) {
                        tmpvar = (Variable) orders.get(i);
                        tmpexpr = tmpvar;
                    } else {
                        tmpexpr = (XQueryExpression) orders.get(i);
                        tmpvar = (Variable) ((LocatedExpression) tmpexpr).getExpression();
                    }
                    if (doit || varReOrderList.contains(tmpvar)) {
                        if (!doit)
                            doit = true;
                        neworders.add(tmpexpr);
                    }
                }
            if (!neworders.isEmpty())
                operator = new OpSort(plan, operator, neworders);
        }
        if (!isQDB())
            operator.isLet(islet);
        return operator;
    }

    public Object clone() throws CloneNotSupportedException {
        AlgFLWR newobj = (AlgFLWR) super.clone();
        newobj.parentDepFLWR = parentDepFLWR;
        if (subDepFLWRs == null)
            newobj.subDepFLWRs = null;
        else {
            newobj.subDepFLWRs = new ArrayList();
            for (int i = 0; i < subDepFLWRs.size(); i++)
                newobj.subDepFLWRs.add(((AlgFLWR) subDepFLWRs.get(i)).clone());
        }
        return newobj;
    } // ************************************************************************

    // * METHODS
    // ************************************************************************
    /**
     * 
     * @return
     */
    public ArrayList getPaths() {
        if (paths == null)
            paths = new ArrayList();
        if (paths.isEmpty()) {
            if (algChildren != null) {
                for (int i = 0; i < algChildren.size(); i++) {
                    ArrayList tmplist = ((Algebra) algChildren.get(i)).getPaths();
                    if (tmplist != null)
                        paths.addAll(tmplist);
                }
            }
            if (subDepFLWRs != null) {
                for (int i = 0; i < subDepFLWRs.size(); i++) {
                    ArrayList tmplist = ((AlgFLWR) subDepFLWRs.get(i)).getPaths();
                    if (tmplist != null)
                        paths.addAll(tmplist);
                }
            }
        }
        return paths;
    }

    /*
     * public ArrayList getVariables() { return ((FLWRExpression) expression).getVariables(); }
     */

    public ArrayList getAllVariables() {
        ArrayList res = (ArrayList) ((FLWRExpression) expression).getVariables().clone();
        if (subDepFLWRs != null)
            for (int i = 0; i < subDepFLWRs.size(); i++)
                res.addAll(((AlgFLWR) subDepFLWRs.get(i)).getAllVariables());
        return res;
    }

    public AlgFLWR getParentDepFLWR() {
        return parentDepFLWR;
    }

    public ArrayList getAllOrderBys() {
        ArrayList parentres = ((FLWRExpression) expression).getOrderBy();
        if (subDepFLWRs == null)
            return parentres;
        ArrayList res = new ArrayList();
        for (int i = 0; i < subDepFLWRs.size(); i++) {
            ArrayList tmpres = ((AlgFLWR) subDepFLWRs.get(i)).getAllOrderBys();
            if (tmpres != null) {
                res.addAll(tmpres);
                if (parentres != null)
                    res.removeAll(parentres);
            }
        }
        if (parentres != null)
            res.addAll(0, parentres);
        if (res.isEmpty())
            return null;
        return res;
    } // ******************************************************************

    // * DEBUGGING
    // ******************************************************************
    public String toCompleteString() {
        return toCompleteString(0);
    }

    public String toCompleteString(int indent) {
        StringBuffer buf = new StringBuffer();
        buf.append(Utils.makeIndent(indent) + "<" + getClass().getName() + ">\n");
        buf.append(Utils.makeIndent(indent + 1) + "<variables> " + variables + " </variables>\n");
        buf.append(Utils.makeIndent(indent + 1) + "<sources> " + getSources() + " </sources>\n");
        buf.append(Utils.makeIndent(indent + 1) + "<isLet> " + islet + " </isLet>\n");
        buf.append(Utils.makeIndent(indent + 1) + "<monosource> " + isMonoSource() + " </monosource>\n");
        buf.append(Utils.makeIndent(indent + 1) + "<hasIdentifier> " + hasIdentifier + " </hasIdentifier>\n");
        if (hinttree != null) {
            buf.append(Utils.makeIndent(indent + 1) + "<HintTree>\n");
            buf.append(Utils.makeIndent(indent + 2) + hinttree.toString() + "\n");
            buf.append(Utils.makeIndent(indent + 1) + "</HintTree>\n");
        }
        buf.append(Utils.makeIndent(indent + 1) + "<Expression type=\"" + expression.getClass() + "\">\n");
        buf.append(Utils.makeIndent(indent + 2) + expression + "\n");
        buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n");
        if (algChildren != null && !algChildren.isEmpty()) {
            buf.append(Utils.makeIndent(indent + 1) + "<Leaves>\n");
            for (int i = 0; i < algChildren.size(); i++) {
                buf.append(Utils.makeIndent(indent + 2) + "<Leave>\n");
                Algebra depchild = (Algebra) algChildren.get(i);
                buf.append(depchild.toCompleteString(indent + 3));
                buf.append(Utils.makeIndent(indent + 2) + "</Leave>\n");
            }
            buf.append(Utils.makeIndent(indent + 1) + "</Leaves>\n");
        }
        if (referenceDefinitionVars != null && !referenceDefinitionVars.isEmpty()) {
            buf.append(Utils.makeIndent(indent + 1) + "<referenceDefinitionVars>\n");
            for (int i = 0; i < referenceDefinitionVars.size(); i++) {
                buf.append(Utils.makeIndent(indent + 2) + "<Variables>" + referenceDefinitionVars.get(i) + "</Variables>\n");
            }
            buf.append(Utils.makeIndent(indent + 1) + "</referenceDefinitionVars>\n");
        }
        if (dependDefinitionVars != null && !dependDefinitionVars.isEmpty()) {
            buf.append(Utils.makeIndent(indent + 1) + "<dependDefinitionVars>\n");
            for (int i = 0; i < dependDefinitionVars.size(); i++) {
                buf.append(Utils.makeIndent(indent + 2) + "<Variables>" + dependDefinitionVars.get(i) + "</Variables>\n");
            }
            buf.append(Utils.makeIndent(indent + 1) + "</dependDefinitionVars>\n");
        }
        if (paths != null && !paths.isEmpty()) {
            buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n");
            for (int i = 0; i < paths.size(); i++)
                buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n");
            buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n");
        }
        if (subDepFLWRs != null) {
            buf.append(Utils.makeIndent(indent + 1) + "<subDepFLWRs>\n");
            for (int i = 0; i < subDepFLWRs.size(); i++)
                buf.append(Utils.makeIndent(indent + 2) + "<depmem>" + ((AlgFLWR) subDepFLWRs.get(i)).getVariables() + "</depmem>\n");
            buf.append(Utils.makeIndent(indent + 1) + "</subDepFLWRs>\n");
        }
        buf.append(Utils.makeIndent(indent) + "</" + getClass().getName() + ">\n");
        return buf.toString();
    }

    public void execute(ExecutionPlan plan) throws MediatorException {
    }

}