/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan ;

import java.util.ArrayList;

import org.xquark.mediator.DOMUtils.Tuple;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.parser.XQueryExpression;

public class OpCartesian extends OpBin {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.7 $";
    private static final String RCSName = "$Name:  $";

    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************

    // ***********************************************************************
    // * INITIALIZATION
    // ***********************************************************************

    public OpCartesian(ExecutionPlan plan, XQueryExpression expression,Operator left, Operator right) throws MediatorException {
        super(plan, expression, left, right) ;
        init();
    }

    public OpCartesian(ExecutionPlan plan, XQueryExpression expression, Operator[] operators) throws MediatorException {
        super(plan, expression, operators) ;
        init();
    }

    private void init() {
        // Sometimes, it can happen that the two members of cartesian
        // product have common attributes. In that case, we must mark them
        // as common. So when we will generate the tuples, we can access
        // them efficiently.
        ArrayList childpaths1 = getLeftAlgebra().getPathsList() ;
        ArrayList childpaths2 = getRightAlgebra().getPathsList() ;
        size = getLeftAlgebra().getSize() + getRightAlgebra().getSize();

//        equivalences = new ArrayList() ;
//        int index = -1;
//        for (int i = 0 ; i < childpaths2.size() ; i ++) {
//            if ((index = childpaths1.indexOf(childpaths2.get(i))) != -1) {
//                addEquivalence(equivalences,index, i) ;
//                size--;
//                break ;
//            }
//        }

//        idequivalences = new ArrayList() ;
//        ArrayList childvars1 = getLeftAlgebra().getVariables() ;
//        ArrayList childvars2 = getRightAlgebra().getVariables() ;
        idsize = getLeftAlgebra().getIdSize() + getRightAlgebra().getIdSize();
//        for (int i=0;i<childvars2.size();i++) {
//            Variable var2i = (Variable)childvars2.get(i);
//            if ((index = childvars1.indexOf(var2i)) != -1) {
//                addEquivalence(idequivalences,index, i) ;
//                idsize--;
//            }
//        }

        ordertype = ORDER_PARALLELIZATION ;
    }

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(OperatorVisitor visitor) throws MediatorException {
		visitor.visit(this);
	}

    // ***********************************************************************
    // * EXECUTE QUERY
    // ***********************************************************************
    /**
     *
     */
    public ResultSet getResultSet(DynamicContext context, OperatorRunnable[] runnables) throws MediatorException {
        try {
            return new CartesianResultSet(this, context, runnables) ;
        }
        catch (XMLDBCException e) {
            throw new MediatorException("Can't construct " + getClass() + ": " + e.getMessage(), e) ;
        }
    }

    // **********************************************************************
    // * OPTIMIZE
    // **********************************************************************


    // **********************************************************************
    // * DEBUG
    // **********************************************************************
    /**
     *
     */
}

class CartesianResultSet extends BinResultSet {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.7 $";
    private static final String RCSName = "$Name:  $";    // * CLASS VARIABLES
    // ***********************************************************************
	private int index = -1;
	private Tuple tuplepivot = null;
	ArrayList tuplelist = null;
    // ***********************************************************************
    // * INITIALIZATION
    // ***********************************************************************
    /**
     *
     */
    public CartesianResultSet(OpCartesian operator, DynamicContext context, OperatorRunnable[] children) throws XMLDBCException {
        super(operator, context, children) ;
    }

    // ***********************************************************************
    // * EVALUATE IMPLEMENTATION
    // ***********************************************************************
    /**
     * In a cartesian product, we can all do in parallel, so we
     * must generate all results of both source and then make the cartesian
     * product.
     *
     * @param non_blocking if the evaluation is blocking, we must
     *        generate all the results before passing.
     *		  A cartesian product is necessary blocking.
     */
    
    protected void evaluate(boolean non_blocking) throws MediatorException {
		while (true) {
			Tuple tuple = null;
			if (index == -1) {
				// verify that tuplepivot is not null and set it (if not finish)
				if (tuplepivot == null) {
					if (children[0].getResultSet().hasNext())
						tuplepivot = children[0].getResultSet().next();
					else {
						setFinishedWhenEmpty();
						children[0].getResultSet().close();
						children[1].getResultSet().close();
						return;
					}
				}
				if (children[1].getResultSet().hasNext()) {
					tuple = children[1].getResultSet().next();
					if (tuplelist == null)
						tuplelist = new ArrayList();
					tuplelist.add(tuple);
				} else {
					if (tuplelist != null && children[0].getResultSet().hasNext()) {
						tuplepivot = children[0].getResultSet().next();
						index = 0;
						tuple = (Tuple) tuplelist.get(index++);
					} else {
						setFinishedWhenEmpty();
						children[0].getResultSet().close();
						children[1].getResultSet().close();
						return;
					}
				}
			} else {
				if (index < tuplelist.size())
					tuple = (Tuple) tuplelist.get(index++);
				else {
					if (children[0].getResultSet().hasNext()) {
						tuplepivot = children[0].getResultSet().next();
						index = 0;
						tuple = (Tuple) tuplelist.get(index++);
					} else {
						setFinishedWhenEmpty();
						children[0].getResultSet().close();
						return;
					}
				}
			}
			buftuples.add(tuplepivot.merge(tuple, operator));
			if (non_blocking)
				break;
		}   	    	
    }
}
