/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan ;

import java.util.ArrayList;

import org.xquark.mediator.decomposer.Utils;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.xquery.parser.XQueryExpression;

public abstract class OpUn extends Operator {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.5 $";
    private static final String RCSName = "$Name:  $";
    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************
    protected Operator childOperator = null ;
    
    // ***********************************************************************
    // * INITIALIZATION
    // ***********************************************************************
    /**
     *
     */
    public OpUn(ExecutionPlan plan, XQueryExpression expression, Operator childOperator) throws MediatorException {
        super(plan, expression) ;
        this.childOperator = childOperator ;
        if (childOperator != null) {
			childOperator.setParentOperator(this);
            addPaths(childOperator.getPaths());
        }
    }
    
	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(OperatorVisitor visitor) throws MediatorException {
		visitor.visit(this);
	}

    // ***********************************************************************
    // * EXECUTE QUERY
    // ***********************************************************************
    /**
     *
     */
    public ResultSet executeQuery(DynamicContext context) throws MediatorException {
        if (childOperator != null) {
			OperatorRunnable or = childOperator.getOperatorRunnable(context);
			or.run() ;
			return getResultSet(context, or);
        }
        return getResultSet(context, null);
    }
    
    /**
	 * @param or
	 * @return
	 */
	protected abstract ResultSet getResultSet(DynamicContext context, OperatorRunnable or) throws MediatorException;

	// ***********************************************************************
    // * METHODS
    // ***********************************************************************
    /**
     *
     */
    public Operator getChildOperator() { return childOperator ; }
    
    // **********************************************************************
    // * OPTIMIZE
    // **********************************************************************
    /**
     * If the *exactly* same set of source is used for an operation,
     * then the subtree can be passed entierely to the source.
     * Beware of processing this from the leaf to the node.
     *
     * <p>Note: If this method is not overloaded or if it returns null,
     * it means that _this_ node will not be optimized.</p>
     */
//    public Operator optimizeSource() throws MediatorException {
//        Operator workexp = childOperator.optimizeSource() ;
//        if (workexp != null) childOperator = workexp ;
//                
//        /*
//        if (childalgebra.isSource()) {
//            Algebra newnode = mergeWithSub() ;
//            return newnode ;
//        }
//        */
//        return null ;
//    }
    
    /**
     * Add projections on attributes the soonest as possible in order to
     * manipulate the less data as possible. As soon an attribute isn't
     * necessary, it must be unselected (by a projection on the complementary
     * variables).
     *
     * <p>Note: If this method is not overloaded or if it returns null,
     * it means that _this_ node will not be optimized.</p>
     */
//    public Operator optimizeProjection() throws MediatorException {
//        return null ;
//    }
    
    /**
     *
     *
     * <p>Note: If this method is not overloaded or if it returns null,
     * it means that _this_ node will not be optimized.</p>
     */
//    public Operator optimizeRestriction() throws MediatorException {
//        return null ;
//    }
    
    
    /**
     * If the set of sources handled by this node are exactly the same (order
     * not important) then return true.
     *
     * An AlgUnOp is alway the same value than the value of its children.
     */
//    protected boolean shareSameSources() {
//        return childOperator.shareSameSources() ;
//    }
    
    /**
     * Return true if the node is of type source. An AlgUnOp is not a source.
     * @return false.
     */
    public boolean isSource() {
        return false ;
    }
    
    /**
     * Return all the sources depending of this node.
     */
    public ArrayList getSources() {
        if (sources == null)
            sources = childOperator.getSources() ;
        return sources ;
    }
    
    public void terminate() {
    	if (childOperator != null) 
    		childOperator.terminate();
    }

	public void setPrepared() throws MediatorException {
		if (prepared) return;
		prepared = true;
		if (childOperator != null) 
			childOperator.setPrepared();
	}
    
    public String toCompleteString(int indent) {
        StringBuffer buf = new StringBuffer() ;
        buf.append(Utils.makeIndent(indent) + "<" + getClass().getName() + " size=" + size + " id size=" + idsize + " isLet=\"" + islet + "\" valueDepends=\"" + valueDepends + "\" whereDepends=\"" + whereDepends + "\">\n") ;
        buf.append(Utils.makeIndent(indent + 1) + "<Expression>\n") ;
        buf.append(Utils.makeIndent(indent + 2) + expression + "\n") ;
        buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n") ;
//        if (variables != null) {
//            buf.append(Utils.makeIndent(indent + 1) + "<Variables>\n") ;
//            for (int i = 0 ; i < variables.size() ; i ++) {
//                buf.append(Utils.makeIndent(indent + 2) + "<Variable>" + variables.get(i) + "</Variable>\n") ;
//            }
//            buf.append(Utils.makeIndent(indent + 1) + "</Variables>\n") ;
//        }
        if (paths != null) {
            buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n") ;
            for (int i = 0 ; i < paths.size() ; i ++) {
                buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n") ;
            }
            buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n") ;
        }
        buf.append(Utils.makeIndent(indent + 1) + "<Element>\n") ;
        if (this.childOperator != null) buf.append(this.childOperator.toCompleteString(indent + 2)) ;
        buf.append(Utils.makeIndent(indent + 1) + "</Element>\n") ;        
        buf.append(Utils.makeIndent(indent) + "</" + getClass().getName() + ">\n") ;
        return buf.toString() ;
    }
}
