/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan;

import java.util.ArrayList;

import org.xquark.mediator.algebra.Algebra;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.schema.SchemaManager;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.typing.TypeVisitor;
public abstract class Operator {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.12 $";
	private static final String RCSName = "$Name:  $";

	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	public final static int ORDER_SERIALIZATION = 0;
	public final static int ORDER_PARALLELIZATION = 1;

	protected XQueryExpression expression = null;
	protected ExecutionPlan plan = null;
	protected Operator parentOperator = null;
	protected ArrayList paths = null;
	protected ArrayList pathslist = null;

	protected ArrayList sources = null;
	protected boolean islet = false;

	protected int size = 0;
	//protected ArrayList equivalences = null;
	protected int idsize = 0;
	//protected ArrayList idequivalences = null;

	protected ArrayList indexesSpec = null;
	protected boolean valueDepends = false;
	protected boolean whereDepends = false;
	
	protected boolean prepared = false;
	
	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public Operator(ExecutionPlan plan, XQueryExpression expression) throws MediatorException {
		super();
		this.plan = plan;
		this.expression = expression;
		resetPaths();
	}

	// ***********************************************************************
	// * GET/SET METHODS
	// ***********************************************************************

	public XQueryExpression getExpression() {
		return expression;
	}

	public void setExpression(XQueryExpression expression) {
		this.expression = expression;
	}

	public int getSize() {
		return size;
	}
	public int getIdSize() {
		return idsize;
	}
	public void setDepends(Algebra depnode) {
		valueDepends = !depnode.getVarsDependingOn().isEmpty();
		whereDepends = !depnode.getVarsWhereOn().isEmpty();
	}
	public boolean getValueDepends() {
		return valueDepends;
	}
	public void setValueDepends(boolean valueDepends) {
		this.valueDepends = valueDepends;
	}
	public boolean getWhereDepends() {
		return whereDepends;
	}
	public void setWhereDepends(boolean whereDepends) {
		this.whereDepends = whereDepends;
	}
	public boolean getDepends() {
		return valueDepends || whereDepends;
	}
	// methods for indexes handling
	public ArrayList getIndexesSpec() {
		return this.indexesSpec;
	}
	public ArrayList getIndexSpec(int index) {
		if (indexesSpec == null || indexesSpec.size() <= index)
			return null;
		return (ArrayList) this.indexesSpec.get(index);
	}
	public int addIndexSpec(ArrayList list) {
		if (this.indexesSpec == null)
			this.indexesSpec = new ArrayList();
		this.indexesSpec.add(list);
		return this.indexesSpec.size();
	}
	public int addIndexSpecs(ArrayList list) {
		if (this.indexesSpec == null)
			this.indexesSpec = new ArrayList();
		this.indexesSpec.addAll(list);
		return this.indexesSpec.size();
	}
	// methods to handle parent operator
	public void setParentOperator(Operator operator) {
		this.parentOperator = operator;
	}
	public Operator getParentOperator() {
		return parentOperator;
	}
	// methods to handle prepared status
	public abstract void setPrepared() throws MediatorException;
	
//	public boolean getPrepared() {
//		return prepared;
//	}
	
	public OperatorRunnable getOperatorRunnable(DynamicContext ctx) {
		return new OperatorRunnable(this,ctx);		
	}
	
	// ***********************************************************************
	// * EXECUTE QUERY
	// ***********************************************************************
	/**
	 *
	 */
	protected abstract ResultSet executeQuery(DynamicContext context) throws MediatorException;

	public Operator addCondition(ExecutionPlan plan, XQueryExpression expr) throws MediatorException {
		return new OpRestrict(plan, expr, this);
	}

	// **********************************************************************
	// * DEBUG
	// **********************************************************************
	/**
	 *
	 */
	public String toCompleteString() {
		return toCompleteString(0);
	}

	/**
	 *
	 */
	public String toCompleteString(int indent) {
		StringBuffer buf = new StringBuffer();
		buf.append(Utils.makeIndent(indent) + "<Class>" + getClass() + "</Class>");
		return buf.toString();
	}

	// ***********************************************************************
	// * METHODS
	// ***********************************************************************
	public void resetPaths() {
		if (this.paths != null)
			this.paths.clear();
		else
			this.paths = new ArrayList();
		if (this.pathslist != null)
			this.pathslist.clear();
		else
			this.pathslist = new ArrayList();
	}
	public ArrayList getPaths() {
		return paths;
	}
	public ArrayList getPathsList() {
		return pathslist;
	}
	public void addPath(XQueryExpression pathi) {
		if (pathi == null)
			return;
		if (!pathslist.contains(pathi.getStringValue())) {
			paths.add(pathi);
			pathslist.add(pathi.getStringValue());
		}
	}
	public void addPaths(ArrayList addpaths) {
		if (addpaths == null || addpaths.isEmpty())
			return;
		for (int i = 0; i < addpaths.size(); i++)
			addPath((XQueryExpression) addpaths.get(i));
	}
	public ExecutionPlan getPlan() {
		return plan;
	}
	public TypeVisitor getTypeVisitor() {
		return plan.getTypeVisitor();
	}
	public SchemaManager getSchemaManager() {
		return plan.getSchemaManager();
	}

	// **********************************************************************
	// * OPTIMIZE
	// **********************************************************************
	/**
	 *
	 */
//	public Operator optimize() throws MediatorException {
//		Operator newnode = this;
//		Operator worknode = this;
//
//		// make projection as soon as possible
//		worknode = newnode.optimizeProjection();
//		if (worknode != null)
//			newnode = worknode;
//
//		// make clever restriction
//		worknode = newnode.optimizeRestriction();
//		if (worknode != null)
//			newnode = worknode;
//
//		// group by source(s).
//		// don't work for now
//		//worknode = newnode.optimizeSource () ;
//		//if (worknode != null) newnode = worknode ;
//
//		return newnode;
//	}

	/**
	 * If the *exactly* same set of source is used for an operation,
	 * then the subtree can be passed entierely to the source.
	 * Beware of processing this from the leaf to the node.
	 *
	 * <p>Note: If this method is not overloaded or if it returns null,
	 * it means that _this_ node will not be optimized.</p>
	 */
//	public Operator optimizeSource() throws MediatorException {
//		return null;
//	}

	/**
	 * Add projections on attributes the soonest as possible in order to
	 * manipulate the less data as possible. As soon an attribute isn't
	 * necessary, it must be unselected (by a projection on the complementary
	 * variables).
	 *
	 * <p>Note: If this method is not overloaded or if it returns null,
	 * it means that _this_ node will not be optimized.</p>
	 */
//	public Operator optimizeProjection() throws MediatorException {
//		return null;
//	}

	/**
	 *
	 *
	 * <p>Note: If this method is not overloaded or if it returns null,
	 * it means that _this_ node will not be optimized.</p>
	 */
//	public Operator optimizeRestriction() throws MediatorException {
//		return null;
//	}

	/**
	 * Merge *this* node with the subnode(s). As it should be called
	 * only from source, the returned node must be of type AlgZeroOp.
	 * If it weren't called from a source, it would be a possible case of bug.
	 * As I don't know yet how to test this case, be careful when using
	 * this method.
	 * The return node is a AlgSource where the XQueryExpression is the merging
	 * of all child(ren) nodes' expression.
	 */
	//protected abstract Algebra mergeWithSub() throws MediatorException ;

	/**
	 * If the set of sources handled by this node are exactly the same (order
	 * not important) then return true.
	 * An AlgZeroOp is alway true.
	 * An AlgUnOp is alway the same value than the value of its children.
	 * An AlgBinValue is true if both of its child are true and IF the
	 * set of sources handled by its children are exactly the same.
	 */
//	protected abstract boolean shareSameSources();

	/**
	 * Return true if the node is of type source.
	 * (AlgSource or AlgValue  ie. all nodes of type AlgZeroOp).
	 */
	public abstract boolean isSource();

	/**
	 * Return all the sources depending of this node.
	 */
	public abstract ArrayList getSources();

	public void isLet(boolean islet) {
		this.islet = islet;
	}
	public boolean isLet() {
		return islet;
	}
	public abstract void terminate();

//	protected void addEquivalence(ArrayList equivalences, int i, int j) {
//		int[] couple = new int[2];
//		couple[0] = i;
//		couple[1] = j;
//		equivalences.add(couple);
//	}
}
