/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan;

import org.xml.sax.ContentHandler;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.schema.validation.PSVInfoSetProvider;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;

public class PlanResultSet implements XDBCResultSetInterface {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.10 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	protected ContentHandler handler = null;
	protected org.xml.sax.ErrorHandler error_handler = null;
	protected org.xml.sax.ext.LexicalHandler lexical_handler = null;

	protected ResultSet opResultset;

	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public PlanResultSet(ResultSet opResultset) throws XMLDBCException {
		this.opResultset = opResultset;
	}

	// ***********************************************************************
	// * XDBCResultSetInterface IMPLEMENTATION
	// ***********************************************************************
	/**
	 * Return the value of the element specified by the path.
	 * @param path
	 * @param nodeAccessor
	 * @param loopID the ID of the loop concerned if any
	 * @return value of the element
	 * @throws XDBCException if a database access error occurs.
	 */
	public String fetch(String path, int nodeAccessor) throws XMLDBCException {
		return fetch(path, nodeAccessor, null);
	}
	public String fetch(String path, int nodeAccessor, String loopID) throws XMLDBCException {
		return fetch(path, nodeAccessor, loopID, null);
	}
	public String fetch(String path, int nodeAccessor, String loopID, PSVInfoSetProvider psvisp) throws XMLDBCException {
		return opResultset.fetch(path, nodeAccessor, loopID, psvisp);
	}

	/**
	 * Generate the value of the result set specified by the path in the handler.
	 * @param path
	 * @param nodeAccessor the type of node to be generated
	 * @param loopID the ID of the loop concerned if any
	 * @throws XDBCException if a database access error occurs.
	 */
	public void generate(String path, int nodeAccessor) throws XMLDBCException {
		generate(path, nodeAccessor, null);
	}
	public void generate(String path, int nodeAccessor, String loopID) throws XMLDBCException {
		generate(path, nodeAccessor, loopID, null);
	}
	public void generate(String path, int nodeAccessor, String loopID, PSVInfoSetProvider psvisp) throws XMLDBCException {
		opResultset.generate(handler, path, nodeAccessor, loopID, psvisp);
	}

	/**
	 * Recover the identifiers of the current result set and stores them in the table <I>identifiers</I>.
	 * @param identifiers the table of identifiers
	 * @throws XDBCException if a database access error occurs.
	 */
	public void getIdentifiers(Object[] identifiers) throws XMLDBCException {
		opResultset.getIdentifiers(identifiers);
	}

	/**
	 * Recover the identifiers of the next result set and stores them in the table <I>identifiers</I>.
	 * @param identifiers the table of identifiers
	 * @throws XDBCException if a database access error occurs.
	 */
	public void getNextIdentifiers(Object[] identifiers) throws XMLDBCException {
		try {
			opResultset.getNextIdentifiers(identifiers);
		} catch (MediatorException me) {
			throw new XMLDBCException(me.getMessage(), me);
		}
	}

	/**
	 * Give the new current result set (the next one).
	 * @return true if the new current result set is valid, false otherwise
	 * @throws XDBCException if a database access error occurs.
	 */
	public boolean next() throws XMLDBCException {
		try {
			opResultset.next();
		} catch (MediatorException me) {
			throw new XMLDBCException(me.getMessage(), me);
		}
		return true;
	}

	/**
	 * Returns true if the result set has more elements.
	 * @return true if the result set has more elements, false otherwise
	 */
	public boolean hasNext() throws XMLDBCException {
		try {
			return opResultset.hasNext();
		} catch (MediatorException me) {
			throw new XMLDBCException(me.getMessage(), me);
		}
	}

	/**
	 * To set a content handler (SAX 2) to intercept events produce when retrieves results by SAX 2 way.
	 * @param handler the content handler implementation (see org.xml.sax.ContentHandler)
	 * @throws XDBCException if a database access error occurs.
	 */
	public void setContentHandler(ContentHandler handler) {
		this.handler = handler;
	}

	/**
	 * To set a content handler (SAX 2) to intercept events produce when
	 * retrieves results by SAX 2 way.
	 * @param handler the error handler implementation
	 * (see org.xml.sax.ErrorHan dler)
	 * @throws XDBCException if a database access error occurs.
	 */
	public void setErrorHandler(org.xml.sax.ErrorHandler handler) {
		this.error_handler = handler;
	}

	/**
	 * To set a lexical handler (SAX 2) to intercept events produce when
	 * retrieves results by SAX 2 way.
	 * @param handler the lexical handler implementation
	 * (see org.xml.sax.ext.LexicalHandler)
	 * @throws XDBCException if a database access error occurs.
	 */
	public void setLexicalHandler(org.xml.sax.ext.LexicalHandler handler) {
		this.lexical_handler = handler;
	}

	/**
	 * close the associated algresultset
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() throws XMLDBCException {
		try {
			opResultset.close();
		} catch (MediatorException me) {
			throw new XMLDBCException(me.getMessage(), me);
		}
	}
}
