/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan;

import java.util.ArrayList;

import org.xml.sax.ContentHandler;
import org.xquark.mediator.DOMUtils.*;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.schema.validation.PSVInfoSetProvider;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.parser.Step;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.typing.QAtomicSerializer;

public abstract class ResultSet {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.16 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	protected TypedDOMReader dom2sax = null;
	protected QAtomicSerializer ser = null;

	protected XQueryExpression expression = null;
	protected Operator operator;
	private boolean first_time = true;
	private boolean finished_when_empty = false;

	protected BufferTuple buftuples = null;
	private Tuple cur_tuple = null;
	private Tuple next_tuple = null;
	protected ArrayList paths;

	protected DynamicContext context = null;
	
	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public ResultSet(Operator operator, DynamicContext context) {
		this.operator = operator;
		this.context = context;
		dom2sax = new TypedDOMReader();
		ser = new QAtomicSerializer();
		dom2sax.setQAtomicSerializer(ser);
		init(operator);
	}

	public void init(Operator operator) {
		if (operator != null) {
			expression = operator.getExpression();
			paths = operator.getPaths();
			buftuples = new BufferTuple(paths, operator.getIdSize());
			buftuples.setIndexesSpec(operator.getIndexesSpec());
		}
	}

	protected void generateTuple() throws MediatorException {
		// if the buffer is not empty, use it before generating the next elements
		if (!buftuples.isEmpty())
			return;

		// a LET is a blocking case.
		evaluate(!operator.isLet());
	}

	/**
	 * Fill the buffer buftuples with tuple.
	 *
	 * @param non_blocking if the evaluation is blocking, we must
	 *        generate all the results before passing.
	 */
	protected abstract void evaluate(boolean non_blocking) throws MediatorException;

	protected void setFinishedWhenEmpty() throws MediatorException {
		this.finished_when_empty = true;
	}

	public Operator getOperator() {
		return operator;
	}

	public SortedKeyMap getIndex(int index) {
		return buftuples.getIndex(index);
	}

	/**
	 * Give the new current result set (the next one).
	 * @return true if the new current result set is valid, false otherwise
	 * @throws XDBCException if a database access error occurs.
	 */
	public Tuple next() throws MediatorException {
		if (next_tuple != null) {
			cur_tuple = next_tuple;
			next_tuple = null;
			return cur_tuple;
		}

		if (first_time) {
			hasNext();
			if (buftuples.isEmpty())
				throw new MediatorException("No more results");
		}

		// CHANGE   04/02/2003
		if (buftuples.isEmpty()) {
			generateTuple();
			if (buftuples.isEmpty())
				throw new MediatorException("No more results");
		}

		if (!buftuples.isEmpty()) {
			cur_tuple = (Tuple) buftuples.remove(0);
			return cur_tuple;
		}

		return cur_tuple;
	}

	/**
	 * Returns true if the result set has more elements.
	 * @return true if the result set has more elements, false otherwise
	 */
	public boolean hasNext() throws MediatorException {

		if (next_tuple != null)
			return true;

		if (first_time) {
			generateTuple();
			first_time = false;
		}

		if ((finished_when_empty) && (buftuples == null || buftuples.isEmpty()))
			return false;

		if (buftuples.isEmpty())
			generateTuple();

		return (!((finished_when_empty) && (buftuples == null || buftuples.isEmpty())));
	}

	/**
	 * Releases this result set object's resources immediately instead of waiting
	 * for this to happen when it is automatically closed.<BR>
	 * <B>Note:</B> A XResultSet object is automatically closed by the XStatement
	 * object that generated it when that XStatement object is closed,
	 * re-executed, or is used to retrieve the next result from a sequence of
	 * multiple results.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() throws MediatorException {
		buftuples = null;
	}

	/**
	 * Return the value of the element specified by the path.
	 * @param path
	 * @param nodeAccessor
	 * @param loopID the ID of the loop concerned if any
	 * @return value of the element
	 * @throws XDBCException if a database access error occurs.
	 */
	public String fetch(String path, int nodeAccessor, String loopID, PSVInfoSetProvider psvisp) throws XMLDBCException {
		String alias = path;
		while ((alias = operator.getPlan().getAlias(path)) != null) {
			path = alias;
		}
		//if (alias != null) path = alias;
		int pos = buftuples.getPos(path);
		if (pos != -1)
			return cur_tuple.makeStringPath(pos, ser);
		switch (nodeAccessor) {
			case Step.CHILD_TEXT :
			//case XDBCResultSetInterface.NODE_ACCESSOR_TEXT :
				pos = buftuples.getPos(path + "/child::text()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_PROCESSING_INSTRUCTION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_COMMENT : break;
				*/
			case Step.CHILD_NODE :
			//case XDBCResultSetInterface.NODE_ACCESSOR_NODE :
				pos = buftuples.getPos(path + "/child::node()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_LOCAL_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PARENT : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_STRING_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_BASE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_CHILDREN : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_ATTRIBUTES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_DECLARATION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PREFIX : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_URI : break;
				*/
			case Step.SELF_TEXT :
			//case XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT :
				pos = buftuples.getPos(path + "/self::text()");
				break;
			case Step.SELF_NODE :
			//case XDBCResultSetInterface.NODE_ACCESSOR_SELF_NODE :
				pos = buftuples.getPos(path + "/self::node()");
				break;
		}
		if (pos == -1)
			pos = buftuples.getPos(path);
		if (pos == -1)
			throw new XMLDBCException("No column " + path + " in tuple.");
		return cur_tuple.makeStringPath(pos, ser);
		//throw new XMLDBCException("Shouldn't be called: a tree can't be cast to a String") ;
	}

	/**
	 * Generate the value of the result set specified by the path in the handler.
	 * @param path
	 * @param nodeAccessor the type of node to be generated
	 * @param loopID the ID of the loop concerned if any
	 * @throws XDBCException if a database access error occurs.
	 */
	public void generate(ContentHandler handler, String path, int nodeAccessor, String loopID, PSVInfoSetProvider psvisp) throws XMLDBCException {
		String alias = path;
		while ((alias = operator.getPlan().getAlias(path)) != null) {
			path = alias;
		}
		//if (alias != null) path = alias;
		int pos = -1;
		switch (nodeAccessor) {
			case Step.CHILD_TEXT :
			//case XDBCResultSetInterface.NODE_ACCESSOR_TEXT :
				pos = buftuples.getPos(path + "/child::text()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_PROCESSING_INSTRUCTION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_COMMENT : break;
				*/
			case Step.CHILD_NODE :
			//case XDBCResultSetInterface.NODE_ACCESSOR_NODE :
				pos = buftuples.getPos(path + "/child::node()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_LOCAL_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PARENT : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_STRING_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_BASE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_CHILDREN : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_ATTRIBUTES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_DECLARATION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PREFIX : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_URI : break;
				*/
			case Step.SELF_TEXT :
			//case XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT :
				pos = buftuples.getPos(path + "/self::text()");
				break;
			case Step.SELF_NODE :
			//case XDBCResultSetInterface.NODE_ACCESSOR_SELF_NODE :
				pos = buftuples.getPos(path + "/self::node()");
				break;
		}
		if (pos == -1)
			pos = buftuples.getPos(path);
		if (pos == -1)
			throw new XMLDBCException("No column " + path + " in tuple.");
		try {
			dom2sax.setPSVInfoSetProvider(psvisp);
			cur_tuple.makeEventPath(dom2sax, handler, pos);
		} catch (XQueryException e) {
			throw new XMLDBCException(e.getMessage());
		}
	}

	/**
	 * Recover the identifiers of the current result set and stores them in the table <I>identifiers</I>.
	 * @param identifiers the table of identifiers
	 * @throws XDBCException if a database access error occurs.
	 */
	public void getIdentifiers(Object[] identifiers) {
		if (cur_tuple == null)
			return;
		cur_tuple.getIdentifiers(identifiers);
	}

	/**
	 * Recover the identifiers of the next result set and stores them in the tab
	 * le <I>identifiers</I>.
	 * @param identifiers the table of identifiers
	 * @throws XDBCException if a database access error occurs.
	 */
	public void getNextIdentifiers(Object[] identifiers) throws MediatorException {
		// because 'getNextIdentifiers' is very special (comprendre
		// very chiant). We must get identifier of next tuple, but keep
		// the cursor on the current one.
		// we solve this by using a "cache" tuple
		if (next_tuple != null) {
			next_tuple.getIdentifiers(identifiers);
			return;
		}
		if (!hasNext()) {
			next_tuple = null;
			return;
		}
		Tuple old_tuple = cur_tuple;
		next();
		next_tuple = cur_tuple;
		next_tuple.getIdentifiers(identifiers);
		cur_tuple = old_tuple;
	}
}
