/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.runtime;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Properties;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.mediator.XDBCWrapper;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDriver;
import org.xquark.xml.xdbc.XMLDriverManager;

/**
 * This class is used to parse a mediator accessor file.
 * 
 */

public class MediatorConfiguration extends DefaultHandler {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************
	private final static String LABEL_ACCESSOR = "accessor";
	private final static String LABEL_NAME = "name";
	private final static String LABEL_DRIVER = "driver";
	private final static String LABEL_CONNECTION = "connection";
	private final static String LABEL_SUBACCESSOR = "subaccessor";

	private final static String LABEL_LOGIN = "login";
	private final static String LABEL_PASSWD = "password";

	public final static String KEY_NAME = "key_name";
	public final static String KEY_SUBACCESSORS = "key_subaccessors";
	public final static String KEY_BASE_URI = "key_base_uri";
	public final static String KEY_DRIVER = "key_driver";
	public final static String KEY_CONNECTION = "key_connection";
	public final static String KEY_LOGIN = "key_login";
	public final static String KEY_PASSWD = "key_passwd";


	private static SAXParserFactory spf = null;
	private static SchemaManager schemamanager = new SchemaManager();
	public static final String MEDIATOR_XSD = "/org/xquark/mediator/resources/Mediator.xsd";

	static {
		// Create a JAXP SAXParserFactory and configure it
		spf = SAXParserFactory.newInstance();
		spf.setNamespaceAware(true);
		spf.setValidating(false);
		
		try {
			URL url = MediatorConfiguration.class.getResource(MEDIATOR_XSD);
			InputSource source = new InputSource(url.toString());
			schemamanager.loadSchema(source);
		} catch (SAXException e) {
			// do nothing !!!
		}
	}

	private Properties xmlproperties = new Properties();
	private ArrayList wrappers = new ArrayList();
	private URL baseDir = null;
	private Locator locator = null;

	protected StringBuffer cur_car = new StringBuffer();
	protected String sub_name = null;
	protected String sub_driver = null;
	protected String sub_url = null;
	protected String sub_user = null;
	protected String sub_password = null;
	

	// **********************************************************************
	// * CONSTRUCTOR
	// **********************************************************************
	/**
	 *
	 */
	public MediatorConfiguration() {
		xmlproperties.put(KEY_SUBACCESSORS, wrappers);
	}

	public Properties getProperties() {
		return xmlproperties;
	}
	
	public static Properties loadConfiguration(String configURI) throws XMLDBCException, MalformedURLException {
		URL url = new URL(configURI);
		try {
			MediatorConfiguration handler = new MediatorConfiguration();
			// Get the encapsulated SAX XMLReader
			XMLReader reader = spf.newSAXParser().getXMLReader();
			ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, new SchemaValidationContext(schemamanager));
			filter.setContentHandler(handler);
			filter.setErrorHandler(handler);
			filter.parse(configURI);
			return handler.getProperties();
		} catch (SAXParseException ex) {
			String msg = "Error in parsing file " + ex.getSystemId() + " at line " + Integer.toString(ex.getLineNumber()) +
						 " : " + ex.getMessage();
			throw new XMLDBCException(msg, ex);
		} catch (SAXException ex) {
			throw new XMLDBCException("Could not parse configuration file", ex);
		} catch (IOException ex) {
			throw new XMLDBCException("Could not read configuration file", ex);
		} catch (ParserConfigurationException ex) {
			throw new XMLDBCException("Could not allocate XML parser", ex);
		}
	}
	
	// **********************************************************************
	// * TOOLS METHODS
	// **********************************************************************
	/**
	 * Return  a property of the accessor.
	 * 
	 * @return a property of the accessor.
	 */
	private String getProperty(String key) {
		return (String) xmlproperties.getProperty(key);
	}

	/**
	 * Set a property of the accessor.
	 *
	 * @param a property of the accessor.
	 */
	private void setProperty(String key, String value) {
		if (key == null) {
			System.err.println("Key " + key + " is null.");
			return;
		}
		if (value == null) {
			System.err.println("Value " + value + " of " + key + " is null.");
			return;
		}
		xmlproperties.setProperty(key, value);
	}

	// **********************************************************************
	// * SAX IMPLEMENTATION
	// **********************************************************************

	public void setDocumentLocator(Locator loc) {
		try {
			baseDir = new URL(loc.getSystemId());
		} catch (MalformedURLException e) {
			// should not occur
		}
	}

	/**
	 *
	 */
	public void startElement(String namespaceURI, String localName, String qualifiedName, Attributes atts) throws SAXException {
		cur_car.setLength(0);
		if (localName.equalsIgnoreCase(LABEL_ACCESSOR)) {
			setProperty(KEY_NAME, atts.getValue(LABEL_NAME));
		} else if (localName.equalsIgnoreCase(LABEL_SUBACCESSOR)) {
			sub_name = atts.getValue(LABEL_NAME);
			sub_driver = null;
			sub_url = null;
			sub_user = null;
			sub_password = null;
		}
	}

	/**
	 *
	 */
	public void characters(char ch[], int start, int length) throws SAXException {
		cur_car.append(ch, start, length);
	}

	/**
	 *
	 */
	public void endElement(String namespaceURI, String localName, String qualifiedName) throws SAXException {
		String value = cur_car.toString().trim();
		cur_car.setLength(0);
		if (localName.equalsIgnoreCase(LABEL_DRIVER)) {
			sub_driver = value;
			try {
				Class.forName(sub_driver);
			} catch (ClassNotFoundException ex) {
				throw new SAXParseException("Could not load driver class", locator, ex);
			}
		} else if (localName.equalsIgnoreCase(LABEL_CONNECTION)) {
			sub_url = value;
			try {
				XMLDriver d = XMLDriverManager.getDriver(sub_url);
				if (d != null) {
					String localPart = d.getSpecificPart(sub_url);
					if (localPart.startsWith("file:")) {
						URL url = new URL(baseDir, localPart.substring(5));
						String driverPart = sub_url.substring(0, sub_url.indexOf(localPart));
						sub_url = driverPart+url.toString();
					}
				}
			} catch (XMLDBCException e) {
				throw new SAXParseException("No suitable driver found for "+sub_url, locator, e);
			} catch (MalformedURLException e) {
				// ignore and leave sub_url as it is
            }
		} else if (localName.equalsIgnoreCase(LABEL_LOGIN)) {
			sub_user = value;
		} else if (localName.equalsIgnoreCase(LABEL_PASSWD)) {
			sub_password = value;
		} else if (localName.equalsIgnoreCase(LABEL_SUBACCESSOR)) {
			wrappers.add(new XDBCWrapper(sub_name, sub_driver, sub_url, sub_user, sub_password));
		}
	}

	public void error(SAXParseException spee) throws SAXException {
		System.err.println("Warning : " + spee.toString());
	}
}
