/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.runtime;

import java.io.PrintStream;
import java.io.StringReader;
import java.util.*;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xquark.mediator.algebra.AlgebraManager;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.mediator.plan.OpSource;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.MetaDataImpl;
import org.xquark.xquery.metadata.VarCounter;
import org.xquark.xquery.normalize.TransformXQueryExpression;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.typing.TypeVisitor;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;

/**
 * This class implements XMLStatement interface that is used for executing
 * constant XQuery or XPath statements
 * and obtaining the results produced (update statements are not
 * supported yet).<BR>
 *
 */
public class MediatorStatement implements XMLStatement {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.14 $";
	private static final String RCSName = "$Name:  $";

	private static Log log = LogFactory.getLog(MediatorStatement.class);

	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************

	private _MediatorConnection connection = null;
	private boolean qaOnly = false;
	HashMap preparedStatemements = new HashMap();
	XMLResultSet currentResultSet = null;

	//public XDBCResultSetInterface debug_resultset = null ;
	//public XQueryExpression debug_qmem = null ;
	private String baseURI = null;
	protected boolean isClosed = false;

	// ************************************************************************
	// * INITIALISATION
	// ************************************************************************
	public MediatorStatement(_MediatorConnection connection) throws XMLDBCException {
		this.connection = connection;
	}

	public HashMap getVariables() {
		return null;
	}

	// ************************************************************************
	// * XMLStatement IMPLEMENTATION
	// ************************************************************************

	/*********************************************************************/
	/*************************** Queries execution management ************/
	/*********************************************************************/
	/**
	 * Executes a query statement of the type specified when this
	 * object was built with the createStatement method (see XMLConnection interface)
	 * that may return results.
	 * If the query generates results, use the <I>getResultSet()</I> or
	 * <I>getDocumentSet()</I> methods to get the set of results.
	 * @param query the query string.
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public boolean execute(String query) throws XMLDBCException {
		if (currentResultSet != null) {
			currentResultSet.close();
		}
		currentResultSet = executeQuery(query);
		return true;
	}

	/**
	 * Executes a query statement of the specified type
	 * that may return results.
	 * If the query generates results, use the <I>getResultSet()</I> or
	 * <I>getDocumentSet()</I> methods to get the set of results.
	 * @param query the query string.
	 * @param queryType the type of query for this statement (see constants in XMLConnection interface).
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public boolean execute(String query, int queryType) throws XMLDBCException {
		return execute(query);
	}

	/**
	 * Executes a query statement of the type specified when this
	 * object was built with the createStatement method (see XMLConnection interface)
	 * that returns results.
	 * @param query the query string
	 * @return an XMLResultSet object that contains the data produced by the given query.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLResultSet executeQuery(String query) throws XMLDBCException {
		try {
			ExecutionPlan plan = compileXQuery(query, baseURI);
			if (!preparedStatemements.isEmpty()) {
				closePreparedStatements();
			}
			if (currentResultSet != null) {
				currentResultSet.close();
			}
			currentResultSet = plan.executeQuery(this);
			return currentResultSet;
		} catch (XQueryException e) {
			throw new XMLDBCException("MediatorStatement.executeQuery : " + e.getMessage(), e);
		}
	}
	// for regression purposes -> printing the plan
	private PrintStream plnout = null;
	public XMLResultSet executeQuery(String query, PrintStream plnout) throws XMLDBCException {
		this.plnout = plnout;
		return executeQuery(query, XMLConnection.XQUERY_STRING_TYPE);
	}

	protected ExecutionPlan compileXQuery(String query, String baseURI) throws XMLDBCException {
		try {
			XQueryParser ql = new XQueryParser(new StringReader(query));
			ql.setBaseURI(baseURI);
			ql.setFactory("org.xquark.xquery.parser.ParserFactory");
			MetaDataImpl metadata = connection.getMetadataAccess();

			XQueryModule module = ql.Start(metadata, metadata.getSchemaManager(), metadata.getModuleManager(), new VarCounter(), qaOnly);
			ArrayList expressions = module.getExpressions();
			if (expressions == null || expressions.isEmpty())
				return null;
			TypeVisitor typevisitor = ql.getTypeVisitor();

			// get declarations

			// if only one source is concerned, no need to bother about mediating, just pipeline requests to it.
			// but first normalize and restructure
			//----------------------------------------------------------
			// START --> FOR SINGLE SOURCE VERIFICATION
			//----------------------------------------------------------
			// comment this bloc if you want the mediator not to overpass single source optimization (for debugging on a simple example)

			module.normalize(typevisitor, connection.getSourceName());
			// TODO handle several expressions with union !!!
			if (expressions.size() > 1) {
				throw new XMLDBCException("Fusion only handles modules with a single query (for now...)");
			}
			for (int j = 0; j < expressions.size(); j++) {
				XQueryExpression exprj = (XQueryExpression) expressions.get(j);
				Utils.fillSources(exprj, metadata, true);
			}

			Set sources = module.getSourceNames();
			Set urls = module.getUrls();
			String xQueryStringImports = null;
			if (!qaOnly && sources != null && sources.size() == 1 && (urls == null || urls.isEmpty())) {
				xQueryStringImports = module.toString(false, true);
			}

			//----------------------------------------------------------
			// STOP --> FOR SINGLE SOURCE VERIFICATION
			//----------------------------------------------------------

			// !!!!!! danger, only one expression ....			
			ExecutionPlan plan = null;
			//for (int j = 0; j < expressions.size(); j++) {
			XQueryExpression exprj = (XQueryExpression) expressions.get(0);
			String xQueryString = module.toString(false, false);

			TransformXQueryExpression txe = exprj.getParentModule().canonize(exprj, typevisitor);
			plan = new ExecutionPlan(metadata, typevisitor, txe.getIdVarList(), xQueryStringImports, xQueryString, module, baseURI, qaOnly);

			AlgebraManager depmanager = new AlgebraManager(plan, metadata, typevisitor, txe.getQMEM(), txe.getQDB(), txe.getIdVarList());

			ArrayList qdbs = txe.getQDB();

			if (log.isDebugEnabled()) {
				for (int si = 0; si < qdbs.size(); si++)
					log.debug("QDB[" + si + "]: " + ((Variable) qdbs.get(si)).getExpression());
				log.debug("QMEM: " + txe.getQMEM());
			}

			XDBCResultSetInterface resultset = null;
			if (!qdbs.isEmpty()) {
				depmanager.generate();
				depmanager.computeVarPaths();
				if (log.isDebugEnabled()) {
					log.debug("Algebra Tree: " + depmanager);
				}
				depmanager.FillPlan();
				if (log.isDebugEnabled()) {
					log.info("Execution Tree: " + plan);
				}
				if (plnout != null) {
					plnout.print(plan.toString());
				}
			}
			plan.setQMEM(txe.getQMEM());
			if (!module.getExternalVariables().isEmpty() && plan.getRoot() != null)
				plan.getRoot().setPrepared();

			//}
			return plan;
		} catch (XQueryException xe) {
			throw new XMLDBCException(xe.getMessage(), xe);
		} catch (ParseException xe) {
			throw new XMLDBCException(xe.getMessage(), xe);
		}

	}
	/**
	 * Executes a query statement of the specified type
	 * that returns results.
	 * If the query generates results, use the <I>getResultSet()</I> or
	 * <I>getDocumentSet()</I> methods to get the set of results.
	 * @param query the query string.
	 * @param queryType the type of query for this statement (see constants in XMLConnection interface).
	 * @return true if there are available resuts, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLResultSet executeQuery(String query, int queryType) throws XMLDBCException {
		return executeQuery(query);
	}

	/**
	 * Executes a query statement of the type specified when this
	 * object was built with the createStatement method (see XMLConnection interface)
	 * that returns complete XML documents. It is a error to call this method with a
	 * query that returns document fragments.
	 * @param query a static and read-only query in the specified type.
	 * @return an XMLResultSet object that contains the data produced by the given query.
	 * @throws XMLDBCException if a data source access error occurs, or if the returned
	 * results are not complete XML documents.
	 */
	public XMLDocumentSet executeDocumentQuery(String query) throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/*********************************************************************/
	/*************************** Results management **********************/
	/*********************************************************************/

	/**
	 * Returns the current statement results as an XMLResultSet object.
	 * @return the current results as an XMLResultSet object or null if there is no result.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLResultSet getResultSet() throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/**
	 * Returns the current statement results as an XMLDocumentSet object.
	 * It is a error to call this method for a statement that returned
	 * document fragments.
	 * @return the current results as an XMLDocumentSet object or null if there is no result.
	 * @throws XMLDBCException if a data source access error occurs, or if the returned
	 * results are not complete XML documents.
	 */
	public XMLDocumentSet getDocumentSet() throws XMLDBCException {
		throw new XMLDBCNotSupportedException("Not yet implemented.");
	}

	/*********************************************************************/
	/*************************** Statement management ********************/
	/*********************************************************************/

	/**
	 * Returns the connection object that produced this statement.
	 * @return the connection object that produced this statement.
	 */
	public XMLConnection getConnection() {
		return connection.getConnection();
	}

	public _MediatorConnection getMediatorConnection() {
		return connection;
	}

	/**
	 * Releases the statement XMLDBC resources immediately instead of waiting for this
	 * to happen when it is automatically closed.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() throws XMLDBCException {
		if (!isClosed) {
			if (currentResultSet != null) {
				try {
					currentResultSet.close();
				} catch (XMLDBCException xe) {
				}
				currentResultSet = null;
			}
			connection.statementClosed(this);
			isClosed = true;
		}
	}

	public boolean isClosed() throws XMLDBCException {
		return isClosed;
	}

	public void setBaseURI(String baseURI) {
		this.baseURI = baseURI;
	}

	public String getBaseURI() {
		return baseURI;
	}

	public void setQAOnly(boolean bool) {
		qaOnly = bool;
	}

	public void closePreparedStatements() throws XMLDBCException {
		for (Iterator it = preparedStatemements.values().iterator(); it.hasNext();) {
			PreparedXMLStatement stat = (PreparedXMLStatement) it.next();
			if (stat != null)
				stat.close();
		}
		preparedStatemements.clear();
	}

	public PreparedXMLStatement getPreparedXMLStatement(OpSource opSource, String sourceName, String requestStr) throws XMLDBCException {
		//if (true) return null;
		PreparedXMLStatement stat = (PreparedXMLStatement) preparedStatemements.get(opSource);
		if (!preparedStatemements.containsKey(opSource)) {
			XMLConnection xmlConnection = null;
			try {
				xmlConnection = connection.getSubConnectionByName(sourceName);
			} catch (MediatorException me) {
				throw new XMLDBCException(me.getMessage(), me);
			}
			try {
				stat = xmlConnection.prepareStatement(requestStr);
			} catch (XMLDBCNotSupportedException xe) {
				stat = null;
			}
			preparedStatemements.put(opSource, stat);
		}
		return stat;
	}
}
