/*
 * This file belongs to the XQuark distribution. Copyright (C) 2004 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307. You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.runtime;

import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.util.*;

import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.schema.SchemaException;
import org.xquark.schema.SimpleType;
import org.xquark.schema.Type;
import org.xquark.xml.xdbc.PreparedXMLStatement;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLResultSet;
import org.xquark.xpath.datamodel.TypedDocumentImpl;
import org.xquark.xpath.datamodel.TypedValue;
import org.xquark.xpath.datamodel.TypedValueImpl;
import org.xquark.xquery.parser.*;

public class PreparedMediatorStatement extends MediatorStatement implements PreparedXMLStatement {

    ExecutionPlan plan = null;

    String query = null;

    // external variables
    HashMap variables = new HashMap();

    protected static TypedDocumentImpl docimpl = null;
    static {
        docimpl = new TypedDocumentImpl();
    }

    /**
     * @param connection
     * @throws XMLDBCException
     */
    public PreparedMediatorStatement(_MediatorConnection connection, String baseURI, String query, Map cachedStatements) throws XMLDBCException {
        super(connection);
        this.query = query;
        if (cachedStatements != null) {
            plan = (ExecutionPlan) cachedStatements.get(query);
        }
        if (plan == null) {
            plan = compileXQuery(query, baseURI);
            if (cachedStatements != null) {
                cachedStatements.put(query, plan);
            }
        }
    }

    public HashMap getVariables() {
        return variables;
    }

    private Variable getVariable(String ns, String varName) throws XMLDBCException {
        HashMap varMap = plan.getModule().getVariables();
        if (varMap == null)
            throw new XMLDBCException("Variable $" + ((ns == null) ? "" : "{" + ns + "}") + varName + " is not defined.");
        varMap = (HashMap) varMap.get(ns);
        if (varMap == null)
            throw new XMLDBCException("Variable $" + ((ns == null) ? "" : "{" + ns + "}") + varName + " is not defined.");
        Variable var = (Variable) varMap.get(varName);
        if (var == null)
            throw new XMLDBCException("Variable $" + ((ns == null) ? "" : "{" + ns + "}") + varName + " is not defined.");
        return var;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setExternalVariable(java.lang.String, java.lang.String, java.lang.String)
     */
    public void setExternalVariable(String ns, String varName, String value) throws XMLDBCException {
        Variable var = getVariable(ns, varName);
        if (!var.getQType().isAtom())
            throw new XMLDBCException("Variable " + var.getStringValue() + " cannot be set with string value.");
        SimpleType st = var.getQType().getSimpleType();
        Object obj = null;
        try {
            obj = st.convert(value);
        } catch (SchemaException se) {
            throw new XMLDBCException("Illegal string value " + value + "for variable {" + ns + "}" + varName + " of type " + st.getPrimitive(), se);
        }
        HashMap map = (HashMap) variables.get(ns);
        if (map == null) {
            map = new HashMap();
            variables.put(ns, map);
        }
        TypedValue typedValue = new TypedValueImpl(docimpl, value);
        typedValue.setTypedValue(obj);
        typedValue.setType(st);
        map.put(varName, typedValue);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setExternalVariable(java.lang.String, java.lang.String)
     */
    public void setExternalVariable(String varName, String value) throws XMLDBCException {
        setExternalVariable(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setObject(java.lang.String, java.lang.String, java.lang.Object)
     */
    public void setObject(String ns, String varName, Object value) throws XMLDBCException {
        Variable var = getVariable(ns, varName);
        if (!var.getQType().isAtom())
            throw new XMLDBCException("Variable " + var.getStringValue() + " complex type not supported yet.");
        SimpleType st = var.getQType().getSimpleType();
        // test for sequence
        if (value instanceof ArrayList) {
            if (true)
                throw new XMLDBCException("Variable " + var.getStringValue() + " sequence not supported yet.");
            ArrayList list = (ArrayList) value;
            if (list.size() > 1 && !var.getQType().isMultiple())
                throw new XMLDBCException("Illegal value " + value + "for variable {" + ns + "}" + varName + " of type " + st.getPrimitive());
            ArrayList typeValues = new ArrayList(list.size());
            for (int i = 0; i < list.size(); i++) {
                Object valuei = list.get(i);
                Object obj = null;
                try {
                    obj = st.convertObject(valuei);
                } catch (SchemaException se) {
                    throw new XMLDBCException("Illegal value " + valuei + "for variable {" + ns + "}" + varName + " of type " + st.getPrimitive(), se);
                }
                TypedValue typedValue = new TypedValueImpl(docimpl, valuei.toString());
                typedValue.setTypedValue(obj);
                typedValue.setType(st);
                typeValues.add(typedValue);
            }
            HashMap map = (HashMap) variables.get(ns);
            if (map == null) {
                map = new HashMap();
                variables.put(ns, map);
            }
            map.put(varName, typeValues);
        } else {
            Object obj = null;
            try {
                obj = st.convertObject(value);
            } catch (SchemaException se) {
                throw new XMLDBCException("Illegal value " + value + "for variable {" + ns + "}" + varName + " of type " + st.getPrimitive(), se);
            }
            HashMap map = (HashMap) variables.get(ns);
            if (map == null) {
                map = new HashMap();
                variables.put(ns, map);
            }
            TypedValue typedValue = new TypedValueImpl(docimpl, value.toString());
            typedValue.setTypedValue(obj);
            typedValue.setType(st);
            map.put(varName, typedValue);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setObject(java.lang.String, java.lang.Object)
     */
    public void setObject(String varName, Object value) throws XMLDBCException {
        setObject(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setBigDecimal(java.lang.String, java.lang.String, java.math.BigDecimal)
     */
    public void setBigDecimal(String ns, String varName, BigDecimal value) throws XMLDBCException {
        setObject(ns, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setBigDecimal(java.lang.String, java.math.BigDecimal)
     */
    public void setBigDecimal(String varName, BigDecimal value) throws XMLDBCException {
        setBigDecimal(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setInputStream(java.lang.String, java.lang.String, java.io.InputStream, int)
     */
    public void setInputStream(String ns, String varName, InputStream value, int length) throws XMLDBCException {
        byte[] bytes = new byte[length];
        try {
            int res = value.read(bytes, 0, length);
            setBytes(ns, varName, bytes);
        } catch (IOException ioe) {
            throw new XMLDBCException(ioe.getMessage(), ioe);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setInputStream(java.lang.String, java.io.InputStream, int)
     */
    public void setInputStream(String varName, InputStream value, int length) throws XMLDBCException {
        setInputStream(null, varName, value, length);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setBytes(java.lang.String, java.lang.String, byte[])
     */
    public void setBytes(String ns, String varName, byte[] value) throws XMLDBCException {
        setObject(ns, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setBytes(java.lang.String, byte[])
     */
    public void setBytes(String varName, byte[] value) throws XMLDBCException {
        setBytes(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setBoolean(java.lang.String, java.lang.String, boolean)
     */
    public void setBoolean(String ns, String varName, boolean value) throws XMLDBCException {
        setObject(ns, varName, new Boolean(value));
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setBoolean(java.lang.String, boolean)
     */
    public void setBoolean(String varName, boolean value) throws XMLDBCException {
        setBoolean(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setDate(java.lang.String, java.lang.String, java.util.Date)
     */
    public void setDate(String ns, String varName, Date value) throws XMLDBCException {
        setObject(ns, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setDate(java.lang.String, java.util.Date)
     */
    public void setDate(String varName, Date value) throws XMLDBCException {
        setDate(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setDouble(java.lang.String, java.lang.String, double)
     */
    public void setDouble(String ns, String varName, double value) throws XMLDBCException {
        setObject(ns, varName, new Double(value));
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setDouble(java.lang.String, double)
     */
    public void setDouble(String varName, double value) throws XMLDBCException {
        setDouble(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setFloat(java.lang.String, java.lang.String, float)
     */
    public void setFloat(String ns, String varName, float value) throws XMLDBCException {
        setObject(ns, varName, new Float(value));
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setFloat(java.lang.String, float)
     */
    public void setFloat(String varName, float value) throws XMLDBCException {
        setFloat(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setLong(java.lang.String, java.lang.String, long)
     */
    public void setLong(String ns, String varName, long value) throws XMLDBCException {
        setObject(ns, varName, new Long(value));
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setLong(java.lang.String, long)
     */
    public void setLong(String varName, long value) throws XMLDBCException {
        setLong(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setString(java.lang.String, java.lang.String, java.lang.String)
     */
    public void setString(String ns, String varName, String value) throws XMLDBCException {
        setObject(ns, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setString(java.lang.String, java.lang.String)
     */
    public void setString(String varName, String value) throws XMLDBCException {
        setString(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setSequence(java.lang.String, java.lang.String, java.lang.String)
     */
    public void setSequence(String ns, String varName, List value) throws XMLDBCException {
        setObject(ns, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#setSequence(java.lang.String, java.lang.String)
     */
    public void setSequence(String varName, List value) throws XMLDBCException {
        setSequence(null, varName, value);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#clearVariables()
     */
    public void clearVariables() throws XMLDBCException {
        variables.clear();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#execute()
     */
    public boolean execute() throws XMLDBCException {
        // TODO Auto-generated method stub
        throw new XMLDBCException("Not supported yet");
        //return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.xquark.xml.xdbc.PreparedXMLStatement#executeQuery()
     */
    public XMLResultSet executeQuery() throws XMLDBCException {
        if (currentResultSet != null) {
            currentResultSet.close();
        }
        if (plan == null) {
            currentResultSet = null;
            return null;
        }
        try {
            currentResultSet = plan.executeQuery(this);
            return currentResultSet;
        } catch (XQueryException xe) {
            currentResultSet = null;
            throw new XMLDBCException(xe.getMessage(), xe);
        }
    }

    public void close() throws XMLDBCException {
        super.close();
        super.closePreparedStatements();
    }

    public void closePreparedStatements() throws XMLDBCException {
        // do nothing
    }

    public Collection getParameters() {
        return variables.values();
    }

    public javax.xml.namespace.QName[] getParameterNames() {
        ArrayList names = new ArrayList();
        XQueryModule module = plan.getModule();
        if (module.getDeclList() != null) {
            for (int i = 0; i < module.getDeclList().size(); i++) {
                QName name = (QName) module.getDeclList().get(i);
                HashMap map = null;
                if (module.getVariables() != null) {
                    map = (HashMap) module.getVariables().get(name.getNameSpace());
                    if (map != null) {
                        Variable vari = (Variable) map.get(name.getLocalName());
                        if (vari != null && vari.getExpression() == null) {
                            names.add(new javax.xml.namespace.QName(vari.getVarName().getNameSpace(), vari.getVarName().getLocalName()));
                        }
                    }
                }
            }
        }
        javax.xml.namespace.QName qnames[] = new javax.xml.namespace.QName[names.size()];
        return (javax.xml.namespace.QName[]) names.toArray(qnames);
    }

    public javax.xml.namespace.QName getParameterType(javax.xml.namespace.QName varName) {
        XQueryModule module = plan.getModule();
        org.xquark.xquery.parser.ExternalVariable extVar = (org.xquark.xquery.parser.ExternalVariable) module.getExternalVariables().get(varName);
        Type type = extVar.getQType().getSimpleType().getBaseType();
        return new javax.xml.namespace.QName(type.getNamespace(), type.getName());
    }

}