/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xqueryevaluator.eval;

import java.io.IOException;
import java.io.InputStream;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

/**
 * XML Pull Parser reader that allow the incremental processing of an XML document,
 * while providing content informatioon through the standard SAX2 ContentHandler and
 * LexicalHandler APIs
 */

public class XPPReader implements Locator, Attributes {
    protected XmlPullParser xpp;
    protected boolean beforeFirstTag = true;

    protected ContentHandler contentHandler = null;
    protected ErrorHandler errorHandler = null;
    protected LexicalHandler lexicalHandler = null;
    protected String systemId;
    protected InputStream input = null;

    private int[] charsBounds = new int[2];

    protected static XmlPullParserFactory factory;

    static {
        try {
            factory = XmlPullParserFactory.newInstance();
            factory.setNamespaceAware(true);
        } catch (XmlPullParserException ex) {}
    }

    public XPPReader() throws SAXException {
        if (factory == null)
            throw new SAXException("No XmlPullParser factory available");
        try {
			xpp = factory.newPullParser();
        } catch (XmlPullParserException ex) {
        	throw new SAXException(ex);
        }
    }

    // -- Attributes interface

    public String getURI(int index) {
        return xpp.getAttributeNamespace(index);
    }

    public String getLocalName(int index) {
        return xpp.getAttributeName(index);
    }

    public String getQName(int index) {
        String prefix = xpp.getAttributePrefix(index);
        if (prefix != null) {
            return prefix + ':' + xpp.getAttributeName(index);
        } else {
            return xpp.getAttributeName(index);
        }
    }

    public String getValue(String uri, String localName) {
        return xpp.getAttributeValue(uri, localName);
    }

    public String getValue(String qName) {
        return xpp.getAttributeValue(null, qName);
    }

    public String getType(int index) {
        return xpp.getAttributeType(index);
    }

    public String getValue(int index) {
        return xpp.getAttributeValue(index);
    }

    public int getLength() {
        return xpp.getAttributeCount();
    }

    public int getIndex(String uri, String localName) {
        for (int i = 0; i < xpp.getAttributeCount(); i++) {
            if (xpp.getAttributeNamespace(i).equals(uri) && xpp.getAttributeName(i).equals(localName)) {
                return i;
            }
        }
        return -1;
    }

    public int getIndex(String qName) {
        for (int i = 0; i < xpp.getAttributeCount(); i++) {
            if (xpp.getAttributeName(i).equals(qName)) {
                return i;
            }
        }
        return -1;
    }

    public String getType(String uri, String localName) {
        int index = getIndex(uri, localName);
        if (index == -1)
            return null;
        else
            return xpp.getAttributeType(index);
    }

    public String getType(String qName) {
        int index = getIndex(qName);
        if (index == -1)
            return null;
        else
            return xpp.getAttributeType(index);
    }

    // -- Locator interface

    public String getPublicId() {
        return null;
    }

    public String getSystemId() {
        return systemId;
    }

    public int getLineNumber() {
        return xpp.getLineNumber();
    }

    public int getColumnNumber() {
        return xpp.getColumnNumber();
    }

    public void setContentHandler(ContentHandler handler) {
        this.contentHandler = handler;
    }

    public ContentHandler getContentHandler() {
        return contentHandler;
    }

    public void setLexicalHandler(LexicalHandler handler) {
        this.lexicalHandler = handler;
    }

    public LexicalHandler getLexicalHandler() {
        return lexicalHandler;
    }

    public void setErrorHandler(ErrorHandler handler) {
        this.errorHandler = handler;
    }

    public ErrorHandler getErrorHandler() {
        return errorHandler;
    }

    public void setInput(String systemId, InputStream input) throws SAXException {
        this.systemId = systemId;
        try {
			xpp.setInput(input, null);
        } catch (XmlPullParserException ex) {
        	throw new SAXException(ex);
        }
    }

    public boolean isAtStart() throws SAXException {
    	try {
			return xpp.getEventType() == XmlPullParser.START_DOCUMENT;
    	} catch (XmlPullParserException ex) {
    		throw new SAXException(ex);
    	}
    }
    
    public boolean parseNextElement() throws SAXException, IOException {
        try {
            int prevCount = 0;
            int curCount = 0;
            int depth = 0;
            char[] chars;
            while (true) {
				switch (xpp.nextToken()) {
					case XmlPullParser.START_TAG :
					    beforeFirstTag = false;
						depth = xpp.getDepth();
						prevCount = xpp.getNamespaceCount(depth - 1);
						curCount = xpp.getNamespaceCount(depth);
						for (int i = prevCount; i < curCount; i++) {
							contentHandler.startPrefixMapping(xpp.getNamespacePrefix(i), xpp.getNamespaceUri(i));
						}
						contentHandler.startElement(xpp.getNamespace(), xpp.getName(), null, this);
						break;
					case XmlPullParser.TEXT :
					case XmlPullParser.IGNORABLE_WHITESPACE :
					case XmlPullParser.CDSECT :
					    if (!beforeFirstTag) {
							chars = xpp.getTextCharacters(charsBounds);
							contentHandler.characters(chars, charsBounds[0], charsBounds[1]);
					    }
						break;
					case XmlPullParser.COMMENT :
						if (lexicalHandler != null) {
							chars = xpp.getTextCharacters(charsBounds);
							lexicalHandler.comment(chars, charsBounds[0], charsBounds[1]);
						}
						break;
					case XmlPullParser.PROCESSING_INSTRUCTION :
						String pi = xpp.getText().trim();
						int index = pi.indexOf(' ');
						if (index != -1) contentHandler.processingInstruction(pi.substring(0, index), pi.substring(index+1));
						else contentHandler.processingInstruction(pi, "");
						break;
					case XmlPullParser.END_TAG :
						contentHandler.endElement(xpp.getNamespace(), xpp.getName(), null);
						depth = xpp.getDepth();
						prevCount = xpp.getNamespaceCount(depth - 1);
						curCount = xpp.getNamespaceCount(depth);
						for (int i = curCount - 1; i >= prevCount; i--) {
							contentHandler.endPrefixMapping(xpp.getNamespacePrefix(i));
						}
						return false;
					case XmlPullParser.END_DOCUMENT :
						return true;
					default :
						break;
				}
            }
        } catch (XmlPullParserException ex) {
            SAXParseException saxException = new SAXParseException("", this, ex);
            if (errorHandler != null)
                errorHandler.fatalError(saxException);
            throw saxException;
        }
    }

}
