/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xqueryevaluator.eval;

import java.util.ArrayList;

public class XPathExpr implements Cloneable {
	
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName = "$Name:  $";
    private java.util.ArrayList fields = new java.util.ArrayList(4);

    public Object clone() {
    	XPathExpr newExpr = new XPathExpr();
    	newExpr.fields = (ArrayList)fields.clone();
    	return newExpr;
    }
    
    public void addElementStep(String namespaceURI, String localName, boolean isAnyLevel, boolean isSelf) {
    	XPathExpr.Step last = removeSelfNodeStep();
    	if (last != null) isAnyLevel = isAnyLevel || last.isAnyLevel();
        fields.add(new XPathExpr.Step(namespaceURI, localName, isAnyLevel, isSelf, false));
    }

	public void addAttributeStep(String namespaceURI, String localName, boolean isAnyLevel) {
		XPathExpr.Step last = removeSelfNodeStep();
		if (last != null) isAnyLevel = isAnyLevel || last.isAnyLevel();
		fields.add(new XPathExpr.Step(namespaceURI, localName, isAnyLevel, false, true));
	}
	
	public void addTextStep(boolean isAnyLevel, boolean isSelf) {
		XPathExpr.Step last = removeSelfNodeStep();
		if (last != null) isAnyLevel = isAnyLevel || last.isAnyLevel();
		fields.add(new XPathExpr.Step(isAnyLevel, isSelf, true));
	}

	public void addNodeStep(boolean isAnyLevel, boolean isSelf) {
		XPathExpr.Step last = removeSelfNodeStep();
		if (last != null) isAnyLevel = isAnyLevel || last.isAnyLevel();
		fields.add(new XPathExpr.Step(isAnyLevel, isSelf, false));
	}

    public int getStepCount() {
        return fields.size();
    }

	public boolean matchElement(String uri, String localName, int index) {
		if (index >= fields.size())
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isElement() && step.matches(uri, localName);    	
	}
    
    public boolean matchAttribute(String uri, String localName, int index) {
		if (index != fields.size() - 1)
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isAttribute() && step.matches(uri, localName);    	
    }
    
	public boolean matchText(int index) {
		if (index != fields.size() - 1)
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isText();    	
	}
    
    public boolean isAnyLevel(int index) {
		if (index >= fields.size())
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isAnyLevel();
    }

	public boolean isSelf(int index) {
		if (index >= fields.size())
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isSelf();
	}

	public boolean isSelfElement(int index) {
		if (index >= fields.size())
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isSelfElement();
	}
	
	public boolean isSelfNode(int index) {
		if (index >= fields.size())
			return false;
		XPathExpr.Step step = (XPathExpr.Step) fields.get(index);
		return step.isSelfNode();
	}

    public void removeLastSelfNodeStep() {
		int lastIndex = fields.size()-1;
		if (lastIndex >= 0) {
			XPathExpr.Step last = (XPathExpr.Step) fields.get(lastIndex);
			if (last.isSelfNode() && !last.isAnyLevel())
				fields.remove(lastIndex); 
		}
    }

	private XPathExpr.Step removeSelfNodeStep() {
		int lastIndex = fields.size()-1;
		if (lastIndex < 0) return null;
		XPathExpr.Step last = (XPathExpr.Step) fields.get(lastIndex);
		if (last.isSelfNode()) {
			fields.remove(lastIndex); 
			return last;
		} else {
			return null;
		}
 	}
	
    static class Step {
        private static final String RCSRevision = "$Revision: 1.1 $";
        private static final String RCSName = "$Name:  $";
        private static final int ELEMENT = 0;
		private static final int ATTR = 1;
		private static final int TEXT = 2;
		private static final int NODE = 3;
        
		int nodeKind;
        String namespaceURI;
        String localName;
        boolean isAny = false;
        boolean anyNamespace = false;
        boolean anyLevel = false;
        boolean isSelf = false;

        Step(String namespaceURI, String localName, boolean isAnyLevel, boolean isSelf, boolean isAttribute) {
            this.namespaceURI = namespaceURI;
            this.localName = localName;
            this.anyLevel = isAnyLevel;
            this.isSelf = isSelf;
            this.nodeKind = isAttribute ? ATTR : ELEMENT;
            if ("*".equals(namespaceURI)) {
            	anyNamespace = true;
            }
            if ("*".equals(localName)) {
                this.localName = null;
                if (namespaceURI == null || anyNamespace)
                    isAny = true;
            }
        }

		Step(boolean isAnyLevel, boolean isSelf, boolean isText) {
			this.anyLevel = isAnyLevel;
			this.isSelf = isSelf;
			this.nodeKind = isText ? TEXT : NODE;
			this.isAny = !isText;
		}

		boolean isAttribute() {
			return nodeKind == ATTR;		
		}
		
		boolean isElement() {
			return nodeKind == ELEMENT || nodeKind == NODE;
		}
		
		boolean isText() {
			return nodeKind == TEXT || nodeKind == NODE;
		}
		
        boolean matches(String ns, String name) {
            if (isAny) {
                return true;
            } else if (localName == null) {
                return namespaceURI.equals(ns);
            } else {
                return localName.equals(name)
                    && (anyNamespace
                        || (namespaceURI == null && ns == null)
                        || (namespaceURI != null && namespaceURI.equals(ns)));
            }
        }

		boolean isAnyLevel() {
			return anyLevel;
		}
		
		boolean isSelf() {
			return isSelf;
		}

		boolean isSelfNode() {
			return isSelf && nodeKind == NODE;
		}

		boolean isSelfElement() {
			return isSelf && nodeKind == ELEMENT;
		}
    }
}
