/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.jdbc.datasource;

import java.io.PrintWriter;
import java.sql.*;
import java.util.Properties;

import javax.sql.DataSource;

import org.apache.commons.dbcp.*;
import org.apache.commons.pool.impl.GenericKeyedObjectPool;
import org.apache.commons.pool.impl.GenericKeyedObjectPoolFactory;
import org.apache.commons.pool.impl.GenericObjectPool;


public class JDBCDataSource implements DataSource {
	
	/**
	 * Default max cached prepared statements
	 */
	private static final int DEFAULT_MAX_OPEN_STATEMENTS = 100;

	/**
	 * The JDBC driver used by the DriverConnectionFactory to establish a connection
	 */
	protected Driver driver = null;
	
	/**
	 * The connection URL to be passed to our JDBC driver to establish
	 * a connection.
	 */
	protected String url = null;

	/**
	 * The connection properties that will be sent to the JDBC driver when
	 * establishing new connections.
	 */
	protected Properties connectionProperties = null;

	/**
	 * The object pool that internally manages connections.
	 */
	protected GenericObjectPool connectionPool = null;

	/**
	 * The connection factory used by the pool to create physical connections
	 */
	protected PoolableConnectionFactory connectionFactory = null;
	
	/**
	 * The data source we will use to manage connections.  
	 */
	protected PoolingDataSource dataSource = null;
	
	JDBCDataSource(Driver driver, String url, Properties connectionProperties) throws SQLException {
		this(driver, url, connectionProperties, DEFAULT_MAX_OPEN_STATEMENTS);
	}
	
	JDBCDataSource(Driver driver, String url, Properties connectionProperties, int maxOpenPreparedStatements) 
	throws SQLException
	{
		this.driver = driver;
		this.url = url;
		this.connectionProperties = connectionProperties;

		connectionPool = new GenericObjectPool();
		connectionPool.setWhenExhaustedAction(GenericObjectPool.WHEN_EXHAUSTED_GROW);
		GenericKeyedObjectPoolFactory statementPoolFactory = null;
		if (maxOpenPreparedStatements >= 0) {
			statementPoolFactory = new GenericKeyedObjectPoolFactory(
						null, 
						-1, // unlimited maxActive (per key)
						GenericKeyedObjectPool.WHEN_EXHAUSTED_FAIL,
						0, // maxWait
						1, // maxIdle (per key) 
						maxOpenPreparedStatements); 
		}

		DriverConnectionFactory driverConnectionFactory =
			new DriverConnectionFactory(driver, url, connectionProperties);

		connectionFactory =
			new PoolableConnectionFactory(driverConnectionFactory,
										  connectionPool,
										  statementPoolFactory,
										  null,
										  false,
										  true);
		if (connectionFactory == null) {
			throw new SQLException("Cannot create PoolableConnectionFactory");
		}
		dataSource = new PoolingDataSource(connectionPool);
	}

	/**
	 * Close and release all connections that are currently stored in the
	 * connection pool associated with our data source.
	 *
	 * @exception SQLException if a database error occurs
	 */
	public void close() throws SQLException {
		GenericObjectPool oldpool = connectionPool;
		connectionPool = null;
		dataSource = null;
		try {
			if (oldpool != null) {
				oldpool.close();
			}
		} catch(SQLException e) {
			throw e;
		} catch(RuntimeException e) {
			throw e;
		} catch(Exception e) {
			throw new SQLNestedException("Cannot close connection pool", e);
		}
	}

	/**
	 * Create (if necessary) and return a connection to the database.
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public Connection getConnection() throws SQLException {
		return dataSource.getConnection();
	}

	/**
	 * Create (if necessary) and return a connection to the database.
	 *
	 * @param username Database user on whose behalf the Connection
	 *   is being made
	 * @param password The database user's password
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public Connection getConnection(String username, String password) throws SQLException {
		return dataSource.getConnection(username, password);
	}

	/**
	 * Return the login timeout (in seconds) for connecting to the database.
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public int getLoginTimeout() throws SQLException {
		return dataSource.getLoginTimeout();
	}

	/**
	 * Return the log writer being used by this data source.
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public PrintWriter getLogWriter() throws SQLException {
		return dataSource.getLogWriter();
	}

	/**
	 * Set the login timeout (in seconds) for connecting to the database.
	 *
	 * @param loginTimeout The new login timeout, or zero for no timeout
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public void setLoginTimeout(int loginTimeout) throws SQLException {
		dataSource.setLoginTimeout(loginTimeout);
	}

	/**
	 * Set the log writer being used by this data source.
	 *
	 * @param logWriter The new log writer
	 *
	 * @exception SQLException if a database access error occurs
	 */
	public void setLogWriter(PrintWriter logWriter) throws SQLException {
		dataSource.setLogWriter(logWriter);
	}

	public Driver getDriver() {
		return this.driver;
	}
	
	public String getUrl() {
		return this.url;
	}

	public String getUsername() {
		return this.connectionProperties.getProperty("user");
	}

	public String getPassword() {
		return this.connectionProperties.getProperty("password");
	}

	public void setDefaultAutoCommit(boolean defaultAutoCommit) {
		connectionFactory.setDefaultAutoCommit(defaultAutoCommit);
	}

	public void setDefaultReadOnly(boolean defaultReadOnly) {
		connectionFactory.setDefaultReadOnly(defaultReadOnly);
	}

	public void setDefaultTransactionIsolation(int defaultTransactionIsolation) {
		connectionFactory.setDefaultTransactionIsolation(defaultTransactionIsolation);
	}

	public void setDefaultCatalog(String defaultCatalog) {
		connectionFactory.setDefaultCatalog(defaultCatalog);
	}

	public void setValidationQuery(String validationQuery) {
		if ((validationQuery != null) && (validationQuery.trim().length() > 0)) {
			connectionFactory.setValidationQuery(validationQuery);
		} else {
			connectionFactory.setValidationQuery(null);
		}
	}

	public boolean isAccessToUnderlyingConnectionAllowed() {
		return dataSource.isAccessToUnderlyingConnectionAllowed();
	}

	public void setAccessToUnderlyingConnectionAllowed(boolean allow) {
		dataSource.setAccessToUnderlyingConnectionAllowed(allow);
	}

	// Connection pool management
	
	/**
	 * [Read Only] The current number of active connections that have been
	 * allocated from this data source.
	 */
	public int getNumActive() {
		if (connectionPool != null) {
			return connectionPool.getNumActive();
		} else {
			return 0;
		}
	}

	/**
	 * [Read Only] The current number of idle connections that are waiting
	 * to be allocated from this data source.
	 */
	public int getNumIdle() {
		if (connectionPool != null) {
			return connectionPool.getNumIdle();
		} else {
			return 0;
		}
	}

	public int getMaxActive() {
		return connectionPool.getMaxActive();
	}

	public void setMaxActive(int maxActive) {
		connectionPool.setMaxActive(maxActive);
	}

	public int getMaxIdle() {
		return connectionPool.getMaxIdle();
	}

	public void setMaxIdle(int maxIdle) {
		connectionPool.setMaxIdle(maxIdle);
	}

	public int getMinIdle() {
		return connectionPool.getMinIdle();
	}

	public void setMinIdle(int minIdle) {
		connectionPool.setMinIdle(minIdle);
	}

	public long getMaxWait() {
		return connectionPool.getMaxWait();
	}

	public void setMaxWait(long maxWait) {
		connectionPool.setMaxWait(maxWait);
	}
	
	public byte getWhenExhaustedAction() {
		return connectionPool.getWhenExhaustedAction();
	}
	
	public void setWhenExhaustedAction(byte whenExhaustedAction) {
		connectionPool.setWhenExhaustedAction(whenExhaustedAction);
	}

	public boolean getTestOnBorrow() {
		return connectionPool.getTestOnBorrow();
	}

	public void setTestOnBorrow(boolean testOnBorrow) {
		connectionPool.setTestOnBorrow(testOnBorrow);
	}

	public boolean getTestOnReturn() {
		return connectionPool.getTestOnReturn();
	}

	public void setTestOnReturn(boolean testOnReturn) {
		connectionPool.setTestOnReturn(testOnReturn);
	}
	
	public long getTimeBetweenEvictionRunsMillis() {
		return connectionPool.getTimeBetweenEvictionRunsMillis();
	}

	public void setTimeBetweenEvictionRunsMillis(long timeBetweenEvictionRunsMillis) {
		connectionPool.setTimeBetweenEvictionRunsMillis(timeBetweenEvictionRunsMillis);
	}

	public int getNumTestsPerEvictionRun() {
		return connectionPool.getNumTestsPerEvictionRun();
	}

	public void setNumTestsPerEvictionRun(int numTestsPerEvictionRun) {
		connectionPool.setNumTestsPerEvictionRun(numTestsPerEvictionRun);
	}

	public long getMinEvictableIdleTimeMillis() {
		return connectionPool.getMinEvictableIdleTimeMillis();
	}

	public void setMinEvictableIdleTimeMillis(long minEvictableIdleTimeMillis) {
		connectionPool.setMinEvictableIdleTimeMillis(minEvictableIdleTimeMillis);
	}

	public boolean getTestWhileIdle() {
		return connectionPool.getTestWhileIdle();
	}

	public void setTestWhileIdle(boolean testWhileIdle) {
		connectionPool.setTestWhileIdle(testWhileIdle);
	}

}
