/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.jdbc.typing;

/**
 * This interface gives acess to meta information about a column in a relational
 * table. This information is useful to verify the integrity constraints
 * in the the relational tables.
 */
public interface ColumnMetaData
{
	/**
	 * Accessor to the column name.
	 * @return the name of the column.
	 */
	public String getColumnName();

	/**
	 * Accessor to the column's table name.
	 * @return the name of the column's table.
	 */
	public String getTableName();
	
	/**
	 * Whether this column can be null.
	 * @return true if the column is nullable in the RDBMS.
	 */
	public boolean isNullable();
	
	/**
	 * Whether filling this column is optional.
	 * @return true if column is nullable or has a default value.
	 */
	public boolean isOptional();
	
	/**
	 * Returns the base SQL type corresponding to the column. The base SQL
	 * types is a subset of SQL data types defined in {@link java.sql.Types}
	 * which represent a family of types, aka VARCHAR, VARBINARY, DECIMAL,
	 * REAL, DOUBLE, TIMESTAMP. Whithin a family mainly size limits differ.
	 * @return a constant defined in {@link java.sql.Types}.
	 */
	public short getDataType();
	
	/**
	 * Returns information on the native SQL type.
	 * @return an object containing many information on the RDBMS specific type.
	 */
	public TypeMap.SQLType getSqlType();
	
	/**
	 * Accessor to the native type name.
	 * @return a native DDL data type.
	 */
	public String getTypeName();
	
	/**
	 * Gets column size or precision. Given in bytes, bits, characters
	 * or digits depending on the base type.
	 * @return -1 if not significant.
	 */
	public long getColumnSize();
	
	/**
	 * Gets scale for the DECIMAL family types.
	 * @return -1 if not significant.
	 */
	public int getDecimalDigits();
	
	/**
	 * Returns the default value defined in the RDBMS for this column.
	 * @return the default value as a String
	 */
	public String getDefaultValue();
	
	/**
	 * Gets the RDBMS position of the column within the table. May be used for
	 * "SELECT *" statements.
	 * @return a positive non-nul integer.
	 */
	public int getOrdinalPosition();
	
	/**
	 * Whether the column belongs to the RDBMS Primary Key of the table.
	 * @return true if column is part of the PK.
	 */
	public boolean isPrimaryKey();
	
	/**
	 * Whether the column type belongs to the SQL types BLOB, CLOB, LONG,
	 * TEXT or more generally types allowing large amount of data to be stored
	 * and that are not indexed in the standard way in RDBMS. With most JDBC
	 * implementations these type must be accessed through streams.
	 * @return true if column has the corresponding type.
	 */
	public boolean isLongType();
	
	/**
	 * Whether the column type belongs to the SQL types BLOB, CLOB.
	 * @return true if column has the corresponding type.
	 */
	public boolean isLOB();

	/**
	 * Gives the position in the table Primary Key.
	 * @return -1 if does not belong to PK.
	 */
	public short getPrimaryKeySeq();
	
	/**
	 * Returned the reference table if the column belongs to a Foreign Key.
	 * @return the reference table name or null if not relevant.
	 */
	public String getRefTable();
	
	/**
	 * Returned the reference column if the column belongs to a Foreign Key.
	 * @return the reference column name or null if not relevant.
	 */
	public String getRefColumn();

	/**
	 * Returns the DDL statement fragment corresponding to the creation of
	 * a column with the same specifications (default value is ignored).
	 * @return a native DDL fragment.
	 */
	public String getTypeCreationString();
	
	public DbType getType();
}
