/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.jdbc.typing;

import java.sql.*;

import org.xquark.jdbc.typing.TypeMap;

/**
 * This class maintains the meta information about a column in a relational
 * table. This information is useful to verify the integrity constraints
 * in the the relational tables.
 *
 * This class is the default JDBC implementation of ColumnMetaData that uses
 * standard JDBC metadata functions.
 */
public class ColumnMetaDataImpl implements ColumnMetaData
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name:  $";

    /** Whether the column belongs to the primary key */
    protected short primaryKeySeq = -1;

    /** Name of the primary key table when the column is a foreign key */
    protected String refTable = null;

    /** Name of the primary key column when the column is a foreign key */
    protected String refColumn = null;

    /** The name of the table that owns this column */
    protected String tableName;

    /** The name of this column */
    protected String columnName;

    /** The data type of this column */
    protected DbType type = null;
    
    /** The native data type of this column */
    protected TypeMap.SQLType sqlType = null;

    /** Whether this column can be null */
    protected boolean nullable;
    
    /** Default value */
    protected String columnDef;

    /** The ordinal position of this column within the table's definition */
    protected int ordinalPosition;

    /** Type creation string */
    protected String typeDDLString;
    
    protected ColumnMetaDataImpl(){};
    
    /**
     * Constructor.
     *
     * The constructor of the ColumnMetaData is called by AbstractConnection
     * implementation.
     * @param rs the result set of DatabaseMetadata.getColumns()
     */
    public ColumnMetaDataImpl(ResultSet rs, TypeMap typeMap,
							  JDBCProperties properties)
    throws SQLException
    {
        tableName = rs.getString(3);
        columnName = rs.getString(4);
		sqlType = typeMap.getType(rs.getString(6));
        if (properties.useStaticMappingFromNativeType())
            type = new DbType(rs.getInt(7), rs.getInt(9), sqlType);
        else
            type = new DbType(rs.getInt(5), rs.getInt(7), rs.getInt(9));
        int tmp = rs.getInt(11);
        nullable = (rs.getInt(11) == DatabaseMetaData.columnNullable)
        			|| sqlType.isAutoFill();	// process it like a nullable (no need to fill)
        columnDef = rs.getString(13);
        if (columnDef != null) {
        	if (columnDef.length() == 0)
        		columnDef = null;	// for MySQL that returns "" instead of null, bet it won't bother to generalize...
        	else if (columnDef.charAt(0) == '\'' && columnDef.charAt(columnDef.length() - 1) == '\'')
        		columnDef = columnDef.substring(1, columnDef.length() - 1); // Oracle encloses string literal in quotes maybe distinguish it from numbers. 
        }
        ordinalPosition = rs.getInt(17);
        
        switch (type.getJDBCType())
        {
            case Types.CLOB:
			case Types.BLOB:
                type.setLOB(true);
                // no break
            case Types.LONGVARCHAR:
			case Types.LONGVARBINARY:
                type.setLongType(true);
                break;
            default:
        }
        
        // type creation string
        typeDDLString = getTypeCreationString(
                        getDataType(),
                        typeMap,
                        getColumnSize(),
                        (int)getColumnSize(),
                        getDecimalDigits());
    }

	public ColumnMetaDataImpl(ResultSetMetaData rs, int i, TypeMap typeMap,
							  JDBCProperties properties)
	throws SQLException
	{
		tableName = rs.getTableName(i);
		columnName = rs.getColumnName(i);
		sqlType = typeMap.getType(rs.getColumnTypeName(i));
		if (properties.useStaticMappingFromNativeType())
		    type = new DbType(sqlType);
		else
		    type = new DbType(rs.getColumnType(i));
		nullable = (rs.isNullable(i) == ResultSetMetaData.columnNullable)
					|| sqlType.isAutoFill();	// process it like a nullable (no need to fill)
		ordinalPosition = i;
		type.setLongType(rs.isSearchable(i));
		type.setLOB(type.isLOB());
        
		type.setLength(rs.getColumnDisplaySize(i));
		if (type.getLength() <= 0)
		    type.setLength(rs.getPrecision(i));
		type.setScale(rs.getScale(i));

		// type creation string
		typeDDLString = getTypeCreationString(
						getDataType(),
						typeMap,
						getColumnSize(),
						(int)getColumnSize(),
						getDecimalDigits());
	}
	
    /**
     * Whether this column can be null.
     */
    public boolean isNullable()
    {
        return nullable;
    }

    public boolean isOptional()
    {
        return isNullable() || getDefaultValue() != null;
    }

    public String getTableName() { return tableName; }

    public String getColumnName() { return columnName; }

    public DbType getType() {
        return type;
    }
    
    public short getDataType() { return (short)type.getJDBCType(); }

    public String getTypeName() { return type.getNativeType(); }

    public long getColumnSize() { return type.getLength(); }

    public int getDecimalDigits() { return type.getScale(); }

    public String getDefaultValue() { return columnDef; }

    public int getOrdinalPosition() { return ordinalPosition; }

    public boolean isPrimaryKey() { return primaryKeySeq != -1; }

    public short getPrimaryKeySeq() { return primaryKeySeq; }

    public String getRefTable() { return refTable; }

    public String getRefColumn() { return refColumn; }

    public String toString()
    {
        return "#<Column "+getTableName()+"."+getColumnName()+">";
    }

    public void setPrimaryKeySeq(short seq) { primaryKeySeq = seq; }

    public void setRefTable(String name) { refTable = name; }

    public void setRefColumn(String name) { refColumn = name; }
    
    public String getTypeCreationString()
    {
        return typeDDLString;
    }
    
    public boolean isLongType()
    {
        return type.isLongType();
    }
    
    public boolean isLOB()
    {
        return type.isLOB();
    }
    
	//////////////////////////////////////////////////////////////////////////
	//                          PUBLIC STATIC UTILITIES
	//////////////////////////////////////////////////////////////////////////
	public static int findRealisticType(int type, long size, TypeMap typeMap)
	{
		switch (type) // switch according to "base" target JDBC type
		{   // and try to analyse facets(restrictions) to find a 'smaller one'
			case Types.VARCHAR :
			case Types.LONGVARCHAR :
				if (size <= typeMap.getSize(Types.VARCHAR))
					type = Types.VARCHAR;
				else
					type = Types.LONGVARCHAR;
				break;
			case Types.VARBINARY :
			case Types.LONGVARBINARY :
				if (size <= typeMap.getSize(Types.VARBINARY))
					type = Types.VARBINARY;
				else
					type = Types.LONGVARBINARY;
				break;
			default :   // Nothing to do with VARCHAR, TIMESTAMP, REAL
		}
		return type;
	}
    
	/* Generate the native SQL statement for column type specification. The 
	 * optimum native type is chosen according to size parameters.
	 */
	public static String getTypeDDL(int dataType, TypeMap typeMap,
			long maxLength)
	{
		return getTypeDDLStatement(dataType, typeMap, maxLength, -1, -1);
	}
    
	public static String getTypeDDL(int dataType, TypeMap typeMap,
			int precision, int scale)
	{
		return getTypeDDLStatement(dataType, typeMap, -1, precision, scale);
	}
    
	public static String getTypeDDL(int dataType, TypeMap typeMap,
			int precision)
	{
		return getTypeDDLStatement(dataType, typeMap, -1, precision, -1);
	}
    
	public static String getTypeDDL(int dataType, TypeMap typeMap)
	{
		return getTypeDDLStatement(dataType, typeMap, -1, -1, -1);
	}
    
	/* Warning: no size adaptation is performed (unlike getTypeDDL()).*/
	public static String getTypeCreationString(int dataType, TypeMap typeMap,
	long maxLength, int precision, int scale)
	{
		String typeString;
        
		/* building precision and scale string */
		switch (dataType) // switch according to target JDBC type for primitive type
		{   // and try to analyse facets(restrictions) to find a 'smaller one'
			case Types.CHAR : case Types.BINARY :
			case Types.VARCHAR : case Types.VARBINARY :
			case Types.DOUBLE : case Types.FLOAT :
			case Types.NUMERIC : case Types.DECIMAL :
				typeString = getDefaultTypeCreationString(dataType, typeMap,
				maxLength, precision, scale);
				break;
			default :   // for other type just concatenate the creation metadada param
				typeString = typeMap.getNativeTypeName(dataType);
				if (typeMap.getTypeParam(dataType) != null)
					typeString += typeMap.getTypeParam(dataType);
		}
		return typeString;
	}
    
	//////////////////////////////////////////////////////////////////////////
	//                              PRIVATE
	//////////////////////////////////////////////////////////////////////////
	private static String getTypeDDLStatement(int dataType, TypeMap typeMap,
			long maxLength, int precision, int scale)
	{
		/* Adding maximum size for types needing it */
		if (maxLength < 0)
		{
			switch (dataType) // switch according to target JDBC type for primitive type
			{   // and try to analyse facets(restrictions) to find a 'smaller one'
				case Types.CHAR : case Types.VARCHAR :
				case Types.VARBINARY : case Types.BINARY :
					maxLength = typeMap.getSize(dataType);
					break;
				default :
			}
		}
        
	   return getTypeCreationString(
			findRealisticType(dataType, maxLength, typeMap),
			typeMap,
			maxLength,
			precision,
			scale
			);
	}

	/** Builds the database-specific string for type specification.
	 * It may be constructed by concatenation of typeName, size, and scale
	 * or in any other way.
	 * <B>A call to the static method loadTypeMetadata must have been
	 * performed before this method is able to return something.</B>
	 * @return  a creation string assumed to be constructed
	 * as follows : typeName + '(' + size + ',' + scale + ')'
	 * null if loadTypeMetadata has not been called.
	 */
	private static String getDefaultTypeCreationString(int dataType,
	TypeMap typeMap, long maxLength, int precision, int scale)
	{
		String ret = null;
		if (typeMap != null)
		{
			ret = typeMap.getNativeTypeName(dataType);
			if (maxLength > 0)
				ret += "(" + maxLength + ")";
			else if (precision > 0)
			{
				ret += "(" + precision;
				if (scale > 0)
					ret += "," + scale;
				ret += ")";
			}
			else if (typeMap.getTypeParam(dataType) == null)    // take the maximum authorized by the rdbms
			{
				switch (dataType) // switch according to target JDBC type for primitive type
				{
					case Types.NUMERIC : case Types.FLOAT :
					case Types.DOUBLE : case Types.DECIMAL :
						// do not add precision (max is implicit) // TO IMPROVE : build a flag for param needed
						break;
					default :   // for other type just concatenate the creation metadada param
						ret += "(" + typeMap.getSize(dataType) + ")";
				}
			}
		}
		return ret;
	}
	/**
	 * @return Returns the sqlType.
	 */
	public TypeMap.SQLType getSqlType() {
		return sqlType;
	}
}
