/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.jdbc.typing;

import org.xquark.schema.SimpleType;
import org.xquark.schema.datatypes.PrimitiveType;

/** This class describes, for a datum to map, features of the database
 * corresponding type such as DDL type creation string.
 * <p>This class is used for generating native dbms type creation string.</p>
 */
public class JavaTypeInfo extends XMLTypeInfo
{
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName     = "$Name:  $";
    
    //////////////////////////////////////////////////////////////////////////
    //                          CONSTANTS
    //////////////////////////////////////////////////////////////////////////
    public final static int JAVA_STRING = 0;
    public final static int JAVA_BOOLEAN = 1;
	public final static int JAVA_BIG_DECIMAL = 2;
    public final static int JAVA_FLOAT = 3;
    public final static int JAVA_DOUBLE = 4;
	public final static int JAVA_LONG = 5;
	
	//below not recognized by JDBC
	
	public final static int JAVA_LIST = 6;
    public final static int JAVA_UNKNOWN = 7;
    
	public final static int SCHEMA_DURATION = 8;
	public final static int SCHEMA_DATETIME = 9;
	public final static int SCHEMA_BYTE_ARRAY = 10;
	public final static int SCHEMA_URI = 11;
	public final static int SCHEMA_QNAME = 12;

    /** mapping between schema type and java types assumed by
     * PrimitiveType.Convert() & toXMLString()
     */
    public static int JAVA_TYPE_BINDINGS[] = new int[33]; // (19 base primitive types + 3, non-dense + index 0 unused)
    
    static
    {
        // mapping between schema type and java types assumed by PrimitiveType.Convert() & toXMLString()
        JAVA_TYPE_BINDINGS[PrimitiveType.STRING] = JAVA_STRING;
        JAVA_TYPE_BINDINGS[PrimitiveType.BOOLEAN] = JAVA_BOOLEAN;
        JAVA_TYPE_BINDINGS[PrimitiveType.DECIMAL] = JAVA_BIG_DECIMAL;   // or Long if possible
        JAVA_TYPE_BINDINGS[PrimitiveType.FLOAT] = JAVA_FLOAT;
        JAVA_TYPE_BINDINGS[PrimitiveType.DOUBLE] = JAVA_DOUBLE;
        JAVA_TYPE_BINDINGS[PrimitiveType.QNAME] = SCHEMA_QNAME;
        JAVA_TYPE_BINDINGS[PrimitiveType.NOTATION] = SCHEMA_QNAME;
        JAVA_TYPE_BINDINGS[PrimitiveType.DURATION] = SCHEMA_DURATION;
        JAVA_TYPE_BINDINGS[PrimitiveType.DATE_TIME] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.TIME] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.DATE] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.GYEAR_MONTH] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.GYEAR] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.GMONTH_DAY] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.GDAY] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.GMONTH] = SCHEMA_DATETIME;
        JAVA_TYPE_BINDINGS[PrimitiveType.HEX_BINARY] = SCHEMA_BYTE_ARRAY;
        JAVA_TYPE_BINDINGS[PrimitiveType.BASE64_BINARY] = SCHEMA_BYTE_ARRAY;
        JAVA_TYPE_BINDINGS[PrimitiveType.ANY_URI] = SCHEMA_URI;
        JAVA_TYPE_BINDINGS[PrimitiveType.UNION] = JAVA_UNKNOWN;
        JAVA_TYPE_BINDINGS[PrimitiveType.LIST] = JAVA_LIST;
        JAVA_TYPE_BINDINGS[PrimitiveType.ANYSIMPLETYPE] = JAVA_STRING;
        
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                          DATA
    //////////////////////////////////////////////////////////////////////////
    
    /** Object maintaining information on mapping between XML and Java types.
     */
    protected int javaType = JAVA_STRING; // for default mapping
    
    
    //////////////////////////////////////////////////////////////////////////
    //                 CONSTRUCTORS & ACCESSORS
    //////////////////////////////////////////////////////////////////////////
	protected JavaTypeInfo() {}
	
	public JavaTypeInfo(String XMLType) { setXMLType(XMLType);}
    
    public JavaTypeInfo(SimpleType XMLType) { setXMLType(XMLType);}
    
	public void setXMLType(String XMLType)
	{
		super.setXMLType(XMLType);
		initialize();
	}
    
	public void setXMLType(SimpleType XMLType)
	{
		super.setXMLType(XMLType);
		initialize();
	}
    
    /** Set metadata info from mapping between JDBC and native SQL types
     * @param typeMap a loaded TypeMap
     */
    private void initialize()
    {
        javaType = JAVA_TYPE_BINDINGS[sType.getPrimitive().getType()];
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                      PUBLIC METHODS
    //////////////////////////////////////////////////////////////////////////
    public int getJavaType()
    {
        return javaType;
    }
    
    //////////////////////////////////////////////////////////////////////////
    //                      UTILITIES
    //////////////////////////////////////////////////////////////////////////
    public String toString()
    {
        StringBuffer buf = new StringBuffer();
        buf.append("# NativeTypeInfo\n");
        buf.append("\tXML = ");
        
        PrimitiveType pType = sType.getPrimitive();
        buf.append(pType);
        buf.append("[maxLen=");
        buf.append(pType.getMaxLength());
        buf.append(", totDig=");
        buf.append(pType.getTotalDigits());
        buf.append(", fracDig=");
        buf.append(pType.getFractionDigits());
        buf.append(", min=");
        buf.append(pType.getMinValue());
        buf.append(", max=");
        buf.append(pType.getMaxValue());
        
        buf.append("]\n");
        buf.append("\tJAVA = ");
        buf.append(javaType);
        buf.append("\n\n");
        
        return buf.toString();
    }
}
