/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.util.ArrayList;
import java.util.List;

/**
 * class for Identity Constraint.
 *
 * <p>This is the implementation of the Identity Constraint.
 * The following properties are defined:</p>
 * <p> 1. name : a NCName, defined in SchemaComponent.</p>
 * <p> 2. target namespace : null or a namespace name, defined in SchemaComponent.</p>
 * <p> 3. category : one of {CATEGORY_KEY, CATEGORY_KEYREF, CATEGORY_UNIQUE}.</p>
 * <p> 4. selector : a restricted XPath expression. </p>
 * <p> 5. fields : a non-empty list of restricted XPath expressions. </p>
 * <p> 6. referenced key : required if category is KEYREF, otherwise forbidden.
 * the referenced key, an identity constraint definition with category equal to 
 * key or unique and the name is equal to referenced name</p>
 *
 * <p>See XML Schema Part 1: Structures 3.11
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 * @see org.xquark.xml.schema.ElementDeclaration
 */
public class IdentityConstraint extends SchemaComponent {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";

  // constant
  public static final int CATEGORY_KEY = 0;
  public static final int CATEGORY_KEYREF = 1;
  public static final int CATEGORY_UNIQUE = 2;
  
  // fields resovled in Loader
  private int category;
  private List selector = null;
  private ArrayList fields = null;
  private IdentityConstraint referencedKey = null;  // initialization : keyref is resovled by keyrefName
  private ElementDeclaration contextDecl;
  
  /**
   * Create a new IdentityConstraint.
   *
   * @param schema The schema which contains this identity constraint.
   * @param name The name of identity constraint, a NCName.
   * @param category The category,  one of {CATEGORY_KEY, CATEGORY_KEYREF, CATEGORY_UNIQUE}.
   * @param contextDecl The context element declaration of identity constraint.
   */
  public IdentityConstraint(Schema schema, String name, int category, ElementDeclaration contextDecl) {
    super(schema, name);
    this.category = category;
    this.contextDecl = contextDecl;
  }
  
  public IdentityConstraint(Schema schema, String name, ElementDeclaration contextDecl) {
    this(schema, name, CATEGORY_UNIQUE, contextDecl);
  }
  
  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  } 
  
  /**
   * Look up the category
   *
   * @return The category, one of {CATEGORY_KEY, CATEGORY_KEYREF, CATEGORY_UNIQUE}.
   */
  public int getCategory() {
    return this.category;
  }
  public void setCategory(int category) {
    this.category = category;
  }

  /**
   * Look up the context element declaration
   *
   * @return the context element declaration
   */
  public ElementDeclaration getContextDeclaration() {
    return this.contextDecl;
  }
 
 /**
   * Look up the selector
   *
   * @return the list of XPath expressions defining the selector
   */
  public List getSelector() {
    return this.selector;
  }
  public void setSelector(List selector) {
    this.selector = selector;
  }

  /**
   * Look up the identity constraint of referenced key
   *
   * @return The name of identity constraint of referenced key, if category is keyref; null, otherwise
   */
  public IdentityConstraint getReferencedKey() {
    return referencedKey;
  }
  public void setReferencedKey(IdentityConstraint refer) {
    this.referencedKey = refer;
  }  
  
  /**
   * Look up the fields
   *
   * @return A list of lists of XPath expressions, each sublist defining a particular field
   * 
   */
  public ArrayList getFields() {
    return this.fields;
  }	
  public void addField(List field) {
    if (fields == null)
      fields = new ArrayList();
    fields.add(field); 
  }
  public void deleteFields() {
    fields.clear();
  }
  
  public boolean isEqualToKeyRefFields() {
    if ( this.category == CATEGORY_KEYREF ) {
      if ( this.fields != null && referencedKey != null && referencedKey.fields != null
           && this.fields.size() == referencedKey.fields.size() )
        return true;
      else
        return false;
    }
    else
      return true;
  }
      
}
