/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.util.*;

/**
 * class for Particle.
 *
 * <p>This is the implementation of the Particle.
 * The following properties are defined:</p>
 * <p> 1. min occurs : a non-negative integer.</p>
 * <p> 2. max occurs : either a non-negative integer or unbounded.</p>
 * <p> 3. term : one of a model group, a wildcard, or an element declaration. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.9
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.ModelGroup
 * @see org.xquark.xml.schema.Wildcard
 * @see org.xquark.xml.schema.ElementDeclaration
 */
public final class Particle implements SchemaVisitable, Cloneable {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
  
  private int minOccurs = 1;
  private int maxOccurs = 1;
  private Object term;

  private int minExtent = 0;
  private int maxExtent = 0;
  private ArrayList firstElements = new ArrayList();
  private ArrayList firstWildcards = null;

  /**
   * Create a new Particle.
   *
   * @param term The term, one of a model group, a wildcard, or an element declaration.
   */
  public Particle(Object term) {
    this.term = term;
  }

  public Object clone() throws CloneNotSupportedException{
    Object result = null;
    result = super.clone();
    ((Particle)result).firstElements = new ArrayList();
    return result;
  }

  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }

  public String toString() {
    String result = "#<Particle "+term;
    Iterator it = firstElements.iterator();
    if (it.hasNext()) result+=("["+it.next());
    while (it.hasNext()) result+=(", "+it.next());
    return result+"]>";
  }
  
  public boolean equals(Particle p) {
    if ( this == p ) return true;
    if ( p == null ) return false;
    if ( this.minOccurs != p.minOccurs || this.maxOccurs != p.maxOccurs ||
         this.minExtent != p.minExtent || this.maxExtent != p.maxExtent )
      return false;
    
    if ( this.term == p.term ) return true;
    
    if ( this.term instanceof ElementDeclaration && p.term instanceof ElementDeclaration )
    {
       if ( ((ElementDeclaration)this.term).equals((ElementDeclaration)p.term) ) return true;
    }
    
    if ( this.term instanceof Wildcard && p.term instanceof Wildcard ) {
       if ( ((Wildcard)this.term).equals((Wildcard)p.term) ) return true;
    }
    
    if ( (this.term instanceof AllModelGroup && p.term instanceof AllModelGroup) ||
         (this.term instanceof ChoiceModelGroup && p.term instanceof ChoiceModelGroup) ||
         (this.term instanceof SequenceModelGroup && p.term instanceof SequenceModelGroup) ) {
      ModelGroup mg1 = (ModelGroup)this.term;
      ModelGroup mg2 = (ModelGroup)p.term;
      if ( mg1.size() != mg2.size() ) return false;
      java.util.Iterator it1 = mg1.iterator();
      java.util.Iterator it2 = mg2.iterator();
      while ( it1.hasNext() ) {
        Particle p1 = (Particle) it1.next();
        Particle p2 = (Particle) it2.next();
        if ( p1.equals(p2) == false ) return false;
      }
      return true;
    }
    
    return false;
  }
  
  /**
   * Look up the min occurs
   *
   * @return The min occurs, a non-negative integer.
   */
  public int getMinOccurs() {
    return minOccurs;
  }

  /**
   * Look up the max occurs
   * max occurs is a non-negative integer. If max occurs is the maxmum number of integer,
   * it means that max occurs is unbounded.
   *
   * @return The max occurs.
   */
  public int getMaxOccurs() {
    return maxOccurs;
  }

  public boolean isMaxOccursUnbounded() {
    return maxOccurs == Integer.MAX_VALUE;
  }

  /**
   * Look up the term
   *
   * @return The term, one of a model group, a wildcard, or an element declaration.
   */
  public Object getTerm() {
    return term;
  }

  public int getMinExtent() {
    return minExtent;
  }

  public int getMaxExtent() {
    return maxExtent;
  }

  public boolean isEmptiable() {
    return minExtent == 0;
  }

  public void setMinOccurs(int min) {
    minOccurs = min;
  }

  public void setMaxOccurs(int max) {
    maxOccurs = max;
  }

  public void setMaxOccursUnbounded() {
    setMaxOccurs(Integer.MAX_VALUE);
  }
  
  void setTerm(Object term) {
    this.term = term;
  }
  
  void setMinExtent(int extent) {
    minExtent = extent;
  }

  void setMaxExtent(int extent) {
    maxExtent = extent;
  }

  void setMaxExtentUnbounded(int extent) {
    setMaxExtent(Integer.MAX_VALUE);
  }

  void addFirstElement(ElementDeclaration decl) {
    firstElements.add(decl);
  }
  
  void addFirstElements(Collection elements) {
    firstElements.addAll(elements);
  }
  
  void addFirstWildcard(Wildcard w) {
    if (firstWildcards == null) firstWildcards = new ArrayList();
    firstWildcards.add(w);
  }
  
  void addFirstWildcards(Collection wildcards) {
    if (wildcards == null) return;
    if (firstWildcards == null) firstWildcards = new ArrayList();
    firstWildcards.addAll(wildcards);
  }
  
  List getFirstElements() {
    return firstElements;
  }
  
  public List getFirstWildcards() {
    return firstWildcards;
  }

  public List firstValidElements() {
    ArrayList result = new ArrayList();
    fillFirstValidElements(result);
    return result;
  }

  void fillFirstValidElements(List result) {
    Iterator it = firstElements.iterator();
    while (it.hasNext()) {
      ElementDeclaration decl = (ElementDeclaration)it.next();
      // add the element in firstElements
      if (!decl.isAbstract()) result.add(decl);
      // add substitution elements
      decl.fillAllConcreteSubstitutionElements(result);
    }
  }

	//            Created
	//            add substitutionGroup treating
	//
  public boolean hasNextElement(String namespace, String localName) {
    Iterator it = firstElements.iterator();
    while (it.hasNext()) {
      ElementDeclaration decl = (ElementDeclaration)it.next();
      if (decl.hasName(namespace, localName)) return true;
      if (decl.getSubstitutionElement(namespace, localName) != null) return true;
    }
    if (firstWildcards != null) {
      it = firstWildcards.iterator();
      while (it.hasNext()) {
        Wildcard w = (Wildcard)it.next();
        if (w.isAllowed(namespace)) return true;
      }
    }
    
    return false;
  }
  
  private Particle substitutionTreating() {
    if ( this.term instanceof Wildcard ) return this;
    else if ( this.term instanceof ElementDeclaration ) {
      List subList = ((ElementDeclaration)this.term).getSubstitutionElementList();
      if ( subList == null || subList.size() == 0 ) return this;
      ChoiceModelGroup choiceGroup = new ChoiceModelGroup();
      Particle choiceParticle = new Particle(choiceGroup);    
      choiceParticle.minOccurs = this.minOccurs;
      choiceParticle.maxOccurs = this.maxOccurs;
      choiceParticle.minExtent = this.minExtent;
      choiceParticle.maxExtent = this.maxExtent;
      Particle newThis = new Particle(this.term);
      newThis.minOccurs = 1;
      newThis.maxOccurs = 1;
      newThis.minExtent = 1;
      newThis.maxExtent = 1;
      choiceGroup.add(newThis);
      Iterator it = subList.iterator();
      while ( it.hasNext() ) {
        Particle p = new Particle(it.next());
        p.minOccurs = 1;
        p.maxOccurs = 1;
        p.minExtent = 1;
        p.maxExtent = 1;
        choiceGroup.add(p);
      }
      return choiceParticle;
    }
    else {
      boolean changed = false;
      Particle newParticle = new Particle(((ArrayList)this.term).clone());
      newParticle.minOccurs = this.minOccurs;
      newParticle.maxOccurs = this.maxOccurs;
      newParticle.minExtent = this.minExtent;
      newParticle.maxExtent = this.maxExtent;
      List it = (List)this.term;
      for ( int i = 0; i < it.size(); i++ ) {
        Particle childParticle = (Particle)it.get(i);
        Particle subPaticle = childParticle.substitutionTreating();
        if ( subPaticle != childParticle ) {
          changed = true;
          ((ArrayList)newParticle.term).set(i, subPaticle);
        }
      }
      if ( changed ) {
        newParticle.optimizationParticle();
        return newParticle;
      }
      else return this;
    }
  }
  
  private void optimizationParticle() {
    if ( this.term instanceof ModelGroup ) {
      ModelGroup group = (ModelGroup)this.term;
      java.util.ListIterator it = group.listIterator();
      while (it.hasNext()) {
        Particle p = (Particle)it.next();
        it.set(p);
        if (p == null) it.remove();
        else if (p.getTerm() instanceof ModelGroup) {
          ModelGroup subgroup = (ModelGroup)p.getTerm();
          if (p.getMinOccurs() == 1 && p.getMaxOccurs() == 1
              && subgroup.getCompositor() == group.getCompositor()) {
            //System.err.println("Simplification");
            it.remove();
            Iterator it2 = subgroup.iterator();
            while (it2.hasNext()) it.add(it2.next());
          }
        }
      }
    }
  }
  
  public void checkParticleValidRestriction(Particle ancestor) throws SchemaException {  
    if ( ancestor == null ) {
      // Particle can't be restricted from its base particle
      throw new SchemaException("cos-particle-restrict.2", this);
    }
    Particle subThis = this.substitutionTreating();
    Particle subAncestor = ancestor.substitutionTreating();
    subThis.checkParticleDerivation(subAncestor);
  }
  
  private void checkParticleDerivation(Particle ancestor) throws SchemaException { 
    // cos-particle-restrict
    if ( this.equals(ancestor) ) return;
        
    if ( this.term instanceof ElementDeclaration )
      this.checkParticleDerivationOK_Elt(ancestor);
    else if ( this.term instanceof Wildcard ) 
      this.checkParticleDerivationOK_Any(ancestor);
    else if ( this.term instanceof AllModelGroup ) 
      this.checkParticleDerivationOK_All(ancestor);
    else if ( this.term instanceof ChoiceModelGroup ) 
      this.checkParticleDerivationOK_Choice(ancestor);
    else if ( this.term instanceof SequenceModelGroup ) 
      this.checkParticleDerivationOK_Sequence(ancestor);
  }
  
  private void checkParticleDerivationOK_Elt(Particle ancestor) throws SchemaException {   
     if ( ancestor.term instanceof ElementDeclaration )
      checkEltElt_NameAndTypeOK(ancestor);
    else if ( ancestor.term instanceof Wildcard )
      checkEltAny_NSCompat(ancestor);
    else
      checkEltAllChoiceSequence_RecursAsIfGroup(ancestor);
   }
  
  private void checkParticleDerivationOK_Any(Particle ancestor) throws SchemaException {
    // cos-particle-restrict
    if ( ancestor.term instanceof Wildcard )
      checkAnyAny_NSSubset(ancestor);
    else 
      // Particle can't be restricted from its base particle
      throw new SchemaException("cos-particle-restrict.2", this);
  }  
  
  private void checkParticleDerivationOK_All(Particle ancestor) throws SchemaException {
    // cos-particle-restrict
    if ( ancestor.term instanceof Wildcard )
      checkAllChoiceSequenceAny_NSRecurseCheckCardinality(ancestor);
    else if ( ancestor.term instanceof AllModelGroup )
      checkAllAllSequenceSequence_Recurse(ancestor);
    else 
      // Particle can't be restricted from its base particle
      throw new SchemaException("cos-particle-restrict.2", this);
  }  
  
  private void checkParticleDerivationOK_Choice(Particle ancestor) throws SchemaException {
    // cos-particle-restrict
    if ( ancestor.term instanceof Wildcard )
      checkAllChoiceSequenceAny_NSRecurseCheckCardinality(ancestor);
    else if ( ancestor.term instanceof ChoiceModelGroup )
      checkChoiceChoice_RecurseLax(ancestor);
    else 
      // Particle can't be restricted from its base particle
      throw new SchemaException("cos-particle-restrict.2", this);
  }  
  
  private void checkParticleDerivationOK_Sequence(Particle ancestor) throws SchemaException {
    // cos-particle-restrict
    if ( ancestor.term instanceof Wildcard )
      checkAllChoiceSequenceAny_NSRecurseCheckCardinality(ancestor);
    else if ( ancestor.term instanceof AllModelGroup )
      checkSequenceAll_RecurseUnordered(ancestor);
    else if ( ancestor.term instanceof ChoiceModelGroup )
      checkSequenceChoice_MapAndSum(ancestor);
    else if ( ancestor.term instanceof SequenceModelGroup )
      checkAllAllSequenceSequence_Recurse(ancestor);
    else 
      // Particle can't be restricted from its base particle
      throw new SchemaException("cos-particle-restrict.2", this);

  }
  
  private String checkOccurrenceRangeOK(int min, int max, int minBase, int maxBase) {
    if ( min < minBase )
      // Particle's minOccurs must be greater than or equal to its base particle's
      return "range-ok.1";

    if ( max > maxBase )
      // Particle's maxOccurs must be less than or equal to its base particle's
      return "range-ok.2.2";
    
    return null;
  }
  
  private void checkEltElt_NameAndTypeOK(Particle ancestor) throws SchemaException {
    ElementDeclaration elt = (ElementDeclaration)this.getTerm();
    ElementDeclaration eltBase = (ElementDeclaration)ancestor.getTerm();
    if ( elt == eltBase ) return;
    if ( elt.getName().equals(eltBase.getName()) == false )
      // Element declaration's name and namespace must be the same as its base's
      throw new SchemaException("rcase-NameAndTypeOK.1", this);    
    else if ( elt.getNamespace() == null && eltBase.getNamespace() != null )
      throw new SchemaException("rcase-NameAndTypeOK.1", this);
    else if ( elt.getNamespace() != null && !(elt.getNamespace().equals(eltBase.getNamespace())) )
      throw new SchemaException("rcase-NameAndTypeOK.1", this);
    
    if ( !( elt.isNillable() == false || eltBase.isNillable() ) )
      // Element declaration's {nillable} must be false, or its base element's {nillable} must true
      throw new SchemaException("rcase-NameAndTypeOK.2", this);
    
    String errCode = checkOccurrenceRangeOK(this.minOccurs, this.maxOccurs, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-NameAndTypeOK.3", this, new SchemaException(errCode));
    
    if ( !( eltBase.getFixedValue() == null || 
         eltBase.getFixedValue().equals(elt.getFixedValue()) ) )
      // If base declaration's {value contraint} is fixed, this declaration's {value constraint} must be fixed with the same value
      throw new SchemaException("rcase-NameAndTypeOK.4", this);
    
    if ( ( (Integer.MAX_VALUE^elt.getBlock()) & eltBase.getBlock()) != 0 ) {
      // Element declaration's {disallowed substitutions} must be a superset of its base declaration's
      throw new SchemaException("rcase-NameAndTypeOK.5", this);
    }
    
    errCode = elt.getType().checkTypeDerivationOK(eltBase.getType(), eltBase.getType().getFinal(), false);
    if ( errCode != null ) 
      // Element declaration's type must be valid derived from its base declaration's type
      throw new SchemaException("rcase-NameAndTypeOK.7", this, new SchemaException(errCode, elt.getType()));
  }
  
  private void checkEltAny_NSCompat(Particle ancestor) throws SchemaException {
    ElementDeclaration elt = (ElementDeclaration)this.getTerm();
    Wildcard wild = (Wildcard)ancestor.getTerm();
    if ( wild.isAllowed(elt.getNamespace()) == false )
      // Element declaration's {target namespace} must be valid with respect to its base wildcard's {namespace constraint}
      throw new SchemaException("rcase-NSCompat.1", this);
    
    String errCode = checkOccurrenceRangeOK(this.minOccurs, this.maxOccurs, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-NSCompat.2", this, new SchemaException(errCode));
 
  }
  
  private void checkEltAllChoiceSequence_RecursAsIfGroup(Particle ancestor) throws SchemaException {
    // rcase-RecurseAsIfGroup
    try {
      if ( ancestor.getTerm() instanceof SequenceModelGroup ) {
        SequenceModelGroup sequence = new SequenceModelGroup();
        sequence.add(this);
        Particle sequenceParticle = new Particle(sequence);
        sequenceParticle.checkAllAllSequenceSequence_Recurse(ancestor);
      }
      else if ( ancestor.getTerm() instanceof ChoiceModelGroup ) {
        ChoiceModelGroup choice = new ChoiceModelGroup();
        choice.add(this);
        Particle choiceParticle = new Particle(choice);    
        choiceParticle.checkChoiceChoice_RecurseLax(ancestor);
      }
      else { // AllModelGroup
        AllModelGroup all = new AllModelGroup();
        all.add(this);
        Particle allParticle = new Particle(all);    
        allParticle.checkAllAllSequenceSequence_Recurse(ancestor);
      }
    }
    catch ( SchemaException se ) {    
      // second strategy
        Particle eltParticle = new Particle(this.getTerm());
        if ( ancestor.getTerm() instanceof SequenceModelGroup ) {
          SequenceModelGroup sequence = new SequenceModelGroup();
          sequence.add(eltParticle);
          Particle sequenceParticle = new Particle(sequence);
          sequenceParticle.minOccurs = this.minOccurs;
          sequenceParticle.maxOccurs = this.maxOccurs;
          sequenceParticle.minExtent = this.minExtent;
          sequenceParticle.maxExtent = this.maxExtent;
          sequenceParticle.checkAllAllSequenceSequence_Recurse(ancestor);
        }
        else if ( ancestor.getTerm() instanceof ChoiceModelGroup ) {
          ChoiceModelGroup choice = new ChoiceModelGroup();
          choice.add(eltParticle);
          Particle choiceParticle = new Particle(choice);    
          choiceParticle.minOccurs = this.minOccurs;
          choiceParticle.maxOccurs = this.maxOccurs;
          choiceParticle.minExtent = this.minExtent;
          choiceParticle.maxExtent = this.maxExtent;
          choiceParticle.checkChoiceChoice_RecurseLax(ancestor);
        }
        else { // AllModelGroup
          AllModelGroup all = new AllModelGroup();
          all.add(eltParticle);
          Particle allParticle = new Particle(all);    
          allParticle.minOccurs = this.minOccurs;
          allParticle.maxOccurs = this.maxOccurs;
          allParticle.minExtent = this.minExtent;
          allParticle.maxExtent = this.maxExtent;
          allParticle.checkAllAllSequenceSequence_Recurse(ancestor);
        }
     }      
  }
  
  private void checkAnyAny_NSSubset(Particle ancestor) throws SchemaException {
    Wildcard wild = (Wildcard)this.getTerm();
    Wildcard wildBase = (Wildcard)ancestor.getTerm();
    String errCode = checkOccurrenceRangeOK(this.minOccurs, this.maxOccurs, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-NSSubset.1", this, new SchemaException(errCode));

    errCode = wild.validWildcardSubset(wildBase); 
    if ( errCode != null ) 
      // Wildcard's {namespace contraint} must be a valid subset of its base wildcard's {namespace contraint}
      throw new SchemaException("rcase-NSSubset.2", this, new SchemaException(errCode, wild));
  }
  
  private void checkAllChoiceSequenceAny_NSRecurseCheckCardinality(Particle ancestor) throws SchemaException {
    java.util.Iterator it = ((ModelGroup)this.getTerm()).iterator();
    while ( it.hasNext() ) {
      Particle p = (Particle)it.next();
      try {
        p.checkParticleDerivation(ancestor);
      }
      catch ( SchemaException se ) {
        // Particle must be a valid restriction of its base wildcard
        throw new SchemaException("rcase-NSRecurseCheckCardinality.1", p, se);
      }
    }
    
    String errCode = checkOccurrenceRangeOK(this.minExtent, this.maxExtent, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Effective total Occurence range is invalid
      throw new SchemaException("rcase-NSRecurseCheckCardinality.2", this, new SchemaException(errCode));

  }
  
  private void checkAllAllSequenceSequence_Recurse(Particle ancestor) throws SchemaException {
    String errCode = checkOccurrenceRangeOK(this.minOccurs, this.maxOccurs, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-Recurse.1", this, new SchemaException(errCode));
        
    java.util.Iterator it = ((ModelGroup)this.getTerm()).iterator();
    java.util.Iterator itBase = ((ModelGroup)ancestor.getTerm()).iterator();
    Particle particleBase = null;
    while ( it.hasNext() ) {
      Particle p = (Particle)it.next();
      particleBase = p.findAllAllSequenceSequenceMappedParticle(particleBase, itBase, null);
    }
    
    while ( itBase.hasNext() ) {
      Particle pBase = (Particle)itBase.next();
      if ( pBase.isEmptiable() == false )
        // Un-mapped particle in base particles must be emptible
        throw new SchemaException("rcase-Recurse.2.2", this);
    } 
  }
   
  private Particle findAllAllSequenceSequenceMappedParticle(
                    Particle particleBase, 
                    java.util.Iterator itBase,
                    SchemaException sException) 
  throws SchemaException 
  {
    if ( particleBase == null ) {
      if ( itBase.hasNext() == false ) {
        // Particle must be a valid restriction of its associated base particle
        if ( sException == null )
          throw new SchemaException("rcase-Recurse.2.1", this);
        else
          throw sException;
      }
      particleBase = (Particle)itBase.next();
    }
    try {
      checkParticleDerivation(particleBase);
      if ( particleBase.getTerm() instanceof ModelGroup )
        return particleBase;
      else
        return null;
    }
    catch ( SchemaException se ) {        
      if ( particleBase.isEmptiable() == false ) {
        // Un-mapped particle in base particles must be emptible
        throw new SchemaException("rcase-Recurse.2.2", this, se);
      }
      return findAllAllSequenceSequenceMappedParticle(null, itBase, se);
    }
  }  
     
  private void checkChoiceChoice_RecurseLax(Particle ancestor) throws SchemaException {
    // first, verifier if this particle is a sub-particle of ancestor
    java.util.Iterator it = ((ChoiceModelGroup)ancestor.term).iterator();
    while ( it.hasNext() ) {
      if ( this.equals((Particle)it.next()) ) return;
    }
   
    // second, the specification
    String errCode = checkOccurrenceRangeOK(this.minOccurs, this.maxOccurs, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-RecurseLax.1", this, new SchemaException(errCode));

    it = ((ModelGroup)this.getTerm()).iterator();
    java.util.Iterator itBase = ((ModelGroup)ancestor.getTerm()).iterator();
    Particle particleBase = null;
    while ( it.hasNext() ) {
      Particle p = (Particle)it.next();
      particleBase = p.findChoiceChoiceMappedParticle(particleBase, itBase, null);
    }
  }

  private Particle findChoiceChoiceMappedParticle(
                        Particle particleBase, 
                        java.util.Iterator itBase,
                        SchemaException sException) 
  throws SchemaException 
  {
    if ( particleBase == null ) {
      if ( itBase.hasNext() == false ) {
        // Particle must be a valid restriction of its associated base particle
        if ( sException == null )
          throw new SchemaException("rcase-RecurseLax.2", this);
        else
          throw sException;
      }
      particleBase = (Particle)itBase.next();
    }
    try {
      checkParticleDerivation(particleBase);
      if ( particleBase.getTerm() instanceof ModelGroup )
        return particleBase;
      else
        return null;
    }
    catch ( SchemaException se ) {        
      return findChoiceChoiceMappedParticle(null, itBase, se);
    }
  }
  
  private void checkSequenceAll_RecurseUnordered(Particle ancestor) throws SchemaException {
    String errCode = checkOccurrenceRangeOK(this.minOccurs, this.maxOccurs, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-RecurseUnordered.1", this, new SchemaException(errCode));
    
    java.util.Iterator it = ((ModelGroup)this.getTerm()).iterator();
    java.util.HashSet mappedBaseParticles = new java.util.HashSet();
    while ( it.hasNext() ) {
      Particle p = (Particle)it.next();
      java.util.Iterator itBase = ((ModelGroup)ancestor.getTerm()).iterator();
      p.findSequenceAllMappedParticle(itBase, mappedBaseParticles);
    }
    
    if ( ((ModelGroup)ancestor.getTerm()).size() > mappedBaseParticles.size() ) {
      java.util.Iterator itBase = ((ModelGroup)ancestor.getTerm()).iterator();
      while ( itBase.hasNext() ) {
        Particle p = (Particle)itBase.next();
        if ( p.isEmptiable() == false && !mappedBaseParticles.contains(p) ) 
        {
          // Un-mapped particle in base particles must be emptible
          throw new SchemaException("rcase-RecurseUnordered.2.3", this);
        }
      }
    }
  }
  
  private void findSequenceAllMappedParticle(
                      java.util.Iterator itBase, 
                      java.util.Set mappedBaseParticles) 
  throws SchemaException 
  {
    Particle particleBase = (Particle)itBase.next();
    try {
      checkParticleDerivation(particleBase);
      if ( mappedBaseParticles.contains(particleBase) )
        // Base particle is mapped by more than one particle
        throw new SchemaException("rcase-RecurseUnordered.2.1", this);
      else
        mappedBaseParticles.add(particleBase);
    }
    catch ( SchemaException se ) {        
      if ( itBase.hasNext() == false ) {
        // Particle must be a valid restriction of its associated base particle
        throw new SchemaException("rcase-RecurseUnordered.2.2", this, se);
      }
      findSequenceAllMappedParticle(itBase, mappedBaseParticles);
    }
  }
    
  private void checkSequenceChoice_MapAndSum(Particle ancestor) throws SchemaException {    
    // first, verifier if this particle is a sub-particle of ancestor
    java.util.Iterator it = ((ChoiceModelGroup)ancestor.term).iterator();
    while ( it.hasNext() ) {
      if ( this.equals((Particle)it.next()) ) return;
    }
   
    // second, the specification
    int size = ((SequenceModelGroup)this.term).size();
    int min = this.minOccurs * ((SequenceModelGroup)this.term).size();
    int max = ( this.maxOccurs == Integer.MAX_VALUE ) ? this.maxOccurs : this.maxOccurs * size;
    String errCode = checkOccurrenceRangeOK(min, max, ancestor.minOccurs, ancestor.maxOccurs);
    if ( errCode != null )
     // Occurence range is invalid
      throw new SchemaException("rcase-MapAndSum.2", this, new SchemaException(errCode));  
    
    it = ((ModelGroup)this.getTerm()).iterator();
    while ( it.hasNext() ) {
      Particle p = (Particle)it.next();
      java.util.Iterator itBase = ((ModelGroup)ancestor.getTerm()).iterator();
      Particle particleBase = p.findSequenceChoiceMappedParticle(itBase);
    }
  }
  
  private Particle findSequenceChoiceMappedParticle(java.util.Iterator itBase) 
  throws SchemaException 
  {
    Particle particleBase = (Particle)itBase.next();
    try {
      checkParticleDerivation(particleBase);
      return particleBase;
    }
    catch ( SchemaException se ) {        
      if ( itBase.hasNext() == false ) {
        // Particle must be a valid restriction of its associated base particle
        throw new SchemaException("rcase-MapAndSum.1", this, se);
      }
      return findSequenceChoiceMappedParticle(itBase);
    }
  }
  
  // check unique particle
  // return the overLaped particle
  Particle uniqueParticleAttribution() {
    if ( this.term instanceof ChoiceModelGroup || 
         this.term instanceof AllModelGroup ) {
      ModelGroup group = (ModelGroup)this.term;
      for ( int i = 0; i < group.size() - 1; i++ ) {
        Particle p = (Particle)group.get(i);
        Particle overlappedParticle = p.uniqueParticleAttribution();
        if (  overlappedParticle != null ) return overlappedParticle;
        for ( int j = i + 1; j < group.size(); j++ ) {
          Particle nextParticle = (Particle)group.get(j);
          if ( nextParticle.term instanceof ModelGroup || p.term instanceof ModelGroup ) {
                if ( p.overlapUPAC(nextParticle) ) 
                    return p;
          }
        }
      }
    }
    else if ( this.term instanceof SequenceModelGroup ) {
      ModelGroup group = (ModelGroup)this.term;
      for ( int i = 0; i < group.size(); i++ ) {
        Particle p = (Particle)group.get(i);
        Particle overlappedParticle = p.uniqueParticleAttribution();
        if (  overlappedParticle != null ) return overlappedParticle;
        if ( p.minOccurs < p.maxOccurs ) {
          if ( i < group.size() - 1 ) {
            Particle par = (Particle)group.get(i+1);
            if ( par.term instanceof SequenceModelGroup ) {
                ModelGroup gPar = (ModelGroup)par.term;
                Particle firstPar = (Particle)gPar.get(0);
                if ( p.term instanceof SequenceModelGroup ) {
                    ModelGroup gp = (ModelGroup)p.term;
                    Particle lastP = (Particle)gp.get(gp.size()-1);
                    if ( lastP.minOccurs < lastP.maxOccurs && firstPar.overlapUPAC(lastP) ) return lastP;
                }
                else {
                    if ( p.overlapUPAC(firstPar) ) return firstPar;
                }
            }
            else if ( par.term instanceof ModelGroup || p.term instanceof ModelGroup ) {
                if (p.overlapUPAC(par)) return p;
            }
          }
        }
      }
    }
    
    return null;
  }
  
  private boolean overlapUPAC(Particle p) {
    if ( this.term instanceof ElementDeclaration ) {
      if ( p.term instanceof ModelGroup )
        return p.overlapUPAC(this);
      else
        return ((ElementDeclaration)this.term).overlapUPAC(p.term);
    }
    else if ( this.term instanceof Wildcard ) {
      if ( p.term instanceof ModelGroup )
        return p.overlapUPAC(this);
      else
        return ((Wildcard)this.term).overlapUPAC(p.term);
    }
    else if ( this.term instanceof SequenceModelGroup &&  p.term instanceof SequenceModelGroup ) {
        ModelGroup group = (ModelGroup)this.term;
        if ( group.size() > 1 ) {
            Particle first = (Particle)group.get(0);
            if ( p.overlapUPAC(first) ) return true;
            Particle last = (Particle)group.get(group.size()-1);
            if ( last.minOccurs < last.maxOccurs && p.overlapUPAC(last) ) return true;
        }
        else {
            return p.overlapUPAC((Particle)group.get(0));
        }
    }
    else if ( this.term instanceof SequenceModelGroup && !(p.term instanceof ModelGroup) ) {
        ModelGroup group = (ModelGroup)this.term;
        if ( group.size() > 1 ) {
            Particle first = (Particle)group.get(0);
            if ( p.overlapUPAC(first) ) return true;
        }
    }
    else {
      ModelGroup group = (ModelGroup)this.term;
      java.util.Iterator it = group.iterator();
      while ( it.hasNext() ) {
        Particle par = (Particle)it.next();
        if ( par.overlapUPAC(p) )
            return true;
      }
    }
    return false;
  }
  
  private boolean overlapUPAC_SingleSingle(Particle p) {
    if ( this.term instanceof ElementDeclaration )
      return ((ElementDeclaration)this.term).overlapUPAC(p.term);
    else
      return ((Wildcard)this.term).overlapUPAC(p.term);
  }
    
  private boolean overlapUPAC_SingleModelGroup(Particle p) {
    if ( this.minOccurs == this.minOccurs ) return false;

    java.util.List list = p.firstValidElements();
    java.util.Iterator it = list.iterator();
    while ( it.hasNext() ) {
      Particle par = (Particle)it.next();
      if ( this.overlapUPAC_SingleSingle(par) ) return true;
    }
    
    list = p.getFirstWildcards();
    if ( list == null ) return false;
    it = list.iterator();
    while ( it.hasNext() ) {
      Particle par = (Particle)it.next();
      if ( this.overlapUPAC_SingleSingle(par) ) return true;
    }
    
    return false;
  }
    
}




