/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.util.*;

import org.xquark.schema.loader.SchemaComponentRef;
import org.xquark.schema.loader.TypeRef;

/**
 * <p>Abstract<br>
 * Encapsulate an schema-validation based error or warning.</p>
 *
 * <p>This exception is applied to report schema-validation error. During SAX
 *    based schema validation, there will be two different exceptions :
 *    1. SchemaException generated by SAX parser
 *    2. SchemaException generated by schema validation.
 *    But during DOM or DOM-like based schema validation, only SchemaException
 *    will be generated.
 * </p>
 */
public class SchemaException extends Exception {
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName = "$Name:  $";

    static private ResourceBundle errorStrings =
        ResourceBundle.getBundle("org.xquark.schema.resources.errors", java.util.Locale.getDefault());
    /**
     * @serial The error code.
     */
    private String errorCode = null;

    /**
     * @serial The object responsible for the exception
     */
    private Object invalidObject = null;

    /**
     * @serial The embedded exceptions if tunnelling, or null.
     */
    private ArrayList exceptions = null;

    public SchemaException(String errorCode) {
        this(errorCode, (Object) null);
    }

    public SchemaException(String errorCode, Object object) {
        super(errorCode == null ? null : errorStrings.getString(errorCode));
        this.errorCode = errorCode;
        this.invalidObject = object;
    }

    public SchemaException(String errorCode, Object object, Exception exception) {
        this(errorCode, object);
        add(exception);
    }

    public SchemaException(String errorCode, Object object, List exceptions) {
        this(errorCode, object);
        add(exceptions);
    }

    /**
     * Create a new SchemaException wrapping an existing exception.
     *
     * <p>The existing exception will be embedded in the new
     * one, and its message will become the default message for
     * the SchemaException.</p>
     *
     * @param e The exception to be wrapped in a SchemaException.
     */
    public SchemaException(Exception e) {
        this(null, null, e);
    }

    public SchemaException(List exceptions) {
        this(null, null, exceptions);
    }

    /**
     * Return the error code for this exception.
     */
    public String getErrorCode() {
        return this.errorCode;
    }

    public Object getInvalidObject() {
        return invalidObject;
    }

    public void setInvalidObject(Object object) {
        invalidObject = object;
    }

    /**
     * Return a detail message for this exception.
     *
     * <p>If there is an embedded exception, and if the SchemaException
     * has no detail message of its own, this method will return
     * the detail message from the embedded exception.</p>
     *
     * @return The error or warning message.
     */
    public String getMessage() {
        return getMessage(new SchemaException.PrintVisitor());
    }

    private String getMessage(SchemaException.PrintVisitor visitor) {
        StringBuffer result = new StringBuffer();
        if (errorCode != null)
            result.append(errorCode);
        String message = super.getMessage();
        if (message != null) {
            if (result.length() > 0)
                result.append(": ");
            result.append(message);
        }
        if (invalidObject != null && visitor != null) {
            result.append(" (");
            if (invalidObject instanceof SchemaVisitable) {
                result.append(visitor.toString((SchemaVisitable) invalidObject));
            } else {
                result.append(invalidObject.toString());
            }
            result.append(")");
        }
        if (exceptions != null) {
            Iterator it = exceptions.iterator();
            while (it.hasNext()) {
                if (result.length() > 0)
                    result.append('\n');
                result.append(((SchemaException) it.next()).getMessage(visitor));
            }
        }
        return result.toString();
    }

    /**
     * Return the embedded exception, if any.
     *
     * @return The embedded exception, or null if there is none.
     */
    public List getExceptions() {
        return exceptions;
    }

    public void add(Exception exception) {
        if (exception == null)
            return;
        if (exceptions == null)
            exceptions = new ArrayList();
        exceptions.add(exception);
    }

    public void add(List exceptions) {
        if (exceptions == null)
            return;
        if (this.exceptions == null)
            this.exceptions = new ArrayList();
        this.exceptions.addAll(exceptions);
    }

    public static SchemaException computeException(Particle p) {
        SchemaException.ComputeExceptionVisitor visitor = new SchemaException.ComputeExceptionVisitor();
        try {
            p.accept(visitor);
        } catch (Exception e) {
            // Left blank on purpose
        }
        return visitor.getResult();
    }

    public static SchemaException computeMinExtentException(Particle p) {
        SchemaException.ComputeExceptionVisitor visitor = new SchemaException.ComputeMinExtentExceptionVisitor();
        try {
            p.accept(visitor);
        } catch (Exception e) {
            // Left blank on purpose
        }
        return visitor.getResult();
    }

    static class ComputeExceptionVisitor extends DefaultSchemaVisitor {
        private static final String RCSRevision = "$Revision: 1.1 $";
        private static final String RCSName = "$Name:  $";

        protected SchemaException result = null;

        public SchemaException getResult() {
            return result;
        }

        public void visit(Particle p) throws SchemaException {
            ((SchemaVisitable) p.getTerm()).accept(this);
            result.setInvalidObject(p);
        }

        public void visit(ElementDeclaration decl) throws SchemaException {
            result = new SchemaException("cvc-particle.2.3");
        }

        public void visit(Wildcard wildcard) throws SchemaException {
            result = new SchemaException("cvc-particle.1.3");
        }

        public void visit(ModelGroup group) throws SchemaException {
            result = new SchemaException("cvc-particle.3.3");
        }
    }

    static class ComputeMinExtentExceptionVisitor extends SchemaException.ComputeExceptionVisitor {
        private static final String RCSRevision = "$Revision: 1.1 $";
        private static final String RCSName = "$Name:  $";

        public void visit(Particle p) throws SchemaException {
            if (p.getMinExtent() > 0) {
                super.visit(p);
            }
        }

        public void visit(SequenceModelGroup group) throws SchemaException {
            Iterator it = group.iterator();
            while (it.hasNext()) {
                Particle p = (Particle) it.next();
                p.accept(this);
                if (result != null)
                    break;
            }
            result = new SchemaException("cvc-particle.3.3", null, result);
        }

        public void visit(ChoiceModelGroup group) throws SchemaException {
            ArrayList exceptions = new ArrayList();
            Iterator it = group.iterator();
            while (it.hasNext()) {
                Particle p = (Particle) it.next();
                result = null;
                p.accept(this);
                if (result != null)
                    exceptions.add(result);
            }
            result = new SchemaException("cvc-particle.3.3", null, exceptions);
        }

        public void visit(AllModelGroup group) throws SchemaException {
            ArrayList exceptions = new ArrayList();
            Iterator it = group.iterator();
            while (it.hasNext()) {
                Particle p = (Particle) it.next();
                result = null;
                p.accept(this);
                if (result != null)
                    exceptions.add(result);
            }
            result = new SchemaException("cvc-particle.3.3", null, exceptions);
        }
    }

    static class PrintVisitor extends DefaultSchemaVisitor {
        private static final String RCSRevision = "$Revision: 1.1 $";
        private static final String RCSName = "$Name:  $";
        private String result = null;
        private int minOccurs = -1;
        private int maxOccurs = -1;

        String toString(SchemaVisitable invalidObject) {
            result = null;
            minOccurs = -1;
            maxOccurs = -1;
            try {
                invalidObject.accept(this);
            } catch (Exception e) {
                // Should not occur
            }
            return result;
        }

        private void addOccurrences() {
            if (minOccurs != -1) {
                if (minOccurs != 1)
                    result += " minOccurs=\"" + minOccurs + "\"";
                if (maxOccurs == Integer.MAX_VALUE)
                    result += " maxOccurs=\"unbounded\"";
                else if (maxOccurs != 1)
                    result += " maxOccurs=\"" + maxOccurs + "\"";
            }
        }

        public void visit(SchemaComponentRef ref) throws SchemaException {
            result = "<reference name=\"" + ref.getName() + "\">";
        }

        public void visit(TypeRef ref) throws SchemaException {
            result = "<reference name=\"" + ref.getName() + "\">";
        }

        public void visit(ElementDeclaration decl) throws SchemaException {
            result = "<element name=\"" + decl.getName() + "\"";
            addOccurrences();
            result += ">";
        }

        public void visit(AttributeDeclaration decl) throws SchemaException {
            result = "<attribute name=\"" + decl.getName() + "\">";
        }

        public void visit(NotationDeclaration decl) throws SchemaException {
            result = "<notation name=\"" + decl.getName() + "\">";
        }

        public void visit(Wildcard w) throws SchemaException {
            String constraint = null;
            if (w.isAny())
                constraint = null;
            else if (!w.isInclude())
                constraint = "##other";
            else {
                Iterator it = w.getNamespaces().iterator();
                while (it.hasNext()) {
                    String ns = (String) it.next();
                    if (ns == null)
                        ns = "##local";
                    else if (ns.equals(w.getNamespace()))
                        ns = "##targetNamespace";
                    if (constraint == null)
                        constraint = "";
                    else
                        constraint += " ";
                    constraint += ns;
                }
            }
            if (minOccurs == -1) {
                result = "<anyAttribute";
                if (constraint != null)
                    result += " namespace=\"" + constraint + "\"";
                result += ">";
            } else {
                result = "<any";
                if (constraint != null)
                    result += " namespace=\"" + constraint + "\"";
                addOccurrences();
                result += ">";
            }
        }

        public void visit(ComplexType type) throws SchemaException {
            result = "<complexType";
            if (type.getName() != null)
                result += " name=\"" + type.getName() + "\"";
            result += ">";
        }

        public void visit(SimpleType type) throws SchemaException {
            result = "<simpleType";
            if (type.getName() != null)
                result += " name=\"" + type.getName() + "\"";
            result += ">";
        }

        public void visit(Particle p) throws SchemaException {
            minOccurs = p.getMinOccurs();
            maxOccurs = p.getMaxOccurs();
            ((SchemaVisitable) p.getTerm()).accept(this);
            minOccurs = maxOccurs = -1;
        }

        public void visit(SequenceModelGroup group) throws SchemaException {
            result = "<sequence";
            addOccurrences();
            result += ">";
        }

        public void visit(ChoiceModelGroup group) throws SchemaException {
            result = "<choice";
            addOccurrences();
            result += ">";
        }

        public void visit(AllModelGroup group) throws SchemaException {
            result = "<all";
            addOccurrences();
            result += ">";
        }

        public void visit(IdentityConstraint constraint) throws SchemaException {
            switch (constraint.getCategory()) {
                case IdentityConstraint.CATEGORY_UNIQUE :
                    result = "<unique name=\"" + constraint.getName() + "\">";
                    break;
                case IdentityConstraint.CATEGORY_KEY :
                    result = "<key name=\"" + constraint.getName() + "\">";
                    break;
                case IdentityConstraint.CATEGORY_KEYREF :
                    result =
                        "<keyref name=\""
                            + constraint.getName()
                            + "\" refer=\""
                            + constraint.getReferencedKey().getName()
                            + "\">";
                    break;
            }
        }
    }
}
