/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.loader.Loader;

public class SchemaManager extends DefaultSchemaLocator
implements SchemaConstants {
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName     = "$Name:  $";
	
	// Predefined schemas
	private static SchemaManager predefined = null;
	// Schema cache based on URLs, shared by all managers
	private static HashMap loadedURLs = null;
	
	private static SimpleType IDType = null;
	private static SimpleType IDREFType = null;
	private static SimpleType IDREFSType = null;
	private static SimpleType QNameType = null;
	private static SimpleType NCNameType = null;
	private static SimpleType timeType = null;
	private static SimpleType dateType = null;
	private static SimpleType dateTimeType = null;
	private static SimpleType decimalType = null;
	private static SimpleType integerType = null;
	private static SimpleType doubleType = null;
	private static SimpleType floatType = null;
	private static SimpleType booleanType = null;
	private static SimpleType stringType = null;
	private static SimpleType anySimpleType = null;
	
	private static HashMap loadedFiles = new HashMap();
	
	static {
		predefined = new SchemaManager();
		loadedURLs = new HashMap();
		try {
			predefined.loadSimpleDataTypes();
			// then, load schema for schemas
			predefined.loadSchemaForSchemas();
			predefined.loadSchemaForInstance();
		}
		catch ( SAXException e ) {
			predefined = null;
			if ( e instanceof SAXParseException ) {
				SAXParseException ex = (SAXParseException)e;
				System.out.println("Error at line "+ex.getLineNumber()+" in "+ex.getSystemId());
			}
			e.getException().printStackTrace();
			throw new RuntimeException(e.getMessage());
		}
		IDType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "ID");
		IDREFType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "IDREF");
		IDREFSType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "IDREFS");
		QNameType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "QName");
		NCNameType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "NCName");
		timeType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "time");
		dateType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "date");
		dateTimeType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "dateTime");
		decimalType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "decimal");
		integerType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "integer");
		doubleType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "double");
		floatType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "float");
		booleanType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "boolean");
		stringType = (SimpleType)predefined.getType(XMLSCHEMA_URI, "string");
		anySimpleType = new SimpleType(predefined.getSchema(XMLSCHEMA_URI), "anySimpleType", null);
        anySimpleType.setPrimitive(PrimitiveType.createType("anySimpleType"));
	}
	
	// Schema cache based on namespaces
	private HashMap schemas = new HashMap();
	private boolean useCache = true;
	
	public SchemaManager() {
		this(predefined, true);
	}
	
	public SchemaManager(boolean useCache) {
		this(predefined, useCache);
	}
	
	public SchemaManager(SchemaManager parent) {
		this(parent, true);
	}
	
	public SchemaManager(SchemaManager parent, boolean useCache) {
		this.useCache = useCache;
		if (parent == null) parent = predefined;
		if (parent != null) {
			Iterator it = parent.getSchemas().iterator();
			while (it.hasNext()) {
				Schema s = (Schema)it.next();
// start add temporary  
				if (s != null)
// end add temporary   
					schemas.put(s.getNamespace(), s);
			}
		}
	}
	
	private Schema loadSimpleDataTypes() throws SAXException {
		Schema result = null;
		try {
			URL url = getClass().getResource("resources/simpleDatatypes.xsd");
			Loader loader = new Loader(this);
			result = loader.load(url);
			schemas.put(result.getNamespace(), result);
		}
		catch (SAXException ex) {
			throw ex;
		}
		catch ( Exception ex ) {
			throw new SAXException(ex);
		}
		return result;
	}
	
	// history
	//              created
	//
	private Schema loadSchemaForSchemas() throws SAXException {
		Schema result = null;
		try {
			URL url = getClass().getResource("resources/XMLSchema.xsd");
			Loader loader = new Loader(this);
			result = loader.load(url);
			schemas.put(result.getNamespace(), result);
		}
		catch (SAXException ex) {
			throw ex;
		}
		catch ( Exception ex ) {
			throw new SAXException(ex);
		}
		
		return result;
	}
	
	private Schema loadSchemaForInstance() throws SAXException {
		Schema result = null;
		try {
			URL url = getClass().getResource("resources/XMLSchema-instance.xsd");
			Loader loader = new Loader(this);
			result = loader.load(url);
			schemas.put(result.getNamespace(), result);
		}
		catch (SAXException ex) {
			throw ex;
		}
		catch ( Exception ex ) {
			throw new SAXException(ex);
		}
		
		return result;
	}
	
	public SimpleType getIDType() {
		return IDType;
	}
	
	public SimpleType getIDREFType() {
		return IDREFType;
	}
	
	public SimpleType getIDREFSType() {
		return IDREFSType;
	}
	
	public SimpleType getQNameType() {
		return QNameType;
	}
	
	public SimpleType getNCNameType() {
		return NCNameType;
	}
	
	public SimpleType getTimeType() {
		return timeType;
	}
	
	public SimpleType getDateType() {
		return dateType;
	}
	
	public SimpleType getDateTimeType() {
		return dateTimeType;
	}
	
	public SimpleType getIntegerType() {
		return integerType;
	}
	
	public SimpleType getDecimalType() {
		return decimalType;
	}
	
	public SimpleType getDoubleType() {
		return doubleType;
	}

	public SimpleType getFloatType() {
		return floatType;
	}

	public SimpleType getBooleanType() {
		return booleanType;
	}

	public SimpleType getAnySimpleType() {
		return anySimpleType;
	}

	public SimpleType getStringType() {
		return stringType;
	}

	public ElementDeclaration getElementDeclaration(String namespace,
	String localName) {
		Schema schema = getSchema(namespace);
		if (schema == null) return null;
		return schema.getElementDeclaration(localName);
	}
	
	public AttributeDeclaration getAttributeDeclaration(String namespace,
	String localName) {
		Schema schema = getSchema(namespace);
		if (schema == null) return null;
		return schema.getAttributeDeclaration(localName);
	}
	
	public Type getType(String namespace, String localName) {
		Schema schema = getSchema(namespace);
		if (schema == null) return null;
		return schema.getType(localName);
	}
	
	public boolean isRegistered(String namespace) {
		return schemas.containsKey(namespace);
	}
	
	public Schema getSchema(String namespace) {
		return (Schema)schemas.get(namespace);
	}
	
	public Collection getSchemas() {
		return schemas.values();
	}
	
	public Schema removeSchema(String namespace) {
		return (Schema)schemas.remove(namespace);
	}
	
	private void checkNamespace(String ns, Schema schema)
	throws SAXException {
		if ( (ns != null && !ns.equals(schema.getNamespace()))
		|| (ns == null && schema.getNamespace() != null) )
			throw new SAXException("Loaded schema target namespace "
			+ schema.getNamespace()
			+ " does not match expected URI: " + ns);
	}
	
	public Schema loadSchema(String namespace)
	throws SAXException {
		return loadSchema(this, namespace);
	}
	
	public Schema loadSchema(SchemaLocator locator, String namespace)
	throws SAXException {
		// Look into schema cache
		if (isRegistered(namespace)) return getSchema(namespace);
		
		Schema result = null;
		InputSource source = null;
		// Load schema from its locations as strings
		Iterator it1 = locator.getSchemaLocations(namespace);
		if (it1 != null) {
			while (it1.hasNext()) {
				source = locator.resolveLocation(namespace, (String)it1.next());
				if (source != null) result = loadSchema(locator, namespace, source);
				if (result != null) return result;
			}
		}
		
		// Load schema from its URI as string
		source = locator.resolveLocation(namespace, namespace);
		if (source != null) result = loadSchema(locator, namespace, source);
		if (result != null) return result;
		
		// Load schema from its locations as URLs
		Iterator it2 = locator.getSchemaLocations(namespace);
		if (it2 != null) {
			while (it2.hasNext()) {
				try {
					URL url = new URL((String)it2.next());
					source = locator.resolveLocation(namespace, url);
					if (source != null) result = loadSchema(locator, namespace, source);
					if (result != null) return result;
				} catch (MalformedURLException ex) {
					// continue
				}
			}
		}
		
		// Load schema from its URI as URL
/*		try {
			URL url = new URL(namespace);
			source = locator.resolveLocation(namespace, url);
			if (source != null) result = loadSchema(locator, namespace, source);
			if (result != null) return result;
		} catch (MalformedURLException ex) {
			// continue
		}
*/		
		// result is null
		schemas.put(namespace, result);
		return result;
	}
	
	private Schema loadSchema(SchemaLocator locator, String namespace, InputSource source)
	throws SAXException {
		Schema result = null;
		// Look into schema cache
		if (isRegistered(namespace)) return getSchema(namespace);
		
		if (source == null) return null;
		// Look into URL cache
		String systemId = source.getSystemId();
		if (systemId != null && useCache) result = (Schema)loadedURLs.get(systemId);
		
		// look if this file is loading ...
		// Algorithm :
		// 1. if this file is first to load, save it and current time to loadedFiles
		// 2. if this file isn't first to load, get its last loading time.
		// 3. if the difference of current time and last loading time is less than 5 millisecond,
		//    it's a "File Loading-Loop" problem.
		// 4. else, save this file and current time to loadedFiles
		//
		// Why this algorithm:
		// 1. User may load a alread loadded file, it's not a "File Loading-Loop"
		// 2. User's schema file may has a problem, so the first schema file loading doesn't work,
		//    but User may reload corrected schema file later. It's not a "File Loading-Loop"
		//
		// Potential problems :
		// 1. In multithread, two thread try to load the same file nearly the same time
		// 2. In monothread or multithread, the schema file loading time is very long
		//
		// noted  Fev 21, 2002
/*		if ( systemId != null && loadedURLs.get(systemId) == null ) {
			Long currentLoadingTime = new Long(System.currentTimeMillis());
			if ( loadedFiles.get(systemId) != null ) {
				Long lastLoadingTime = (Long)loadedFiles.get(systemId);
				if ( currentLoadingTime.longValue() - lastLoadingTime.longValue() < 500 )
					throw new SAXException("File loading-loop : " + systemId);
			}
			loadedFiles.put(systemId, currentLoadingTime);
		}
*/		
		try {
			if (result == null) {
				Loader loader = new Loader(this, locator);
				result = loader.load(source);
				checkNamespace(namespace, result);
				// Add schema to URL cache
				if (systemId != null && useCache) loadedURLs.put(systemId, result);
			}
		} catch (IOException ex) {
			throw new SAXException(ex);
		} finally {
			// Add schema to namespace cache
			schemas.put(namespace, result);
		}
		return result;
	}
	
	public Schema loadSchema(String namespace, InputSource source)
	throws SAXException {
		return loadSchema(this, namespace, source);
	}
	
	public Schema loadSchema(SchemaLocator locator, String namespace, URL location)
	throws SAXException {
		return loadSchema(locator, namespace, locator.resolveLocation(namespace, location));
	}
	
	public Schema loadSchema(String namespace, URL location)
	throws SAXException {
		return loadSchema(this, namespace, location);
	}
	
	public Schema loadSchema(SchemaLocator locator, String namespace, String location)
	throws SAXException {
		return loadSchema(locator, namespace, locator.resolveLocation(namespace, location));
	}
	
	public Schema loadSchema(String namespace, String location)
	throws SAXException {
		return loadSchema(this, namespace, location);
	}
	
	public Schema loadSchema(SchemaLocator locator, InputSource source)
	throws SAXException {
		Schema result = null;
		if (source == null) return null;
		
		// Look into URL cache
		//String systemId = source.getSystemId(); @@@AM: Disabled for the time being
		//if (systemId != null) result = (Schema)loadedURLs.get(systemId);
		
		if (result == null) {
			Loader loader = new Loader(this, locator);
			try {
				result = loader.load(source);
				// Add schema to URL cache
				//if (systemId != null) loadedURLs.put(systemId, result);
			} catch (IOException ex) {
				throw new SAXException(ex);
			}
		}
		String namespace = result.getNamespace();
		// @@@AM: Commented out to enable schema replacement
		//if (isRegistered(namespace)) {
		//	return getSchema(namespace);
		//}
		// Add schema to namespace cache
		schemas.put(namespace, result);
		return result;
	}
	
	public Schema loadSchema(InputSource source)
	throws SAXException {
		return loadSchema(this, source);
	}
	
	public boolean putSchema(Schema schema)
	throws SAXException {
		if ( isRegistered(schema.getNamespace()) ) return false;
		schemas.put(schema.getNamespace(), schema);
		return true;
	}
	
	
}
