/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.util.ArrayList;

import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.validation.DefaultValidationInfo;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.schema.validation.ValidationInfo;

/**
 * class for Simple Type Definition.
 *
 * <p>This is the implementation of the Simple Type Definition.
 * The following properties are defined:</p>
 * <p> 1. name : optional, null or a NCName, defined in SchemaComponent.</p>
 * <p> 2. target namespace : null or a namespace name, defined in SchemaComponent.</p>
 * <p> 3. base type defintion : a simple type defintion, defined in Type.</p>
 * <p> 4. facets : a set of constraining facets. </p>
 * <p> 5. variety : one of {VARIETY_ATOMIC, VARIETY_LIST, VARIETY_UNION}. </p>
 * <p> 6. primitive type definition : if variety is VARIETY_ATOMIC, a built-in primitive type must be defined. </p>
 * <p> 7. item type definition : if variety is VARIETY_LIST, an item type (which is a simple type) must be defined. </p>
 * <p> 8. member type definitions : if variety is VARIETY_UNION, a non-empty sequence of simple type definitions must be defined. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.14
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 * @see org.xquark.xml.schema.Type
 */
public final class SimpleType extends Type {
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName = "$Name:  $";
    // constants used by SimpleType
    public static final int VARIETY_ATOMIC = 1;
    public static final int VARIETY_LIST = 2;
    public static final int VARIETY_UNION = 3;

    private PrimitiveType primitive = null;
    private java.util.HashMap facets = new java.util.HashMap(10);

    // nVariety : VARIETY_ATOMIC, VARIETY_LIST, VARIETY_UNION
    private int nVariety = VARIETY_ATOMIC;

    private Type itemType = null; // used only when nVariety = VARIETY_LIST
    private ArrayList memberTypes = null; // used only when nVariety = VARIETY_UNION

    /**
     * Create a new SimpleType.
     *
     * @param schema The schema which contains this attribute declaration.
     * @param name The simple type name, null or a NCName.
     * @param baseType The base type definition, null or a simple type.
     */
    public SimpleType(Schema schema, String name, Type baseType) {
        super(schema, name, baseType);
    }

    public void accept(SchemaVisitor visitor) throws SchemaException {
        visitor.visit(this);
    }

    public int getContentType() {
        return TEXT_ONLY;
    }

    public SimpleType getValueType() {
        return this;
    }

    public void setVariety(int nVariety) {
        switch (nVariety) {
            case VARIETY_ATOMIC :
            case VARIETY_LIST :
                this.nVariety = nVariety;
                break;
            case VARIETY_UNION :
                this.nVariety = nVariety;
                if (memberTypes == null)
                    memberTypes = new ArrayList();
                break;
            default :
                // use the default value if nVariety is unfound
                this.nVariety = VARIETY_ATOMIC;
        }
    }

    /**
     * Return the type's variety. One of {VARIETY_ATOMIC, VARIETY_LIST, VARIETY_UNION}. 
     *
     * @return The variety.
     */
    public int getVariety() {
        return nVariety;
    }

    public int getDerivationMethod() {
        return RESTRICTION;
    }

    /**
     * @return null if valid, a string containing the error code otherwise
     */

    public String checkTypeDerivationOK(Type ancestor, int exclusions, boolean useBlock) {
        if (SIMPLE_UR_TYPE.equals(ancestor.getName()))
            return null;
        if (!ancestor.isSimpleType()) {
            if (ANY_TYPE.equals(ancestor.getName()))
                return null;
            else
                return "cos-st-derived-ok.2.2";
        }
        SimpleType simpleBase = (SimpleType) ancestor;
        if (this == simpleBase)
            return null;
        java.util.Collection memberTypes = simpleBase.getMemberTypes();
        if (memberTypes != null) {
            java.util.Iterator it = memberTypes.iterator();
            while (it.hasNext()) {
                if (it.next() == this)
                    return null;
            }
        }
        if ((getDerivationMethod() & exclusions) != 0)
            return "cos-st-derived-ok.2.1";
        Type parent = getBaseType();
        if (parent == null)
            return "cos-st-derived-ok.2.2";
        return parent.checkTypeDerivationOK(ancestor, exclusions, useBlock);
    }

    public String normalizedValue(String value) {
        return primitive.normalizedValue(value);
    }

    public StringBuffer normalizedValue(StringBuffer value) {
        return primitive.normalizedValue(value);
    }

    public StringBuffer normalizedValue(StringBuffer value, StringBuffer out) {
        return primitive.normalizedValue(value, out);
    }

    public String toXMLString(Object value, ValidationContextProvider context) {
        return primitive.toXMLString(value, context);
    }

    public ValidationInfo validate(String value, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
        try {
        	if (nVariety == VARIETY_UNION) return validateUnion(value, normalize, context);
        	else {
        		DefaultValidationInfo info = new DefaultValidationInfo();
        		info.setValidationType(this);
        		if (normalize) value = primitive.normalizedValue(value);
        		info.setNormalizedValue(value);
        		info.setActualValue(primitive.convert(value, false, context));
        		return info;
        	}
        } catch (SchemaException ex) {
            throw new SchemaException("cvc-simple-type", this, ex);
        }
    }

    public ValidationInfo validate(StringBuffer value, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
			try {
				if (nVariety == VARIETY_UNION) return validateUnion(value, normalize, context);
				else {
					DefaultValidationInfo info = new DefaultValidationInfo();
					info.setValidationType(this);
					if (normalize) value = primitive.normalizedValue(value);
					info.setNormalizedValueBuffer(value);
					info.setActualValue(primitive.convert(value, false, context));
					return info;
				}
			} catch (SchemaException ex) {
				throw new SchemaException("cvc-simple-type", this, ex);
			}
    }

	private ValidationInfo validateUnion(String value, boolean normalize, ValidationContextProvider context) throws SchemaException {
		ValidationInfo info = null;
		ArrayList exceptions = null;
		primitive.checkPattern(value);
		for (int i = 0; i < memberTypes.size(); i++) {
			try {
				SimpleType member = (SimpleType) memberTypes.get(i);
				info = member.validate(value, normalize, context);
				break;
			} catch (SchemaException ex) {
				if (exceptions == null) exceptions = new ArrayList();
				exceptions.add(ex);
			}
		}
		if (info != null) {
			primitive.checkFacets(info.getActualValue());
			return info;
		} else {
			throw new SchemaException("cvc-datatype-valid.1.2.3", this, exceptions);
		}
	}
	
	private ValidationInfo validateUnion(StringBuffer value, boolean normalize, ValidationContextProvider context) throws SchemaException {
		ValidationInfo info = validateUnion(value.toString(), normalize, context);
		value.setLength(0);
		value.append(info.getNormalizedValue());
		return info;
	}
	
    public Object convert(String value, boolean normalize, ValidationContextProvider context) throws SchemaException {
        try {
            return primitive.convert(value, normalize, context);
        } catch (SchemaException ex) {
            throw new SchemaException("cvc-simple-type", this, ex);
        }
    }

    public Object convert(StringBuffer value, boolean normalize, ValidationContextProvider context)
        throws SchemaException {
        try {
            return primitive.convert(value, normalize, context);
        } catch (SchemaException ex) {
            throw new SchemaException("cvc-simple-type", this, ex);
        }
    }
    
    public Object convertObject(Object data) throws SchemaException {
    	return primitive.convertObject(data);
    }

    public String toCanonicalForm(String value, ValidationContextProvider context) throws SchemaException {
    	ValidationInfo info = validate(value, true, context);
        SimpleType realType = info.getValidationType();
        Object obj = info.getActualValue();
        return realType.toXMLString(obj, context);
    }

    public String validateDefault(String value, ValidationContextProvider context) throws SchemaException {
        try {
            return toCanonicalForm(value, context);
        } catch (SchemaException ex) {
            throw new SchemaException("cos-valid-default.1", this, ex);
        }
    }

    /**
     * Return the type's built-in primitive type;
     *
     * @return The PrimitiveType.
     */
    public PrimitiveType getPrimitive() {
        return primitive;
    }

    /**
     * Look up all contraining facets.
     * example: type is an instance of SimpleType
     *   Collection facets = type.getAllFacets();
     *   Iterator it = facets.iterator();
     *   // find facet name and vale
     *   while ( it.hasNext() ) {
     *     Facet fct = (Facet)it.next();
     *     String name = fct.getName();
     *     String value = fct.getValue();
     *     Boolean fixed = fct.getFixed(); // if the facet value is fixed, return true
     *     HashSet enums = fct.getEnumFacets(); // get all enumeration values
     *   }
     *
     * @return A collection view of contraining facets.
     */
    public java.util.Collection getAllFacets() {
        return facets.values();
    }

    public java.util.HashMap getFacets() {
        return facets;
    }

    public void setFacet(Facet facet) throws SchemaException {
        String facetName = facet.getName();

        // facet : enumeration
        if (ENUMERATION_TAG.equals(facetName)) {
            Facet enumFacet = (Facet) facets.get(facetName);
            if (enumFacet == null) {
                enumFacet = facet;
                facets.put(facetName, enumFacet);
            }
            enumFacet.addFacet(facet.getValue());
        }

        // facet : pattern
        else if (PATTERN_TAG.equals(facetName)) {
            Facet pattternFacet = (Facet) facets.get(facetName);
            if (pattternFacet == null) {
                pattternFacet = facet;
                facets.put(facetName, pattternFacet);
            } else {
                String pat = pattternFacet.getValue();
                String facetValue = facet.getValue();
                if (pat != null)
                    facetValue = pat + "|" + facetValue;
                pattternFacet.setValue(facetValue);
            }
        }

        // other facet
        else {
            if (facets.put(facetName, facet) != null) {
                //      "Facet "+facetName+" cannot appear twice in a simple type restriction";
                throw new SchemaException("src-single-facet-value", this);
            }
        }

    }

    public void setPrimitive(PrimitiveType primitive) {
        this.primitive = primitive;
    }

    /**
     * If variety is VARIETY_LIST, return the type's item type;
     * else, return null.
     *
     * @return The item type, a simple type definition.
     */
    public Type getItemType() {
        return itemType;
    }

    // used only when nVariety = VARIETY_LIST
    public void setItemType(Type type) {
        itemType = type;
    }

    /**
     * If variety is VARIETY_UNION, return a non-empty sequence of simple type defintions;
     * else, return null.
     *
     * @return The member types, an ArrayList which contains simple type definition.
     */
    public ArrayList getMemberTypes() {
        return memberTypes;
    }

    public SimpleType getMemberType(int index) {
        if (index < memberTypes.size())
            return (SimpleType) memberTypes.get(index);
        else
            return null;
    }

    // used only when nVariety = VARIETY_UNION
    public void addMemberType(Type type) {
        if (memberTypes == null)
            memberTypes = new ArrayList();
        memberTypes.add(type);
    }

    // used only when nVariety = VARIETY_UNION
    public void addMemberTypes(java.util.List memberTypes) {
        if (this.memberTypes == null)
            this.memberTypes = new ArrayList();
        this.memberTypes.addAll(memberTypes);
    }

    // TYPING STUFF
    public boolean isSimpleType() {
        return true;
    }

}
