/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import org.xquark.schema.validation.ContentIterator;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.schema.validation.ValidationInfo;

/**
 * Abstract class for SimpleType and ComplexType.
 *
 * <p>This is the implementation of the common propertoes of Simple Type
 * and Complex Type. The following properties are defined :</p>
 * <p> 1. base type defintion : a simple type or a complex type defintion.
 * if a type is an instance of Simple Type, this base type must be a simple type</p>
 * <p> 2. abstract : a boolean, only a complex type can be abstract </p>
 *
 * <p>See XML Schema Part 1: Structures 3.4, 3.14
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 * @see org.xquark.xml.schema.SimpleType
 * @see org.xquark.xml.schema.ComplexType
 */
public abstract class Type extends SchemaComponent {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
  
  private Type baseType;
  private int fin = 0;
  
  /**
   * Create a new Type.
   *
   * @param schema The schema which contains this type.
   * @param name The type name, optional, a NCName as defined by [XML-Namespace].
   * @param baseType The base type definition, either a simple type definition or 
   * a complex type definition. if a type is an instance of Simple Type, 
   * this base type must be a simple type.
   */
  protected Type(Schema schema, String name, Type baseType) {
    super(schema, name);
    this.baseType = baseType;
  }

  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }

  /**
   * Return the type's base type definition, if any
   *
   * @return The base type definition, or null if there is none.
   */
  public Type getBaseType() {
    return baseType;
  }
   
  /**
   * This is a final type, which can't have substitions.
   *
   * @return The type final definition, a subset of {extension, restriction}..
   */
  public int getFinal() {
    return fin;
  }

  public boolean isAbstract() {
    return false;
  }

  public int getBlock() {
    return 0;
  }
  
  public int getDerivationMethod() {
    return RESTRICTION;
  }

  public boolean isDerivedFrom(Type ancestor) {
    if (ancestor == this) return true;
    if (baseType == null) return false;
    else return baseType.isDerivedFrom(ancestor);
  }

  /**
   * @return null if valid, a string containing the error code otherwise
   */

  public String checkTypeDerivationOK(Type ancestor, int exclusions, boolean useBlock) {
    return null;
  }

  public abstract int getContentType();

  public abstract SimpleType getValueType();

  public ElementDeclaration getElementDeclaration(String namespace, String name) {
    return null;
  }

  public java.util.Collection getElementDeclarations() {
    return null;
  }
  
  public AttributeDeclaration getAttributeDeclaration(String namespace, String name) {
    return null;
  }  
  
  public java.util.Collection getAttributeDeclarations() {
    return null;
  } 
  
  public Wildcard getAttributeWildcard() {
    return null;
  }

  public ContentIterator contentIterator() {
    return null;
  }

  public ContentIterator childIterator() { 
    return null;
  }

  public String normalizedValue(String value) {
    return value;
  }
 
  public StringBuffer normalizedValue(StringBuffer value) {
    return value;
  }

  public StringBuffer normalizedValue(StringBuffer value, StringBuffer out) {
    if (out == value) return value;
    if (out == null) out = new StringBuffer();
    out.append(value.toString());
    return out;
  }
  
  public String toXMLString(Object value, ValidationContextProvider context) {
    return value.toString();
  }

  public String toXMLString(Object value) {
    return toXMLString(value, null);
  }

  public abstract ValidationInfo validate(String value, boolean normalize, ValidationContextProvider context)
    throws SchemaException;
	
  public abstract ValidationInfo validate(StringBuffer value, boolean normalize, ValidationContextProvider context)
    throws SchemaException;
	
  public Object convert(String value, boolean normalize, ValidationContextProvider context) 
    throws SchemaException
  {
    return value;
  }
   
  public Object convert(StringBuffer value, boolean normalize, ValidationContextProvider context)
    throws SchemaException
  {
    return value;
  }
	
  public ValidationInfo validate(String value) throws SchemaException {
    return validate(value, true, null);
  }	
	
  public ValidationInfo validate(StringBuffer value) throws SchemaException {
    return validate(value, true, null);
  }
	
  public Object convert(String value) throws SchemaException {
    return convert(value, true, null);
  }
   
  public Object convert(StringBuffer value) throws SchemaException {
    return convert(value, true, null);
  }
	
  public String validateDefault(String value, ValidationContextProvider context) throws SchemaException {
    return value;
  }
  
  public String toCanonicalForm(String value, ValidationContextProvider context) throws SchemaException {
    return value;
  }
  
  public String toCanonicalForm(String value) throws SchemaException {
    return toCanonicalForm(value, null);
  }
  
  public void setBaseType(Type type) {
    baseType = type;
  }

  public void setFinal(int fin) {
    this.fin = fin;
  }
  
  // TYPING STUFF
  public boolean isSimpleType() { return false; }
  
}











