/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

/**
 * class for Wildcard.
 *
 * <p>This is the implementation of the Wildcard.
 * The following properties are defined:</p>
 * <p> 1. any : a boolean.</p>
 * <p> 2. namespaces : a set of namespaces.</p>
 * <p> 3. process contents : one of {SKIP, LAX, STRICT}.</p>
 *
 * <p>See XML Schema Part 1: Structures 3.10
 * W3C Recommendation 2 May 2001</p>
 *
 */
public final class Wildcard extends SchemaComponent {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";

  private boolean any;
  private boolean exclude = false;
  private java.util.HashSet namespaces = new java.util.HashSet();
  private int processContents = STRICT;
  /**
   * Create a new Wildcard.
   * no parameter, in the case, any = false, by default
   * @param
   */
  Wildcard(Schema schema) {
    this(schema, false);
  }

  /**
   * Create a new Wildcard.
   *
   * @param any The any, a boolean.
   */
  public Wildcard(Schema schema, boolean any) {
    super(schema, null);
    this.any = any;
  }
  
  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }

  public boolean isAllowed(String namespace) {
    if (any) return true;
    if (exclude) 
      return (namespace != null && !namespaces.contains(namespace));
    return namespaces.contains(namespace);
  }

  public boolean isUnqualifiedAllowed() {
    return isAllowed(null);
  }

  public String checkNamespaceAllowed(String namespace) {
    if (any) return null;
    if (exclude) {
      if (namespaces.contains(namespace))
        return "cvc-wildcard-namespace.2.2";
      if (namespace == null)
        return "cvc-wildcard-namespace.2.3";
      return null;
    }
    if (!namespaces.contains(namespace))
      return "cvc-wildcard-namespace.3";
    return null;
  }
  
  public String validWildcardSubset(Wildcard ancestor) {
    if ( ancestor.any ) return null;
    if ( this.any ) return "cos-ns-subset";
    if ( this.exclude ) {
      if ( ancestor.exclude == false ) {
        // Parent wildcard constraint must ba a pair of not and a namespace
        return "cos-ns-subset.2.2";
      }
      java.util.Iterator it =  this.namespaces.iterator();
      while ( it.hasNext() ) {
        Object ns = it.next();
        if ( ancestor.namespaces.contains(ns) == false ) {
          // Parent wildcard constraint must be a pair of not and the same namespace
          return "cos-ns-subset.2";
        }
      }
    }
    else {
      java.util.Iterator it =  this.namespaces.iterator();
      while ( it.hasNext() ) {
        Object ns = it.next();
        if ( ancestor.isAllowed((String)ns) == false ) {
          // All namespaces in namespace list must be allowed in parent wildcard.
          return "cos-ns-subset.3";
        }
      }
    }
    return null;
  }

  /**
   * Look up the process contents.
   *
   * @return The process contents, one of {SKIP, LAX, STRICT}.
   */
  public int getProcessContents() {
    return processContents;
  }
  
  /**
   * Look up the any.
   *
   * @return The any, a boolean.
   */
  boolean isAny() { 
    return any;
  }

  boolean isInclude() {
    return !exclude;
  }

  /**
   * Look up the namespaces.
   *
   * @return A set of namespaces.
   */
  java.util.Set getNamespaces() {
    return namespaces;
  }
  
  public void add(String namespace) {
    any = false;
    namespaces.add(namespace);
  }

  public void exclude(String namespace) {
    any = false;
    exclude = true;
    namespaces.add(namespace);
  }

  public void setProcessContents(int processContents) {
    this.processContents = processContents;
  }
  
  boolean equals(Wildcard wild) {
    if ( this == wild ) return true;
    if ( wild == null ) return false;
    if ( this.any != wild.any ) return false;
    if ( this.exclude != wild.exclude ) return false;
    if ( this.processContents != wild.processContents ) return false;
    if ( this.namespaces.size() != wild.namespaces.size() ) return false;
    java.util.Iterator it1 = this.namespaces.iterator();
    java.util.Iterator it2 = wild.namespaces.iterator();
    while ( it1.hasNext() ) {
        Object ns1 = it1.next();
        Object ns2 = it2.next();
        if ( ns1 != null ) {
            if ( ns1.equals(ns2) == false ) return false;
        }
        else if ( ns2 != null )
            return false;
    }
    return true;
  }
  
  boolean overlapUPAC(Object obj) {
    if ( obj == null ) return false;
    if ( obj instanceof ElementDeclaration )
      return overlapUPAC((ElementDeclaration)obj);
    else if ( obj instanceof Wildcard ) 
       return overlapUPAC((Wildcard)obj);
    else
      return false;
 }
  
  boolean overlapUPAC(ElementDeclaration decl) {
    if ( decl == null ) return false;
    else return decl.overlapUPAC(this);
  }
   
  boolean overlapUPAC(Wildcard wild) {
    if ( wild == null ) return false;
    if ( this == wild ) return true;
    
    try {
      Wildcard intersectionWild = this.wildcardIntersection(wild);
      if ( intersectionWild.getNamespaces().size() > 0 ) return true;
    }
    catch ( SchemaException se ) {
      // it's empty
    }
    
    return false;
  }
  
  Wildcard wildcardIntersection(Wildcard wild)
    throws SchemaException
  {
    if ( this.isAny() ) return wild;
    if (wild == null || wild.isAny()) return this;
    
    Wildcard result = new Wildcard(schema);
    result.setProcessContents(this.getProcessContents());
    if (this.isInclude()) {
      java.util.Iterator it = this.getNamespaces().iterator();
      while (it.hasNext()) {
        String ns = (String)it.next();
        if (wild.isAllowed(ns)) result.add(ns);
      }
    } else if (wild.isInclude()) {
      java.util.Iterator it = wild.getNamespaces().iterator();
      while (it.hasNext()) {
        String ns = (String)it.next();
        if (this.isAllowed(ns)) result.add(ns);
      }
    } else {
//    "Attribute wildcards cannot be combined";
      throw new SchemaException("cos-aw-intersect.5", this);
    }
    return result;
  }
    
  Wildcard wildcardUnion(Wildcard wild)
    throws SchemaException
  {
    if (wild == null) return this;
    Wildcard result = new Wildcard(schema, true);
    result.setProcessContents(this.getProcessContents());
    if (this.isAny() || wild.isAny()) {
      // result is any
    } else if (this.isInclude()) {
      if (wild.isInclude()) { 
        java.util.Iterator it = this.getNamespaces().iterator();
        while (it.hasNext()) {
          String ns = (String)it.next();
          result.add(ns);
        }
        it = wild.getNamespaces().iterator();
        while (it.hasNext()) {
          String ns = (String)it.next();
          if (!this.isAllowed(ns)) result.add(ns);
        }
      } else {
        java.util.Iterator it = wild.getNamespaces().iterator();
        String ns = (String)it.next();
        if (this.isAllowed(ns)) {
          // any
        } else {
          result.exclude(ns);
        }
      }
    } else if (wild.isInclude()) {
      java.util.Iterator it = this.getNamespaces().iterator();
      String ns = (String)it.next();
      if (wild.isAllowed(ns)) {
        // any
      } else {
        result.exclude(ns);
      }
    } else {
//    "Attribute wildcards cannot be combined";
      throw new SchemaException("cos-aw-union.4", this);
    }
    return result;
  }
  
}
