/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.datatypes;

import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;


abstract class ComparableType extends EnumerableType {
	
    class ComparableConstraints implements Cloneable {
        private static final String RCSRevision = "$Revision: 1.2 $";
        private static final String RCSName = "$Name:  $";
        private int minType = -1;
        private Comparable min = null;
        private int maxType = -1;
        private Comparable max = null;
    
        public Object clone() throws CloneNotSupportedException {
            return super.clone();
        }

		private boolean checkMin(Comparable value) {
				return minType == -1 || value.compareTo(min) >= minType;   
		}
		
		private boolean checkMax(Comparable value) {
				return maxType == -1 || max.compareTo(value) >= maxType;   
		}
		
        void checkParentMinExclusive(Comparable value) throws SchemaException {
            if (!checkMin(value)) {
                if (minType == 1) {
                    throw new SchemaException("minExclusive-valid-restriction.1");
                } else {
                    throw new SchemaException("minExclusive-valid-restriction.3");
                }
            }
        }
    
        void checkParentMinInclusive(Comparable value) throws SchemaException {
			if (!checkMin(value)) {
                if (minType == 1) {
                    throw new SchemaException("minInclusive-valid-restriction.1");
                } else {
                    throw new SchemaException("minInclusive-valid-restriction.3");
                }
            }
        }
    
        void checkParentMaxExclusive(Comparable value) throws SchemaException {
            if (!checkMax(value)) {
                if (maxType == 1) {
                    throw new SchemaException("maxExclusive-valid-restriction.2");
                } else {
                    throw new SchemaException("maxExclusive-valid-restriction.1");
                }
            }
    
        }
    
        void checkParentMaxInclusive(Comparable value) throws SchemaException {
            if (!checkMax(value)) {
                if (maxType == 1) {
                    throw new SchemaException("maxInclusive-valid-restriction.2");
                } else {
                    throw new SchemaException("maxInclusive-valid-restriction.1");
                }
            }
        }
    
        void setMinExclusive(Comparable value, boolean fixed) throws SchemaException {
            if ((isFixed(PrimitiveType.MIN_EXCLUSIVE_FACET) && !value.equals(min)) || isFixed(PrimitiveType.MIN_INCLUSIVE_FACET)) {
                throw new SchemaException("unknown-error-code.1");
            }
    
            if (!checkMax(value)) {
                if (maxType == 1) {
                    throw new SchemaException("minExclusive-valid-restriction.4");
                } else {
                    throw new SchemaException("minExclusive-valid-restriction.2");
                }
            }
    
            minType = 1;
            min = value;
            setFixed(PrimitiveType.MIN_EXCLUSIVE_FACET, fixed);
        }
    
        void setMinInclusive(Comparable value, boolean fixed) throws SchemaException {
            if ((isFixed(PrimitiveType.MIN_INCLUSIVE_FACET) && !value.equals(min)) || isFixed(PrimitiveType.MIN_EXCLUSIVE_FACET)) {
                throw new SchemaException("unknown-error-code.1");
            }
    
            if (!checkMax(value)) {
                if (maxType == 1) {
                    throw new SchemaException("minInclusive-valid-restriction.2");
                } else {
                    throw new SchemaException("minInclusive-valid-restriction.4");
                }
            }
    
            minType = 0;
            min = value;
            setFixed(PrimitiveType.MIN_INCLUSIVE_FACET, fixed);
        }
    
        void setMaxExclusive(Comparable value, boolean fixed) throws SchemaException {
            if ((isFixed(PrimitiveType.MAX_EXCLUSIVE_FACET) && !value.equals(max)) || isFixed(PrimitiveType.MAX_INCLUSIVE_FACET)) {
                throw new SchemaException("unknown-error-code.1");
            }
    
            if (!checkMin(value)) {
                if (minType == 1) {
                    throw new SchemaException("maxExclusive-valid-restriction.4");
                } else {
                    throw new SchemaException("maxExclusive-valid-restriction.3");
                }
            }
    
            maxType = 1;
            max = value;
            setFixed(PrimitiveType.MAX_EXCLUSIVE_FACET, fixed);
        }
    
        void setMaxInclusive(Comparable value, boolean fixed) throws SchemaException {
            if ((isFixed(PrimitiveType.MAX_INCLUSIVE_FACET) && !value.equals(max)) || isFixed(PrimitiveType.MAX_EXCLUSIVE_FACET)) {
                throw new SchemaException("unknown-error-code.1");
            }
    
            if (!checkMin(value)) {
                if (minType == 1) {
                    throw new SchemaException("maxInclusive-valid-restriction.4");
                } else {
                    throw new SchemaException("maxInclusive-valid-restriction.3");
                }
            }
    
            maxType = 0;
            max = value;
            setFixed(PrimitiveType.MAX_INCLUSIVE_FACET, fixed);
        }
    
        void checkConstraints(Comparable value) throws SchemaException {
            if (!checkMin(value)) {
                if (minType == 1)
                    invalidFacet("cvc-minExclusive-valid", min, value.toString());
                else
                    invalidFacet("cvc-minInclusive-valid", min, value.toString());
            }
            if (!checkMax(value)) {
                if (maxType == 1)
                    invalidFacet("cvc-maxExclusive-valid", max, value.toString());
                else
                    invalidFacet("cvc-maxInclusive-valid", max, value.toString());
            }
        }
    
        void setMinValue(Comparable value) {
            min = value;
        }
    
        void setMaxValue(Comparable value) {
            max = value;
        }
    
		void setMinValue(Comparable value, boolean inclusive) {
			min = value;
			minType = inclusive ? 0 : 1;
		}
    
		void setMaxValue(Comparable value, boolean inclusive) {
			max = value;
			maxType = inclusive ? 0 : 1;
		}
    
        Comparable getMinValue() {
            return min;
        }
    
        Comparable getMaxValue() {
            return max;
        }
    
        boolean isMinInclusive() {
            return minType == 0;
        }
    
        boolean isMaxInclusive() {
            return maxType == 0;
        }
    
    }
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName = "$Name:  $";

    private ComparableConstraints constraints = null;

    ComparableType(String name, int type) {
        super(name, type);
    }

    public Object clone() throws CloneNotSupportedException {
        ComparableType result = (ComparableType) super.clone();
        if (constraints != null)
            result.constraints = (ComparableConstraints) constraints.clone();
        return result;
    }

    private void initConstraints() {
        if (constraints == null)
            constraints = new ComparableConstraints();
    }

    protected void setMinExclusive(String value, boolean fixed) throws SchemaException {
        initConstraints();
        Comparable min = toComparable(super.normalizedValue(value));
        constraints.checkParentMinExclusive(min);
        constraints.setMinExclusive(min, fixed);
    }

    protected void setMinInclusive(String value, boolean fixed) throws SchemaException {
        initConstraints();
		Comparable min = toComparable(super.normalizedValue(value));
		constraints.checkParentMinInclusive(min);
		constraints.setMinInclusive(min, fixed);
    }

    protected void setMaxExclusive(String value, boolean fixed) throws SchemaException {
        initConstraints();
		Comparable max = toComparable(super.normalizedValue(value));
		constraints.checkParentMaxExclusive(max);
		constraints.setMaxExclusive(max, fixed);
    }

    protected void setMaxInclusive(String value, boolean fixed) throws SchemaException {
        initConstraints();
		Comparable max = toComparable(super.normalizedValue(value));
		constraints.checkParentMaxInclusive(max);
		constraints.setMaxInclusive(max, fixed);
    }

    protected void setMaxValue(Comparable value) {
        constraints.setMaxValue(value);
    }

    protected void setMinValue(Comparable value) {
        constraints.setMinValue(value);
    }


    protected Comparable checkComparable(Comparable comp) throws SchemaException {
        if (constraints != null)
            constraints.checkConstraints(comp);
        return comp;
    }

    protected abstract Comparable toComparable(String value) throws SchemaException;

    public void checkFacets(Object valueSpace) throws SchemaException {
        super.checkFacets(valueSpace);
        checkComparable((Comparable) valueSpace);
    }

    protected Object toValueSpace(String value, ValidationContextProvider context) throws SchemaException {
        return toComparable(value);
    }

    public Comparable getMinValue() {
        if (constraints == null)
            return null;
        return constraints.getMinValue();
    }

    public Comparable getMaxValue() {
        if (constraints == null)
            return null;
        return constraints.getMaxValue();
    }

    public boolean isMinInclusive() {
        if (constraints == null)
            return false;
        return constraints.isMinInclusive();
    }

    public boolean isMaxInclusive() {
        if (constraints == null)
            return false;
        return constraints.isMaxInclusive();
    }


}
