/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.datatypes;

import java.math.BigDecimal;

import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;


class DecimalType extends ComparableType {
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName = "$Name:  $";
    
    private static final BigDecimal MAX_LONG = new BigDecimal(Long.MAX_VALUE);
	private static final BigDecimal MIN_LONG = new BigDecimal(Long.MIN_VALUE);

    private int totalDigits = -1;
    private int fractionDigits = -1;
    private ComparableConstraints longConstraints = null;

    DecimalType() {
        super("decimal", PrimitiveType.DECIMAL);
    }

	public Object toValidType(Object data) {
		if ((data instanceof Long) || (data instanceof BigDecimal)) {
			return data;
		} else if (data instanceof Number) {
			return new Long(((Number)data).longValue());
		} else {
			return (Long)data;
		}
	}
	

	public Object clone() throws CloneNotSupportedException {
		DecimalType result = (DecimalType) super.clone();
		initLongConstraints();
		return result;
	}

    protected void setTotalDigits(String value, boolean fixed) throws SchemaException {
        int val;
        try {
            val = Integer.parseInt(value.trim());
        } catch (NumberFormatException nfe) {
            //      "illegal facet value";
            throw new SchemaException("unknown-error-code.2");
        }
        if (val <= 0) {
            //      "illegal facet value";
            throw new SchemaException("unknown-error-code.2");
        }
        if ((totalDigits != -1 && val > totalDigits) || (super.isFixed(PrimitiveType.TOTALDIGITS_FACET) && val != totalDigits)) {
            //      "totalDigits is " + val + ", but parent totalDigits is " + totalDigits;
            throw new SchemaException("totalDigits-valid-restriction");
        }
        if (fractionDigits != -1 && fractionDigits > val) {
            //      "fractionDigits is " + fractionDigits + ", but totalDigits is " + val;
            throw new SchemaException("fractionDigits-totalDigits");
        }
        totalDigits = val;
        super.setFixed(PrimitiveType.TOTALDIGITS_FACET, fixed);
    }

    protected void setFractionDigits(String value, boolean fixed) throws SchemaException {
        int val;
        try {
            val = Integer.parseInt(value.trim());
        } catch (NumberFormatException nfe) {
            //      "illegal facet value";
            throw new SchemaException("unknown-error-code.2");
        }

        if (val < 0) {
            //      "illegal facet value";
            throw new SchemaException("unknown-error-code.2");
        }

        if ((fractionDigits != -1 && val > fractionDigits)
            || (super.isFixed(PrimitiveType.FRACTIONDIGITS_FACET) && val != fractionDigits)) {
            //      "fractionDigits is " + val + ", but parent fractionDigits is " + totalDigits;
            throw new SchemaException("fractionDigits-valid-restriction.1");
        }
        if (totalDigits != -1 && val > totalDigits) {
            //      "It is an error for fractionDigits to be greater than totalDigits";
            throw new SchemaException("fractionDigits-totalDigits");
        }
        fractionDigits = val;
        super.setFixed(PrimitiveType.FRACTIONDIGITS_FACET, fixed);

        initLongConstraints();
    }

	private void initLongConstraints() {
		if (fractionDigits == 0) {
			BigDecimal min = (BigDecimal)super.getMinValue();
			BigDecimal max = (BigDecimal)super.getMaxValue();
			if (min != null && min.compareTo(MIN_LONG) >= 0 && max != null && max.compareTo(MAX_LONG) <= 0) {
				longConstraints = new ComparableConstraints();
				longConstraints.setMinValue(new Long(min.longValue()), super.isMinInclusive());
				longConstraints.setMaxValue(new Long(max.longValue()), super.isMaxInclusive());
			}
		}
	}

	protected void setMinExclusive(String value, boolean fixed) throws SchemaException {
		longConstraints = null;
		super.setMinExclusive(value, fixed);
		initLongConstraints();
	}

	protected void setMinInclusive(String value, boolean fixed) throws SchemaException {
		longConstraints = null;
		super.setMinInclusive(value, fixed);
		initLongConstraints();
	}

	protected void setMaxExclusive(String value, boolean fixed) throws SchemaException {
		longConstraints = null;
		super.setMaxExclusive(value, fixed);
		initLongConstraints();
	}

	protected void setMaxInclusive(String value, boolean fixed) throws SchemaException {
		longConstraints = null;
		super.setMaxInclusive(value, fixed);
		initLongConstraints();
	}

    protected Comparable toComparable(String value) throws SchemaException {
    	Long longValue = toLong(value);
    	if (longValue != null) return longValue;
    	else return toBigDecimal(value);
    }

	private Long toLong(String value) throws SchemaException {
		if (longConstraints == null) return null;
		Long result = null;
		try {
			result = new Long(value);
		} catch (NumberFormatException ex) {
			result = null;
		}
		return result;
	}
	
    private BigDecimal toBigDecimal(String value) throws SchemaException {
//        if (value.indexOf('e') != -1 || value.indexOf('E') != -1) {
//            super.invalidValue(value);
//        }
        BigDecimal result = null;
        try {
            result = new BigDecimal(value);
        } catch (NumberFormatException ex) {
            super.invalidValue(value);
        }
        return result;
    }

	protected Comparable checkComparable(Comparable comp) throws SchemaException {
		if (longConstraints != null && comp instanceof Long) {
			longConstraints.checkConstraints(comp);
		} else {
			super.checkComparable(comp);
		}
		return comp;
	}
	
    protected void checkDecimal(Number dec) throws SchemaException {
    	int sc = 0;
        if (dec instanceof BigDecimal) {
            BigDecimal bigDec = (BigDecimal) dec;
            sc = bigDec.scale();
			if (fractionDigits != -1 && sc > fractionDigits)
				super.invalidFacet("cvc-fractionDigits-valid", new Integer(fractionDigits), bigDec.toString());
        }
		if (totalDigits != -1) {
			double maxVal = Math.pow(10, totalDigits-sc);
			if (Math.abs(dec.doubleValue()) >= maxVal)
				super.invalidFacet("cvc-totalDigits-valid", new Integer(totalDigits), dec.toString());
		}
    }

    public void checkFacets(Object valueSpace) throws SchemaException {
        super.checkFacets(valueSpace);
        checkDecimal((Number)valueSpace);
    }

    public int getTotalDigits() {
        return totalDigits;
    }

    public int getFractionDigits() {
        return fractionDigits;
    }

	public String toXMLString(Object obj, ValidationContextProvider context) {
		if (obj == null) return null;
		if (obj instanceof Long) return obj.toString();
		else {
			String result = null;
			BigDecimal data = (BigDecimal)obj;
			result = data.toString();
			if (fractionDigits == 0) {
				int index = result.indexOf('.');
				if (index != -1) return result.substring(0, index);
				else return result;
			} else {
				int index = result.indexOf('.');
				int last = result.length() - 1;
				if (index == -1) {
					return result+".0";
				} else if (index == last) {
					return result+"0";
				} else {
					while (last > index+1 && result.charAt(last) == '0') last--;
					return result.substring(0, last+1);
				}
			}
		}
	}
	
}
