/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.validation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.xquark.schema.*;

public class ChoiceContentIterator extends ContentIterator {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";

  private ChoiceModelGroup group;
  private int count = 0;
  private ContentIterator iterator = null;
  
  public ChoiceContentIterator(Particle particle) 
  {
    super(particle);
    group = (ChoiceModelGroup)particle.getTerm();
  }

  public ElementDeclaration getMatchedDeclaration() {
    if (iterator != null) 
      return iterator.getMatchedDeclaration();
    else return null;
  }

  public ElementDeclaration getModelDeclaration() {
    if (iterator != null) 
      return iterator.getModelDeclaration();
    else return null;
  }

  public int getProcessContents() {
    if (iterator != null) 
      return iterator.getProcessContents();
    else return LAX;
  }
  
  private ContentIterator selectContentIterator(String namespace, String localName) 
  {
    Iterator it = group.iterator();
    ContentIterator result = null;
    
    if ( reporter != null && count < maxOccurs && parentParticle.hasNextElement(namespace, localName) ) 
      reporter.startGroup(parentParticle, ModelGroup.CHOICE, count);
    
    while (it.hasNext()) {
      Particle particle = (Particle)it.next();
      if (particle.hasNextElement(namespace, localName)) {
	result = ContentIteratorFactory.createIterator(particle, reporter);
      } 
      // it's necessary to test if the parentParticle has this element
      // if not, too many particle will be skiped
      // noted Dec 07 2001
      else if (reporter != null && parentParticle.hasNextElement(namespace, localName) ) {
	reporter.alternates(particle);
      }
    }
    return result;
  }
  
  protected int nextElement(String namespace, String localName)
  {
    // Iterator already in use
    ArrayList previousExceptions = null;
    if (iterator != null) {
      switch (iterator.nextElement(namespace, localName)) {
      case MATCHED:
	return MATCHED;
      case UNKNOWN:
        return UNKNOWN;
      case INVALID:
	if (iterator.getMatchedDeclaration() != null) 
	  return invalidElement(iterator.getExceptions());
	else
	  return invalidElement("cvc-particle.3.3", parentParticle, iterator.getExceptions());
      case TOO_MANY:
	previousExceptions = iterator.getExceptions();
	// try next particle
      case DONE:
	if (reporter != null)
	  reporter.endGroup(parentParticle, ModelGroup.CHOICE);      
	count++;
	break;
	// try next particle
      }
    }

    iterator = selectContentIterator(namespace, localName);
    if (iterator == null) {
      if (previousExceptions != null) {
	// Previous iterator had too many occurrences, and no new iterator matched
	// Report previous error
	return invalidElement("cvc-particle.3.3", parentParticle, previousExceptions);
      } else if (count == 0 && parentParticle.getMinExtent() > 0) {
	SchemaException se = SchemaException.computeMinExtentException(parentParticle);
	return invalidElement(se);
      } else {
	// Iterator is finished
	return stopProcessing();
      }
    }
    if (count >= maxOccurs) {
      if (previousExceptions != null) 
	invalidElement("cvc-particle.3.2", parentParticle, previousExceptions);
      else
	invalidElement("cvc-particle.3.2", parentParticle);
      return TOO_MANY;
    }
    switch (iterator.nextElement(namespace, localName)) {
    case MATCHED:
      return MATCHED;
    case UNKNOWN:
      return UNKNOWN;
    case INVALID:
    case TOO_MANY:
    case DONE:
    default:
      if (iterator.getMatchedDeclaration() != null) 
	return invalidElement(iterator.getExceptions());
      else
	return invalidElement("cvc-particle.3.3", parentParticle, iterator.getExceptions());
    }
  }

  public int stopProcessing() {
    int c;
    // Stop current iterator
    if (iterator != null) {
      switch (iterator.stopProcessing()) {
      case INVALID:
	return invalidElement("cvc-particle.3.3", parentParticle, iterator.getExceptions());
      case DONE:
	// Current iterator can be stopped. Continue
      }
      if (reporter != null)
	reporter.endGroup(parentParticle, ModelGroup.CHOICE);
      c = count+1;
    } else {
      c = count;
    }
    if (c < minOccurs && parentParticle.getMinExtent() != 0) {
      return invalidElement("cvc-particle.3.1", parentParticle);      
    } else if (c < maxOccurs && reporter != null) {
      reporter.skipped(parentParticle);
    }
    
    return DONE;
  }

  protected boolean nextValidElements(List elements) {
    int c;
    // Iterator in use
    if (iterator != null) {
      boolean canStop = iterator.nextValidElements(elements);
      // If current iterator cannot be terminated immediately 
      // return false
      if (!canStop) return false;
      c = count+1;
    } else {
      c = count;
    }
    // Add elements of the group
    if (c < maxOccurs && elements != null) {
      Iterator it = group.iterator();
      while (it.hasNext()) {
	Particle particle = (Particle)it.next();
	elements.addAll(particle.firstValidElements());
      }
    }
    // If all elements are optional or minOccurs is reached
    // add null to the elements list
    if (c >= minOccurs || parentParticle.getMinExtent() == 0) return true;
    else return false;
  }
  
}


