/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.validation;

import java.util.ArrayList;
import java.util.List;

import org.xquark.schema.*;

public abstract class ContentIterator implements SchemaConstants {
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName = "$Name:  $";

    public static final int MATCHED = 0;
    public static final int INVALID = 1;
    public static final int DONE = 2;
    public static final int TOO_MANY = 3;
    public static final int UNKNOWN = 4;

    protected boolean endRequired = false;
    protected boolean done = false;
    protected Particle parentParticle = null;
    protected int minOccurs = 1;
    protected int maxOccurs = 1;
    protected ContentReporter reporter = null;
    protected ArrayList exceptions = null;
    protected Type currentType = null;

    public ContentIterator() {}

    public ContentIterator(int minOccurs, int maxOccurs) {
        this.minOccurs = minOccurs;
        this.maxOccurs = maxOccurs;
    }

    public ContentIterator(Particle particle) {
        this(particle.getMinOccurs(), particle.getMaxOccurs());
        parentParticle = particle;
    }

    public boolean isValid() {
        return exceptions == null;
    }

    public ArrayList getExceptions() {
        return exceptions;
    }

    void resetExceptions() {
        exceptions = null;
    }

    public ContentReporter getContentReporter() {
        return reporter;
    }

    public void setContentReporter(ContentReporter reporter) {
        this.reporter = reporter;
    }

    public void startElement(String namespace, String localName, Type localType) throws SchemaException {
        startElement(namespace, localName, localType, localType != null);
    }

    public void startElement(String namespace, String localName, Type localType, boolean localTypePresent)
        throws SchemaException {
        if (done)
            illegalState("Iterator terminated for element {" + namespace + "}" + localName);
        else if (endRequired)
            illegalState("Missing endElement statement before element {" + namespace + "}" + localName);

        switch (nextElement(namespace, localName)) {
            case MATCHED :
                endRequired = true;
                done = false;
                currentType = checkDeclarationAndType(localType, localTypePresent);
                break;
            case DONE :
            case INVALID :
            case TOO_MANY :
                endRequired = false;
                done = true;
                break;
        }
        if (!isValid())
            throw new SchemaException(exceptions);
    }

    public final void startElement(String namespace, String localName) throws SchemaException {
        startElement(namespace, localName, null, false);
    }

    public final void startElement(org.w3c.dom.Element node, Type localType) throws SchemaException {
        startElement(node.getNamespaceURI(), node.getLocalName(), localType, true);
    }

    public final void startElement(org.w3c.dom.Element node) throws SchemaException {
        startElement(node.getNamespaceURI(), node.getLocalName(), null, false);
    }

    public void endElement(String namespace, String localName) throws SchemaException {
        if (done)
            illegalState("Iterator terminated for element {" + namespace + "}" + localName);
        else if (!endRequired)
            illegalState("Unexpected endElement statement for element {" + namespace + "}" + localName);
        endRequired = false;
        done = false;
    }

    public final void endElement(org.w3c.dom.Element node) throws SchemaException {
        endElement(node.getNamespaceURI(), node.getLocalName());
    }

    public final List nextValidElements() {
        if (endRequired || done)
            return null;
        ArrayList result = new ArrayList();
        boolean canStop = nextValidElements(result);
        if (canStop)
            result.add(null);
        return result;
    }

    public final boolean canStopProcessing() {
        if (endRequired || done)
            return true;
        return nextValidElements(null);
    }

    protected final Type checkDeclarationAndType(Type localType, boolean localTypePresent) throws SchemaException {
        ElementDeclaration matchedDecl = getMatchedDeclaration();
        ElementDeclaration origDecl = getModelDeclaration();
        Type result = null;
        if (matchedDecl == null) {
            result = localType;
        } else {
            Type declType = matchedDecl.getType();
            if (matchedDecl.isAbstract())
                invalidElement("cvc-elt.2", matchedDecl);
            if (localType == null) {
                if (localTypePresent)
                    result = null;
                else
                    result = declType;
            } else {
                int exclusions = declType.getBlock() | matchedDecl.getBlock();
                if (origDecl != null)
                    exclusions |= origDecl.getBlock();
                String errCode = localType.checkTypeDerivationOK(declType, exclusions, false);
                if (errCode != null)
                    invalidElement("cvc-elt.4.3", matchedDecl, new SchemaException(errCode, localType));
                result = localType;
            }
        }
        if (result != null && result.isAbstract()) {
            invalidElement("cvc-complex-type.1", result);
        }
        return result;
    }

    protected final int invalidElement(SchemaException se) {
        if (exceptions == null)
            exceptions = new ArrayList();
        exceptions.add(se);
        return INVALID;
    }

    protected final int invalidElement(List exList) {
        if (exceptions == null)
            exceptions = new ArrayList();
        exceptions.addAll(exList);
        return INVALID;
    }

    protected final int invalidElement(String errCode, String errMsg) {
        return invalidElement(new SchemaException(errCode, errMsg));
    }

    protected final int invalidElement(String errCode, Object obj) {
        return invalidElement(new SchemaException(errCode, obj));
    }

    protected final int invalidElement(String errCode, Object obj, SchemaException se) {
        return invalidElement(new SchemaException(errCode, obj, se));
    }

    protected final int invalidElement(String errCode, Object obj, List exList) {
        return invalidElement(new SchemaException(errCode, obj, exList));
    }

    protected final void illegalState(String message) {
        throw new IllegalStateException(message);
    }

    public int getMinOccurs() {
        return minOccurs;
    }

    public int getMaxOccurs() {
        return maxOccurs;
    }

    public Type getType() {
        return currentType;
    }

    public Particle getParticle() {
        return parentParticle;
    }

    void setType(Type type) {
        currentType = type;
    }

    public abstract ElementDeclaration getMatchedDeclaration();

    public ElementDeclaration getModelDeclaration() {
        return null;
    }

    public abstract int getProcessContents();

    protected abstract int nextElement(String namespace, String localName);

    public abstract int stopProcessing();

    protected abstract boolean nextValidElements(List elements);

}
