/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.validation;

import org.xquark.schema.*;

public class ElementTreeContentIterator extends ContentIterator {
    private static final String RCSRevision = "$Revision: 1.2 $";
    private static final String RCSName = "$Name:  $";
    private java.util.LinkedList iterators = new java.util.LinkedList();
    private ContentIterator iterator;
    private ElementDeclaration matchedDecl = null;
    private SchemaManager manager;

    public ElementTreeContentIterator(SchemaManager manager) {
        this(manager, false);
    }

    public ElementTreeContentIterator(SchemaManager manager, boolean strict) {
        this.manager = manager;
        this.iterator = new DefaultContentIterator(manager, strict ? STRICT : LAX);
    }

    public ElementTreeContentIterator(ComplexType type) {
        this.manager = type.getManager();
        DefaultContentIterator dummy = new DefaultContentIterator(manager);
        dummy.setType(type);
        iterators.addFirst(dummy);
        this.iterator = type.childIterator();
    }

    public ElementDeclaration getMatchedDeclaration() {
        return matchedDecl;
    }

    public ElementDeclaration getModelDeclaration() {
        if (iterator != null)
            return iterator.getModelDeclaration();
        else
            return null;
    }

    public int getProcessContents() {
        return iterator.getProcessContents();
    }

    public void startElement(String namespace, String localName, Type localType, boolean localTypePresent)
        throws SchemaException {
        if (done)
            illegalState("Iterator terminated for element {" + namespace + "}" + localName);

        if (endRequired) {
            if (currentType.isSimpleType())
                invalidElement("cvc-type.3.1.2", currentType);
            else if (currentType.getContentType() == EMPTY)
                invalidElement("cvc-complex-type.2.1", currentType);
            else if (currentType.getContentType() == TEXT_ONLY)
                invalidElement("cvc-complex-type.2.2", currentType);
            else
                invalidElement("cvc-complex-type.2.4", currentType);
            endRequired = false;
            iterators.addFirst(iterator);
            iterator = new DefaultContentIterator(manager);
        }
        Type parentType = null;
        currentType = null;
        switch (iterator.nextElement(namespace, localName)) {
            case MATCHED :
                matchedDecl = iterator.getMatchedDeclaration();
                currentType = checkDeclarationAndType(localType, localTypePresent);
                if (currentType != null) {
                    // This is used to retrieve the parent type when an iterator must be replaced by a default one
                    iterator.setType(currentType);
                    ContentIterator nextIterator = currentType.childIterator();
                    if (nextIterator != null) {
                        endRequired = false;
                        iterators.addFirst(iterator);
                        iterator = nextIterator;
                    } else {
                        endRequired = true;
                    }
                    checkPendingExceptions();
                    return;
				} else {
					endRequired = false;
					iterators.addFirst(iterator);
					iterator = new DefaultContentIterator(manager);
					checkPendingExceptions();
					return;
				}
            case UNKNOWN :
                // used in default and wildcard iterator when no declaration is found
                matchedDecl = null;
                currentType = checkDeclarationAndType(localType, localTypePresent);
                if (currentType != null) {
                    // This is used to retrieve the parent type when an iterator must be replaced by a default one
                    iterator.setType(currentType);
                    ContentIterator nextIterator = currentType.childIterator();
                    if (nextIterator != null) {
                        endRequired = false;
                        iterators.addFirst(iterator);
                        iterator = nextIterator;
                    } else {
                        endRequired = true;
                    }
                    checkPendingExceptions();
                    return;
                } else {
					endRequired = false;
					iterators.addFirst(iterator);
					iterator = new DefaultContentIterator(manager, iterator.getProcessContents());
					checkPendingExceptions();
					return;
                }
            case INVALID :
            case TOO_MANY :
                if (!iterators.isEmpty())
                    parentType = ((ContentIterator) iterators.getFirst()).getType();
                invalidElement("cvc-complex-type.2.4", parentType, iterator.getExceptions());
                break;
            case DONE :
                if (!iterators.isEmpty())
                    parentType = ((ContentIterator) iterators.getFirst()).getType();
                invalidElement(
                    "cvc-complex-type.2.4",
                    parentType,
                    SchemaException.computeException(iterator.getParticle()));
                break;
        }

        if (parentType == null && iterator != null && iterator instanceof DefaultContentIterator)
            // it's an error, a dead loop
            invalidElement("cvc-complex-type.2.4", matchedDecl);
        else {
            // Replace invalid iterator by default one
            if (parentType != null)
                iterator = new DefaultContentIterator((ComplexType) parentType);
            else
                iterator = new DefaultContentIterator(manager);
            endRequired = false;
            // Try again
            startElement(namespace, localName, localType, localTypePresent);
        }
    }

    private void checkPendingExceptions() throws SchemaException {
        if (!isValid()) {
            SchemaException exception = new SchemaException(getExceptions());
            resetExceptions();
            throw exception;
        }
    }

    public void endElement(String namespace, String localName) throws SchemaException {
        if (done)
            illegalState("Iterator terminated for element {" + namespace + "}" + localName);
        else if (!endRequired) {
            if (iterator.stopProcessing() == INVALID) {
                Type parentType = null;
                if (!iterators.isEmpty())
                    parentType = ((ContentIterator) iterators.getFirst()).getType();
                invalidElement("cvc-complex-type.2.4", parentType, iterator.getExceptions());
            }
            if (iterators.isEmpty()) {
                illegalState("Too many endElement statements before element {" + namespace + "}" + localName);
            } else {
                iterator = (ContentIterator) iterators.removeFirst();
                if (!iterators.isEmpty()) {
                    ContentIterator prev = (ContentIterator) iterators.getFirst();
                    currentType = prev.getType();
                    matchedDecl = prev.getMatchedDeclaration();
                }
            }
        }
        endRequired = false;
        checkPendingExceptions();
    }

    public int getMinOccurs() {
        return iterator.getMinOccurs();
    }

    public int getMaxOccurs() {
        return iterator.getMaxOccurs();
    }

    public Particle getParticle() {
        return iterator.getParticle();
    }

    protected int nextElement(String namespace, String localName) {
        return INVALID;
    }

    public int stopProcessing() {
        return INVALID;
    }

    protected boolean nextValidElements(java.util.List elements) {
        return iterator.nextValidElements(elements);
    }

}
