/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.validation;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.schema.*;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.schema.datatypes.QName;
import org.xquark.util.NamespaceContextHandler;

public class ValidatingSchemaHandler extends NamespaceContextHandler implements ValidationContextProvider {
    private static final String RCSRevision = "$Revision: 1.4 $";
    private static final String RCSName = "$Name:  $";

    private SchemaValidationContext svContext;
    private SchemaManager manager;
    private SchemaLocator locator;
    private ContentIterator iterator = null;
    private boolean nilElement = false;
    private boolean afterStart = false;

    private HashSet namespaces = new HashSet();

    private StringBuffer buffer = new StringBuffer();
    private AttributesImpl attributes = new AttributesImpl();
    private boolean extendsInfoset = false;
    private boolean strictAssess = true;

    private boolean debug = false;

    private Locator docLocator = null;
    private String defaultDocumentBase = null;
    private String documentBase = null;
    private URL base = null;

    private StringBuffer errBuffer = new StringBuffer();
    private boolean isCharactersValid = true;
    private String errCode = null;
    private ElementDeclaration errDecl = null;
    private Type errType = null;

    public ValidatingSchemaHandler(SchemaValidationContext svContext) {
        this(svContext, svContext, true);
    }

    public ValidatingSchemaHandler(SchemaValidationContext svContext, boolean strict) {
        this(svContext, svContext, strict);
    }

    public ValidatingSchemaHandler(SchemaValidationContext svContext, SchemaLocator locator) {
        this(svContext, locator, true);
    }

    public ValidatingSchemaHandler(SchemaValidationContext svContext, SchemaLocator locator, boolean strict) {
        super();
        this.svContext = svContext;
        this.manager = svContext.getSchemaManager();
        this.locator = locator;
        strictAssess = strict;
		iterator = new ElementTreeContentIterator(manager, strictAssess);
    }

    public void setDebug(boolean b) {
        debug = b;
    }

    public boolean isDebug() {
        return debug;
    }

    public void setDocumentLocator(Locator locator) {
        docLocator = locator;
        super.setDocumentLocator(locator);
    }

    public String getDocumentBase() {
        if (documentBase != null)
            return documentBase;
        else
            return defaultDocumentBase;
    }

    public void setDocumentBase(String documentBase) {
        defaultDocumentBase = documentBase;
    }

    public java.util.Map getNotationDeclarations() {
        return null;
    }

    public void startDocument() throws SAXException {
        super.startDocument();
        iterator = new ElementTreeContentIterator(manager, strictAssess);
        nilElement = false;
        buffer.setLength(0);
        documentBase = defaultDocumentBase;
        if (docLocator != null && docLocator.getSystemId() != null)
            documentBase = docLocator.getSystemId();
        if (documentBase != null) {
            try {
   				base = new URL(documentBase);
            } catch (MalformedURLException ex) {
                //reportError(new SAXParseException("Invalid document base: "+documentBase, docLocator));
            }
        }
        namespaces.clear();
        svContext.reset(this);
        locator.clearSchemaLocations();
    }

    public void endDocument() throws SAXException {
        documentBase = null;
        base = null;
        super.endDocument();
    }

    public void startElement(String namespaceURI, String localName, String qName, Attributes atts)
        throws SAXException {
        if (debug)
            System.err.println("startElement: " + localName);

        // test if there is an error in precedent element
        if (!isCharactersValid) {
            isCharactersValid = true;
            if (errCode.indexOf("elt") == -1)
                invalidValue(errBuffer.toString(), errDecl, errType, new SchemaException(errCode, errType));
            else
                invalidValue(errBuffer.toString(), errDecl, errType, new SchemaException(errCode, errDecl));
            errBuffer.setLength(0);
        }

        // Normalize empty namespace to null
        String ns = null;
        if (namespaceURI != null && namespaceURI.length() == 0)
            ns = null;
        else
            ns = namespaceURI;

        // Retrieves any schema locations specified in attributes
        addSchemaLocations(atts);
        // Add namespace to list of encountered namespaces (used to avoid schema locations appearing after ns)
        namespaces.add(ns);
        // Retrieve enclosing element info
        Type type = svContext.getElementType();
        ElementDeclaration matchedDecl = svContext.getElementDeclaration();
        // This is applied to the enclosing element
        if (nilElement) {
            // unexpected element
            invalidElement("cvc-elt.3.2.1", matchedDecl);
            nilElement = false;
        }

        svContext.pushElementInfoset(ns, localName);
        Type localType = null;
        String xsiType = atts.getValue(SchemaConstants.XSI_URI, SchemaConstants.XSI_TYPE_ATTR);
        // Used to hold a type resolution exception until the matched declaration is known
        SchemaException typeResolutionException = null;
        if (xsiType != null) {
            try {
                localType = resolveTypeRef(xsiType);
            } catch (SchemaException se) {
                typeResolutionException = se;
            }
        }

        // Load schema if needed
        try {
            manager.loadSchema(locator, ns);
        } catch (SAXException ex) {
            reportInvalidSchemaError(ns, ex);
        }

        try {
            iterator.startElement(ns, localName, localType, xsiType != null);
        } catch (SchemaException se) {
            // matchedDecl refers to the enclosing element
            invalidContent(localName, matchedDecl, type, se);
        }

        // use the newly found validation info
        type = iterator.getType();
        matchedDecl = (ElementDeclaration) iterator.getMatchedDeclaration();

        // Handle pending exception on local type resolution
        if (typeResolutionException != null) {
            typeResolutionException.setInvalidObject(matchedDecl);
            invalidType(xsiType, typeResolutionException);
        }

        // Check process contents type
        int process = iterator.getProcessContents();
        if (strictAssess && process == SchemaConstants.STRICT && matchedDecl == null && type == null) {
            invalidElement("cvc-assess-elt.1.1", null);
            invalidElement("cvc-assess-elt.1.2", null);
        }

        // check if it's a nillable element
        nilElement =
            SchemaConstants.TRUE_VALUE.equals(atts.getValue(SchemaConstants.XSI_URI, SchemaConstants.XSI_NIL_ATTR));
        if (nilElement && matchedDecl != null) {
            if (!matchedDecl.isNillable()) {
                // Illegal nil element
                invalidElement("cvc-elt.3.1", matchedDecl);
            }
            if (matchedDecl.getFixedValue() != null) {
                invalidElement("cvc-elt.3.2.2", matchedDecl);
            }
        }
        svContext.initElementValidationInfo(matchedDecl, type, nilElement, process);
        atts = validateAttributes(matchedDecl, type, atts);
        super.startElement(namespaceURI, localName, qName, atts);
        // Used to handle mixed case 
        // (normalized value is buffered only when element has no children)
        afterStart = true;
        buffer.setLength(0);
    }

    private void addSchemaLocations(Attributes atts) throws SAXException {
        String schemaLocations = atts.getValue(SchemaConstants.XSI_URI, SchemaConstants.XSI_SCHEMA_LOCATION_ATTR);
        String noNamespaceSchemaLocation =
            atts.getValue(SchemaConstants.XSI_URI, SchemaConstants.XSI_NO_NAMESPACE_SCHEMA_LOCATION_ATTR);
        if (schemaLocations != null) {
            StringTokenizer tok = new StringTokenizer(schemaLocations);
            while (tok.hasMoreTokens()) {
                String ns = tok.nextToken();
                if (namespaces.contains(ns)) {
                    String errMsg =
                        "Location information for schema " + ns + " cannot appear after assessment has begun";
                    throw new SAXException(errMsg);
                }
                if (!tok.hasMoreTokens()) {
                    String errMsg = "Odd number of values in schemaLocation attribute. " + schemaLocations;
                    throw new SAXException(errMsg);
                }
                String loc = tok.nextToken();
                URL uri = null;
                try {
                    uri = new URL(base, loc);
                    locator.addSchemaLocation(ns, uri.toString());
                } catch (MalformedURLException ex) {
                    // if invalid URI, ignore it
                }
            }
        }
        if (noNamespaceSchemaLocation != null) {
            if (namespaces.contains(null)) {
                String errMsg =
                    "Location information for schema without namespace cannot appear after assessment has begun";
                throw new SAXException(errMsg);
            }
            URL uri = null;
            try {
                uri = new URL(base, noNamespaceSchemaLocation);
                locator.addSchemaLocation(null, uri.toString());
            } catch (MalformedURLException ex) {
                // if invalid URI, ignore it
            }
        }
    }

    private Type resolveTypeRef(String xsiType) throws SchemaException, SAXException {
        SimpleType validator = manager.getQNameType();
        xsiType = validator.normalizedValue(xsiType);
        QName qName = null;
        try {
            qName = (QName) validator.convert(xsiType, false, this);
        } catch (SchemaException se) {
            throw new SchemaException("cvc-elt.4.1", se);
        }
        String uri = qName.getNamespaceURI();
        String typeName = qName.getLocalName();
        Schema schema = null;
        Type result = null;
        try {
            schema = manager.loadSchema(locator, uri);
        } catch (SAXException ex) {
            reportInvalidSchemaError(uri, ex);
        }
        if (schema != null)
            result = schema.getType(typeName);
        if (result == null) {
            throw new SchemaException("cvc-elt.4.2", new SchemaException("cvc-resolve-instance.1", (Object) xsiType));
        } else {
            return result;
        }
    }

    public void characters(char[] ch, int start, int len) throws SAXException {
        if (debug)
            System.err.println("Characters: " + new String(ch, start, len));
        if (len == 0)
            return;

        Type type = svContext.getElementType();
        ElementDeclaration matchedDecl = svContext.getElementDeclaration();

        // check nilElement
        if (nilElement) {
            // unexpected content
            isCharactersValid = false;
            errBuffer.append(ch, start, len);
            errCode = "cvc-elt.3.2.1";
            errDecl = matchedDecl;
            errType = type;
        }

        if (type == null) {
            super.characters(ch, start, len);
            return;
        }

        switch (type.getContentType()) {
            case SchemaConstants.EMPTY :
                isCharactersValid = false;
                errBuffer.append(ch, start, len);
                errCode = "cvc-complex-type.2.1";
                errDecl = matchedDecl;
                errType = type;
                super.characters(ch, start, len);
                break;
            case SchemaConstants.ELEMENT_ONLY :
                boolean bWhitespaceOnly = true;
                for (int i = start; i < start + len; i++) {
                    if (!PrimitiveType.isXMLWhitespace(ch[i]))
                        bWhitespaceOnly = false;
                }
                if (bWhitespaceOnly) {
                    super.ignorableWhitespace(ch, start, len);
                } else {
                    isCharactersValid = false;
                    errBuffer.append(ch, start, len);
                    errCode = "cvc-complex-type.2.3";
                    errDecl = matchedDecl;
                    errType = type;
                    super.characters(ch, start, len);
                }
                break;
            case SchemaConstants.MIXED :
                if (afterStart && matchedDecl != null && matchedDecl.getFixedValue() != null)
                    buffer.append(ch, start, len);
                super.characters(ch, start, len);
                break;
            case SchemaConstants.TEXT_ONLY :
                // content type is simple type
                buffer.append(ch, start, len);
                super.characters(ch, start, len);
                break;
        }
    }

    public void endElement(String namespaceURI, String localName, String qName) throws SAXException {
        if (debug)
            System.err.println("endElement: " + localName);

        // test if there is an error in precedent element
        if (!isCharactersValid) {
            isCharactersValid = true;
            if (errCode.indexOf("elt") == -1)
                invalidValue(errBuffer.toString(), errDecl, errType, new SchemaException(errCode, errType));
            else
                invalidValue(errBuffer.toString(), errDecl, errType, new SchemaException(errCode, errDecl));
            errBuffer.setLength(0);
        }

        Type type = svContext.getElementType();
        ElementDeclaration matchedDecl = svContext.getElementDeclaration();

        validateValue(type, matchedDecl);

        try {
            iterator.endElement(namespaceURI, localName);
        } catch (SchemaException se) {
            if (!nilElement)
                invalidContent(localName, matchedDecl, type, se);
        }
        try {
            svContext.checkIdentityConstraints();
        } catch (SchemaException se) {
            reportValidationError(se);
        }
        // This step computes the final validity of the element
        svContext.computeValidationInfo();

        super.endElement(namespaceURI, localName, qName);

        svContext.popElementInfoset();
        nilElement = false;
        afterStart = false;
        buffer.setLength(0);
    }

    private void validateValue(Type type, ElementDeclaration matchedDecl) throws SAXException {
        if (type == null
            || type.getContentType() == SchemaConstants.EMPTY
            || type.getContentType() == SchemaConstants.ELEMENT_ONLY
            || nilElement) {
            buffer.setLength(0);
            return;
        }
        if (!afterStart) {
            if (matchedDecl != null && matchedDecl.getFixedValue() != null)
                invalidElement("cvc-elt.5.2.2.1", matchedDecl);
            buffer.setLength(0);
            return;
        }

        Object actualValue = null;
        if (buffer.length() == 0 && matchedDecl != null && matchedDecl.getValueConstraint() != null) {
            // Use default value
            String normalizedValue = matchedDecl.getValueConstraint();
            if (type != matchedDecl.getType()) {
                try {
                    normalizedValue = type.validateDefault(normalizedValue, this);
                } catch (SchemaException se) {
                    invalidElement("cvc-elt.5.1.1", matchedDecl, se);
                }
            }
            if (extendsInfoset)
                super.characters(normalizedValue.toCharArray(), 0, normalizedValue.length());
            buffer.append(normalizedValue);
            if (type.getContentType() == SchemaConstants.TEXT_ONLY) {
                try {
                    actualValue = checkValueValid(type.getValueType(), buffer, -1);
                    svContext.setSchemaSpecified(-1, true);
                } catch (SchemaException se) {
                    invalidElement("cvc-elt.5.1.2", matchedDecl, se);
                }
            }
        } else {
            // use infoset
            if (type.getContentType() == SchemaConstants.TEXT_ONLY) {
                try {
                    actualValue = checkValueValid(type.getValueType(), buffer, -1);
                    svContext.setSchemaSpecified(-1, false);
                } catch (SchemaException se) {
                    invalidValue(buffer.toString(), matchedDecl, type, se);
                }
                if (actualValue != null && matchedDecl != null && matchedDecl.getFixedValue() != null) {
                    if (!matchedDecl.getFixedValue().equals(type.toXMLString(actualValue)))
                        invalidValue(
                            buffer.toString(),
                            matchedDecl,
                            type,
                            new SchemaException("cvc-elt.5.2.2.2.2", matchedDecl));
                }
            } else if (
                matchedDecl != null
                    && matchedDecl.getFixedValue() != null
                    && !buffer.toString().equals(matchedDecl.getFixedValue())) {
                invalidValue(
                    buffer.toString(),
                    matchedDecl,
                    type,
                    new SchemaException("cvc-elt.5.2.2.2.1", matchedDecl));
            }
        }

        // check if this element is defined in IdentityConstraint
        if (actualValue != null && matchedDecl != null) {
            try {
                svContext.checkKeyValue(actualValue);
            } catch (SchemaException se) {
                reportValidationError("Error while processing value -> '" + actualValue + "'", se);
            }
        }
    }

    /**
     * This method modifies its value parameter
     */
    private Object checkValueValid(SimpleType type, StringBuffer value, int attrIndex) throws SchemaException {
    	ValidationInfo info = type.validate(value, true, this);
        SimpleType realType = info.getValidationType();
        Object result = info.getActualValue();
        if (realType != type)
            svContext.setMemberType(attrIndex, realType);
        svContext.setNormalizedValue(attrIndex, info.getNormalizedValue());
        svContext.setActualValue(attrIndex, result);
        return result;
    }

    private Attributes validateAttributes(ElementDeclaration matchedDecl, Type type, Attributes atts)
        throws SAXException {
        if (atts == null)
            return null;
        if (type == null || type.isSimpleType()) {
            StringBuffer buffer = new StringBuffer();
            attributes.clear();
            for (int i = 0; i < atts.getLength(); i++) {
                String namespace = atts.getURI(i);
                String localName = atts.getLocalName(i);
                String attrValue = atts.getValue(i);
                attributes.addAttribute(namespace, localName, atts.getQName(i), atts.getType(i), attrValue);
                if (SchemaConstants.XMLNS_URI.equals(namespace) || SchemaConstants.XMLNS_PREFIX.equals(localName))
                    continue;
                AttributeDeclaration attrDecl = null;
                if (!SchemaConstants.XSI_URI.equals(namespace)) {
                    if (namespace.length() == 0)
                        namespace = null;
                    if (type != null) {
                        invalidAttribute(
                            namespace,
                            localName,
                            atts.getValue(i),
                            matchedDecl,
                            type,
                            new SchemaException("cvc-type.3.1.1", type));
                    } else {
                        Schema schema = null;
                        try {
                            schema = manager.loadSchema(locator, namespace);
                        } catch (SAXException ex) {
                            //reportInvalidSchemaError(namespace, ex);
                        }
                        if (schema != null)
                            attrDecl = schema.getAttributeDeclaration(localName);
                    }
                } else {
                    attrDecl = manager.getAttributeDeclaration(namespace, localName);
                    if (attrDecl == null) {
                        invalidAttribute(
                            namespace,
                            localName,
                            attrValue,
                            matchedDecl,
                            null,
                            new SchemaException("cvc-assess-attr.1", (Object) null));
                    }
                }
                int index = svContext.addAttributePSVI(namespace, localName, attrDecl, true);
                if (attrDecl != null) {
                    buffer.setLength(0);
                    buffer.append(attrValue);
                    Object actualValue = validateAttribute(matchedDecl, null, attrDecl, buffer, index);
                    if (actualValue != null) {
                        svContext.setSchemaSpecified(index, false);
                        attributes.setValue(i, buffer.toString());
                        if (attrDecl.getFixedValue() != null) {
                            Type attrType = attrDecl.getType();
                            if (!attrDecl.getFixedValue().equals(attrType.toXMLString(actualValue))) {
                                SchemaException attributeException = new SchemaException("cvc-attribute.4", attrDecl);
                                invalidAttribute(
                                    namespace,
                                    localName,
                                    attrValue,
                                    matchedDecl,
                                    type,
                                    attributeException);
                                svContext.addErrorCodes(index, attributeException);
                            }
                        }
                    }
                    svContext.computeAttributeValidationInfo(index);
                }
            }
            return attributes;
        } else {
            ComplexType ctype = (ComplexType) type;
            boolean wildIDFound = false;
            StringBuffer buffer = new StringBuffer();
            attributes.clear();
            for (int i = 0; i < atts.getLength(); i++) {
                String namespace = atts.getURI(i);
                String localName = atts.getLocalName(i);
                String attrValue = atts.getValue(i);
                attributes.addAttribute(namespace, localName, atts.getQName(i), atts.getType(i), attrValue);
                if (SchemaConstants.XMLNS_URI.equals(namespace) || SchemaConstants.XMLNS_PREFIX.equals(localName))
                    continue;
                AttributeDeclaration attrDecl = null;
                boolean strictWildcard = false;
                if (!SchemaConstants.XSI_URI.equals(namespace)) {
                    if (namespace.length() == 0)
                        namespace = null;
                    attrDecl = ctype.getAttributeDeclaration(namespace, localName);
                    if (attrDecl == null) {
                        Wildcard wildcard = ctype.getAttributeWildcard();
                        Schema schema = null;
                        if (wildcard == null) {
                            invalidAttribute(
                                namespace,
                                localName,
                                attrValue,
                                matchedDecl,
                                ctype,
                                new SchemaException("cvc-complex-type.3.2.1", ctype));
                        } else if (!wildcard.isAllowed(namespace)) {
                            SchemaException wildEx =
                                new SchemaException(wildcard.checkNamespaceAllowed(namespace), wildcard);
                            invalidAttribute(
                                namespace,
                                localName,
                                attrValue,
                                matchedDecl,
                                ctype,
                                new SchemaException("cvc-complex-type.3.2.2", ctype, wildEx));
                        } else if (wildcard.getProcessContents() != SchemaConstants.SKIP) {
                            try {
                                schema = manager.loadSchema(locator, namespace);
                            } catch (SAXException ex) {
                                reportInvalidSchemaError(namespace, ex);
                            }
                            if (schema != null)
                                attrDecl = schema.getAttributeDeclaration(localName);
                            if (wildcard.getProcessContents() == SchemaConstants.STRICT) {
                                strictWildcard = true;
                                if (strictAssess && attrDecl == null) {
                                    invalidAttribute(
                                        namespace,
                                        localName,
                                        attrValue,
                                        matchedDecl,
                                        ctype,
                                        new SchemaException("cvc-assess-attr.1", (Object) null));
                                }
                            }
                            boolean isID = attrDecl != null && attrDecl.getType().isDerivedFrom(manager.getIDType());
                            if (isID) {
                                if (wildIDFound)
                                    invalidAttribute(
                                        namespace,
                                        localName,
                                        attrValue,
                                        matchedDecl,
                                        ctype,
                                        new SchemaException("cvc-complex-type.5.1", ctype));
                                else {
                                    if (ctype.getIDAttribute() != null)
                                        invalidAttribute(
                                            namespace,
                                            localName,
                                            attrValue,
                                            matchedDecl,
                                            ctype,
                                            new SchemaException("cvc-complex-type.5.2", ctype));
                                    wildIDFound = true;
                                }
                            }
                        }
                    }
                } else {
                    strictWildcard = true;
                    attrDecl = manager.getAttributeDeclaration(namespace, localName);
                    if (attrDecl == null) {
                        invalidAttribute(
                            namespace,
                            localName,
                            attrValue,
                            matchedDecl,
                            null,
                            new SchemaException("cvc-assess-attr.1", (Object) null));
                    }
                }
                int index = svContext.addAttributePSVI(namespace, localName, attrDecl, strictWildcard);
                if (attrDecl != null) {
                    buffer.setLength(0);
                    buffer.append(attrValue);
                    Object actualValue = validateAttribute(matchedDecl, ctype, attrDecl, buffer, index);
                    if (actualValue != null) {
                        svContext.setSchemaSpecified(index, false);
                        attributes.setValue(i, buffer.toString());
                        if (attrDecl.getFixedValue() != null) {
                            Type attrType = attrDecl.getType();
                            if (!attrDecl.getFixedValue().equals(attrType.toXMLString(actualValue))) {
                                SchemaException attributeException = new SchemaException("cvc-attribute.4", attrDecl);
                                invalidAttribute(
                                    namespace,
                                    localName,
                                    attrValue,
                                    matchedDecl,
                                    ctype,
                                    attributeException);
                                svContext.addErrorCodes(index, attributeException);
                            }
                        }
                    }
                    svContext.computeAttributeValidationInfo(index);
                }
            }
            Iterator it = ctype.getRequiredAndDefaultAttributes().iterator();
            while (it.hasNext()) {
                AttributeDeclaration decl = (AttributeDeclaration) it.next();
                String ns = decl.getNamespace();
                if (ns == null)
                    ns = "";
                int index = atts.getIndex(ns, decl.getName());
                if (index == -1) {
                    if (decl.getUse() == SchemaConstants.REQUIRED) {
                        invalidContent(
                            decl.getName(),
                            matchedDecl,
                            ctype,
                            new SchemaException("cvc-complex-type.4", ctype));
                    }
                    if (decl.getValueConstraint() != null) {
                        if (extendsInfoset)
                            attributes.addAttribute(ns, decl.getName(), "", "CDATA", decl.getValueConstraint());
                        int attrIndex = svContext.addAttributePSVI(decl.getNamespace(), decl.getName(), decl, true);
                        buffer.setLength(0);
                        buffer.append(decl.getValueConstraint());
                        if (validateAttribute(matchedDecl, ctype, decl, buffer, attrIndex) != null)
                            svContext.setSchemaSpecified(attrIndex, true);
                    }
                }
            }
            return attributes;
        }
    }

    private Object validateAttribute(
        ElementDeclaration matchedDecl,
        ComplexType ctype,
        AttributeDeclaration decl,
        StringBuffer buffer,
        int index)
        throws SAXException {
        Object actualValue = null;
        Type attrType = decl.getType();
        try {
            actualValue = checkValueValid(attrType.getValueType(), buffer, index);
            if (ctype != null && decl == ctype.getNotationAttribute()) {
                svContext.setNotation(ctype.getSchema().getNotationDeclaration(buffer.toString()));
            }
        } catch (SchemaException se) {
            SchemaException attributeException = new SchemaException("cvc-attribute.3", decl, se);
            invalidAttribute(
                decl.getNamespace(),
                decl.getName(),
                buffer.toString(),
                matchedDecl,
                ctype,
                attributeException);
            svContext.addErrorCodes(index, attributeException);
        }
        // check if this element is defined in IdentityConstraint
        if (actualValue != null && !(SchemaConstants.XSI_URI.equals(decl.getNamespace()))) {
            try {
                svContext.checkKeyValue(decl, actualValue);
            } catch (SchemaException se) {
                reportValidationError("Error while processing value -> '" + actualValue + "'", se);
            }
        }
        return actualValue;
    }

    private void invalidContent(String localName, ElementDeclaration decl, Type type, SchemaException se)
        throws SAXException {
        if (se.getErrorCode() == null && se.getExceptions() != null) {
            java.util.Iterator it = se.getExceptions().iterator();
            while (it.hasNext())
                invalidContent(localName, decl, type, (SchemaException) it.next());
            return;
        }
        SchemaException exception = se;
        if (!exception.getErrorCode().startsWith("cvc-elt")) {
            if (!exception.getErrorCode().startsWith("cvc-type"))
                exception = new SchemaException("cvc-type.3.2", type, exception);
            if (decl != null)
                exception = new SchemaException("cvc-elt.5.2.1", decl, exception);
        }
        svContext.addErrorCodes(-1, exception);
        reportValidationError("Error while processing localName -> " + localName, exception);
    }

    private void invalidValue(String value, ElementDeclaration decl, Type type, SchemaException se)
        throws SAXException {
        SchemaException exception = se;
        if (!exception.getErrorCode().startsWith("cvc-elt")) {
            if (!exception.getErrorCode().startsWith("cvc-type")) {
                if (type.isSimpleType()) {
                    exception = new SchemaException("cvc-type.3.1.3", type, exception);
                } else {
                    if (!exception.getErrorCode().startsWith("cvc-complex-type"))
                        exception = new SchemaException("cvc-complex-type.2.2", type, exception);
                    exception = new SchemaException("cvc-type.3.2", type, exception);
                }
            }
            if (decl != null)
                exception = new SchemaException("cvc-elt.5.2.1", decl, exception);
        }
        svContext.addErrorCodes(-1, exception);
        reportValidationError("Error while processing value -> '" + value + "'", exception);
    }

    private void invalidAttribute(
        String namespace,
        String localName,
        String value,
        ElementDeclaration decl,
        Type type,
        SchemaException se)
        throws SAXException {
        SchemaException exception = se;
        if (type != null && !exception.getErrorCode().startsWith("cvc-elt")) {
            if (!exception.getErrorCode().startsWith("cvc-type")) {
                if (!type.isSimpleType()) {
                    if (!exception.getErrorCode().startsWith("cvc-complex-type"))
                        exception = new SchemaException("cvc-complex-type.3.1", type, exception);
                    exception = new SchemaException("cvc-type.3.2", type, exception);
                }
            }
            if (decl != null)
                exception = new SchemaException("cvc-elt.5.2.1", decl, exception);
        }
        svContext.addErrorCodes(-1, exception);
        String message = "Error while processing attribute -> ";
        if (namespace != null && namespace.length() > 0) {
            String prefix = getPrefix(namespace);
            if (prefix != null && prefix.length() > 0)
                message += prefix + ":";
        }
        message += localName + "=\"" + value + "\"";
        reportValidationError(message, exception);
    }

    private void invalidType(String type, SchemaException exception) throws SAXException {
        svContext.addErrorCodes(-1, exception);
        String message = "Error while resolving type -> " + type;
        reportValidationError(message, exception);
    }

    private void invalidElement(String errCode, Object object) throws SAXException {
        invalidElement(new SchemaException(errCode, object));
    }

    private void invalidElement(String errCode, Object object, SchemaException e) throws SAXException {
        invalidElement(new SchemaException(errCode, object, e));
    }

    private void invalidElement(SchemaException se) throws SAXException {
        svContext.addErrorCodes(-1, se);
        reportValidationError(se);
    }

    private void reportInvalidSchemaError(String namespace, SAXException ex) throws SAXException {
        reportError(new SAXParseException("Invalid schema for namespace {" + namespace + "}", docLocator, ex));
    }

    private void reportValidationError(SchemaException se) throws SAXException {
        reportValidationError(null, se);
    }

    private void reportValidationError(String message, SchemaException se) throws SAXException {
        if (message != null)
            message += "\n" + se.getMessage();
        reportError(new SAXParseException(message, docLocator, se));
    }

    private void reportError(SAXParseException ex) throws SAXException {
        if (getPluggedErrorHandler() != null) {
            getPluggedErrorHandler().error(ex);
        }
    }

}
