/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

/*
 * XMLSerializer.java
 *
 */

package org.xquark.serialize;

import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;

import org.xquark.util.NSPrefixDecorator;

/**
 * A simple SAX2 XML serializer implementing SAX ContentHandler and
 * LexicalHandler. <B>Warning: </B>Default behavior is to perform automatic
 * indentation and to ignore ignorable whitespace generated by the XML
 * processor.
 * 
 * <B>IMPORTANT </B>: SAX2 XMLReader
 * <code>http://xml.org/sax/features/namespaces</code> feature is supposed to
 * be set to true which is the default.
 */
public class XMLSerializer extends NSPrefixDecorator {
	private static final String RCSRevision = "$Revision: 1.5 $";

	private static final String RCSName = "$Name:  $";

	private BasicXMLSerializer serializer = new BasicXMLSerializer();

	private XMLSerializer() {
		setContentHandler(serializer);
		setLexicalHandler(serializer);
		setEnabled(true);
	}

	/**
	 * Creates a new XMLSerializer
	 * 
	 * @param out
	 *            the OutputStream where the serializer writes bytes. For
	 *            instance a FileOutputStream. The Unicode character encoding
	 *            used will be the platform default.
	 * @throws UnsupportedEncodingException
	 *             if the UTF-8 encoding is not supported by the JAVA platform
	 */
	public XMLSerializer(OutputStream out) throws UnsupportedEncodingException {
		this();
		serializer.setOutputStream(out);
	}

	/**
	 * Creates a new XMLSerializer
	 * 
	 * @param out
	 *            the OutputStream where the serializer writes bytes. For
	 *            instance a FileOutputStream.
	 * @param encoding
	 *            a string specifying the character encoding, which must be
	 *            supported by the JDK. Example: UTF-8, ISO-8859-1
	 * @throws UnsupportedEncodingException
	 *             if the provided encoding is not supported by the JAVA
	 *             platform
	 * @see java.io.OutputStream
	 */
	public XMLSerializer(OutputStream out, String encoding)
			throws UnsupportedEncodingException {
		this();
		serializer.setDefaultEncoding(encoding);
		serializer.setOutputStream(out);
	}

	/**
	 * Creates a new XMLSerializer
	 * 
	 * @param out
	 *            the writer where the serializer writes characters. For
	 *            instance a FileWriter.
	 */
	public XMLSerializer(Writer out) {
		this();
		serializer.setWriter(out);
	}

	/**
	 * Creates a new XMLSerializer
	 * 
	 * @param out
	 *            the writer where the serializer writes characters. For
	 *            instance a FileWriter.
	 * @param encoding
	 *            a string specifying the character encoding (must be compatible
	 *            with the encoding used by the writer). This value is used to
	 *            generate the encoding information at the beginning of the XML
	 *            file
	 * @see java.io.OutputStream
	 */
	public XMLSerializer(Writer out, String encoding) {
		this();
		serializer.setDefaultEncoding(encoding);
		serializer.setWriter(out);
	}

	/**
	 * Creates a new XMLSerializer
	 * 
	 * @param encoding
	 *            a string for the character encoding (must be supported by the
	 *            JDK used). Example: UTF-8, ISO-8859-1
	 * @param indent
	 *            if true, tabs are used at the beginning of lines to indent the
	 *            output file.
	 */
	public XMLSerializer(String encoding, boolean indent) {
		this();
		serializer.setDefaultEncoding(encoding);
		serializer.setIndent(indent);
	}

	/**
	 * Set the OutputStream where the serializer writes bytes.
	 * 
	 * When this method is called, the output Writer is created with the
	 * currently specified encoding. Further calls to setEncoding will not
	 * change the encoding used by the Writer.
	 * 
	 * @param out
	 *            the OutputStream where the serializer writes bytes. For
	 *            instance a FileOutputStream.
	 * @throws UnsupportedEncodingException
	 *             if the provided encoding is not supported by the JAVA
	 *             platform
	 */
	public void setOutputStream(OutputStream out)
			throws UnsupportedEncodingException {
		serializer.setOutputStream(out);
	}

	/**
	 * Set the Writer where the serializer writes characters.
	 * 
	 * @param out
	 *            the writer where the serializer writes characters.
	 */
	public void setWriter(Writer out) {
		serializer.setWriter(out);
	}

	/**
	 * Set the indenting mode. <B>Warning: </B>Default behavior is to perform
	 * automatic indentation ( ignorable whitespaces generated by the XML
	 * processor are ignored).
	 * 
	 * @param indent
	 *            if true, tabs are used at the beginning of lines to indent the
	 *            output file.
	 */
	public void setIndent(boolean indent) {
		serializer.setIndent(indent);
	}

	/**
	 * Set the ordering mode for attributes. The default behavior is to perform
	 * no ordering.
	 * 
	 * @param mode
	 *            If true, ordering is performed following the
	 *            {@link <a href="http://www.w3.org/TR/xml-c14n">Canonical XML<a>}
	 *            W3C recommendation. If false, attributes are serialized as
	 *            passed by the SAX2 XMLReader.
	 */
	public void setCanonicalOutput(boolean mode) {
		serializer.setCanonicalOutput(mode);
	}

	/**
	 * turn on or off the use of ignorable whitespace for indentation.
	 * <B>Warning: </B>This flag is ignored if automatic indentation is on.
	 * 
	 * @param use
	 *            if true (the default), ignorable whitespace are used for
	 *            indentation unless automatic indentation is on.
	 */
	public void setUseIgnorableWhitespaces(boolean use) {
		serializer.setUseIgnorableWhitespaces(use);
	}

	/**
	 * Set the character encoding.
	 * 
	 * @param encoding
	 *            a string specifying the character encoding, which must be
	 *            supported by the JDK. Example: UTF-8, ISO-8859-1.
	 */
	public void setEncoding(String encoding) {
		serializer.setDefaultEncoding(encoding);
	}

	/**
	 * Enable XML declaration generation on startDocument() event.
	 * 
	 * @param enable
	 *            XML Declaration is generated if true. Default is true.
	 */
	public void setGenerateXMLDeclaration(boolean enable) {
		serializer.setGenerateXMLDeclaration(enable);
	}

	/**
	 * Enable encoding attribute generation in XML declaration on
	 * startDocument() event. Considered only if XML decleration enabled.
	 * 
	 * @param enable
	 *            encoding declaration generation if true. Default is true.
	 */
	public void setGenerateEncodingDeclaration(boolean enable) {
		serializer.setGenerateEncodingDeclaration(enable);
	}

	/**
	 * Enable automatic close of the output user stream on endDocument() event.
	 * 
	 * @param close
	 *            Output stream will be closed if true. Default is false.
	 */
	public void setAutoStreamClose(boolean close) {
		serializer.setAutoStreamClose(close);
	}

	/**
	 * Accessor to the current indenting mode.
	 * 
	 * @return the current indenting mode.
	 */
	public boolean getIndent() {
		return serializer.getIndent();
	}

	/**
	 * Accessor to the current ordering mode for attributes.
	 * 
	 * @return the current ordering mode for attributes.
	 */
	public boolean getCanonicalOutput() {
		return serializer.getCanonicalOutput();
	}

	/**
	 * Accessor to the current the status of the use of ignorable whitespace by
	 * the serializer.
	 * 
	 * @return the status of the use of ignorable whitespace by the serializer.
	 */
	public boolean getUseIgnorableWhitespaces() {
		return serializer.getUseIgnorableWhitespaces();
	}

	/**
	 * Accessor to the current character encoding in use.
	 * 
	 * @return the character encoding in use, if previously specified.
	 */
	public String getEncoding() {
		return serializer.getDefaultEncoding();
	}

	/**
	 * Accessor to the current flag for XML declaration generation.
	 * 
	 * @return if true, the XML declaration will be generated.
	 */
	public boolean getGenerateXMLDeclaration() {
		return serializer.getGenerateXMLDeclaration();
	}

	/**
	 * Return the value set for stream automatic close feature.
	 * 
	 * @see #setAutoStreamClose(boolean).
	 */
	public boolean getAutoStreamClose() {
		return serializer.getAutoStreamClose();
	}

	/**
	 * Closes the potentially open element start tag waiting to know by the
	 * following events if the tag is empty. By this additional event the user
	 * indicates to the serializer that the current element must not be
	 * processed with an open-close tag.
	 */
	public void completeStartTag() {
		serializer.completeStartTag();
	}

	/**
	 * Flushes the user stream.
	 */
	public void flush() {
		serializer.flush();
	}

	/**
	 * Reset internal state before reuse.
	 */
	public void reset() {
		serializer.reset();
	}

	/**
	 * Closes the user stream.
	 */
	public void close() {
		serializer.close();
	}
}