/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

/**
 * An XMLFilter implementation used to perform empty character events removal
 * and consecutive character events aggregation.
 */
public class CharNormalizeFilter extends DefaultXMLFilter
{
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
    private boolean lastEventIsChar = false;
    private StringBuffer chars = new StringBuffer();
    private boolean removeWhitespace = false;   // if true removes all character data containing only whitespaces

    /** Constructor
     * @param reader The parent XMLReader that is filtered.
     * @param removeWhitespace If true, packets of characters containing only
     * whitespace characters are removed. On the other
     * hand, aggregation is always active.
     */
    public CharNormalizeFilter(XMLReader reader,boolean removeWhitespace) {
        super(reader);
        this.removeWhitespace = removeWhitespace;
    }

    ////////////////////////////////////////////////////////////////////
    //    Implementation of LexicalHandler interface.
    ////////////////////////////////////////////////////////////////////
    public void comment (char ch[], int start, int length)
      throws SAXException
    {
        saveWaitingChars();
        super.comment(ch, start, length);
    }

    ////////////////////////////////////////////////////////////////////
    //    Implementation of ContentHandler interface.
    ////////////////////////////////////////////////////////////////////
    public void startElement(String namespaceURI, String localName, String qName, Attributes atts)
    throws SAXException
    {
        saveWaitingChars();
        super.startElement(namespaceURI, localName, qName, atts);
    }
    
    public void endElement (String uri, String localName, String rawName)
    throws SAXException
    {
        saveWaitingChars();
        super.endElement(uri, localName, rawName);
    }

    public void processingInstruction (String target, String data)
    throws SAXException
    {
        saveWaitingChars();
        super.processingInstruction(target, data);
    }
    
    public void characters (char ch[], int start, int length)
    throws SAXException
    {
        lastEventIsChar = true;
        
        // removes (if option set) white space data passed by the parser because there is no DTD
        if (removeWhitespace && isWhitespace(ch, start, length))
            return;
        
        // Store the characters in the StringBuffer
        chars.append(ch, start, length);
    }
    
    ////////////////////////
    //  Private Utilities
    ////////////////////////
    
    private void saveWaitingChars() throws SAXException
    {
      if (chars.length() > 0) {
        char array[] = chars.toString().toCharArray();
        super.characters(array, 0, array.length);
        
        lastEventIsChar = false;
        chars.setLength(0);
      }
    }
    
    /**
     * Test if the character array is whitespace according to XML 1.0 specifications.
     *
     * @return the result of the test.
     */
    private boolean isWhitespace(char ch[], int start, int length)
    {
        int i, max = start+length;
        for (i = start; i < max; i++)
        {
            switch (ch[i])
            {
                case 0x20 : case 0x9 : case 0xD : case 0xA :
                    break;
                default :
                    return false;
            }
        }
        return true; // Only found WS : text ignored
    }
    
}
