/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import java.io.IOException;
import java.util.Iterator;

import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;

/**
 * Define a DOM-tree walker that produces SAX events.
 *
 * <p><B> WARNING : DOM must support namespaces !!!</p>
 * <p>Special Feature:</p>
 * <code>http://wwww.xquark.org/sax/features/fragment</code> disables SAX
 * startDocument() and endDocument() events triggering
 *
 * @see org.xml.sax.XMLReader, org.w3c.dom.Element
 */
public class DOMReader implements XMLReader, Locator, SAXConstants
{
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name:  $";
    
    protected boolean fragment = false;
    /**
     * The DOM document or element that will be parsed.
     */
    protected Node root;
    
    /**
     * The actual ContentHandler
     */
	protected ContentHandler contentHandler = null;
    
    /**
     * The actual DTDHandler
     */
	protected DTDHandler dtdHandler = null;
    
    /**
     * The actual ErrorHandler
     */
	protected ErrorHandler errorHandler = null;
    
    /**
     * The actual EntityResolver
     */
	protected EntityResolver entityResolver = null;
    
    /**
     * The actual LexicalHandler
     */
	protected LexicalHandler lexicalHandler = null;
    
    /**
     * The DOM to SAX wrapper used for attributes
     */
    protected DOM2SAXAttrs wAtts = new DOM2SAXAttrs();
    
    public DOMReader(){}
    
    public DOMReader(Document doc)
    {setDocument(doc);}    // SR 25/08/2000 Better customize contructor than overload XMLReader interface.
    
    public DOMReader(Element root)
    {setRoot(root);}    // SR 25/08/2000 Better customize contructor than overload XMLReader interface.
    
    public void setDocument(Document doc)
    {this.root = doc;}
    
    public void setRoot(Node root)
    {this.root = root;}
    
    public ContentHandler getContentHandler()
    {
        return contentHandler;
    }
    
    public DTDHandler getDTDHandler()
    {
        return dtdHandler;
    }
    
    public EntityResolver getEntityResolver()
    {
        return entityResolver;
    }
    
    public ErrorHandler getErrorHandler()
    {
        return errorHandler;
    }
    
    public boolean getFeature(String name)
    throws SAXNotRecognizedException,
    SAXNotSupportedException
    {
        if (name.equals(FRAGMENT_FEATURE))
            return fragment;
        if (name.equals(SAX_NAMESPACE_FEATURE))
            return true;
        if (name.equals(SAX_PREFIX_FEATURE))
            return false;
        throw new SAXNotRecognizedException("Don't recognize feature \""+name+"\"");
    }
    
    /**
     * Return the current lexical handler used by this object
     */
/*	SR 28/08/2000 : Not in XMLReader interface. Use GetProperty and SetProperty
        public LexicalHandler getLexicalHandler()
        {
                return lexicalHandler;
        }
 */
    public Object getProperty(String name)
    throws SAXNotRecognizedException,
    SAXNotSupportedException
    {
        if (name.equals(SAX_LEXICAL_PROPERTY))
            return lexicalHandler;
        else
            throw new SAXNotRecognizedException("Property \""+name+"\" not recognized");
    }
    
    public void parse(InputSource input)    // SR 25/08/2000 To be full SAX XMLReader compliant : see constructor.
    throws SAXException,
    IOException
    {
        // input source is ignored
        parse("");
    }
    
    public void parse(String systemId)
    throws SAXException,
    IOException
    {
        // systemId is ignored
        if ((contentHandler != null) && !fragment)
        {
            contentHandler.setDocumentLocator(this);
            contentHandler.startDocument();
        }
        if (root == null)
            throw new SAXException("A DOM document or element must be provided.");
        else
        {
            switch (root.getNodeType())
            {
                case Node.ELEMENT_NODE:
				// add   06/06/2003
				case Node.TEXT_NODE:
                    parseNode(root);
                    break;
                case Node.DOCUMENT_FRAGMENT_NODE:
                case Node.DOCUMENT_NODE:
                    parseChildNodes(root);
                    break;
                default:
                    throw new SAXException("The root node provided for browsing must be an Element, a Document or a Document fragment.");
            }
            if ((contentHandler != null) && !fragment)
                contentHandler.endDocument();
        }
    }
        
    /**
     * Parse Node childs and report their content in the ContentHandler.
     *
     * @param node the node
     */
    protected void parseChildNodes(Node node) throws SAXException {
        NodeList nl=node.getChildNodes();
        int nbchilds=nl.getLength();
        for (int i=0; i<nbchilds; i++)
            parseNode(nl.item(i));
    }
    
    /**
     * Parse a Node and report is content in the ContentHandler.
     *
     * @param node the node to parse
     */
    protected void parseNode(Node node) throws SAXException {
        switch (node.getNodeType())
        {
            case Node.ELEMENT_NODE:
                if (contentHandler != null) {
                	generateStartElement(node);
                }
                parseChildNodes(node);
                if (contentHandler != null)
                {
                    generateEndElement(node);
                }
                break;
            case Node.PROCESSING_INSTRUCTION_NODE:
                if (contentHandler != null)
                	generateProcessingInstruction(node);
                break;
            case Node.TEXT_NODE:
                if (contentHandler != null)
                	generateText(node);
                break;
            case Node.CDATA_SECTION_NODE:
				if (contentHandler != null)
					generateCDataSection(node);
                break;
            case Node.COMMENT_NODE:
                if (lexicalHandler != null)
                	generateComment(node);
                break;
            case Node.ENTITY_REFERENCE_NODE:
                if (contentHandler != null)
                	generateEntityReference(node);
                break;
            case Node.DOCUMENT_FRAGMENT_NODE:
            case Node.DOCUMENT_NODE:
            case Node.DOCUMENT_TYPE_NODE:
            case Node.ENTITY_NODE:
            case Node.NOTATION_NODE:
            case Node.ATTRIBUTE_NODE:
                break;
        }
    }
    
    protected void generateStartElement(Node node) throws SAXException {
		String uri = node.getNamespaceURI();
		if (uri == null) uri = "";
		String local = node.getLocalName();
		if (local == null) local = "";
		String prefix = node.getPrefix();
		String qname = prefix == null ? local : prefix + ":" +local;
		// OPTIM SR 23/05/2001				DOMAttributes atts=null;
		wAtts.setDOMAttributes(node.getAttributes());
		Iterator it = wAtts.getPrefixIterator(); // OPTIM SR 23/05/2001
		if (it != null)
			while (it.hasNext()) {
				Attr a = (Attr)it.next();
				contentHandler.startPrefixMapping(
				a.getLocalName().equals(XMLNS_PREFIX) ? "": a.getLocalName(), // default prefix
				a.getValue());
			}
		contentHandler.startElement(uri, local, qname, wAtts);
    }
    
    protected void generateEndElement(Node node) throws SAXException {
		String uri = node.getNamespaceURI();
		if (uri == null) uri = "";
		String local = node.getLocalName();
		if (local == null) local = "";
		String prefix = node.getPrefix();
		String qname = prefix == null ? local : prefix + ":" +local;
		contentHandler.endElement(uri, local, qname);
		// Need to reset wAtts after recursive call to parseChildNodes
		wAtts.setDOMAttributes(node.getAttributes());
		Iterator it = wAtts.getPrefixIterator(); // OPTIM SR 23/05/2001
		if (it != null)
			// WARNING : endPrefixMapping ordered has not been reversed
			while (it.hasNext()) {
				Attr a = (Attr)it.next();
				contentHandler.endPrefixMapping(a.getLocalName().equals(XMLNS_PREFIX) ? "": a.getLocalName());
			}
	}

	protected void generateProcessingInstruction(Node node) throws SAXException {    
		contentHandler.processingInstruction(node.getNodeName(), node.getNodeValue());
	}
	
	protected void generateText(Node node)  throws SAXException {
		String value = node.getNodeValue();
		if (value != null)
			contentHandler.characters(value.toCharArray(), 0, value.length());
	}
	
	protected void generateComment(Node node) throws SAXException  { 
		String value = node.getNodeValue();
		if (value != null)
			lexicalHandler.comment(value.toCharArray(), 0, value.length());
	}
	
	protected void generateCDataSection(Node node) throws SAXException  {
		if (lexicalHandler != null)	lexicalHandler.startCDATA();
		String value = node.getNodeValue();
		if (value != null)
			contentHandler.characters(value.toCharArray(), 0, value.length());
		if (lexicalHandler != null)	lexicalHandler.endCDATA();
	}
	
	protected void generateEntityReference(Node node) throws SAXException  {
		contentHandler.skippedEntity(node.getNodeName());
	}
	
    public void setContentHandler(ContentHandler handler)
    {
        contentHandler = handler;
    }
    
    public void setDTDHandler(DTDHandler handler)
    {
        dtdHandler = handler;
    }
    
    public void setEntityResolver(EntityResolver resolver)
    {
        entityResolver = resolver;
    }
    
    public void setErrorHandler(ErrorHandler handler)
    {
        errorHandler = handler;
    }
    
    public void setFeature(String name, boolean value)
    throws SAXNotRecognizedException,
    SAXNotSupportedException
    {
        if (name.equals(FRAGMENT_FEATURE))
            fragment = value;
        else if (name.equals(SAX_NAMESPACE_FEATURE))
        {
            if (!value)
                throw new SAXNotSupportedException
                ("Feature: \""
                + name
                + "="
                + value
                + "\" not supported");
        }
        else if (name.equals(SAX_PREFIX_FEATURE))
        {
            if (value)
                throw new SAXNotSupportedException
                ("Feature: \""
                + name
                + "="
                + value
                + "\" not supported");
        }
        else
            throw new SAXNotRecognizedException("Feature \""+name+"\" not recognized.");
    }
    
    /**
     * Set the current lexical handler
     */
/*	SR 28/08/2000 : Not in XMLReader interface. Use GetProperty and SetProperty
        public void setLexicalHandler (LexicalHandler handler)
        {
                lexicalHandler=handler;
        }
 */
    public void setProperty(String name, Object value)
    throws SAXNotRecognizedException,
    SAXNotSupportedException
    {
        if (name.equals("http://xml.org/sax/properties/lexical-handler"))
            lexicalHandler = (LexicalHandler)value;
        else
            throw new SAXNotRecognizedException("Don't recognize property \""+name+"\"");
    }
    
    public int getColumnNumber()
    { return -1; }
    
    public int getLineNumber()
    { return -1; }
    
    public String getPublicId()
    { return null; }
    
    public String getSystemId()
    { return null; }
}


