/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import java.io.*;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import org.xml.sax.*;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.serialize.XMLSerializer;

/**
 * This handler processes specific XML documents that under the root element
 * have multiple elements. It dispatches these subelements into 2 streams
 * according to the fact that these elements are valid or not. When not, the
 * schema error message is written inline as a comment.
 */
public class DocumentCleaner extends ValidatingSchemaHandler
{
    // output streams
	private OutputStream validOutput = null;
	private OutputStream invalidOutput = null;
	
	private XMLSerializer serializer = null;
	private ByteArrayOutputStream buffer = null;
	
	private byte depth = 0;
	private SAXParseException lastElementException = null;

	public DocumentCleaner(OutputStream validOutput, OutputStream invalidOutput,
						   SchemaValidationContext context, String encoding) {
		super(context);
		this.validOutput = validOutput;
		this.invalidOutput = invalidOutput;
		buffer = new ByteArrayOutputStream();
		try {
			serializer = new XMLSerializer(buffer, encoding);
		}
		catch (UnsupportedEncodingException e) {
		}
		setErrorHandler(new MyErrorHandler());
		setContentHandler(serializer);
		setLexicalHandler(serializer);
	}
	
	public DocumentCleaner(OutputStream validOutput, OutputStream invalidOutput) {
		this(validOutput, invalidOutput, new SchemaValidationContext(), "UTF-8");
	}
	
	public void endDocument() throws SAXException
	{
		super.endDocument();
		// serialize remaining 1st-level element
		flushBuffer2Both();
		try {
            validOutput.close();
			invalidOutput.close();
        }
        catch (IOException e) {
            throw new SAXException("Exception when closing streams", e);
        }
	}
	
	private void flushBuffer2Both() throws SAXException
	{
	    serializer.flush();
		try {
			buffer.writeTo(validOutput);
			buffer.writeTo(invalidOutput);
		}
		catch (IOException e) {
			throw new SAXException("Exception when serializing XML prolog", e);
		}
		buffer.reset();
	}

	public void startElement(String str, String str1, String str2, Attributes attributes)
	throws SAXException
	{
	    depth++;
	    switch (depth)
	    {
	        case 1:
	            super.startElement(str, str1, str2, attributes);
		        serializer.completeStartTag();
				// serialize xml prolog and 1st-level element start
				flushBuffer2Both();
	            break;
	            
	        case 2:
	            lastElementException = null;
	            super.startElement(str, str1, str2, attributes);
	            break;
	            
	        default :
	            super.startElement(str, str1, str2, attributes);
		}
	}
	
	public void endElement(String str, String str1, String str2)
	throws SAXException
	{
		super.endElement(str, str1, str2);
		if (depth== 2){
		    try {
				if (lastElementException == null) {
					serializer.flush();
				    buffer.writeTo(validOutput);
				}
				else {
					StringBuffer errorMsg = new StringBuffer();
					errorMsg.append("\nOriginal location: line ");
					errorMsg.append(lastElementException.getLineNumber());
					errorMsg.append(", column ");
					errorMsg.append(lastElementException.getColumnNumber());
					errorMsg.append('\n');
					errorMsg.append(lastElementException.getMessage());
					errorMsg.append('\n');
					char[] comment = new char[errorMsg.length()];
					errorMsg.getChars(0, comment.length, comment, 0);
					super.comment(comment, 0, comment.length);
					serializer.flush();
				    buffer.writeTo(invalidOutput);
				}
				buffer.reset();
			}
			catch (IOException e) {
				throw new SAXException("Exception when serializing XML prolog", e);
			}
		}
		depth--;
	}

	private class MyErrorHandler implements ErrorHandler {

        public void error(SAXParseException arg0) throws SAXException {
			lastElementException = arg0;
        }

        public void fatalError(SAXParseException arg0) throws SAXException {}
        public void warning(SAXParseException arg0) throws SAXException {}
	    
	}
}
