/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import java.util.ArrayList;

import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;


public class MultiplexerDecorator extends HandlerDecorator {

	private ContentHandler[] contentHandlers;
	private LexicalHandler[] lexicalHandlers;
	private ErrorHandler[] errorHandlers;
	
	public MultiplexerDecorator(ArrayList handlers) {
		ArrayList chList = new ArrayList();
		ArrayList lhList = new ArrayList();
		ArrayList ehList = new ArrayList();
		for (int i = 0; i < handlers.size(); i++) {
			Object handler = handlers.get(i);
			if (handler instanceof ContentHandler)
				chList.add(handler);
			if (handler instanceof LexicalHandler)
				lhList.add(handler);
			if (handler instanceof ErrorHandler)
				ehList.add(handler);
		}
		contentHandlers = (ContentHandler[]) chList.toArray(new ContentHandler[0]);
		lexicalHandlers = (LexicalHandler[]) lhList.toArray(new LexicalHandler[0]);
		errorHandlers = (ErrorHandler[]) ehList.toArray(new ErrorHandler[0]);
	}
	//////////////////////////////////////////////////////
	//      Implementation of ContentHandler interface.
	//////////////////////////////////////////////////////
	public void startDocument() throws SAXException
	{
		super.startDocument();
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].startDocument();
	}
    
	public void startElement(String str, String str1, String str2, Attributes attributes)
	throws SAXException
	{
		super.startElement(str, str1, str2, attributes);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].startElement(str, str1, str2, attributes);
	}
    
	public void setDocumentLocator(Locator locator)
	{
		super.setDocumentLocator(locator);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].setDocumentLocator(locator);
	}
    
	public void endDocument() throws SAXException
	{
		super.endDocument();
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].endDocument();
	}
    
	public void characters(char[] values, int param, int param2)
	throws SAXException
	{
		super.characters(values, param, param2);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].characters(values, param, param2);
	}
    
	public void processingInstruction(String str, String str1)
	throws SAXException
	{
		super.processingInstruction(str, str1);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].processingInstruction(str, str1);
	}
    
	public void endElement(String str, String str1, String str2)
	throws SAXException
	{
		super.endElement(str, str1, str2);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].endElement(str, str1, str2);
	}
    
	public void startPrefixMapping(String str, String str1) throws SAXException
	{
		super.startPrefixMapping(str, str1);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].startPrefixMapping(str, str1);
	}
    
	public void ignorableWhitespace(char[] values, int param, int param2)
	throws SAXException
	{
		super.ignorableWhitespace(values, param, param2);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].ignorableWhitespace(values, param, param2);
	}
    
	public void endPrefixMapping(String str)  throws SAXException
	{
		super.endPrefixMapping(str);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].endPrefixMapping(str);
	}
    
	public void skippedEntity(String str) throws SAXException
	{
		super.skippedEntity(str);
		for (int i = 0; i < contentHandlers.length; i++) 
			contentHandlers[i].skippedEntity(str);
	}
    
	//////////////////////////////////////////////////////
	//      Implementation of LexicalHandler interface.
	//////////////////////////////////////////////////////
	public void startDTD(String name, String publicId,  String systemId)
	throws SAXException
	{
		super.startDTD(name, publicId, systemId);
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].startDTD(name, publicId, systemId);
	}
    
	public void endDTD() throws SAXException
	{
		super.endDTD();
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].endDTD();
	}
    
	public void startEntity(String name) throws SAXException
	{
		super.startEntity(name);
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].startEntity(name);
	}
    
	public void endEntity(String name) throws SAXException
	{
		super.endEntity(name);
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].endEntity(name);
	}
    
	public void startCDATA() throws SAXException
	{
		super.startCDATA();
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].startCDATA();
	}
    
	public void endCDATA() throws SAXException
	{
		super.endCDATA();
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].endCDATA();
	}
    
	public void comment(char ch[], int start, int length) throws SAXException
	{
		super.comment(ch, start, length);
		for (int i = 0; i < lexicalHandlers.length; i++) 
			lexicalHandlers[i].comment(ch, start, length);
	}
    
	//////////////////////////////////////////////////////
	//      Implementation of EventHandler interface.
	//////////////////////////////////////////////////////
	/**
	 * Filter a warning event.
	 *
	 * @param e The warning as an exception.
	 * @exception org.xml.sax.SAXException The client may throw
	 *            an exception during processing.
	 * @see org.xml.sax.ErrorHandler#warning
	 */
	public void warning(SAXParseException e)
	throws SAXException
	{
		super.warning(e);
		for (int i = 0; i < errorHandlers.length; i++) 
			errorHandlers[i].warning(e);
	}
    
    
	/**
	 * Filter an error event.
	 *
	 * @param e The error as an exception.
	 * @exception org.xml.sax.SAXException The client may throw
	 *            an exception during processing.
	 * @see org.xml.sax.ErrorHandler#error
	 */
	public void error(SAXParseException e)
	throws SAXException
	{
		super.error(e);
		for (int i = 0; i < errorHandlers.length; i++) 
			errorHandlers[i].error(e);
	}
    
    
	/**
	 * Filter a fatal error event.
	 *
	 * @param e The error as an exception.
	 * @exception org.xml.sax.SAXException The client may throw
	 *            an exception during processing.
	 * @see org.xml.sax.ErrorHandler#fatalError
	 */
	public void fatalError(SAXParseException e)
	throws SAXException
	{
		super.fatalError(e);
		for (int i = 0; i < errorHandlers.length; i++) 
			errorHandlers[i].fatalError(e);
	}

}
