/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import java.util.Iterator;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;

/**
 * A ContentHandler and LexicalHandler implementation used to perform qualified
 * SAX names (rawnames) reconstruction from prefix mapping events, intended
 * repository only support the default values. It also performs xmlns attributes
 * reconstruction.
 */
public class NSPrefixDecorator extends NamespaceContextHandler implements SAXConstants
{
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName     = "$Name:  $";
    
    private boolean enabled = false;
    private AttributesImpl attributes = null;
    
    public final static String NS_PREFIX = "xmlns";
    public final static String XML_PREFIX = "xml";
    
    public NSPrefixDecorator()
    {}
    
    private String getQualifiedName(String uri, String localName, String qName) throws SAXException
    {
        if (localName == null || localName.length() == 0)
            return qName;
        String prefix = getPrefix(uri); // ATTENTION : No guaranty it is the last prefix.
        if (prefix == null || prefix.length() == 0)
            return localName;
        else
            return prefix + ":" + localName;
    }
    
    public final void startElement(String namespaceURI, String localName, String qName, Attributes atts)
    throws SAXException
    {
        if (enabled)
        {
            attributes.clear();
            
            /* add xmlns attributes */
            Iterator enum = getDeclaredPrefixes().iterator();
            String uri;
            String prefix;
            while (enum.hasNext())
            {
                prefix = (String)enum.next();
                if (!prefix.equals(XML_PREFIX))
                {
                    uri = getNamespaceURI(prefix);
                    attributes.addAttribute(XMLNS_URI, prefix, prefix.length() == 0 ? NS_PREFIX : NS_PREFIX+":"+prefix, "CDATA", uri);
                }
            }
            
            int len = atts.getLength();
            for (int i=0; i<len; i++)
            {
                attributes.addAttribute(
                atts.getURI(i),
                atts.getLocalName(i),
                getQualifiedName(atts.getURI(i), atts.getLocalName(i), atts.getQName(i)), // browsing must be performed for this reason
                atts.getType(i),
                atts.getValue(i));
            }
            
            super.startElement(namespaceURI, localName, getQualifiedName(namespaceURI, localName, qName), attributes);
        }
        else
            super.startElement(namespaceURI, localName, qName, atts);
    }
    
    public final void endElement(String namespaceURI, String localName, String qName)
    throws SAXException
    {
        if (enabled)
            super.endElement(namespaceURI, localName, getQualifiedName(namespaceURI, localName, qName));
        else
            super.endElement(namespaceURI, localName, qName);
    }
    /**
     * True if raw names reconstruction is enabled.
     * @return
     */
    public boolean isEnabled() {
        return enabled;
    }

    /**
     * Enables raw names reconstruction.
     * @param enabled
     */
    public void setEnabled(boolean enabled) {
    	if (enabled && attributes == null)
			attributes = new AttributesImpl();
        this.enabled = enabled;
    }

}
