/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

/*
 * NSSupportFilter.java
 *
 * Created on 12 novembre 2001, 21:53
 */

package org.xquark.util;

import java.util.*;

/** 
 * Helper for XML prefix declaration management
 */
public class NamespaceContextStack implements NamespaceContext
{
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";

	/**
	 *	Stack of NSContextInfo objects
	 */
	private ArrayList contexts = new ArrayList();
	/**
	 *
	 */
	private NSContextInfo current = null;
	/**
	 * Feature the XML depth (in fact the number of calls to pushContext()
	 */
	private int level = 0;
	/**
	 * The stack size (the number of times a real push() was performed after
	 * a pushContext() was called assumed one is performed at the biginning
	 * for xml).
	 */
	private int count = 0;

	public NamespaceContextStack()
	{
		reset();
	}

	public void reset()
	{
		level = 0;
		count = 0;
		current = push();
		current.declarePrefix("xml", "http://www.w3.org/XML/1998/namespace");
	}

	public void declarePrefix(String prefix, String uri)
	{
		if (current.getLevel() < level)
			current = push();
		current.declarePrefix(prefix, uri);
	}

	public void pushContext()
	{
		level++;
	}

	public void popContext()
	{
		level--;
		if (current.getLevel() > level)
			current = pop();
	}

	///////////////////////////////////////////////////////////////////////////
	//                              PRIVATE
	///////////////////////////////////////////////////////////////////////////
	private NSContextInfo push()
	{
		NSContextInfo context = null;
		if (count == contexts.size())
		{
			// allocate new context
			context = new NSContextInfo();
			contexts.add(context);
		}
		else
		{
			// reuse a previously allocated context
			context = (NSContextInfo) contexts.get(count);
		}
		count++;
		context.initialize(level, current);
		return context;
	}

	private NSContextInfo pop()
	{
		// hack to avoid popping the last context
		if (count > 1)
			count--;
		else
			reset();
		return (NSContextInfo) contexts.get(count - 1);
	}

	public NamespaceContext getNamespaceContext()
	{
		return current.getContext(true);
	}

	public String getNamespaceURI(String prefix)
	{
		return current.getContext().getNamespaceURI(prefix);
	}
	
	public Collection getNamespaceURIs() {
		return current.getContext().getNamespaceURIs();
	}

	public String getPrefix(String uri)
	{
		return current.getContext().getPrefix(uri);
	}

	public List getDeclaredPrefixes()
	{
		return current.getDeclaredPrefixes(level);
	}

	public Collection getPrefixes()
	{
		return current.getContext().getPrefixes();
	}

	public Collection getPrefixes(String uri)
	{
		return current.getContext().getPrefixes(uri);
	}

	///////////////////////////////////////////////////////////////////////////
	//                              PRIVATE
	///////////////////////////////////////////////////////////////////////////
	static class NSContextInfo
	{
		private static final String RCSRevision = "$Revision: 1.1 $";
		private static final String RCSName = "$Name:  $";
		static private final List EMPTY_LIST = new ArrayList(1);

		private NamespaceContextImpl context = new NamespaceContextImpl();
		private int level = -1;
		private boolean locked = false;

		void initialize(int level, NSContextInfo parent)
		{
			this.level = level;
			if (parent != null)
			{
				if (locked)
					context = new NamespaceContextImpl();
				context.initialize(parent.getContext());
			}
			locked = false;
		}

		int getLevel()
		{
			return level;
		}

		NamespaceContextImpl getContext()
		{
			return getContext(false);
		}

		NamespaceContextImpl getContext(boolean lock)
		{
			if (!locked)
				locked = lock;
			return context;
		}

		void declarePrefix(String prefix, String uri)
		{
			context.declarePrefix(prefix, uri);
		}

		public List getDeclaredPrefixes(int currentLevel)
		{
			if (currentLevel == level)
				return context.getDeclaredPrefixes();
			else
				return EMPTY_LIST;
		}
	}

	static class NamespaceContextImpl implements NamespaceContext
	{
		private static final String RCSRevision = "$Revision: 1.1 $";
		private static final String RCSName = "$Name:  $";
		private HashMap prefixMap = new HashMap();
		private HashMap namespaceMap = new HashMap();
		private ArrayList declarations = new ArrayList();

		void initialize(NamespaceContextImpl parent)
		{
			prefixMap.clear();
			namespaceMap.clear();
			declarations.clear();
			prefixMap.putAll(parent.getPrefixMap());
			namespaceMap.putAll(parent.getNamespaceMap());
		}

		HashMap getPrefixMap()
		{
			return prefixMap;
		}

		HashMap getNamespaceMap()
		{
			return namespaceMap;
		}

		void declarePrefix(String prefix, String uri)
		{
			prefixMap.put(prefix, uri);
			namespaceMap.put(uri, prefix);
			declarations.add(prefix);
		}

		public String getNamespaceURI(String prefix)
		{
			return (String) prefixMap.get(prefix);
		}

		public String getPrefix(String uri)
		{
			return (String) namespaceMap.get(uri);
		}

		public Collection getPrefixes()
		{
			return Collections.unmodifiableCollection(prefixMap.keySet());
		}

		public Collection getNamespaceURIs() {
			return Collections.unmodifiableCollection(namespaceMap.keySet());
		}

		public List getDeclaredPrefixes() {
			return Collections.unmodifiableList(declarations);
		}
		
		public Collection getPrefixes(String uri)
		{
			ArrayList result = new ArrayList();
			Iterator it = prefixMap.entrySet().iterator();
			while (it.hasNext())
			{
				Map.Entry entry = (Map.Entry) it.next();
				if (uri.equals(entry.getValue()))
					result.add(entry.getKey());
			}
			return result;
		}
	}

}
