/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

import java.io.PrintStream;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class describes a XMLDBC (Xml DataBase Connectivity) exception for
 * various use in XMLDBC API. <BR>
 * All other specific XMLDBC exceptions extend this exception. <BR>
 * <BR>
 */
public class XMLDBCException extends java.lang.Exception {

	//*************************** Constants declaration:

	/** Error code for exception that have no associated error code. */
	public static final int NO_CODE = 0;

	//*************************** Properties declaration:

	/** The error code of this exception. */
	protected int errorCode = NO_CODE;

	/** The underlying exception (or error) of this exception. */
	protected java.lang.Throwable exception;

	//*************************** Constructor declaration:

	/**
	 * Constructor with error code.
	 * 
	 * @param code
	 *            the code featuring the kind of error (see constants).
	 */
	public XMLDBCException(int code) {
		this(code, "", null);
	}

	/**
	 * Constructor with message.
	 * 
	 * @param msg
	 *            a message to describe the exception.
	 */
	public XMLDBCException(java.lang.String msg) {
		this(XMLDBCException.NO_CODE, msg, null);
	}

	/**
	 * Constructor with error code and message.
	 * 
	 * @param code
	 *            the code featuring the kind of error (see constants).
	 * @param msg
	 *            a message to describe the exception.
	 */
	public XMLDBCException(int code, java.lang.String msg) {
		this(code, msg, null);
	}

	/**
	 * Constructor with message and underlying exception.
	 * 
	 * @param msg
	 *            a message to describe the exception.
	 * @param exception
	 *            the underlying exception (or error).
	 */
	public XMLDBCException(java.lang.String msg, java.lang.Throwable exception) {
		this(XMLDBCException.NO_CODE, msg, exception);
	}

	/**
	 * Constructor with error code, message and underlying exception.
	 * 
	 * @param code
	 *            the code featuring the kind of error (see constants).
	 * @param msg
	 *            a message to describe the exception.
	 * @param exception
	 *            the underlying exception (or error).
	 */
	public XMLDBCException(int code, java.lang.String msg,
			java.lang.Throwable exception) {
		super(msg);
		this.errorCode = code;
		this.exception = exception;
	}

	//*************************** Other methods declaration:

	/**
	 * To get the code of this exception/error.
	 * 
	 * @return the error code.
	 */
	public int getCode() {
		return this.errorCode;
	}

	/**
	 * To get the exception of this exception/error.
	 * 
	 * @return the underlying exception.
	 */
	public java.lang.Throwable getException() {
		return this.exception;
	}

	/**
	 * Returns the exception message, or the embedded exception message if none
	 * is present
	 * 
	 * @return the message
	 */
	public String getMessage() {
		String message = super.getMessage();
		if ((message == null || message.length() == 0) && exception != null)
			message = this.exception.getMessage();

		return message;
	}

	/**
	 * Print an error message processing embedded expeptions to the standard
	 * error output.
	 */
	// TODO: put it in a common tools library ?
	public static void printError(Throwable e, boolean pst) {
		printError(e, System.err, pst);
	}
	
	/**
	 * Print an error message processing embedded expeptions.
	 */
	// TODO: put it in a common tools library ?
	public static void printError(Throwable e, PrintStream ps, boolean pst) {
		if (e != null) {
			if (e.getMessage() != null) {
				/* print exception message */
				ps.println("Exception: " + e.getMessage());
			}

			/* printing location info */
			if (e instanceof SAXParseException) {
				SAXParseException spex = (SAXParseException) e;
				ps.println("XML error location : " + spex.getSystemId() + " ["
						+ spex.getLineNumber() + ':' + spex.getColumnNumber()
						+ ']');
			}
			/* Print stack trace */
			if (pst)
				e.printStackTrace(ps);

			/* underlying exceptions */
			if (e instanceof XMLDBCException)
				printError(((XMLDBCException) e).getException(), ps, pst);
			else if (e instanceof SAXException)
				printError(((SAXException) e).getException(), ps, pst);
		}
	}

}