/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

import java.io.Writer;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.ext.LexicalHandler;

/**
 * This interface describes a document object which provides access to XML data
 * in several ways. When the underlying object is a stored document, it also
 * supports a number of document management methods also available in the
 * XMLCollection interface.<BR><BR>
 */
public interface XMLDocument
{


    
    /**
     * Returns the document identifier in the XML collection.
     * return the document identifier as a String, or null if the document is not a stored
     * object.
     */
    public String getIdentifier();
    
    /**
     * Changes the document identifier in the XML collection. This change is persistent in the
     * data source.
     * Note: Depending on the auto-commit mode of the underlying data source connection, the
     * change may only become definitive after a commit operation.
     * @param identifier the new document identifier
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the renaming operation is not supported.
     */
    public void setIdentifier(String identifier) throws XMLDBCException, XMLDBCNotSupportedException;
    
    /**
     * Returns the XML collection to which this document belongs. This may be null if
     * the document is not a stored object.
     * @return an XMLCollection object, or null if the document is not a stored object.
     */
    public XMLCollection getCollection();
    
    /**
     * Removes the document from the collection to which it belongs. The object should
     * not be accessed any more after this method is called.
     * Note: Depending on the auto-commit mode of the underlying data source connection, the
     * change may only become definitive after a commit operation.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the remove operation is not supported.
     */
    public void remove() throws XMLDBCException, XMLDBCNotSupportedException;
    
    /**
     * Sets a (SAX2) content handler to intercept events produced when retrieving
     * the document content.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the content handler implementation (see org.xml.sax.ContentHandler)
     */
    public void setContentHandler(ContentHandler handler);
    
    /**
     * Sets a (SAX2) lexical handler to intercept events produced when retrieving
     * the document content.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the lexical handler implementation (see org.xml.sax.ext.LexicalHandler)
     */
    public void setLexicalHandler(LexicalHandler handler);
    
    /**
     * Sets a (SAX2) error handler to intercept error events produced when retrieving
     * the document content.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the error handler implementation (see org.xml.sax.ErrorHandler)
     */
    public void setErrorHandler(ErrorHandler handler);
    
    /**
     * Retrieves the current (SAX2) content handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @return the content handler implementation (see org.xml.sax.ContentHandler)
     */
    public ContentHandler getContentHandler();
    
    /**
     * Retrieves the current (SAX2) lexical handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @return the lexical handler implementation (see org.xml.sax.ext.LexicalHandler)
     */
    public LexicalHandler getLexicalHandler();
    
    /**
     * SRetrieves the current (SAX2) error handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @return the error handler implementation (see org.xml.sax.ErrorHandler)
     */
    public ErrorHandler getErrorHandler();
    
    /**
     * Returns the document content as a DOM2 Document.
     * @return the DOM2 Document containing the document data.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public Document getAsDocument() throws XMLDBCException;
    
    /**
     * Returns the document content as a DOM2 Document.
     * @return the DOM2 Document containing the document data.
     * @throws XMLDBCException if a data source access error occurs.
     * @deprecated New method name is {@link #getAsDocument()}
     */
    public Document getAsDOM() throws XMLDBCException;
    
    /**
     * Attaches a DOM2 fragment corresponding to the XML document to the Element
     * node passed as a parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void getAsDOM(Element parent) throws XMLDBCException;
    
    /**
     * Returns the document content as an XML string.
     * @return the XML string containing the document data.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String getAsString() throws XMLDBCException;
    
    /**
     * Returns the document content in a provided character stream.
     * @param out a {@link java.io.Writer Writer} object assuming user is
     * responsible for the character encoding.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void getAsStream(Writer out) throws XMLDBCException;
    
    /**
     * Returns the document content as a SAX2 event flow.
     * This method uses the various SAX2 handlers set by the setXXXHandler() methods.
     * <P>*** ONLY FOR SAX2 ***</P>
     * @throws XMLDBCException if a data source access error occurs
     * or if the user has not set a content handler.
     * @throws org.xml.sax.SAXException if a handler exception occurs.
     */
    public void getAsSAX() throws XMLDBCException, org.xml.sax.SAXException;
}
