/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

import java.io.Writer;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.schema.validation.PSVInfoSetProvider;

/**
 * This interface represents the set of results of a query.<BR>
 * This interface has setContentHandler(...), setLexicalHandler(...) and setErrorHandler(...)
 * methods allowing the client to get the results as SAX2 events.<BR><BR>
 */
public interface XMLResultSet
{
    /*********************************************************************/
    /*************************** SAX management functions ****************/
    /*********************************************************************/
    
    /**
     * Sets a (SAX2) content handler to intercept events produced when retrieving results as SAX.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the content handler implementation (see org.xml.sax.ContentHandler)
     */
    public void setContentHandler(ContentHandler handler);
    
    /**
     * Sets a (SAX2) lexical handler to intercept events produced when retrieving results as SAX.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the lexical handler implementation (see org.xml.sax.ext.LexicalHandler)
     */
    public void setLexicalHandler(LexicalHandler handler);
    
    /**
     * Sets a (SAX2) error handler to intercept error events produced when retrieving results as SAX.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the error handler implementation (see org.xml.sax.ErrorHandler)
     */
    public void setErrorHandler(ErrorHandler handler);
    
    /**
     * Returns the current content handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     */
    public ContentHandler getContentHandler();
    
    /**
     * Returns the current lexical handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     */
    public LexicalHandler getLexicalHandler();
    
    /**
     * Returns the current error handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     */
    public ErrorHandler getErrorHandler();
    
    /*********************************************************************/
    /*************************** ResultSet management ********************/
    /*********************************************************************/
    
    /**
     * Returns the statement that produced this result set.
     * @return the XStatement object that produced this XResultSet object.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLStatement getStatement() throws XMLDBCException;
    
    /**
     * Returns the result set metadata (the XML Schema that models the returned data)
     * as an XMLDocument.
     * This schema describes the XML type of each result in the set.
     * @return the result set metadata (a XML Schema).
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDocument getMetaData() throws XMLDBCException;
    
    /**
     * Returns the result set prefix map (as prefix-namespace pairs).
     * This map is used to produce each result in the set.
     * @return the result set prefix map.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public Map getPrefixMap() throws XMLDBCException;
    
    /**
     * Releases this result set object's resources immediately instead of waiting
     * for this to happen when it is automatically closed.<BR>
     * <B>Note:</B> A XResultSet object is automatically closed by the XStatement
     * object that generated it when that XStatement object is closed,
     * re-executed, or is used to retrieve the next result from a sequence of
     * multiple results.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void close() throws XMLDBCException;
    
    /**
     * Indicates whether the cursor is before the first result in this result set.
     * @return true if the cursor is before the first result; false otherwise.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean isBeforeFirst() throws XMLDBCException;
    
    /**
     * Returns the next result as a DOM2 document and increments by 1 the cursor
     * position.
     * The cursor is initially positioned before the first result; the first
     * call to the method nextAsDOM() makes the first result the current result.
     * @return the next result (as a DOM level 2 document) of the result set or null
     * if there is no more result in the result set.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public Document nextAsDocument() throws XMLDBCException;
    
    /**
     * Returns the next result as a DOM2 node and increments by 1 the cursor position.
     * The cursor is initially positioned before the first result; the first
     * call to the method nextAsDOM() makes the first result the current result.
     * @deprecated Prefer {@link #nextAsDocument()} or {@link #nextAsDOM(Element)}
     * becuse a result may return multiple root nodes.
     * @return the next result (as a DOM level 2 node) of the result set or
     * null if there is no more result in the result set.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public Node nextAsDOM() throws XMLDBCException;
    
    /**
     * Attaches the next result nodes as children of the Element node passed
     * as a parameter to and increments by 1 the cursor position.
     * The cursor is initially positioned before the first result; the first
     * call to the method nextAsDOM() makes the first result the current result.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void nextAsDOM(Element parent) throws XMLDBCException;
    
    /**
     * Returns the next result as an XML string and increments by 1 the cursor position.
     * The cursor is initially positioned before the first result; the first
     * call to the method nextAsString() makes the first result the current result.
     * @return the next result (as an XML string) of the result set or null if
     * there is no more result in the result set.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String nextAsString() throws XMLDBCException;
    
    /**
     * Returns the next result in a provided character stream and increments
     * by 1 the cursor position.
     * The cursor is initially positioned before the first result; the first
     * call to the method nextAsString() makes the first result the current result.
     * @param out a {@link java.io.Writer Writer} object assuming user is
     * responsible for the character encoding.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void nextAsStream(Writer out) throws XMLDBCException;
    
    /**
     * Returns the next result as a SAX2 event flow and increments by 1 the cursor position.
     * The cursor is initially positioned before the first result; the first
     * call to the method nextAsSAX makes the first result the current result.
     * This method uses the various SAX2 handlers set by the setXXXHandler() methods.
     * <P>*** ONLY FOR SAX2 ***</P>
     * @throws XMLDBCException if a data source access error occurs
     * or if user has not set a content handler.
     * @throws org.xml.sax.SAXException if a handler exception occurs.
     */
    public void nextAsSAX() throws XMLDBCException, org.xml.sax.SAXException;
    
    /**
     * Indicates if there is a next Result after the current Result in ResultSet.
     * @return true if there is a next Result, false otherwise.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean hasNext() throws XMLDBCException;
    
    /**
     * Retrieves the current cursor position in ResultSet. The first position is number 1.
     * @return the current cursor position in ResultSet (-1 if the cursor is not on a Result).
     * @throws XMLDBCException if a data source access error occurs.
     */
    public int getPosition() throws XMLDBCException;
    
    /**
     * Returns the fragments contained in the result set as a single XML document, with
     * a document root element having the specified namespace, local name and qualified name.
     * Note: this convenience method can only be called if the cursor is before the
     * first element. It will process the complete result set in one step.
     * @param namespace the namespace of the document root element.
     * @param localName the local name of the document root element.
     * @param qName the qualified name of the document root element.
     * @return an XML document containing the fragments of the result set, as children of
     * the specified document root element.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDocument getFragmentsAsDocument(String namespace, String localName, String qName)
    throws XMLDBCException;

	/**
	 * 
	 * @return true if a root tag is present, false otherwise.
	 */
	public boolean hasRootTag();
	
	/**
	 * 
	 * @return true if a document is present, false otherwise.
	 */
	public boolean isDocument();
	
	/**
	 * 
	 * @return the PSVInfoSetProvider if present null otherwise
	 */
	public PSVInfoSetProvider getPSVInfoSetProvider();
    
}
