/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2004 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

import org.xml.sax.InputSource;
import org.xquark.schema.SchemaManager;
import org.xquark.xquery.metadata.StaticContext;
import org.xquark.xquery.metadata.VarCounter;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.typing.TypeVisitor;

public class ModuleManager extends DefaultModuleLocator {
	private static final String RCSRevision = "$Revision: 1.4 $";
	private static final String RCSName = "$Name:  $";


	// XQueryModule cache based on namespaces
	private HashMap modules = new HashMap();
	private boolean useCache = true;

	private StaticContext staticContext = null;

	public ModuleManager() {
		this(null, true);
	}

//	public ModuleManager(boolean useCache) {
//		this(null, useCache);
//	}

	public ModuleManager(ModuleManager parent) {
		this(parent, true);
	}

	public ModuleManager(ModuleManager parent, boolean useCache) {
		this.useCache = useCache;
		if (parent != null) {
			Iterator it = parent.getModules().iterator();
			while (it.hasNext()) {
				XQueryModule s = (XQueryModule) it.next();
				// start add temporary  
				if (s != null)
					// end add temporary   
					modules.put(s.getNamespace(), s);
			}
		}
	}

	public void setStaticContext(StaticContext staticContext) {
		this.staticContext = staticContext;
	}

	public boolean isRegistered(String namespace) {
		return modules.containsKey(namespace);
	}

	public XQueryModule getModule(String namespace) {
		return (XQueryModule) modules.get(namespace);
	}

	public Collection getModules() {
		return modules.values();
	}

	public XQueryModule removeModule(String namespace) {
		return (XQueryModule) modules.remove(namespace);
	}

	private void checkNamespace(String ns, XQueryModule module) throws XQueryException {
		if ((ns != null && !ns.equals(module.getNamespace())) || (ns == null && module.getNamespace() != null))
			throw new XQueryException("Loaded module target namespace " + module.getNamespace() + " does not match expected URI: " + ns);
	}

	public XQueryModule loadModule(String namespace, SchemaManager sc, TypeVisitor typeVisitor) throws XQueryException {
		return loadModule(namespace, sc, typeVisitor, new VarCounter());
	}
	public XQueryModule loadModule(String namespace, SchemaManager sc, TypeVisitor typeVisitor, VarCounter varCounter) throws XQueryException {
		if (namespace.startsWith("java:")) {
			return new JavaModule(namespace.substring(5), typeVisitor);
		} else if (namespace.startsWith("wsdl:")) {
			return new WSDLModule(namespace.substring(5));
		}
		return loadModule(this, namespace, sc, varCounter);
	}

	public XQueryModule loadModule(String namespace, SchemaManager sc) throws XQueryException {
		return loadModule(namespace, sc, new VarCounter());
	}
	public XQueryModule loadModule(String namespace, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		return loadModule(this, namespace, sc, varCounter);
	}

	public XQueryModule loadModule(ModuleLocator locator, String namespace, SchemaManager sc) throws XQueryException {
		return loadModule(locator, namespace, sc, new VarCounter());
	}
	public XQueryModule loadModule(ModuleLocator locator, String namespace, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		// Look into module cache
		if (isRegistered(namespace))
			return getModule(namespace);

		XQueryModule result = null;
		InputSource source = null;
		// Load module from its locations as strings
		Iterator it1 = locator.getModuleLocations(namespace);
		if (it1 != null) {
			while (it1.hasNext()) {
				source = locator.resolveLocation(namespace, (String) it1.next());
				if (source != null)
					result = loadModule(locator, namespace, source, sc, varCounter);
				if (result != null)
					return result;
			}
		}

		// Load module from its URI as string
		source = locator.resolveLocation(namespace, namespace);
		if (source != null)
			result = loadModule(locator, namespace, source, sc, varCounter);
		if (result != null)
			return result;

		// Load module from its locations as URLs
		Iterator it2 = locator.getModuleLocations(namespace);
		if (it2 != null) {
			while (it2.hasNext()) {
				try {
					URL url = new URL((String) it2.next());
					source = locator.resolveLocation(namespace, url);
					if (source != null)
						result = loadModule(locator, namespace, source, sc, varCounter);
					if (result != null)
						return result;
				} catch (MalformedURLException ex) {
					// continue
				}
			}
		}

		// result is null
		throw new XQueryException("Could not find module "+namespace);
	}

	private XQueryModule loadModule(ModuleLocator locator, String namespace, InputSource source, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		// Look into module cache
		if (isRegistered(namespace))
			return getModule(namespace);

		XQueryModule result = null;

		try {
			XQueryParser parser = new XQueryParser(source.getByteStream());
			// parsing request
			parser.setBaseURI(source.getSystemId());
			result = parser.Start(staticContext.getMetaData(), sc, this, varCounter);
			checkNamespace(namespace, result);
		} catch (ParseException ex) {
			throw new XQueryException(ex.getMessage(), ex);
		}
		// Add module to namespace cache
		modules.put(namespace, result);
		return result;
	}
	private XQueryModule loadModule(ModuleLocator locator, InputSource source, MetadataAccess metadata, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		XQueryModule result = null;

		try {
			XQueryParser parser = new XQueryParser(source.getByteStream());
			// parsing request
			parser.setBaseURI(source.getSystemId());
			// test staticContext
			result = parser.Start(metadata, sc, this, varCounter);
		} catch (ParseException ex) {
			throw new XQueryException(ex.getMessage(), ex);
		}
		// Add module to namespace cache
		String namespace = result.getNamespace();
		modules.put(namespace, result);
		return result;
	}

	public XQueryModule loadModule(InputSource source, MetadataAccess metadata, SchemaManager sc) throws XQueryException {
		return loadModule(this, source, metadata, sc, new VarCounter());
	}
	public XQueryModule loadModule(String namespace, InputSource source, SchemaManager sc) throws XQueryException {
		return loadModule(this, namespace, source, sc, new VarCounter());
	}
	public XQueryModule loadModule(String namespace, InputSource source, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		return loadModule(this, namespace, source, sc, varCounter);
	}

	public XQueryModule loadModule(ModuleLocator locator, String namespace, URL location, SchemaManager sc) throws XQueryException {
		return loadModule(locator, namespace, locator.resolveLocation(namespace, location), sc, new VarCounter());
	}
	public XQueryModule loadModule(ModuleLocator locator, String namespace, URL location, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		return loadModule(locator, namespace, locator.resolveLocation(namespace, location), sc, varCounter);
	}

	public XQueryModule loadModule(String namespace, URL location, SchemaManager sc) throws XQueryException {
		return loadModule(this, namespace, location, sc);
	}
	public XQueryModule loadModule(String namespace, URL location, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		return loadModule(this, namespace, location, sc, varCounter);
	}

	public XQueryModule loadModule(ModuleLocator locator, String namespace, String location, SchemaManager sc) throws XQueryException {
		return loadModule(locator, namespace, locator.resolveLocation(namespace, location), sc, new VarCounter());
	}
	public XQueryModule loadModule(ModuleLocator locator, String namespace, String location, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		return loadModule(locator, namespace, locator.resolveLocation(namespace, location), sc, varCounter);
	}

	public XQueryModule loadModule(String namespace, String location, SchemaManager sc) throws XQueryException {
		return loadModule(this, namespace, location, sc);
	}
	public XQueryModule loadModule(String namespace, String location, SchemaManager sc, VarCounter varCounter) throws XQueryException {
		return loadModule(this, namespace, location, sc, varCounter);
	}

//	public XQueryModule loadModule(ModuleLocator locator, InputStream stream, SchemaManager sc) throws XQueryException {
//		return loadModule(locator, stream, sc, new VarCounter());
//	}
//	public XQueryModule loadModule(ModuleLocator locator, InputStream stream, SchemaManager sc, VarCounter varCounter) throws XQueryException {
//		XQueryModule result = null;
//		try {
//			XQueryParser parser = new XQueryParser(stream);
//			// parsing request
//			// setBaseURI
//			result = parser.Start(staticContext.getMetaData(),sc,this,varCounter);
//		} catch (ParseException ex) {
//			throw new XQueryException(ex.getMessage(), ex);
//		}
//		String namespace = result.getNamespace();
//		// Add module to namespace cache
//		modules.put(namespace, result);
//		return result;
//	}
//	public XQueryModule loadModule(InputStream stream, SchemaManager sc) throws XQueryException {
//		return loadModule(this, stream, sc);
//	}
//	public XQueryModule loadModule(InputStream stream, SchemaManager sc, VarCounter varCounter) throws XQueryException {
//		return loadModule(this, stream, sc, varCounter);
//	}

	public boolean putModule(XQueryModule module) throws XQueryException {
		if (isRegistered(module.getNamespace()))
			return false;
		modules.put(module.getNamespace(), module);
		return true;
	}

}
