/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package  org.xquark.xquery.metadata;

import java.io.IOException;
import java.util.ArrayList;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.schema.Schema;
import org.xquark.schema.loader.Loader;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;


// ***************************************************************************
// * Class 'MetaParser'
// ***************************************************************************
/**
 * Class to use to parse metadata represented as an XML file conforming to
 * schema /org/xquark/xml/xdbc/resources/XMLDBCMetaData.xsd
 *
 */
public class MetaParser
extends DefaultHandler {
    
    // **********************************************************************
    // * Constant
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name:  $";
    // labels and attributes name of tag that we parse.
    private final static String TAG_SOURCE = "metadata";
    private final static String TAG_SCHEMAS = "schemas";
    private final static String TAG_SCHEMA = "schema";
    private final static String TAG_SCHEMADEFINITION = "schemadefinition";
    private final static String TAG_COLLECTIONS = "collections";
    private final static String TAG_COLLECTION = "collection";
    private final static String TAG_STEP = "step";
    private final static String ATT_SOURCE_NAME = "source";
    private final static String ATT_COLLECTION_NAME = "name";
    private final static String ATT_STEP_NS = "ns";
    private final static String ATT_STEP_NAME = "name";
    private final static String ATT_STEP_TYPE = "type";
    private final static String ATT_STEP_TYPE_ATT = "attribute";
    private final static String ATT_STEP_TYPE_ELEMENT = "element";
    
    // **********************************************************************
    // * Fields
    // **********************************************************************
    // a reference of the metadata where we will insert metawrapper
    // and new schema we are going to discover.
    private MetaDataImpl metadata = null;
    
    // the current metacollection we arre parsing.
    private MetaCollection metacollection = null;
    
    // the current metawrapper we are parsing.
    private MetaWrapper metawrapper = null;
    
    // the source name of the wrapper.
    private String sourcename = null;
    
    // use during the parsing to see if we are in the context that we are
    // inside the parsing of a schema
    private boolean inSchema = false;
    
    // facilities when parsing schemas contained in the XMLDocument.
    // As prefix mapping are callbacked, *before* the callback
    // 'startElement' "schema", we should keep it for when we will see
    // the startElement 'schema'. In the preix mapping, we consider
    // prefixes and uri's. So we need 2 arraylist of strings to
    // keep them.
    private ArrayList tmppref = null;
    private ArrayList tmpuri = null;
    
    // probably temporary as well as schema doesn't correcty serialize.
    // keep the string representation of the schemas that we will parse.
    // private String strschema = null ;
    private StringBuffer bufSchema = new StringBuffer();
    
    // for generic and static use of SAX parser.
    private static SAXParserFactory spf = null;
    static {
        // Create a JAXP SAXParserFactory and configure it
        spf = SAXParserFactory.newInstance() ;
        spf.setNamespaceAware(true);
        spf.setValidating(false);
    }
    
    private XMLReader reader = null;
    private Loader loader = null;
    private static SchemaValidationContext svContext = null;
    
    
    // **********************************************************************
    // * Constructor
    // **********************************************************************
    /**
     * Construct the metaparser of a wrapper XML description.
     * As a wrapper has to be inserted into a metadata, keep a reference
     * on metadata.
     */
    public MetaParser(MetaDataImpl metadata) {
        this.metadata = metadata;
        svContext = new SchemaValidationContext(metadata.getLoadingManager()) ;
        tmppref = new ArrayList() ;
        tmpuri = new ArrayList() ;
    }
    
    
    // **********************************************************************
    // * Methods
    // **********************************************************************
    public void error(SAXParseException spe) {
        System.err.println("Caught SAXParseException : " + spe.getMessage());
    }
    
    private void createReader() throws SAXException, ParserConfigurationException {
        // Create a JAXP SAXParser
        SAXParser saxParser = spf.newSAXParser();
        
        // Get the encapsulated SAX XMLReader
        reader = saxParser.getXMLReader();
    }
    
    
    // **********************************************************************
    // Create a MetaWrapper
    // **********************************************************************
    /**
     * Create a wrapper from the SAX callbacks.
     */
    public MetaWrapper createWrapper(String uri, boolean validate) throws MetadataException {
        try {
            metawrapper = new MetaWrapper(metadata);
            createReader() ;
            
            ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svContext);
            filter.setContentHandler(this);
            filter.setErrorHandler(this);
            
            filter.parse(uri);
            
            return metawrapper;
        }
        catch (org.xml.sax.SAXParseException e) {}
        catch (SAXException e) {}
        catch (IOException e) {}
        catch (ParserConfigurationException e) {}
        throw  new MetadataException();
    }
    
    
    /**
     * Create a wrapper from the SAX callbacks.
     */
    public MetaWrapper createWrapper(InputSource inputSource, boolean validate) throws MetadataException {
        try {
            metawrapper = new MetaWrapper(metadata);
            createReader() ;
            
            ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svContext);
            filter.setContentHandler(this);
            filter.setErrorHandler(this);
            filter.parse(inputSource);
            return metawrapper;
        }
		catch (org.xml.sax.SAXParseException e) {}
		catch (SAXException e) {}
		catch (IOException e) {}
		catch (ParserConfigurationException e) {}
        throw  new MetadataException();
    }
    
    
    public MetaWrapper createWrapper(boolean validate) throws MetadataException {
        try {
            metawrapper = new MetaWrapper(metadata);
            createReader() ;
            
            ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svContext);
            filter.setContentHandler(this);
            filter.setErrorHandler(this);
            
            return metawrapper;
        }
		catch (org.xml.sax.SAXParseException e) {}
		catch (SAXException e) {}
		catch (ParserConfigurationException e) {}
        throw  new MetadataException();
    }
    
    // **********************************************************************
    /**
     * Receive notification of the beginning of a document.
     *
     * <p>The SAX reader will invoke this method only once, before any
     * other methods in this interface or in DTDHandler (except for
     * setDocumentLocator).</p>
     *
     * @exception SAXException Any SAX exception, possibly
     *			wrapping another exception.
     */
    public void startDocument()
    throws SAXException {
    }
    
    
    public void startPrefixMapping(String prefix, String uri) throws SAXException {
        if (loader != null && inSchema) {
            loader.startPrefixMapping(prefix, uri);
        }
        else {
            tmppref.add(prefix) ;
            tmpuri.add(uri) ;
        }
    }
    
    
    public void endPrefixMapping(String prefix)
    throws SAXException {
        if (loader != null && inSchema) {
            loader.endPrefixMapping(prefix);
        }
    }
    
    
    /**
     * Receive notification of the beginning of an element.
     */
    public void startElement(String uri, String local, String raw, Attributes attrs)
    throws SAXException {
        
        if (local.equalsIgnoreCase(TAG_SCHEMA)) {
            bufSchema.setLength(0);
            inSchema = true;
            loader = new Loader(metadata.getSchemaManager());
            loader.startDocument();
            
            
            // <serialisation of the strschema to keep in the cache>
            bufSchema.append("<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>");
            bufSchema.append("<");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String)tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            for (int i = 0 ; i < tmppref.size() ; i ++) {
                
                bufSchema.append(" xmlns");
                if (tmppref.get(i) != null && ((String)tmppref.get(i)).length() != 0) bufSchema.append(":");
                bufSchema.append(tmppref.get(i));
                bufSchema.append("=\"");
                bufSchema.append(tmpuri.get(i));
                bufSchema.append("\"");
                
                loader.startPrefixMapping((String) tmppref.get(i), (String) tmpuri.get(i));
            }
            for (int i = 0 ; i < attrs.getLength() ; i ++) {
                bufSchema.append(" ");
                bufSchema.append(attrs.getLocalName(i));
                bufSchema.append("=\"");
                bufSchema.append(attrs.getValue(i));
                bufSchema.append("\"");
            }
            bufSchema.append(">" );
            // </serialisation of the strschema to keep in the cache>
            
            loader.startElement(uri, local, (raw == null) ? "" : raw, (attrs == null) ? new AttributesImpl() : attrs);
            
            return;
        }
        
        if (inSchema) {
            bufSchema.append("<");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String)tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            for (int i = 0 ; i < attrs.getLength() ; i ++) {
                bufSchema.append(" ");
                bufSchema.append(attrs.getLocalName(i));
                bufSchema.append("=\"");
                bufSchema.append(attrs.getValue(i));
                bufSchema.append("\"");
            }
            bufSchema.append(">");
            
            loader.startElement(uri, local, (raw == null) ? "" : raw, (attrs ==
            null) ? new AttributesImpl() : attrs);
            return;
        }
        
        if (local.equalsIgnoreCase(TAG_SOURCE)) {
            sourcename = attrs.getValue("", ATT_SOURCE_NAME);
            metawrapper.setSourceName(sourcename);
        }
        else {
            if (local.equalsIgnoreCase(TAG_SCHEMAS)) {
                // do nothing
            }
            else {
                if (local.equalsIgnoreCase(TAG_COLLECTIONS)) {
                    metawrapper.initMetaCollections();
                }
                else {
                    if (local.equalsIgnoreCase(TAG_COLLECTION)) {
                        metacollection = new MetaCollection(metadata, attrs.getValue("", ATT_COLLECTION_NAME));
                    }
                    else {
                        if (local.equalsIgnoreCase(TAG_STEP)) {
                            boolean isattribute = isAttribute(attrs.getValue("", ATT_STEP_TYPE));
                            metacollection.addMetaElement(attrs.getValue("", ATT_STEP_NS), attrs.getValue("", ATT_STEP_NAME), isattribute);
                        }
                    }
                }
            }
        }
    }
    
    
    private boolean isAttribute(String name) {
        if (name == null) {
            return  false;
        }
        if (name.equalsIgnoreCase(ATT_STEP_TYPE_ATT)) {
            return  true;
        }
        return  false;
    }
    
    
    /**
     * Receive notification of the end of an element.
     *
     * <p>The SAX reader will invoke this method at the end of every
     * element in the XML document; there will be a corresponding
     * startElement () event for every endElement () event (even when the
     * element is empty).</p>
     *
     * <p>For information on the names, see startElement.</p>
     *
     * @param uri The Namespace URI, or the empty string if the
     *		element has no Namespace URI or if Namespace
     *		processing is not being performed.
     * @param localName The local name (without prefix), or the
     *		empty string if Namespace processing is not being
     *		performed.
     * @param rawName The raw XML 1.0 name (with prefix), or the
     *		empty string if raw names are not available.
     * @exception SAXException Any SAX exception, possibly
     *			wrapping another exception.
     */
    public void endElement(String uri, String local, String raw) throws SAXException {
        if (local.equalsIgnoreCase(TAG_SCHEMA)) {
            bufSchema.append("</");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String)tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            bufSchema.append(">");
            loader.endElement(uri, local, raw);
            inSchema = false;
            
            tmppref.clear();
            tmpuri.clear();
            
            if (loader != null) {
                loader.endDocument();
                Schema schema = loader.getSchema();
                if (metawrapper != null)
                    metadata.putSchema(schema, bufSchema.toString());
            }
            return;
        }
        
        if (inSchema) {
            loader.endElement(uri, local, raw);
            bufSchema.append("</");
            if (uri != null) {
                int indexPref = tmpuri.indexOf(uri);
                if (indexPref >= 0) {
                    String pref = (String)tmppref.get(indexPref);
                    if (pref != null && pref.length() > 0) {
                        bufSchema.append(pref);
                        bufSchema.append(":");
                    }
                }
            }
            bufSchema.append(local);
            bufSchema.append(">");
            return;
        }
        
        if (local.equalsIgnoreCase(TAG_SOURCE)) {
            // nothing to be done
        }
        else {
            if (local.equalsIgnoreCase(TAG_SCHEMAS)) {
                // nothing to be done
            }
            else {
                if (local.equalsIgnoreCase(TAG_SCHEMADEFINITION)) {
                    // attention, ici, transformer jusqu'\u00e0 la prochaine balise
                    // en XMLReader. (Utiliser le callback character pour cela suivant
                    // un etat boolean pour cela ?)
                }
                else {
                    if (local.equalsIgnoreCase(TAG_COLLECTIONS)) {
                        // nothing to be done
                    }
                    else {
                        if (local.equalsIgnoreCase(TAG_COLLECTION)) {
                            metawrapper.addMetaCollection(metacollection);
                        }
                        else {
                            if (local.equalsIgnoreCase(TAG_STEP)) {
                                metacollection.addEndElement();
                            }
                        }
                    }
                }
            }
        }
    }
    
    
    /**
     * Receive notification of character data.
     *
     * <p>The Parser will call this method to report each chunk of
     * character data.  SAX readers may return all contiguous character
     * data in a single chunk, or they may split it into several
     * chunks; however, all of the characters in any single event
     * must come from the same external entity so that the Locator
     * provides useful information.</p>
     *
     * <p>The application must not attempt to read from the array
     * outside of the specified range.</p>
     *
     * <p>Note that some readers will report whitespace using the
     * ignorableWhitespace () method rather than this one (validating
     * readers <em>must</em> do so).</p>
     *
     * @param ch The characters from the XML document.
     * @param start The start position in the array.
     * @param length The number of characters to read from the array.
     * @exception SAXException Any SAX exception, possibly
     *			wrapping another exception.
     * @see #ignorableWhitespace
     * @see org.xml.sax.Locator
     */
    public void characters(char ch[], int start, int length)
    throws SAXException {
    }
    
    
    /**
     * Receive notification of the end of a document.
     *
     * <p>The SAX reader will invoke this method only once, and it will
     * be the last method invoked during the parse.  The reader shall
     * not invoke this method until it has either abandoned parsing
     * (because of an unrecoverable error) or reached the end of
     * input.</p>
     *
     * @exception SAXException Any SAX exception, possibly
     *			wrapping another exception.
     */
    public void endDocument()
    throws SAXException {
    }
}
