/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * Copyright (C) 2004 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.normalize;

import java.util.ArrayList;

import org.xquark.xpath.Axis;
import org.xquark.xpath.NodeKind;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionCOLLECTION;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDATA;
import org.xquark.xquery.typing.*;

/**
 * This class reduce a located expression.
 * It is used by ExpressionNormalizationVisitor class.
 * <BR>
 */
public class LocatedExpressionReducer {

	/****************************************************************/
	/*************************** Constants declaration **************/
	/****************************************************************/

	/** CVS/QA stuff. */
	private static final String RCSRevision = "$Revision: 1.3 $";

	/** CVS/QA stuff. */
	private static final String RCSName = "$Name:  $";

	/****************************************************************/
	/************************** Properties declaration **************/
	/****************************************************************/

	/** main expression */
	protected LocatedExpression rootExpr;

	/** The list of the expressions on which to apply one or more steps.  */
	protected ArrayList expressions;

	/** The list of steps to apply. */
	protected ArrayList steps;

	/** The current position in the list of steps to apply. */
	protected int currentPos;

	/** Flag to indicate if the current located expression were reduced. */
	protected boolean reduced;

	/** Flag to indicate if the current steps were reduced. */
	//   protected boolean stepsReduced;

	/** To know if this current located expression can be reduce or not. */
	private boolean isUnapplicable;

	/** The list of reduced expressions (could be empty). */
	protected ArrayList reducedExpressions;

	/** The type visitor to build new XQuery expressions. */
	protected TypeVisitor typeVisitor;

	/****************************************************************/
	/************************** Constructors ************************/
	/****************************************************************/

	/**
	 * To construct a new LocatedExpressionReducer.
	 * @param expr the XQuery expression on which apply XPath.
	 * @param steps steps of XPath to apply/reduce.
	 * @param typeVisitor an instance of the type visitor (could be null).
	 * @throws NormalizeException
	 */
	// add   22/08/2003
	protected LocatedExpressionReducer() {
	}

	// add   22/08/2003
	protected void reset(LocatedExpression locExpr, TypeVisitor typeVisitor) throws NormalizeException {
		rootExpr = locExpr;
		XQueryExpression expr = rootExpr.getExpression();
		this.expressions = new ArrayList();
		if (expr instanceof XQueryExpressionSequence) {
			this.expressions = ((XQueryExpressionSequence) expr).getSubExpressions();
		} else {
			this.expressions.add(expr);
		}
		this.steps = locExpr.getSteps();
		this.currentPos = 1;
		this.reduced = false;
		//   this.stepsReduced = false;
		this.isUnapplicable = false;
		if (this.reducedExpressions == null)
			this.reducedExpressions = new ArrayList();
		else
			this.reducedExpressions.clear();
		this.typeVisitor = typeVisitor;
	}

	/****************************************************************/
	/************************** Private methods *********************/
	/****************************************************************/

	/****************************************************************/
	/************************ Protected methods *********************/
	/****************************************************************/

	/**
	 * To retrieve the list of XQuery expressions that were reduce.
	 * @return the list of XQuery expressions that were reduce (could be empty).
	 */
	protected ArrayList getReducedExpressions() {
		return this.reducedExpressions;
	}

	/**
	 * To build a new expression from an ArrayList of expressions.
	 * @param expressions an ArrayList of expressions.
	 * @return a new XQueryExpression (could be null).
	 */
	protected XQueryExpression buildExpression(ArrayList expressions) throws NormalizeException {
		try {
			if ((expressions == null) || (expressions.size() == 0)) {
				return null;
			} else if (expressions.size() == 1) {
				return (XQueryExpression) expressions.get(0);
			} else {
				return new XQueryExpressionSequence(expressions, ((XQueryExpression) expressions.get(0)).getParentModule());
			}
		} catch (XQueryException e) {
			throw new NormalizeException(0, "Unexpected exception in LocatedExpressionReducer.buildExpression(...) :\n==>" + e.toString() + "<==\n", e);
		}
	}

	/**
	 * To retrieve the XQuery expression that were reduce.
	 * @return the XQuery expression that were reduce (could be null if the reduction eliminated the expression).
	 */
	protected XQueryExpression getReducedExpression() throws NormalizeException {
		return this.buildExpression(this.reducedExpressions);
	}

	/**
	 * To retrieve the steps that were reduce.
	 * @return the steps that were reduce (could be an empty if all steps have been appling).
	 */
	protected ArrayList getReducedSteps() {
		return this.steps;
	}

	/**
	 * To know if a step is node test with axis self (i.e. /self::node()).
	 * @param step the step to test.
	 * @return true if the step is /self::node(), false otherwise.
	 */
	protected boolean isStepSelfNode(Step step) throws NormalizeException {
		if ((step != null) && (step.getPredicates() == null) && (step.getAxis() == Axis.SELF)) {
			XQueryExpression test = step.getExpression();
			if ((test != null) && (test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.NODE)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * To know if a step is a node test (with any axis).
	 * @param step the step to test.
	 * @return true if the step is /axis::node(), false otherwise.
	 */
	protected boolean isStepNode(Step step) throws NormalizeException {
		if ((step != null) && (step.getPredicates() == null)) {
			XQueryExpression test = step.getExpression();
			if ((test != null) && (test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.NODE)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * To know if a step is equal to /descendant-or-self::node() (i.e. //).
	 * @param step the step to test.
	 * @return true if the step is equal to /descendant-or-self::node(), false otherwise.
	 */
	protected boolean isSlashSlash(Step step) throws NormalizeException {
		if ((step != null) && (step.getAxis() == Axis.DESCENDANT_OR_SELF) && (step.getPredicates() == null)) {
			XQueryExpression test = step.getExpression();
			if ((test != null) && (test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.NODE)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * To retrieve the next step(s) to apply (null, one or two steps maximum). Valid steps pairs are :<BR>
	 * <UL>
	 * <LI>/descendant-or-self::node(), /self::node()</LI>
	 * <LI>/descendant-or-self::node(), /self::text()</LI>
	 * <LI>/descendant-or-self::node(), /self::nametest</LI>
	 * <LI>/descendant-or-self::node(), /child::text()</LI>
	 * <LI>/descendant-or-self::node(), /child::nametest</LI>
	 * <LI>/descendant-or-self::node(), /@nametest</LI>
	 * <LI>/child:: node(), null</LI>
	 * <LI>/self::text(), null</LI>
	 * <LI>/self::nametest, null</LI>
	 * <LI>/child::text(), null</LI>
	 * <LI>/child::nametest, null</LI>
	 * <LI>/@nametest, null</LI>
	 * </UL>
	 * @return the next step(s) to apply (if null, there is no more steps to apply).
	 */
	protected Step[] getNextSteps() throws NormalizeException {
		Step[] next = new Step[2];
		if (this.currentPos >= this.steps.size()) {
			return null;
		} else if (this.steps.size() - this.currentPos == 1) {
			this.currentPos++;
			next[0] = (Step) this.steps.get(this.currentPos - 1);
		} else {
			Step currentStep = (Step) this.steps.get(this.currentPos);
			if (this.isSlashSlash(currentStep)) {
				next[0] = currentStep;
				this.currentPos++;
				Step secondStep = (Step) this.steps.get(this.currentPos);
				if ((this.steps.size() - this.currentPos >= 1) && ((this.isStepSelfNode(secondStep)) || (!(this.isStepNode(secondStep))))) {
					//All step except /axis::node() (but self::node() is ok)
					next[1] = secondStep;
					this.currentPos++;
				}
			} else {
				next[0] = currentStep;
				this.currentPos++;
			}
		}
		return next;
	}

	/**
	 * Test if two QName match (including when contains wildcard '*') :<BR>
	 * <PRE><CODE>
	namespace1  localname1  namespace2  localname2  match?
	-------------------------------------------------------
	1   null        A           null        A           true
	2   null        A           null        B           false
	3   null        A           *           *           true
	4   null        A           *           A           true
	5   null        A           *           B           false
	6   null        A           ns2         *           false
	7   null        A           ns2         A           false
	8   null        A           ns2         B           false
	-------------------------------------------------------
	9   *           *           null        A           true (idem 3)
	10  *           *           *           *           true
	11  *           *           *           A           true
	12  *           *           ns2         *           true
	13  *           *           ns2         A           true
	-------------------------------------------------------
	14  *           A           null        A           true (idem 4)
	15  *           A           null        B           false (idem 5)
	16  *           A           *           *           true (idem 11)
	17  *           A           *           A           true
	18  *           A           *           B           false
	19  *           A           ns2         *           true
	20  *           A           ns2         A           true
	21  *           A           ns2         B           false
	-------------------------------------------------------
	22  ns1         A           null        A           false (idem 7)
	23  ns1         A           null        B           false (idem 8)
	24  ns1         A           *           *           true (idem 13)
	25  ns1         A           *           A           true (idem 20)
	26  ns1         A           *           B           false (idem 21)
	27  ns1         A           ns2         *           false
	28  ns1         A           ns2         A           false
	29  ns1         A           ns2         B           false
	30  ns1         A           ns1         *           true
	31  ns1         A           ns1         A           true
	32  ns1         A           ns1         B           false
	-------------------------------------------------------
	33  ns1         *           null        A           false (idem 6)
	34  ns1         *           *           *           true (idem 12)
	35  ns1         *           *           A           true (idem 19)
	36  ns1         *           ns2         *           false
	37  ns1         *           ns2         A           false (idem 27)
	38  ns1         *           ns1         *           true
	39  ns1         *           ns1         A           true (idem 30)
	-------------------------------------------------------
	Impossible case : namespace = null, localname = *
	 * </CODE></PRE>
	 * @param q1 a QName to test if it match with pattern.
	 * @param q2 a pattern to test a QName.
	 * @return true if first paramter match with second parameter, false otherwise.
	 */
//	static protected boolean matchQName(QName q1, QName q2) throws NormalizeException {
//		//outln(" Matching test between : "+q1.getPrefixName()+"("+q1.getNameSpace()+"):"+q1.getLocalName()+" and "+q2.getPrefixName()+"("+q2.getNameSpace()+"):"+q2.getLocalName());
//		boolean resp = false; //Default value.
//		//Test impossible case :
//		if (((q1.getNameSpace() == null) && (q1.getLocalName().equals("*"))) || ((q2.getNameSpace() == null) && (q2.getLocalName().equals("*")))) {
//			throw new NormalizeException(0, "Impossible case : a QName can not have namespace == null AND localname '*' !");
//		}
//		//Test match (big test but only one, two or maximum three conditions to answer if QNames match :
//		if (q1.getNameSpace() == null) { //NS1 = null
//			if (q2.getNameSpace() == null) { //NS2 = null
//				resp = q1.getLocalName().equals(q2.getLocalName());
//			} else if (q2.getNameSpace().equals("*")) { //NS2 = *
//				resp = (q2.getLocalName().equals("*")) || (q1.getLocalName().equals(q2.getLocalName()));
//			} //Default NS2 = url
//		} else if (q1.getNameSpace().equals("*")) { //NS1 = *
//			if (q2.getNameSpace() == null) { //NS2 = null
//				resp = (q1.getLocalName().equals("*")) || (q1.getLocalName().equals(q2.getLocalName()));
//			} else if (q2.getNameSpace().equals("*")) { //NS2 = *
//				resp = (q1.getLocalName().equals("*")) || (q2.getLocalName().equals("*")) || (q1.getLocalName().equals(q2.getLocalName()));
//			} else { //Default NS2 = url
//				resp = (q1.getLocalName().equals("*")) || (q2.getLocalName().equals("*")) || (q1.getLocalName().equals(q2.getLocalName()));
//			}
//		} else { //NS1 = url
//			if (q2.getNameSpace() == null) { //NS2 = null
//				//Nothing to do (default = false);
//			} else if (q2.getNameSpace().equals("*")) { //NS2 = *
//				resp = (q1.getLocalName().equals("*")) || (q2.getLocalName().equals("*")) || (q1.getLocalName().equals(q2.getLocalName()));
//			} else if (q1.getNameSpace().equals(q2.getNameSpace())) { //Default NS2 = url (and == NS1)
//				resp = (q1.getLocalName().equals("*")) || (q2.getLocalName().equals("*")) || (q1.getLocalName().equals(q2.getLocalName()));
//			} //Default : NS2 = url and <> NS1
//		}
//		//outln(" RESULT MATCHING = "+resp);
//		return resp;
//	}
	static protected boolean matchStepExpr(XQueryExpression expr1, XQueryExpression expr2) throws NormalizeException {
		//outln(" Matching test between : "+q1.getPrefixName()+"("+q1.getNameSpace()+"):"+q1.getLocalName()+" and "+q2.getPrefixName()+"("+q2.getNameSpace()+"):"+q2.getLocalName());
		boolean resp = false; //Default value.
		String namespace1 = null;
		String localname1 = null;
		String namespace2 = null;
		String localname2 = null;
		if (expr1 instanceof QName) {
			namespace1 = ((QName)expr1).getNameSpace();
			localname1 = ((QName)expr1).getLocalName();
		} else if (expr1 instanceof ValueString) {
			localname1 = ((ValueString)expr1).getValue();
		}
		if (expr2 instanceof QName) {
			namespace2 = ((QName)expr2).getNameSpace();
			localname2 = ((QName)expr2).getLocalName();
		} else if (expr2 instanceof ValueString) {
			localname2 = ((ValueString)expr2).getValue();
		}
		
		// TODO handle this case, not done yet
		if (localname1 == null || localname2 == null) {			
			//return false;
			throw new NormalizeException(0, "Error in normalization, case not yet handled! (matching : " + expr1 + " with " + expr2 + ")");
		}
		//Test impossible case :
		if (((namespace1 == null) && (localname1.equals("*"))) || ((namespace2 == null) && (localname2.equals("*")))) {
			throw new NormalizeException(0, "Impossible case : namespace is null and localname is '*' !");
		}
		//Test match (big test but only one, two or maximum three conditions to answer if QNames match :
		if (namespace1 == null) { //NS1 = null
			if (namespace2 == null) { //NS2 = null
				resp = localname1.equals(localname2);
			} else if (namespace2.equals("*")) { //NS2 = *
				resp = (localname2.equals("*")) || (localname1.equals(localname2));
			} //Default NS2 = url
		} else if (namespace1.equals("*")) { //NS1 = *
			if (namespace2 == null) { //NS2 = null
				resp = (localname1.equals("*")) || (localname1.equals(localname2));
			} else if (namespace2.equals("*")) { //NS2 = *
				resp = (localname1.equals("*")) || (localname2.equals("*")) || (localname1.equals(localname2));
			} else { //Default NS2 = url
				resp = (localname1.equals("*")) || (localname2.equals("*")) || (localname1.equals(localname2));
			}
		} else { //NS1 = url
			if (namespace2 == null) { //NS2 = null
				//Nothing to do (default = false);
			} else if (namespace2.equals("*")) { //NS2 = *
				resp = (localname1.equals("*")) || (localname2.equals("*")) || (localname1.equals(localname2));
			} else if (namespace1.equals(namespace2)) { //Default NS2 = url (and == NS1)
				resp = (localname1.equals("*")) || (localname2.equals("*")) || (localname1.equals(localname2));
			} //Default : NS2 = url and <> NS1
		}
		//outln(" RESULT MATCHING = "+resp);
		return resp;
	}

	/**
	 * Gets a new /descendant-or-self::node() step.
	 * @return a new step.
	 */
	/** Contants for /descendant-or-self::node() step. */
	//static private Step descOrSelf;
	static protected Step getDOSNodeStep() throws NormalizeException {
		try {
			//			if (descOrSelf == null) {
			//				descOrSelf = new Step(true, Axis.DESCENDANT_OR_SELF, new NodeType(NodeKind.NODE, null, null), null, null);
			//				return descOrSelf;
			//			} else {
			//				return (Step) descOrSelf.clone();
			//			}
			return new Step(true, Axis.DESCENDANT_OR_SELF, new NodeTest(NodeKind.NODE, null, null), null, null);
		} catch (XQueryException e) {
			throw new NormalizeException(0, "Unexpected exception in LocatedExpressionReducer.getDOSNode() :\n==>" + e.toString() + "<==\n", e);
		}
	}

	/**
	 * Gets a new /parent::node() step.
	 * @return a new step.
	 */
	/** Contants for /parent::node() step. */
	//static protected Step parent;
	static protected Step getParentNodeStep() throws NormalizeException {
		try {
			//			if (parent == null) {
			//				parent = new Step(true, Axis.PARENT, new NodeType(NodeKind.NODE, null, null), null, null);
			//				return parent;
			//			} else {
			//				return (Step) parent.clone();
			//			}
			return new Step(true, Axis.PARENT, new NodeTest(NodeKind.NODE, null, null), null, null);
		} catch (XQueryException e) {
			throw new NormalizeException(0, "Unexpected exception in LocatedExpressionReducer.getParentNode() :\n==>" + e.toString() + "<==\n", e);
		}
	}

	/**
	 * Gets a new step with the given axis and the same test that the given step (and copy predicates).
	 * @return a new step.
	 */
	static protected Step modifyStep(boolean separator, byte axis, Step aStep) throws NormalizeException {
		aStep.setHasSeparator(separator);
		aStep.setAxis(axis);
		return aStep;
	}

	// add   04/09/2003
	// spilts a step into two steps whenever possible
	static protected void splitStep(ArrayList stepList, int firstStepPosition, int caseFirstStep) throws NormalizeException {
		Step firstStep = (Step) stepList.get(firstStepPosition);
		if (caseFirstStep == 14)
			stepList.set(firstStepPosition, getParentNodeStep());
		else
			stepList.set(firstStepPosition, getDOSNodeStep());
		if ((caseFirstStep == 10) || (caseFirstStep == 11) || (caseFirstStep == 14)) {
			stepList.add(firstStepPosition + 1, modifyStep(true, Axis.SELF, firstStep));
		} else { //caseFirstStep == 7, 8 or 9 :
			stepList.add(firstStepPosition + 1, modifyStep(true, Axis.CHILD, firstStep));
		}
	}

	static protected Step createUnionStep(ArrayList list1, ArrayList list2) throws NormalizeException {
		try {
			((Step) list1.get(0)).setHasSeparator(false);
			((Step) list2.get(0)).setHasSeparator(false);
			// create LocatedExpressions
			LocatedExpression tmpLoc1 = new LocatedExpression(list1, false, null, false);
			LocatedExpression tmpLoc2 = new LocatedExpression(list2, false, null, false);
			// create union
			ListOpUNIONExpression tmpListUnion = new ListOpUNIONExpression(tmpLoc1, tmpLoc2, null);
			tmpListUnion.setParenthesis(true);
			// create new Step									
			return new Step(true, Axis.NONE, tmpListUnion, null, null);
		} catch (XQueryException e) {
			throw new NormalizeException("Could not construct Union Step " + e.getMessage());
		}
	}
	/**
	 * Gets an integer code according to the type of the given step.
	 * @param oneStep a step.
	 * @return an integer witch according to the type of step. The values are in [0, 16] :
	 * <OL start="0">
	 *      <LI>any union step</LI>
	 *      <LI>/self::anynametest</LI>
	 *      <LI>/self::text()</LI>
	 *      <LI>/self::node() or /.</LI>
	 *      <LI>/child::anynametest</LI>
	 *      <LI>/child::text()</LI>
	 *      <LI>/child::node()</LI>
	 *      <LI>/descendant::anynametest</LI>
	 *      <LI>/descendant::text()</LI>
	 *      <LI>/descendant::node()</LI>
	 *      <LI>/descendant-or-self::anynametest</LI>
	 *      <LI>/descendant-or-self::text()</LI>
	 *      <LI>/descendant-or-self::node()</LI>
	 *      <LI>/attribute::anynametest or /@anynametest</LI>
	 *      <LI>/parent::node() or /..</LI>
	 *      <LI>/parent::text</LI>
	 *      <LI>/parent::anynametest</LI>
	 * </OL>
	 */
	static final int SELF_NT = 1;
	static final int SELF_TEXT = 2;
	static final int SELF_NODE = 3;
	static final int CHILD_NT = 4;
	static final int CHILD_TEXT = 5;
	static final int CHILD_NODE = 6;
	static final int DESC_NT = 7;
	static final int DESC_TEXT = 8;
	static final int DESC_NODE = 9;
	static final int DOS_NT = 10;
	static final int DOS_TEXT = 11;
	static final int DOS_NODE = 12;
	static final int ATT_NT = 13;
	static final int PARENT_NT = 14;
	static final int PARENT_TEXT = 15;
	static final int PARENT_NODE = 16;
	// special steps!!!
	static final int LOCATED = 20;
	static final int UNION = 21;

	static protected int getType(Step oneStep, boolean hasRootStep) throws NormalizeException {
		if (hasRootStep && !oneStep.hasSeparator()) {
			throw new NormalizeException(0, "Separator type " + oneStep.hasSeparator() + " of step " + oneStep + " not supported !");
		}
		//Get the type of sub expression :
		int innerExpr = -1; //Values : anynametest => 1, text() => 2 and node() => 3
		XQueryExpression subExpr = oneStep.getExpression();
		if (subExpr instanceof QName) {
			innerExpr = 1;
		} else if (subExpr instanceof NodeTest) {
			NodeTest thisNode = (NodeTest) subExpr;
			if (thisNode.getKind() == NodeKind.TEXT) {
				innerExpr = 2;
			} else if (thisNode.getKind() == NodeKind.NODE) {
				innerExpr = 3;
			} else {
				throw new NormalizeException(0, "Node type " + NodeKind.NODEKINDSTRINGS[thisNode.getKind()] + " of step " + oneStep + " not supported !");
			}
		} else if (subExpr instanceof LocatedExpression) {
			return LOCATED;
		} else if (subExpr instanceof ListOpUNIONExpression) {
			return UNION;
		} else {
			throw new NormalizeException(0, "Expression " + subExpr + " of step " + oneStep + " not supported !");
		}
		//Switch according to axis of given step :
		int axis = oneStep.getAxis();
		switch (axis) {
			case Axis.SELF :
				return innerExpr;
			case Axis.CHILD :
				return innerExpr + 3;
			case Axis.DESCENDANT :
				return innerExpr + 6;
			case Axis.DESCENDANT_OR_SELF :
				return innerExpr + 9;
			case Axis.ATTRIBUTE :
				if (innerExpr == 1)
					return innerExpr + 12;
				throw new NormalizeException(0, "Node test or text test on attribute of step " + oneStep + " not supported !");
			case Axis.PARENT :
				if (innerExpr == 1 || innerExpr == 3)
					return innerExpr + 13;
				throw new NormalizeException(0, "Text test on parent axis of step " + oneStep + " not supported !");
			default :
				if (axis < 0) {
					throw new NormalizeException(0, "Axis (value = " + axis + ") of step " + oneStep + " not supported !");
				} else {
					throw new NormalizeException(0, "Axis " + Axis.AXISSTRINGS[axis] + " of step " + oneStep + " not supported !");
				}
		}
	}

	/**
	 * To simplify (normalize) steps two by two. We use various rules to do this (see StepNormalization_0_2.xls).
	 * Example : /descendant-or-self::A/self::B becomes /descendant-or-self::A if A = B else it can be eliminate !
	 * @return Flag to indicate if the current located expression must be eliminated.
	 * 0 elimination
	 * 1 no elimination and reduced
	 * 2 no elimination not reduced
	 */
	static public boolean eliminateSteps(ArrayList stepList, boolean hasRootStep) throws NormalizeException {
		//outln("\n\n==> OLD STEPS = "+stepList);
		if (stepList == null)
			return true;
		Step firstStep = null;
		Step secondStep = null;
		int stepsReduced = 0;
		//   int size = stepList.size();
		//if (size <= 0) {
		// change   22/08/2003
		if (stepList.size() <= (hasRootStep ? 1 : 0)) {
			//Nothing to do ! impossible ?
		} else if (stepList.size() >= (hasRootStep ? 2 : 1)) {
			//General case :
			//   ArrayList newSteps = new ArrayList();
			int caseFirstStep;
			int caseSecondStep;
			int firstStepPosition = (hasRootStep ? 1 : 0);
			boolean hasPredicate = false;
			//outln("STEP 1 = "+firstStep+" "+caseFirstStep);
			//Loop until there is no more stepList :
			do {
				firstStep = (Step) stepList.get(firstStepPosition);
				caseFirstStep = getType(firstStep, hasRootStep); //See getType(Step) method for step type code !
				hasPredicate = firstStep.hasPredicates();
				caseSecondStep = -1;
				if (stepList.size() > firstStepPosition + 1) {
					secondStep = (Step) stepList.get(firstStepPosition + 1);
					caseSecondStep = getType(secondStep, true); //See getType(Step) method for step type code !
					//outln("STEP 2 = "+secondStep+" "+caseSecondStep);
				}
				firstStepPosition++;
				if (caseFirstStep == 0 || caseSecondStep == 0)
					continue;
				//Switch according to steps types combinaisons (two by two or one if this is the last step) :
				if (caseSecondStep == -1) {
				} else {
					//------------------------------- There are at least two steps :
					switch (caseFirstStep) {
						case SELF_NT :
							switch (caseSecondStep) {
								case SELF_NT :
									if (!matchStepExpr(firstStep.getExpression(),secondStep.getExpression())) {
										//QNames are not matching : Eliminate this Located Expression :
										return true;
									}
									break;
								case SELF_TEXT :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case SELF_TEXT :
							switch (caseSecondStep) {
								case PARENT_NT :
								case PARENT_NODE :
								case SELF_TEXT :
								case SELF_NODE :
								case DOS_NODE :
									break;
								default :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case SELF_NODE :
							break;
						case CHILD_NT :
							switch (caseSecondStep) {
								case SELF_NT :
									if (!matchStepExpr(firstStep.getExpression(),secondStep.getExpression())) {
										//QNames are not matching : Eliminate this Located Expression :
										return true;
									}
									break;
								case SELF_TEXT :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case CHILD_TEXT :
							switch (caseSecondStep) {
								case PARENT_NT :
								case PARENT_NODE :
								case SELF_TEXT :
								case SELF_NODE :
								case DOS_NODE :
									break;
								default :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case CHILD_NODE :
							break;
						case DESC_NT :
							switch (caseSecondStep) {
								case SELF_TEXT :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case DESC_TEXT :
							switch (caseSecondStep) {
								case SELF_TEXT :
								case SELF_NODE :
								case PARENT_NT :
								case PARENT_NODE :
								case DOS_NODE :
									break;
								default :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case DESC_NODE :
							break;
						case DOS_NT :
							switch (caseSecondStep) {
								case SELF_TEXT :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case DOS_TEXT :
							switch (caseSecondStep) {
								case SELF_TEXT :
								case SELF_NODE :
								case PARENT_NT :
								case PARENT_NODE :
									break;
								default :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case DOS_NODE :
							break;
						case ATT_NT :
							switch (caseSecondStep) {
								case SELF_NODE :
								case PARENT_NT :
								case PARENT_NODE :
								case DOS_NODE :
									break;
								default :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case PARENT_NT :
							switch (caseSecondStep) {
								case SELF_TEXT :
									//Eliminate this Located Expression :
									return true;
							}
							break;
						case PARENT_NODE :
							switch (caseSecondStep) {
								case SELF_TEXT :
									//Eliminate this Located Expression :
									return true;
							}
							break;
					}
				}
			} while ((stepList.size() >= firstStepPosition) && (firstStep != null) && (caseSecondStep != -1));
			//   stepList = newSteps;
		}
		//outln("NEW STEPS ="+stepList);
		//Do not eliminate this Located Expression :
		return false;
	}

	static public int simplifySteps(ArrayList steps, boolean inPredicate, XQueryModule parentModule) throws NormalizeException {

		boolean hasRootVar = false;
		try {
			Step step0 = (Step) steps.get(0);
			if (step0.getExpression() instanceof Variable)
				hasRootVar = true;
			if (step0.getAxis() != -1 && step0.hasSeparator()) {
				ArrayList argVect = new ArrayList();
				argVect.add(new ValueString("*:*", parentModule));
				FunctionCOLLECTION funcColExpr = new FunctionCOLLECTION(argVect, parentModule);
				Step step = new Step(false, Axis.NONE, funcColExpr, null, parentModule);
				step.setParentModule(parentModule);
				steps.add(0, step);
			} else if (inPredicate && !step0.hasSeparator()) {
				if (step0.getExpression() instanceof QName || step0.getExpression() instanceof NodeTest) {
					hasRootVar = true;
					if (step0.getAxis() == -1 && step0.getExpression() instanceof QName)
						step0.setAxis(Axis.CHILD);
				}
			}
		} catch (XQueryException xqe) {
			throw new NormalizeException(xqe.getMessage(), xqe);
		}
		return simplifySteps(steps, ((Step) steps.get(0)).getAxis() == Axis.NONE, hasRootVar, parentModule);
	}

	static private int simplifySteps(ArrayList stepList, boolean hasRootStep, boolean hasRootVar, XQueryModule parentModule) throws NormalizeException {
		//outln("\n\n==> OLD STEPS = "+stepList);
		if (stepList == null || eliminateSteps(stepList, hasRootStep))
			return 0;
		Step firstStep = null;
		Step secondStep = null;
		int stepsReduced = 0;
		boolean hasSeparator = ((Step) stepList.get(0)).hasSeparator();
		//   int size = stepList.size();
		//if (size <= 0) {
		// change   22/08/2003
		if (stepList.size() <= (hasRootStep ? 1 : 0)) {
			//Nothing to do ! impossible ?
		} else if (stepList.size() >= (hasRootStep ? 2 : 1)) {
			//General case :
			//   ArrayList newSteps = new ArrayList();
			int caseFirstStep;
			int caseSecondStep;
			int firstStepPosition = (hasRootStep ? 1 : 0);
			int minStepPosition = firstStepPosition;
			boolean hasPredicate = false;
			//outln("STEP 1 = "+firstStep+" "+caseFirstStep);
			//Loop until there is no more stepList :
			do {
				firstStep = (Step) stepList.get(firstStepPosition);
				caseFirstStep = getType(firstStep, hasRootStep); //See getType(Step) method for step type code !
				hasPredicate = firstStep.hasPredicates();
				caseSecondStep = -1;
				if (stepList.size() > firstStepPosition + 1) {
					secondStep = (Step) stepList.get(firstStepPosition + 1);
					caseSecondStep = getType(secondStep, true); //See getType(Step) method for step type code !
					//outln("STEP 2 = "+secondStep+" "+caseSecondStep);
				}
				//Switch according to steps types combinaisons (two by two or one if this is the last step) :
				if (caseSecondStep == -1) {
					switch (caseFirstStep) {
						case SELF_NT :
						case SELF_TEXT :
						case CHILD_NT :
						case CHILD_TEXT :
						case CHILD_NODE :
						case DOS_NODE :
						case ATT_NT :
						case PARENT_NODE :
							 ((Step) stepList.get(0)).setHasSeparator(hasSeparator);
							return (stepsReduced != 0) ? 1 : 2;
						case SELF_NODE :
							// Expr/self::node()[p1] -> Expr[p1] if possible otherwise unchanged
							//We eliminate this step !
							//PB if ($x, "toto")/self::node() => $x but not "toto" (not a node)
							if (firstStepPosition != minStepPosition) {
								stepList.remove(firstStepPosition);
								((Step) stepList.get(firstStepPosition - 1)).addPredicates(firstStep.getPredicates());
								stepsReduced++;
							} else if (!firstStep.hasPredicates() && hasRootStep == true) {
								stepList.remove(firstStepPosition);
								stepsReduced++;
							}
							((Step) stepList.get(0)).setHasSeparator(hasSeparator);
							return (stepsReduced != 0) ? 1 : 2;
						case DESC_NT :
						case DESC_TEXT :
						case DESC_NODE :
						case DOS_NT :
						case DOS_TEXT :
						case PARENT_NT :
							// Expr/descendant::node()[p1] -> Expr/descendant-or-self::node()/child::node()[p1]
							// Expr/descendant::NT[p1] -> Expr/descendant-or-self::node()/child::NT[p1]
							// Expr/descendant::text()[p1] -> Expr/descendant-or-self::node()/child::text()[p1]
							// Expr/descendant-or-self::NT[p1] -> Expr/descendant-or-self::node()/self::NT[p1]
							// Expr/descendant-or-self::text()[p1] -> Expr/descendant-or-self::node()/self::text()[p1]
							// Expr/parent::NT[p1] -> Expr/parent::node()/child::NT[p1]
							splitStep(stepList, firstStepPosition, caseFirstStep);
							if (firstStepPosition > minStepPosition)
								firstStepPosition--;
							// to ensure looping
							caseSecondStep = 0;
							break;
					}
					continue;
				}
				if (caseSecondStep == LOCATED) {
					if ((caseFirstStep == CHILD_NT || caseFirstStep == CHILD_NODE) && !firstStep.hasPredicates()) {
						//---> child::NT[p0] / dos::node()[p1]/parent::node()[p2]
						//---> self::node()[child::NT[p0]] ( self::node() | child::NT/dos::node() ) [child::node()[p1]][p2]

						LocatedExpression tmpLoc = (LocatedExpression) secondStep.getExpression();
						ArrayList tmpSteps = tmpLoc.getSteps();
						// create [child::node()[p1]][p2]
						ArrayList predicates = new ArrayList(2);
						// creating [child::node()[p1]]
						try {
							ArrayList tmpList = new ArrayList(1);
							tmpList.add(modifyStep(false, Axis.CHILD, (Step) tmpSteps.get(0)));
							((Step) tmpList.get(0)).setHasSeparator(false);
							//LocatedExpression pred1 = new LocatedExpression(tmpList, true, typevisitor);
							LocatedExpression pred1 = new LocatedExpression(tmpList, true, null);
//							pred1.setBracket(true);
							predicates.add(pred1);
						} catch (XQueryException e) {
							throw new NormalizeException("Could not construct predicate " + e.getMessage());
						}
						// add [p2]
						secondStep = (Step) tmpSteps.get(1);
						if (secondStep.getPredicates() != null)
							predicates.addAll(secondStep.getPredicates());
						// create self::node()[child::NT[p0]]
						try {
							ArrayList tmpList = new ArrayList(1);
							firstStep.setHasSeparator(false);
							tmpList.add(firstStep);
							//LocatedExpression pred1 = new LocatedExpression(tmpList, true, typevisitor);
							LocatedExpression pred1 = new LocatedExpression(tmpList, true, null);
//							pred1.setBracket(true);
							ArrayList tmpPreds = new ArrayList(1);
							tmpPreds.add(pred1);
							stepList.set(firstStepPosition, new Step(true, Axis.SELF, new NodeTest(NodeKind.NODE, null, null), tmpPreds, null));
						} catch (XQueryException e) {
							throw new NormalizeException("Could not construct step " + e.getMessage());
						}
						//create (self::node() | child::NT/dos::node())
						// construct tpmSteps1
						ArrayList tmpSteps1 = new ArrayList(1);
						try {
							tmpSteps1.add(new Step(false, Axis.SELF, new NodeTest(NodeKind.NODE, null, null), null, null));
						} catch (XQueryException e) {
							throw new NormalizeException("Could not construct step " + e.getMessage());
						}
						// construct tmpSteps2
						ArrayList tmpSteps2 = new ArrayList(2);
						try {
							firstStep = (Step) firstStep.clone();
						} catch (CloneNotSupportedException e) {
							throw new NormalizeException("Could not clone step " + e.getMessage());
						}
						firstStep.setPredicates(null);
						tmpSteps2.add(firstStep);
						tmpSteps2.add(getDOSNodeStep());

						firstStep = createUnionStep(tmpSteps1, tmpSteps2);
						firstStep.setPredicates(predicates);
						stepList.set(firstStepPosition + 1, firstStep);

						if (firstStepPosition > minStepPosition)
							firstStepPosition--;

					} else {
						throw new NormalizeException("Could not handle " + stepList);
					}
					continue;
				}
				if (caseSecondStep == UNION) {
					if (firstStep.hasPredicates()) {
						firstStepPosition++;
						continue;
					}
					// step/unionStep(list1,list2)[p2] -> unionStep(step/list1,step/list2)[p2]
					// handle first list
					ArrayList tmpSteps1 = ((LocatedExpression) ((ListOpUNIONExpression) secondStep.getExpression()).getExpression1()).getSteps();
					((Step) tmpSteps1.get(0)).setHasSeparator(true);
					try {
						tmpSteps1.add(0, firstStep.clone());
					} catch (CloneNotSupportedException e) {
						throw new NormalizeException("Could not clone " + firstStep);
					}
					int retVal1 = simplifySteps(tmpSteps1, false, false, parentModule);
					// handle second list
					ArrayList tmpSteps2 = ((LocatedExpression) ((ListOpUNIONExpression) secondStep.getExpression()).getExpression2()).getSteps();
					((Step) tmpSteps2.get(0)).setHasSeparator(true);
					try {
						tmpSteps2.add(0, firstStep.clone());
					} catch (CloneNotSupportedException e) {
						throw new NormalizeException("Could not clone " + firstStep);
					}
					int retVal2 = simplifySteps(tmpSteps2, false, false, parentModule);
					if (retVal1 == 0 && retVal2 == 0)
						return 0;
					if (retVal1 == 0) {
						stepList.remove(firstStepPosition);
						stepList.addAll(firstStepPosition, tmpSteps2);
						if (firstStepPosition > minStepPosition)
							firstStepPosition--;
					} else if (retVal2 == 0) {
						stepList.remove(firstStepPosition);
						stepList.addAll(firstStepPosition, tmpSteps1);
						if (firstStepPosition > minStepPosition)
							firstStepPosition--;
					} else if (retVal1 == 2 && retVal2 == 2) {
						tmpSteps1.remove(0);
						tmpSteps2.remove(0);
						((Step) tmpSteps1.get(0)).setHasSeparator(false);
						((Step) tmpSteps2.get(0)).setHasSeparator(false);
						firstStepPosition++;
					} else {
						stepList.remove(firstStepPosition);
						((Step) tmpSteps1.get(0)).setHasSeparator(false);
						((Step) tmpSteps2.get(0)).setHasSeparator(false);
						if (firstStepPosition > minStepPosition)
							firstStepPosition--;
					}
					continue;
				}
				switch (caseFirstStep) {
					case UNION :
						// advance one step
						firstStepPosition++;
						break;
					case SELF_NT :
						switch (caseSecondStep) {
							case SELF_NT :
								// self::NT1[p1]/self::NT2[p2] -> self:NT1[p1][p2] if NT1 and NT2 are compatible otherwise ERROR
								if (matchStepExpr(firstStep.getExpression(), secondStep.getExpression())) {
									if (firstStep.getExpression() instanceof QName && ((QName) firstStep.getExpression()).getLocalName().equals("*")) {
										//remove step :
										stepList.remove(firstStepPosition);
										secondStep.addPredicates(firstStep.getPredicates());
										stepsReduced++;
									} else if (secondStep.getExpression() instanceof QName && ((QName) secondStep.getExpression()).getLocalName().equals("*")) {
										//remove step :
										stepList.remove(firstStepPosition + 1);
										firstStep.addPredicates(secondStep.getPredicates());
										stepsReduced++;
									} else {
										//QNames are matching !
										//remove step :
										stepList.remove(firstStepPosition);
										secondStep.addPredicates(firstStep.getPredicates());
										stepsReduced++;
									}
								} else {
									//QNames are not matching : Eliminate this Located Expression :
									return 0;
								}
								break;
							case SELF_TEXT :
								// self::NT1[p1]/self::text() -> ERROR
								//Eliminate this Located Expression :
								return 0;
							case SELF_NODE :
								// self::NT1[p1]/self::node()[p2] -> self::NT1[p1][p2]
								//remove step :
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							case PARENT_NT :
							case PARENT_NODE :
								// self::NT1[p1]/parent::NT2[p2] -> parent::node()/child::NT2[child::node()[p1]][p2]
								// self::NT1[p1]/parent::node()[p2] -> parent::node()[child::node()[p1]][p2]
								stepsReduced++;
								//remove step :
								stepList.remove(firstStepPosition);
								// add predicate
								ArrayList tmpSteps = new ArrayList(1);
								tmpSteps.add(modifyStep(false, Axis.CHILD, firstStep));
								try {
									//LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, typevisitor, false);
									LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
//									tmpLoc.setBracket(true);
									secondStep.addPredicate(0, tmpLoc);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not add predicate " + e.getMessage());
								}
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								break;
								//							case DESC_NODE :
								//							case DESC_TEXT :
								//							case DESC_NT :
								//							case DOS_TEXT :
								//							case DOS_NT :
								//								splitStep(stepList, firstStepPosition + 1, caseSecondStep);
								//								// Advance two steps
								//								firstStepPosition += 2;
							default :
								//Advance one step :
								firstStepPosition++;
								break;
						}
						break;
					case SELF_TEXT :
						switch (caseSecondStep) {
							case PARENT_NT :
							case PARENT_NODE :
								// self::text()[p1]/parent::NT[p2] -> parent::node()/child::NT[child::text()[p1]][p2]
								// self::text()[p1]/parent::node[p2] -> parent::node()[child::text()[p1]][p2]
								stepsReduced++;
								//remove step :
								stepList.remove(firstStepPosition);
								// add predicate
								ArrayList tmpSteps = new ArrayList(1);
								tmpSteps.add(modifyStep(false, Axis.CHILD, firstStep));
								try {
									//LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, typevisitor, false);
									LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
//									tmpLoc.setBracket(true);
									secondStep.addPredicate(0, tmpLoc);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not add predicate " + e.getMessage());
								}
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								break;
							case SELF_TEXT :
							case SELF_NODE :
							case DOS_NODE :
								// self::text()[p1]/self::text()[p2] -> self::text()[p1][p2]
								// self::text()[p1]/self::node()[p2] -> self::text()[p1][p2]
								//remove step :
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							default :
								//Eliminate this Located Expression :
								return 0;
						}
						break;
					case SELF_NODE :
						if (firstStep.hasPredicates()) {
							if (firstStepPosition != minStepPosition) {
								stepList.remove(firstStepPosition);
								((Step) stepList.get(firstStepPosition - 1)).addPredicates(firstStep.getPredicates());
								stepsReduced++;
								firstStepPosition--;
							} else {
								//Advance one step :
								firstStepPosition++;
							}
						} else {
							stepList.remove(firstStepPosition);
							stepsReduced++;
							if (firstStepPosition > minStepPosition)
								firstStepPosition--;
						}
						break;
					case CHILD_NT :
						switch (caseSecondStep) {
							case SELF_NT :
								if (matchStepExpr(firstStep.getExpression(), secondStep.getExpression())) {
									if (firstStep.getExpression() instanceof QName && ((QName) firstStep.getExpression()).getLocalName().equals("*")) {
										// remove step
										stepList.remove(firstStepPosition);
										secondStep.addPredicates(firstStep.getPredicates());
										// modify second step
										stepList.set(firstStepPosition, modifyStep(true, Axis.CHILD, secondStep));
										stepsReduced++;
										//move back one step :
										if (firstStepPosition > minStepPosition)
											firstStepPosition--;
									} else if (secondStep.getExpression() instanceof QName && ((QName) secondStep.getExpression()).getLocalName().equals("*")) {
										//remove step :
										stepList.remove(firstStepPosition + 1);
										firstStep.addPredicates(secondStep.getPredicates());
										stepsReduced++;
									} else {
										//QNames are matching !
										//remove step :
										stepList.remove(firstStepPosition + 1);
										firstStep.addPredicates(secondStep.getPredicates());
										stepsReduced++;
									}
									break;
								} else {
									//QNames are not matching : Eliminate this Located Expression :
									return 0;
								}
							case SELF_TEXT :
								//Eliminate this Located Expression :
								return 0;
							case SELF_NODE :
								//remove step :
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							case PARENT_NT :
							case PARENT_NODE :
								stepsReduced++;
								//remove step :
								stepList.remove(firstStepPosition);
								secondStep = modifyStep(true, Axis.SELF, secondStep);
								stepList.set(firstStepPosition, secondStep);
								// add predicate
								ArrayList tmpSteps = new ArrayList(1);
								tmpSteps.add(modifyStep(false, Axis.CHILD, firstStep));
								try {
									//LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, typevisitor, false);
									LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
//									tmpLoc.setBracket(true);
									secondStep.addPredicate(0, tmpLoc);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not add predicate " + e.getMessage());
								}
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								break;
							default :
								//Advance one step :
								firstStepPosition++;
								break;
						}
						break;
					case CHILD_TEXT :
						switch (caseSecondStep) {
							case PARENT_NT :
							case PARENT_NODE :
								stepsReduced++;
								//remove step :
								stepList.remove(firstStepPosition);
								secondStep = modifyStep(true, Axis.SELF, secondStep);
								stepList.set(firstStepPosition, secondStep);
								// add predicate
								ArrayList tmpSteps = new ArrayList(1);
								firstStep.setHasSeparator(false);
								tmpSteps.add(firstStep);
								try {
									//LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, typevisitor, false);
									LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
//									tmpLoc.setBracket(true);
									secondStep.addPredicate(0, tmpLoc);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not add predicate " + e.getMessage());
								}
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								break;
							case SELF_TEXT :
							case SELF_NODE :
							case DOS_NODE :
								//remove step :
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							default :
								//Eliminate this Located Expression :
								return 0;
						}
						break;
					case CHILD_NODE :
						switch (caseSecondStep) {
							case SELF_NT :
							case SELF_TEXT :
								if (!firstStep.hasPredicates()) {
									stepsReduced++;
									//modify step :
									stepList.set(firstStepPosition, modifyStep(true, Axis.CHILD, secondStep));
									// remove step
									stepList.remove(firstStepPosition + 1);
									//move back one step :
									if (firstStepPosition > minStepPosition)
										firstStepPosition--;
								} else {
									//Advance one step :
									firstStepPosition++;
								}
								break;
							case SELF_NODE :
								stepsReduced++;
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								break;
							case DOS_NODE :
								stepsReduced++;
								// modify step
								if (!firstStep.hasPredicates()) {
									stepList.set(firstStepPosition, secondStep);
									stepList.set(firstStepPosition + 1, firstStep);
									if (firstStepPosition > minStepPosition)
										firstStepPosition--;
								} else {
									//Advance one step :
									firstStepPosition++;
								}
								break;
							case PARENT_NT :
							case PARENT_NODE :
								stepsReduced++;
								//remove step :
								stepList.remove(firstStepPosition);
								secondStep = modifyStep(true, Axis.SELF, secondStep);
								stepList.set(firstStepPosition, secondStep);
								// add predicate
								ArrayList tmpSteps = new ArrayList(1);
								firstStep.setHasSeparator(false);
								tmpSteps.add(firstStep);
								try {
									//LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, typevisitor, false);
									LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
//									tmpLoc.setBracket(true);
									secondStep.addPredicate(0, tmpLoc);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not add predicate " + e.getMessage());
								}
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								break;
							default :
								//Advance one step :
								firstStepPosition++;
								break;
						}
						break;
					case DESC_NT :
						switch (caseSecondStep) {
							case SELF_TEXT :
								//Eliminate this Located Expression :
								return 0;
							default :
								stepsReduced++;
								stepList.set(firstStepPosition, modifyStep(true, Axis.CHILD, firstStep));
								stepList.add(firstStepPosition, getDOSNodeStep());
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								else
									firstStepPosition++;
								break;
						}
						break;
					case DESC_TEXT :
						switch (caseSecondStep) {
							case SELF_TEXT :
							case SELF_NODE :
							case DOS_NODE :
							case PARENT_NT :
							case PARENT_NODE :
								stepsReduced++;
								stepList.set(firstStepPosition, modifyStep(true, Axis.CHILD, firstStep));
								stepList.add(firstStepPosition, getDOSNodeStep());
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								else
									firstStepPosition++;
								break;
							default :
								//Eliminate this Located Expression :
								return 0;
						}
						break;
					case DESC_NODE :
						stepsReduced++;
						stepList.set(firstStepPosition, modifyStep(true, Axis.CHILD, firstStep));
						stepList.add(firstStepPosition, getDOSNodeStep());
						if (firstStepPosition > minStepPosition)
							firstStepPosition--;
						else
							firstStepPosition++;
						break;
					case DOS_NT :
						switch (caseSecondStep) {
							case SELF_TEXT :
								//Eliminate this Located Expression :
								return 0;
							default :
								stepsReduced++;
								stepList.set(firstStepPosition, modifyStep(true, Axis.SELF, firstStep));
								stepList.add(firstStepPosition, getDOSNodeStep());
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								else
									firstStepPosition++;
								break;
						}
						break;
					case DOS_TEXT :
						switch (caseSecondStep) {
							case SELF_TEXT :
							case SELF_NODE :
							case DOS_NODE :
							case PARENT_NT :
							case PARENT_NODE :
								stepsReduced++;
								stepList.set(firstStepPosition, modifyStep(true, Axis.SELF, firstStep));
								stepList.add(firstStepPosition, getDOSNodeStep());
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								else
									firstStepPosition++;
								break;
							default :
								//Eliminate this Located Expression :
								return 0;
						}
						break;
					case DOS_NODE :
						switch (caseSecondStep) {
							case SELF_NODE :
								// remove step
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							case DESC_NT :
							case DESC_TEXT :
							case DESC_NODE :
							case DOS_NT :
							case DOS_TEXT :
							case DOS_NODE :
								if (!firstStep.hasPredicates()) {
									stepsReduced++;
									stepList.remove(firstStepPosition);
									if (firstStepPosition > minStepPosition)
										firstStepPosition--;
								} else {
									firstStepPosition++;
									break;
								}
								break;
								// case PARENT_NT :							
							case PARENT_NODE :

								// dos::node()[p1]/parent::node()[p2]	-> (parent::node() | dos::node())[child::node()[p1]][p2]
								// dos::node()[p1]/parent::NT2()[p2]	-> (parent::node() | dos::node())/self::NT2[child::node()[p1]][p2]
								// test if not first step
								if (firstStepPosition == minStepPosition) {
									if (hasRootVar) {
										// advance of step
										firstStepPosition++;
										// add LARS
										//throw new NormalizeException("Could not handle " + stepList);
									} else {
										// change DOS into DESC
										modifyStep(firstStep.hasSeparator(), Axis.DESCENDANT, firstStep);
									}
									break;
								}
								// for now make special step
								// remove step
								stepList.remove(firstStepPosition + 1);
								// create locatedExpression
								LocatedExpression tmpLoc = null;
								ArrayList tmpSteps = new ArrayList(3);
								firstStep.setHasSeparator(false);
								tmpSteps.add(firstStep);
								tmpSteps.add(secondStep);
								try {
									tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
									firstStep = new Step(true, Axis.NONE, tmpLoc, null, null);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not create located expression " + e.getMessage());
								}
								stepList.set(firstStepPosition, firstStep);
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;

								/*
								// erase one step
								stepList.remove(firstStepPosition + 1);
								stepsReduced++;
								// construct future predicates
								ArrayList predicates = new ArrayList(2);
								LocatedExpression tmpLoc = null;
								ArrayList tmpSteps = new ArrayList(1);
								try {
									tmpSteps.add(modifyStep(false, Axis.CHILD, firstStep));
									tmpLoc = new LocatedExpression(tmpSteps, true, null);
									tmpLoc.setBracket(true);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not construct predicate " + e.getMessage());
								}
								predicates.add(tmpLoc);
								if (secondStep.hasPredicates())
									predicates.addAll(secondStep.getPredicates());
								// construct tpmSteps1
								ArrayList tmpSteps1 = new ArrayList(2);
								secondStep.setHasSeparator(false);
								secondStep.setPredicates(null);
								tmpSteps1.add(secondStep);
								// construct tmpSteps2
								ArrayList tmpSteps2 = new ArrayList(2);
								firstStep = getDOSNodeStep();
								tmpSteps2.add(firstStep);
								firstStep = createUnionStep(tmpSteps1, tmpSteps2);
								firstStep.setPredicates(predicates);
								stepList.set(firstStepPosition, firstStep);
								
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								*/
								// dos::node()[p1]/parent::node()[p2]	-> ( self::node[P1]/parent::node() | descendant::node()/parent::node()[p1]) [p2]

								break;

							default :
								//Advance one step :
								firstStepPosition++;
								break;
						}
						break;
					case ATT_NT :
						switch (caseSecondStep) {
							case SELF_NODE :
							case DOS_NODE :
								// remove step
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							case SELF_NT :
							case SELF_TEXT :
							case CHILD_NT :
							case CHILD_TEXT :
							case CHILD_NODE :
							case DESC_NT :
							case DESC_TEXT :
							case DESC_NODE :
							case DOS_NT :
							case DOS_TEXT :
							case ATT_NT :
								return 0;
								//case PARENT_NT :
							case PARENT_NODE :
								// TODO add parent node in special construct
								stepsReduced++;
								//remove step :
								stepList.remove(firstStepPosition);
								secondStep = modifyStep(true, Axis.SELF, secondStep);
								stepList.set(firstStepPosition, secondStep);
								// add predicate
								ArrayList tmpSteps = new ArrayList(1);
								firstStep.setHasSeparator(false);
								tmpSteps.add(firstStep);
								try {
									//LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, typevisitor, false);
									LocatedExpression tmpLoc = new LocatedExpression(tmpSteps, true, null, false);
//									tmpLoc.setBracket(true);
									secondStep.addPredicate(0, tmpLoc);
								} catch (XQueryException e) {
									throw new NormalizeException("Could not add predicate " + e.getMessage());
								}
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								break;
							default :
								//Advance one step :
								firstStepPosition++;
								break;
						}
						break;
					case PARENT_NT :
						switch (caseSecondStep) {
							case SELF_TEXT :
								//Eliminate this Located Expression :
								return 0;
							default :
								stepsReduced++;
								stepList.set(firstStepPosition, modifyStep(true, Axis.SELF, firstStep));
								stepList.add(firstStepPosition, getParentNodeStep());
								if (firstStepPosition > minStepPosition)
									firstStepPosition--;
								else
									firstStepPosition++;
								break;
						}
						break;
					case PARENT_NODE :
						switch (caseSecondStep) {
							case SELF_TEXT :
								//Eliminate this Located Expression :
								return 0;
							case SELF_NODE :
								//remove step :
								stepList.remove(firstStepPosition + 1);
								firstStep.addPredicates(secondStep.getPredicates());
								stepsReduced++;
								break;
							default :
								//Advance one step :
								firstStepPosition++;
								break;
						}
						break;
				}
			} while ((stepList.size() >= firstStepPosition) && (firstStep != null) && (caseSecondStep != -1));
			//   stepList = newSteps;
		}
		//outln("NEW STEPS ="+stepList);
		//Do not eliminate this Located Expression :
		 ((Step) stepList.get(0)).setHasSeparator(hasSeparator);
		return (stepsReduced != 0) ? 1 : 2;
	}

	/**
	 * To apply a Step recursively (when /descendant-or-self::node() is present).
	 * @param element an element constructor
	 * @param oneFinalStep the step to apply recursively on element.
	 * @return a list of sub expressions reduce whith the given step recursively.
	 */
	protected ArrayList recursiveReduction(Element element, Step oneFinalStep, boolean noAttributes) throws NormalizeException {
		//outln("[recursiveReduction] "+element+"/"+oneFinalStep);
		try {
			ArrayList subExpr = new ArrayList();
			ArrayList subExprs = element.getSubExpressions();
			if (subExprs != null) {
				for (int i = 0; i < subExprs.size(); i++) {
					XQueryExpression subExpri = (XQueryExpression) subExprs.get(i);
					if (noAttributes && subExpri.getQType() != null && subExpri.getQType().getSubClass() == QType.ATTRIBUTE)
						continue;
					subExpr.addAll(this.applyFinalStep(subExpri, oneFinalStep, true, 0, element.getParentModule()));
				}
			}
			return subExpr;
		} catch (XQueryException e) {
			throw new NormalizeException(0, e.getMessage());
		}
	}

	/**
	 * Apply one of theses normalized steps :<BR>
	 * <UL>
	 * <LI>//self::node() (<B>*</B>)</LI>
	 * <LI>//self::text() (<B>*</B>)</LI>
	 * <LI>//self::nametest (<B>*</B>)</LI>
	 * <LI>//child::text() (<B>*</B>)</LI>
	 * <LI>//child::nametest (<B>*</B>)</LI>
	 * <LI>//@nametest (<B>*</B>)</LI>
	 * <LI>/self::node()</LI>
	 * <LI>/self::text()</LI>
	 * <LI>/self::nametest</LI>
	 * <LI>/child::node()</LI>
	 * <LI>/child::text()</LI>
	 * <LI>/child::nametest</LI>
	 * <LI>/@nametest</LI>
	 * <LI>/parent::node()</LI>
	 * <LI>/parent::nametest</LI>
	 * </UL>
	 * on an XQuery expression. The supported (explicitly) XQueryExpressions type are :
	 * <UL>
	 * <LI>XQueryExpressionSequence</LI>
	 * <LI>Element</LI>
	 * <LI>ValueText</LI>
	 * <LI>FLWRExpression</LI>
	 * <LI>LocatedExpression</LI>
	 * <LI>Variable</LI>
	 * <LI>ITEExpression</LI>
	 * <LI>ListOpUNIONExpression</LI>
	 * <LI>ValueString</LI>
	 * <LI>ValueBoolean</LI>
	 * <LI>ValueDouble</LI>
	 * <LI>ValueFloat</LI>
	 * <LI>ValueInteger</LI>
	 * <LI>XQueryBooleanOperatorExpression</LI>
	 * <LI>AttributeValuePair</LI>
	 * <LI>UnOpMinusExpression</LI>
	 * <LI>ListOpArithExpression</LI>
	 * <LI>ListOpCompExpression</LI>
	 * </UL>
	 * Other expressions are supported (by default : can not apply XPath on them).
	 * (Note: E//S1 : when there is //, the parameter slashSlash is true and oneFinalStep
	 * parameter contains only the step without the //, i.e. S1).
	 * @param expressionToNormalize an expression on which to apply a step.
	 * @param oneFinalStep a step to apply on given expression.
	 * @param slashSlash true if the given step if preceded by /descendant-or-self::node(), false otherwise.
	 * @param depth the current depth from the last node on which one applies step /descendant-or-self::node().
	 * @return an ArrayList of reduce expressions (could be null).
	 */
	private ArrayList applyFinalStep(XQueryExpression expressionToNormalize, Step appliedStep, boolean slashSlash, int depth, XQueryModule parentModule) throws NormalizeException {
		//outln("[applyFinalStep] //="+slashSlash+" depth="+depth+" "+oneFinalStep+" "+this.reduced+" "+this.isUnapplicable
		//+" "+expressionToNormalize.getClass().getName()+" "+expressionToNormalize);
		try {
			ArrayList results = new ArrayList();
			if (expressionToNormalize == null) {
				return results;
			}
			//Special case if it is a sequence :
			if (expressionToNormalize instanceof XQueryExpressionSequence) {
				ArrayList subExpr = ((XQueryExpressionSequence) expressionToNormalize).getSubExpressions();
				for (int i = 0; i < subExpr.size(); i++) {
					results.addAll(this.applyFinalStep((XQueryExpression) subExpr.get(i), appliedStep, slashSlash, depth, parentModule));
				}
			} else {
				if (appliedStep == null) {
					results.add(expressionToNormalize);
				} else if (appliedStep.getPredicates() != null) {
					//outln("=========> PREDICATE !!!!!"+oneFinalStep.getPredicates()+" "+oneFinalStep.getPredicates().get(0).getClass().getName());
					this.reduced = false;
					//throw new NormalizeException("Can not apply a step with predicate ("
					//+oneFinalStep.getPredicates()+") in LocatedExpressionReducer !");
				} else {
					//Here we have really a step to apply on  an xqueryexpression :
					XQueryExpression test = null;
					if (expressionToNormalize instanceof ValueText) {
						//----------------------------------------------------
						//----------------------------- ValueText expression :
						//----------------------------------------------------
						test = appliedStep.getExpression();
						if ((appliedStep.getAxis() == Axis.SELF) && (test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.TEXT)) {
							//++++++++++++++++++++ step = /self::text()
							if (appliedStep.getPredicates() == null) {
								this.reduced = true;
								results.add((ValueText) expressionToNormalize);
							}
						} else if ((appliedStep.getAxis() == Axis.SELF) && (test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.NODE)) {
							//++++++++++++++++++++ Step = /self::node()
							if (appliedStep.getPredicates() == null) {
								this.reduced = true;
								results.add(expressionToNormalize);
							}
						} else {
							//Other case (like Step = /self::nametest) : nothing to do !
							//results is empty
							this.reduced = true;
						}
					} else if (expressionToNormalize instanceof Element) {
						//--------------------------------------------------
						//----------------------------- Element expression :
						//--------------------------------------------------
						Element elt = (Element) expressionToNormalize;
						//Switch according to the type of step :
						test = appliedStep.getExpression();
						if ((appliedStep.getAxis() == Axis.ATTRIBUTE) && (test instanceof QName)) {
							//++++++++++++++++++++ step = @nametest
							this.reduced = true;
							ArrayList tmplist = elt.getAttributes();
							if (tmplist != null) {
								for (int i = 0; i < tmplist.size(); i++) {
									AttributeValuePair atti = (AttributeValuePair) tmplist.get(i);
									if (!atti.isXmlns() && matchStepExpr(atti.getAttributeName(), test)) {
										results.add(tmplist.get(i));
									}
								}
							}
							// add LARS 17/09/2003
							// look also in subexpressions!!!
							tmplist = elt.getSubExpressions();
							if (tmplist != null) {
								for (int i = 0; i < tmplist.size(); i++) {
									XQueryExpression expri = (XQueryExpression) tmplist.get(i);
									if (expri.getQType() != null && expri.getQType().getSubClass() == QType.ATTRIBUTE) {
										XQueryExpression expriname = ((QTypeAttribute) expri.getQType()).getName();
										if (matchStepExpr( expriname,  test)) {
											results.add(expri);
										}
									}
								}
							}
							if (slashSlash) { //To apply step // if it is present :
								results.addAll(this.recursiveReduction(elt, appliedStep, false));
							}
						} else if (appliedStep.getAxis() == Axis.SELF) {
							//++++++++++++++++++++ Axis = self :
							if (test instanceof QName) {
								//++++++++++++++++++++ Step = /self::nametest
								this.reduced = true;
								XQueryExpression eltName = elt.getStartTag();
								if (matchStepExpr(eltName, test)) {
									if (slashSlash) { //To apply step // if it is present :
										results.add(elt.clone());
										results.addAll(this.recursiveReduction(elt, appliedStep, true));
									} else {
										results.add(elt);
									}
								} else {
									if (slashSlash) { //To apply step // if it is present :
										results.addAll(this.recursiveReduction(elt, appliedStep, true));
									} //else nothing to do : result is empty !

								}
							} else if ((test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.NODE)) {
								//++++++++++++++++++++ Step = /self::node()
								this.reduced = true;
								if (slashSlash) { //To apply step // if it is present :
									results.add(elt.clone());
									results.addAll(this.recursiveReduction(elt, appliedStep, true));
								} else {
									results.add(elt);
								}
							} else if ((test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.TEXT)) {
								//++++++++++++++++++++ Step = /self::text()
								//Nothing to do !
								//results is empty
								this.reduced = true;
								if (slashSlash) { //To apply step // if it is present :
									results.addAll(this.recursiveReduction(elt, appliedStep, true));
								}
							} else {
								throw new NormalizeException("Unsupported step (" + appliedStep + ") in LocatedExpressionReducer !");
							}
						} else if (appliedStep.getAxis() == Axis.CHILD) {
							//++++++++++++++++++++ Axis = child :
							ArrayList subNodes = elt.getSubExpressions();
							if ((subNodes == null) || (subNodes.size() == 0)) {
								//No children : nothing to do !
								//results is empty
								this.reduced = true;
							} else {
								if ((test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.TEXT)) {
									//++++++++++++++++++++ Step = /child::text()
									for (int i = 0; i < subNodes.size(); i++) {
										if (subNodes.get(i) instanceof ValueText) {
											this.reduced = true;
											results.add((ValueText) subNodes.get(i));
										} else if (!(subNodes.get(i) instanceof Element)) {
											Step newStep = (Step) appliedStep.clone();
											newStep.setAxis(Axis.SELF);
											results.addAll(this.applyFinalStep((XQueryExpression) subNodes.get(i), newStep, slashSlash, depth + 1, parentModule));
										} else {
											//Child is Element !
											this.reduced = true;
											if (slashSlash) { //To apply step // if it is present :
												Step newStep = (Step) appliedStep.clone();
												newStep.setAxis(Axis.SELF);
												results.addAll(this.recursiveReduction((Element) subNodes.get(i), newStep, true));
											}
										}

									}
								} else if ((test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.NODE)) {
									//++++++++++++++++++++ Step = /child::node()
									for (int i = 0; i < subNodes.size(); i++) {
										XQueryExpression subExpri = (XQueryExpression) subNodes.get(i);
										if (subExpri.getQType() != null && subExpri.getQType().getSubClass() == QType.ATTRIBUTE)
											continue;
										if ((subExpri instanceof Element) || (subExpri instanceof ValueText)) {
											this.reduced = true;
											results.add(subExpri);
										} else {
											Step newStep = (Step) appliedStep.clone();
											newStep.setAxis(Axis.SELF);
											results.addAll(this.applyFinalStep(subExpri, newStep, slashSlash, depth + 1, parentModule));
										}
									}
								} else if (test instanceof QName) {
									//++++++++++++++++++++ Step = /child::nametest
									for (int i = 0; i < subNodes.size(); i++) {
										Step newStep = (Step) appliedStep.clone();
										newStep.setAxis(Axis.SELF);
										results.addAll(this.applyFinalStep((XQueryExpression) subNodes.get(i), newStep, slashSlash, depth + 1, parentModule));
									}
								} else {
									throw new NormalizeException("Unsupported step test (" + appliedStep + ") in LocatedExpressionReducer !");
								}
							}
						} else {
							//Other Axis : nothing to do !
							//results is empty
						}
					} else if (expressionToNormalize instanceof FLWRExpression) {
						//----------------------------------------------
						//----------------------------- FLWRExpression :
						//----------------------------------------------
						//Modified this FLWRExpression
						FLWRExpression flwr = (FLWRExpression) expressionToNormalize;
						XQueryExpression returnClause = flwr.getReturnClause();
						ArrayList newReturnClause = this.applyFinalStep(returnClause, appliedStep, slashSlash, depth, parentModule);
						XQueryExpression newReturnExpression = this.buildExpression(newReturnClause);
						if (newReturnExpression != null) {
							flwr.setReturnClause(newReturnExpression);
							results.add(flwr);
						}
					} else if (expressionToNormalize instanceof LocatedExpression) {
						//-------------------------------------------------
						//----------------------------- LocatedExpression :
						//-------------------------------------------------
						// add LARS do not redo locatedexpression with attribute QType
						if (appliedStep.getAxis() == Axis.ATTRIBUTE && expressionToNormalize.getQType() != null && expressionToNormalize.getQType().getSubClass() == QType.ATTRIBUTE) {
							// do nothing
						} else {
							LocatedExpression tmploc = (LocatedExpression) expressionToNormalize;
							//Update this located epxression :
							if (slashSlash) {
								tmploc.addStep(getDOSNodeStep());
							}
							tmploc.addStep(appliedStep);
							if (!this.isUnapplicable) {
								this.reduced = true;
							}
							if (simplifySteps(tmploc.getSteps(), true, (tmploc.getStepNum(0).getExpression() instanceof Variable), tmploc.getStepNum(0).getExpression().getParentModule()) != 0) {
								boolean correctType = true;
								try {
									if (this.typeVisitor != null) {
										expressionToNormalize.accept(this.typeVisitor);
									}
								} catch (TypeException te) {
									//Nothing to do : eliminate this located by default.
									correctType = false;
								}
								if (correctType)
									results.add(expressionToNormalize);
							}
						}
					} else if (expressionToNormalize instanceof Variable) {
						//---------------------------------------------------
						//----------------------------- Variable expression :
						//---------------------------------------------------
						//Modified this Variable :
						this.reduced = true;
						if (this.isStepSelfNode(appliedStep)) {
							results.add(expressionToNormalize);
						} else {
							ArrayList newSteps = new ArrayList();
							newSteps.add(new Step(false, Axis.NONE, expressionToNormalize, null, null));
							if (slashSlash) {
								newSteps.add(getDOSNodeStep());
							}
							newSteps.add(appliedStep);
							try {
								results.add(new LocatedExpression(newSteps, parentModule));
							} catch (TypeException te) {
								//Nothing to do : eliminate this located by default.
							}
						}
					} else if (expressionToNormalize instanceof ITEExpression) {
						//-------------------------------------------------
						//----------------------------- ITEExpression :
						//-------------------------------------------------
						//Modified this ITEExpression :
						ITEExpression ite = (ITEExpression) expressionToNormalize;
						XQueryExpression thenClause = ite.getThenExpression();
						XQueryExpression elseClause = ite.getElseExpression();
						ArrayList newThenClause = this.applyFinalStep(thenClause, appliedStep, slashSlash, depth, parentModule);
						ArrayList newElseClause = this.applyFinalStep(elseClause, appliedStep, slashSlash, depth, parentModule);
						ite.setThenExpression(this.buildExpression(newThenClause));
						ite.setElseExpression(this.buildExpression(newElseClause));
						results.add(ite);
					} else if (expressionToNormalize instanceof ListOpUNIONExpression) {
						//-------------------------------------------------
						//----------------------------- ListOpUNIONExpression :
						//-------------------------------------------------
						//Modified this ListOpUNIONExpression :
						ListOpUNIONExpression loue = (ListOpUNIONExpression) expressionToNormalize;
						XQueryExpression left = loue.getExpression1();
						XQueryExpression rigth = loue.getExpression2();
						ArrayList newLeft = this.applyFinalStep(left, appliedStep, slashSlash, depth, parentModule);
						ArrayList newRigth = this.applyFinalStep(rigth, appliedStep, slashSlash, depth, parentModule);
						loue.setExpression1(this.buildExpression(newLeft));
						loue.setExpression2(this.buildExpression(newRigth));
						results.add(loue);
					} else if (expressionToNormalize instanceof FunctionDATA) {
						test = appliedStep.getExpression();
						if (appliedStep.getAxis() == Axis.SELF && test instanceof NodeTest && ((NodeTest) test).getKind() == NodeKind.TEXT) {
							results.add(expressionToNormalize);
						} else {
							// do nothing, eliminate!
						}
						/*
						} else if (expressionToNormalize instanceof ValueString) {
						} else if (expressionToNormalize instanceof ValueBoolean) {
						} else if (expressionToNormalize instanceof ValueDouble) {
						} else if (expressionToNormalize instanceof ValueFloat) {
						} else if (expressionToNormalize instanceof ValueInteger) {
						*/
					} else if (expressionToNormalize instanceof Value) {
						//Other supported cases : nothing to do (eliminate this by default)
						//results is empty
						this.reduced = true;
					} else if (expressionToNormalize instanceof AggregateFunctionCall) {
						test = appliedStep.getExpression();
						if (appliedStep.getAxis() == Axis.SELF) {
							if ((test instanceof NodeTest) && (((NodeTest) test).getKind() == NodeKind.TEXT)) {
								results.add(expressionToNormalize);
							}
						}
					} else if ((expressionToNormalize instanceof PrimitiveFunctionCall) && !(expressionToNormalize instanceof InputFunctionCall)) {
					} else if (expressionToNormalize instanceof XQueryBooleanOperatorExpression) {
					} else if (expressionToNormalize instanceof AttributeValuePair) {
					} else if (expressionToNormalize instanceof UnOpMinusExpression) {
					} else if (expressionToNormalize instanceof ListOpArithExpression) {
					} else if (expressionToNormalize instanceof ListOpCompExpression) {
						//Other supported cases : nothing to do (eliminate this by default)
						//results is empty
						this.reduced = true;
					} else {
						//Other types of expressions (ex : FunctionCOLLECTION):
						//We should not modify them!
						this.isUnapplicable = true;
						ArrayList newSteps = new ArrayList();
						newSteps.add(new Step(false, Axis.NONE, expressionToNormalize, null, null));
						if (slashSlash) {
							newSteps.add(getDOSNodeStep());
						}
						newSteps.add(appliedStep);
						try {
							results.add(new LocatedExpression(newSteps, parentModule));
						} catch (TypeException te) {
							//Nothing to do : eliminate this located by default.
						}
					}
				}
			}
			//outln("RESULT = "+results);
			return results;
		} catch (CloneNotSupportedException e) {
			throw new NormalizeException(0, e.getMessage(), e);
		} catch (XQueryException e) {
			throw new NormalizeException(0, e.getMessage());
		}
	}

	/**
	 * To reduce (this) located expression (if it is possible).
	 * @return true if expression was reduce, false otherwise.
	 */

	protected boolean reduceLocatedExpression(XQueryModule parentModule) throws NormalizeException {
		try {
			this.reduced = false;
			//   this.stepsReduced = false;
			boolean stepsReduced = false;
			//Normalize steps two by two :
			int simplified = simplifySteps(steps, true, (rootExpr.getExpression() instanceof Variable), parentModule);
			if (simplified == 0) {
				//We must eliminate the current Located Expression !
				this.reduced = true;
			} else {
				stepsReduced = (simplified == 1);
				Step[] next = null;
				boolean firstTime = true;
				//Apply each step on the first step expression :
				do {
					next = this.getNextSteps();
					//Format next one or two steps :
					if (next != null) {
						//boolean slashSlash = false;
						boolean slashSlash = this.isSlashSlash(next[0]);
						boolean special = false;
						if (slashSlash) {
							if (next[1] != null) {
								next[0] = next[1];
							} else {
								//Create "this" step, i.e. : /self::node()
								next[0] = new Step(true, Axis.SELF, new NodeTest(NodeKind.NODE, null, parentModule), null, parentModule);
								special = true;
							}
							next[1] = null;
						}
						//Reduce expressions and/or Steps :
						if (firstTime && (this.expressions.size() == 1)) {
							//Eliminate all cases that can not be reduce, i.e. :
							//1- E/axis::anyTest if E is Variable
							//1- E/axis::anyTest if E is FunctionCall
							if (this.expressions.get(0) instanceof Variable || this.expressions.get(0) instanceof InputFunctionCall) {
								//								//For steps normalization (two by two) :
								//								if (stepsReduced && (this.steps.size() >= 1)) {
								//									Step firstExpr = new Step(false, Axis.NONE, (XQueryExpression) this.expressions.get(0), null, this.typeVisitor);
								//									this.steps.add(0, firstExpr);
								//									this.reducedExpressions.add(new LocatedExpression(this.steps, this.typeVisitor));
								//								} else {
								//									this.reducedExpressions.add(this.expressions.get(0));
								//								}
								if (steps.size() == 1 && ((Step) steps.get(0)).getAxis() == Axis.NONE)
									this.reducedExpressions.add(rootExpr.getExpression());
								else
									this.reducedExpressions.add(rootExpr);
								return this.reduced || stepsReduced;
							}
						}
						firstTime = false;
						for (int i = 0; i < this.expressions.size(); i++) {
							ArrayList partialExprs = this.applyFinalStep((XQueryExpression) this.expressions.get(i), next[0], slashSlash, 0, parentModule);
							if (partialExprs != null && !partialExprs.isEmpty()) {
								XQueryExpression tmpExpr = (XQueryExpression) partialExprs.get(partialExprs.size() - 1);
								//								if (special && tmpExpr instanceof LocatedExpression) {
								//									LocatedExpression tmpLoc = (LocatedExpression) tmpExpr;
								//									tmpLoc.getSteps().remove(tmpLoc.getSteps().size() - 1);
								//								}
								this.reducedExpressions.addAll(partialExprs);
							}
						}
						this.expressions.clear();
						this.expressions.addAll(this.reducedExpressions);
						this.reducedExpressions.clear();
					} else {
						this.reducedExpressions.addAll(this.expressions);
					}
				} while (next != null);
			}
			return this.reduced || stepsReduced;
		} catch (XQueryException e) {
			throw new NormalizeException(0, e.getMessage());
		}
	}

	/****************************************************************/
	/*************** Package visibility methods *********************/
	/****************************************************************/

	/****************************************************************/
	/************************** Public methods **********************/
	/****************************************************************/
}
