/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.normalize;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Stack;

import org.xquark.xpath.Axis;
import org.xquark.xpath.NodeKind;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.FunctionQNAME;
import org.xquark.xquery.parser.util.Constants;
import org.xquark.xquery.typing.QType;
import org.xquark.xquery.typing.TypeException;
import org.xquark.xquery.typing.TypeVisitor;

// Note :
// the visitor return value : resultExpr is null if the expression
// has been transformed without creating a new expression

public class RestructureVisitor extends DefaultParserVisitor {

	private static final String RCSRevision = "$Revision: 1.10 $";
	private static final String RCSName = "$Name:  $";

	private XQueryExpression resultExpr = null;
	private TypeVisitor typeVisitor = null;

	private final Integer NO_CONTEXT = new Integer(0);
	private final Integer IN_RETURN = new Integer(1);
	private final Integer IN_WHERE = new Integer(2);
	private final Integer IN_FUNCTION = new Integer(3);
	private final Integer IN_FLWR = new Integer(4);

	private Stack contextStack = new Stack();
	private Stack inWhereExpressionStack = new Stack();
	private Stack parentFLWRExpressionStack = new Stack();
	private Stack varDefinitionStack = new Stack();
	private int inReturn = 0;

	private int ruleNumber = 0;
	private boolean isMediator = false;
	private boolean isRoot = true;

	private final int[] initOrder = { -1, -1 };

	public RestructureVisitor(TypeVisitor typeVisitor) {
		this.typeVisitor = typeVisitor;
	}

	public RestructureVisitor(TypeVisitor typeVisitor, boolean isMediator) {
		this.typeVisitor = typeVisitor;
		this.isMediator = isMediator;
	}

	public XQueryExpression getResult() {
		return resultExpr;
	}
	public int getRuleNumber() {
		return ruleNumber;
	}

	public void visit(ComputedText arg) throws XQueryException {
		if (isRoot) {
			makeDummyOneFlwr(arg);
			return;
		}
		XQueryExpression tmpExpr = arg.getExpression();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			//change = true;
			arg.setExpression(resultExpr);
		}
		resultExpr = null;
	}

	public void visit(Document arg) throws XQueryException {
		if (isRoot) {
			makeDummyOneFlwr(arg);
			return;
		}
		XQueryExpression tmpExpr = arg.getExpression();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			//change = true;
			arg.setExpression(resultExpr);
		}
		resultExpr = null;
	}

	/*
	--> Rules for class Element
	ELT-1. putting computed namespaces first then attributes
	ELT-2. if not in return clause (root in fact) then create FWLR expression (for cannisation purposes) and restructure it
	ELT-3. if in return iterate on attributes and subexpressions and restructure each of them
	*/
	public void visit(Element arg) throws XQueryException {
		// rule ELT-2.
		if (isRoot) {
			makeDummyOneFlwr(arg);
			return;
		}
		// reorder element subexpressions
		// rule ELT-1.
		ArrayList subExpressions = arg.getSubExpressions();
		if (subExpressions != null) {
			byte nbAtt = 0;
			byte nbNS = 0;
			for (int i = 0; i < subExpressions.size(); i++) {
				XQueryExpression expri = (XQueryExpression) subExpressions.get(i);
				if (expri instanceof ComputedNamespace) {
					if (i > nbNS) {
						subExpressions.remove(i);
						subExpressions.add(nbNS, expri);
					}
					nbNS++;
				} else if (expri instanceof AttributeValuePair) {
					subExpressions.remove(i);
					arg.addAttributes((AttributeValuePair) expri);
				} else if (expri.getQType().getSubClass() == QType.ATTRIBUTE) {
					if (i > nbNS + nbAtt) {
						subExpressions.remove(i);
						subExpressions.add(nbNS + nbAtt, expri);
					}
					nbAtt++;
				}
			}
			arg.setComputedAttNum(nbAtt);
			arg.setComputedNSNum(nbNS);
		}
		// rule ELT-3.
		arg.getStartTag().accept(this);
		if (resultExpr != null)
			arg.setStartTag(resultExpr);
		if (arg.getAttributes() != null) {
			ArrayList newAttributes = new ArrayList(arg.getAttributes().size());
			for (int i = 0; i < arg.getAttributes().size(); i++) {
				AttributeValuePair expri = (AttributeValuePair) arg.getAttributes().get(i);
				expri.accept(this);
				if (resultExpr != null)
					newAttributes.add(resultExpr);
				else
					newAttributes.add(expri);
			}
			arg.setAttributes(newAttributes);
		}
		if (subExpressions != null) {
			ArrayList newSubExpressions = new ArrayList(subExpressions.size());
			for (int i = 0; i < subExpressions.size(); i++) {
				XQueryExpression expri = (XQueryExpression) subExpressions.get(i);
				expri.accept(this);
				if (resultExpr != null)
					newSubExpressions.add(resultExpr);
				else
					newSubExpressions.add(expri);
			}
			arg.setSubExpressions(newSubExpressions);
		}
		// rule ELT-4.
		// this is special for sequences of atomic values -> insert white space
		//		subExpressions = arg.getSubExpressions();
		//		if (subExpressions != null && subExpressions.size() > 1) {
		//			boolean prevAtom = false;
		//			for (int i = 0; i < subExpressions.size(); i++) {
		//				XQueryExpression expri = (XQueryExpression) subExpressions.get(i);
		//				if (expri.getQType().isAtom()) {
		//					if (prevAtom) {
		//						arg.addSubExpressions(i,new ValueText(" ",typeVisitor));
		//						i++;
		//					} else
		//						prevAtom = true;
		//				} else
		//					prevAtom = false;
		//			}
		//		}
		resultExpr = null;
	}

	/*
	--> Rules for class FLWRExpression
	FLWR-1. while return clause is FLWR erase 'return'
	FLWR-2. if return clause is not LocatedExpression, Variable or ElementConstructor create intermediate variable
	FLWR-3. for all possible variable (see condition!) transform for binding into let binding, restructure all variables anyway
	FLWR-4. restructure where clause
	FLWR-5. for all possible variable (see condition!) substitute variable in the other clause and then remove it
	FLWR-6. If no more variable left depending on context recreate FLWR and restructure result
	FLWR-7. If only let clauses add dummy for clause in front
	FLWR-8. restructure return clause
	*/
	public void visit(FLWRExpression arg) throws XQueryException {
		if (isRoot) {
			isRoot = false;
		}
		// handling case when variables are bound on union elements
		// not done yet

		XQueryExpression returnClause = arg.getReturnClause();
		XQueryExpression whereClause = arg.getWhereClause();
		// return clause should be either LocatedExpression, Variable, ElementConstructor, SortedExprression
		// rule FLWR-0.
		//		if (returnClause instanceof SortedExpression) {
		//			returnClause.accept(this);
		//			returnClause = resultExpr;
		//			arg.setReturnClause(resultExpr);
		//			FLWRExpression flwr = (FLWRExpression) returnClause;
		//			if (flwr.getOrderBy() != null)
		//				for (int i=0;i<flwr.getOrderBy().size();i++)
		//					arg.addOrderBy((XQueryExpression)flwr.getOrderBy().get(i));
		//		}
		// rule FLWR-1.
		while (returnClause instanceof FLWRExpression) {
			ArrayList orderBys = ((FLWRExpression) returnClause).getOrderBy();
			if (orderBys == null || orderBys.isEmpty()) {
				ArrayList vars = ((FLWRExpression) returnClause).getVariables();
				for (int i = 0; i < vars.size(); i++)
					arg.addVariable((Variable) vars.get(i));
				if (whereClause == null)
					arg.setWhereClause(((FLWRExpression) returnClause).getWhereClause());
				else
					arg.addWhereClause(((FLWRExpression) returnClause).getWhereClause());
				arg.setReturnClause(((FLWRExpression) returnClause).getReturnClause());
				returnClause = arg.getReturnClause();
			} else
				break;
		}
		// rule FLWR-2.
		if (!(returnClause instanceof XQueryExpressionSequence) && !(returnClause instanceof LocatedExpression) && !(returnClause instanceof Variable) && !(returnClause instanceof Element) && !(returnClause instanceof ITEExpression) && !(returnClause instanceof SortedExpression) && !(returnClause instanceof XMLComment) && !(returnClause instanceof Document) && !(returnClause instanceof XMLProcessingInstruction) && !(returnClause instanceof FLWRExpression) && !(returnClause instanceof ComputedText) && !(returnClause instanceof AttributeValuePair)) {
			Variable var = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, returnClause, Variable.RES_CREATOR);
			arg.addVariable(var);
			arg.setReturnClause(var);
			returnClause = arg.getReturnClause();
		}
		// restructure variables
		// rule FLWR-3.
		ArrayList variables = arg.getVariables();
		for (int j = 0; j < variables.size(); j++) {
			Variable tmpVar = (Variable) variables.get(j);
			// add   22/01/2003
			// erase   23/06/2003
			// add   21/07/2003
			//if (tmpVar.getBindingType() == Constants.FOR_BINDINGTYPE && !tmpVar.isInput() && tmpVar.getExpression().getQType().getOccurence() == QType.OCC_1_1 && variables.size() != 1) {
			if (tmpVar.getCreator() != Variable.RES_CREATOR && tmpVar.getBindingType() == Constants.FOR_BINDINGTYPE && !tmpVar.isInput() && // for testing only(tmpVar.getExpression().getQType().getOccurence() == QType.OCC_1_1 || tmpVar.getExpression().getQType().getOccurence() == QType.OCC_0_1) && 
			tmpVar.getExpression().getQType().getOccurence() == QType.OCC_1_1 && (variables.size() != 1 || arg.getParentExpression() != null)) {
				tmpVar.setBindingType(Constants.LET_BINDINGTYPE);
				ruleNumber++;
			}
			contextStack.push(IN_FLWR);
			parentFLWRExpressionStack.push(arg);
			tmpVar.accept(this);
			contextStack.pop();
			parentFLWRExpressionStack.pop();
		}

		XQueryExpression tmpSubExpr = null;
		variables = arg.getVariables();
		whereClause = arg.getWhereClause();
		returnClause = arg.getReturnClause();
		ArrayList orderBys = arg.getOrderBy();
		// rule FLWR-4.
		if (whereClause != null) {
			parentFLWRExpressionStack.push(arg);
			inWhereExpressionStack.push(arg);
			contextStack.push(IN_WHERE);
			whereClause.accept(this);
			parentFLWRExpressionStack.pop();
			inWhereExpressionStack.pop();
			contextStack.pop();
			XQueryExpression tmpExpr = resultExpr;
			if (tmpExpr != null)
				arg.setWhereClause(tmpExpr);
		}
		// rule FLWR-5.
		boolean noloop = true;
		for (int j = 0; j < variables.size(); j++) {
			Variable tmpVar = (Variable) variables.get(j);
			XQueryExpression tmpVarExpr = tmpVar.getExpression();
			int bindingtype = tmpVar.getBindingType();
			// replacing expressions
			// for $x in value
			// let $x := notcollection && not aggregatefunctioncall
			// let $x := collection && locatedexpression starting with variable
			if (tmpVar.getCreator() != Variable.RES_CREATOR && (tmpVarExpr instanceof Value || // ADD lars 10/02/04 
			//(bindingtype == Constants.LET_BINDINGTYPE && tmpVarExpr instanceof FLWRExpression) ||				
			 (bindingtype == Constants.LET_BINDINGTYPE && !tmpVarExpr.getQType().isMultiple() && //!(tmpVarExpr instanceof AggregateFunctionCall) && 
			!(tmpVarExpr instanceof FunctionCall) && !(tmpVarExpr instanceof ListOpArithExpression) && !(tmpVarExpr instanceof FLWRExpression)) || (bindingtype == Constants.LET_BINDINGTYPE && tmpVarExpr instanceof LocatedExpression && ((LocatedExpression) tmpVarExpr).getSteps().size() != 1 && ((LocatedExpression) tmpVarExpr).getExpression() instanceof Variable && tmpVarExpr.getQType().isMultiple()))) {
				//if (tmpVar.getBindingType() == Constants.LET_BINDINGTYPE && tmpVarExpr instanceof LocatedExpression && tmpVarExpr.getQType().isCollection()) {
				// create substitute map
				HashMap map = new HashMap(1);
				map.put(tmpVar, tmpVarExpr);
				// replace in all LET variable after this one
				for (int i = j + 1; i < variables.size(); i++)
					 ((Variable) variables.get(i)).getExpression().substitute(map, typeVisitor);
				// replace in WHERE clause
				if (whereClause != null) {
					if ((tmpSubExpr = (XQueryExpression) map.get(whereClause)) != null)
						arg.setWhereClause(tmpSubExpr);
					else
						whereClause.substitute(map, typeVisitor);
				}
				//replace in orderby (not necessary -> only mono values accepted)
				if (orderBys != null) {
					for (int i = 0; i < orderBys.size(); i++) {
						XQueryExpression orderby = (XQueryExpression) orderBys.get(i);
						if ((tmpSubExpr = (XQueryExpression) map.get(orderby)) != null)
							orderBys.set(i, tmpSubExpr);
						else
							orderby.substitute(map, typeVisitor);
					}
				}
				// replace in RETURN clause
				if ((tmpSubExpr = (XQueryExpression) map.get(returnClause)) != null)
					arg.setReturnClause(tmpSubExpr);
				else
					returnClause.substitute(map, typeVisitor);
				variables.remove(j);
				j--;
				//			} else if (bindingtype == Constants.FOR_BINDINGTYPE || tmpVarExpr.getQType().isMultiple())
				// change   10/08/2003
			} else if (bindingtype == Constants.FOR_BINDINGTYPE)
				noloop = false;
		}
		// rule FLWR-6.
		if (variables.isEmpty()) {
			Integer tmpInt = this.NO_CONTEXT;
			if (!contextStack.isEmpty())
				tmpInt = (Integer) contextStack.peek();
			if (whereClause != null) {
				if (tmpInt == this.IN_FLWR) {
					((FLWRExpression) this.parentFLWRExpressionStack.peek()).addWhereClause(whereClause);
					resultExpr = arg.getReturnClause();
					ruleNumber++;
				} else {
					// resultExpr = new ITEExpression(whereClause,arg.getReturnClause(),new XQueryVoid(typeVisitor),typeVisitor);
					Variable var = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, new ValueInteger("1", arg.getParentModule()), Variable.RES_CREATOR);
					var.isDummy(true);
					ArrayList vars = new ArrayList(1);
					vars.add(var);
					// CHANGE   30/08/2002 arg.getReturnClause().setRoot(true);
					resultExpr = new FLWRExpression(vars, null, whereClause, arg.getReturnClause(), arg.getParentModule());
					resultExpr.setParentModule(arg.getParentModule());
					isRoot = true;
					ruleNumber++;
				}
			} else {
				resultExpr = arg.getReturnClause();
				isRoot = true;
			}
			// change   19/03/2002
			//resultExpr.clearParentExpression();
			if (resultExpr != null) {
				resultExpr.setParentExpressions(arg.getParentExpression());
				XQueryExpression tmpExpr = resultExpr;
				resultExpr.accept(this);
				if (resultExpr == null)
					resultExpr = tmpExpr;
			}
			return;
		}
		// rule FLWR-7.
		if (noloop) {
			Variable var = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, new ValueInteger("1", arg.getParentModule()), Variable.RES_CREATOR);
			var.isDummy(true);
			arg.insertVariable(var);
		}

		// restructure orderby
		ArrayList orderby = arg.getOrderBy();
		if (orderby != null) {
			for (int i = 0; i < orderby.size(); i++) {
				XQueryExpression orderi = (XQueryExpression) orderby.get(i);
				//				orderi.accept(this);
				//				if (resultExpr != null)
				//					orderby.set(i,resultExpr);
				//				orderi = (XQueryExpression)orderby.get(i);
				if (orderi instanceof Variable || (orderi instanceof LocatedExpression && ((LocatedExpression) orderi).getExpression() instanceof Variable))
					continue;
				// replace with variable
				Variable var = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, orderi, Variable.RES_CREATOR);
				arg.addVariable(var);
				orderby.set(i, var);
			}
		}
		// rule FLWR-8.
		// resructure return clause
		returnClause = arg.getReturnClause();
		inReturn++;
		contextStack.push(IN_RETURN);
		//if (arg.getVariables() != null && !arg.getVariables().isEmpty())
		this.parentFLWRExpressionStack.push(arg);
		returnClause.accept(this);
		//if (arg.getVariables() != null && !arg.getVariables().isEmpty())
		this.parentFLWRExpressionStack.pop();
		inReturn--;
		contextStack.pop();
		XQueryExpression tmpExpr = resultExpr;
		if (tmpExpr != null)
			arg.setReturnClause(tmpExpr);

		// add LARS 09/03/04
		// handling FLWR in function call
		Integer tmpInt = this.NO_CONTEXT;
		if (!contextStack.isEmpty())
			tmpInt = (Integer) contextStack.peek();
		if (tmpInt == this.IN_FUNCTION) {
			// add variable to parentflwr
			FLWRExpression lastflwr = (FLWRExpression) parentFLWRExpressionStack.peek();
			Variable addvar = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, arg, Variable.RES_CREATOR);
			if (varDefinitionStack.isEmpty()) {
				lastflwr.addVariable(addvar);
			} else {
				Variable lastvar = (Variable) varDefinitionStack.peek();
				int prevpos = lastflwr.getVarPosition(lastvar);
				if (prevpos != -1)
					lastflwr.addVariable(prevpos, addvar);
				else
					lastflwr.addVariable(addvar);
			}
			resultExpr = addvar;
			return;
		}

		// add LARS 14/06/04
		// erasing dummy variables for $x in 1 whenever possible
		if (isMediator) {
			while (((Variable) arg.getVariables().get(0)).isDummy()) {
				if (arg.getVariables().size() > 1)
					arg.getVariables().remove(0);
				else
					break;
			}
		}

		// reset result
		resultExpr = null;
	}

	public void visit(ListOpArithExpression arg) throws XQueryException {
		if (isRoot) {
			makeDummyOneFlwr(arg);
			return;
		}
		//resultExpr = null;
		// test if variable should be added
		if (!contextStack.isEmpty()) {
			Integer tmpInt = (Integer) contextStack.peek();
			if (tmpInt.equals(IN_RETURN)) {
				XQueryExpression parent0Expr = (XQueryExpression) arg.getParentExpression().get(0);
				if (!(parent0Expr instanceof Variable)) {
					FLWRExpression lastflwr = (FLWRExpression) parentFLWRExpressionStack.peek();
					Variable addvar = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, arg, Variable.RES_CREATOR);
					if (varDefinitionStack.isEmpty()) {
						lastflwr.addVariable(addvar);
					} else {
						Variable lastvar = (Variable) varDefinitionStack.peek();
						int prevpos = lastflwr.getVarPosition(lastvar);
						if (prevpos != -1)
							lastflwr.addVariable(prevpos, addvar);
						else
							lastflwr.addVariable(addvar);
					}
					resultExpr = addvar;
					return;
				}
			}
		}
		contextStack.push(IN_FUNCTION);
		XQueryExpression tmpExpr = arg.getExpression1();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression1(resultExpr);
		}
		tmpExpr = arg.getExpression2();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression2(resultExpr);
		}
		resultExpr = null;
		contextStack.pop();
	}

	public void visit(ITEExpression arg) throws XQueryException {
		if (isRoot) {
			throw new XQueryException(arg.getClass().getName() + " cannot be root expression of parse tree");
		}
		if (contextStack.peek() != IN_RETURN)
			throw new XQueryException(arg.getClass().getName() + " can only appear in a return clause");
		// add LARS 09/03/04
		// special test for empty and exists functions
		XQueryExpression ifExpr = arg.getIfExpression();
		while (ifExpr instanceof FunctionNOT) {
			arg.setIfExpression(((FunctionNOT) ifExpr).getArgument(0));
			XQueryExpression expr = arg.getElseExpression();
			arg.setElseExpression(arg.getThenExpression());
			arg.setThenExpression(expr);
		}
		ifExpr = arg.getIfExpression();
		if (ifExpr instanceof FunctionEMPTY && ((FunctionEMPTY) ifExpr).getArgument(0).getQType().isMultiple()) {
			ArrayList args = new ArrayList(1);
			args.add(((FunctionEMPTY) ifExpr).getArgument(0));
			arg.setIfExpression(new ListOpCompExpression(new FunctionCOUNT(args, arg.getParentModule()), new ValueInteger("0", arg.getParentModule()), Constants.EQ_COMPOP, arg.getParentModule()));
		} else if (ifExpr instanceof FunctionEXISTS && ((FunctionEXISTS) ifExpr).getArgument(0).getQType().isMultiple()) {
			ArrayList args = new ArrayList(1);
			args.add(((FunctionEXISTS) ifExpr).getArgument(0));
			arg.setIfExpression(new ListOpCompExpression(new FunctionCOUNT(args, arg.getParentModule()), new ValueInteger("0", arg.getParentModule()), Constants.NEQ_COMPOP, arg.getParentModule()));
		} else {
			QType qtype = ifExpr.getQType();
			if (qtype != null && qtype.isMultiple()) {
				ArrayList args = new ArrayList(1);
				args.add(ifExpr);
				arg.setIfExpression(new ListOpCompExpression(new FunctionCOUNT(args, arg.getParentModule()), new ValueInteger("0", arg.getParentModule()), Constants.NEQ_COMPOP, arg.getParentModule()));
			}
		}
		contextStack.push(IN_FUNCTION);
		arg.getIfExpression().accept(this);
		if (resultExpr != null) {
			arg.setIfExpression(resultExpr);
		}
		//		arg.getThenExpression().accept(this);
		//		if (resultExpr != null) {
		//			arg.setThenExpression(resultExpr);
		//		}
		//		arg.getElseExpression().accept(this);
		//		if (resultExpr != null) {
		//			arg.setElseExpression(resultExpr);
		//		}
		contextStack.pop();
		resultExpr = arg;
	}

	/*
	--> Rules for class LocatedExpression
	LOC-1. if no parents or parent is sorted expression create FLWR expression
	LOC-2. if parents and parent FLWR and parent is not FunctionCall and parent is not variable and first step expression is rooNodeFunctionCall then replace expression with intermediaite variable
	LOC-3. if in return and qtype is multiple and (first step expression is not variable or first step expression qtype is multiple) and (no parent or parent is variable) then create new FLWR to iterate on expr
	*/
	public void visit(LocatedExpression arg) throws XQueryException {
		if (isRoot) {
			makeDummyFlwr(arg);
			return;
		}
		resultExpr = null;
		//Step step0 = (Step) arg.getSteps().get(0);
		ArrayList parents = arg.getParentExpression();
		boolean noParents = false;
		if (parents == null || parents.isEmpty())
			noParents = true;
		// rule LOC-1.
		if (noParents || (XQueryExpression) parents.get(0) instanceof SortedExpression) {
			if (arg.getQType() != null && arg.getQType().isMultiple() && !(arg.getExpression() instanceof Variable)) {
				Variable var = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, arg, Variable.RES_CREATOR);
				ArrayList varlist = new ArrayList(1);
				varlist.add(var);
				FLWRExpression tmpFLWR = new FLWRExpression(varlist, null, null, var, arg.getParentModule());
				tmpFLWR.setParentModule(arg.getParentModule());
				resultExpr = tmpFLWR;
				ruleNumber++;
			}
		}
		// rule LOC-2.
		// add   11/06/2003
		else if (!noParents && !parentFLWRExpressionStack.isEmpty() && !((XQueryExpression) parents.get(0) instanceof FunctionCall) && !((XQueryExpression) parents.get(0) instanceof Variable) && arg.getExpression() instanceof InputFunctionCall) {
			Variable var = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, arg, Variable.RES_CREATOR);
			FLWRExpression parentflwr = (FLWRExpression) parentFLWRExpressionStack.peek();
			parentflwr.addVariable(var);
			resultExpr = var;
		}
		// end add   11/06/2003
		// rule LOC-3.
		else if (!contextStack.isEmpty()) {
			Integer tmpInt = (Integer) contextStack.peek();
			if (tmpInt.equals(IN_RETURN)) {
				// condition to be added to be ok in repository
				QType qtype = arg.getQType();
				if (qtype != null && qtype.isMultiple() && (!(arg.getExpression() instanceof Variable) || arg.getExpression().getQType().isMultiple()) && (noParents || !(parents.get(0) instanceof Variable))) {
					Variable tmpVar = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, arg, Variable.RES_CREATOR);
					ArrayList vars = new ArrayList(1);
					vars.add(tmpVar);
					FLWRExpression tmpFLWR = new FLWRExpression(vars, null, tmpVar, arg.getParentModule());
					tmpFLWR.setParentModule(arg.getParentModule());
					resultExpr = tmpFLWR;
					ruleNumber++;
				}
			}
		}

	}

	/*
	--> Rules for class PrimitiveFunctionCall
	PRIM-1. iterate on arguments and replace them if necessary
	*/
	public void visit(PrimitiveFunctionCall arg) throws XQueryException {
		if (isRoot) {
			makeDummyFlwr(arg);
			return;
		}
		resultExpr = null;
		Integer context = null;
		if (!this.contextStack.isEmpty())
			context = (Integer) this.contextStack.peek();
		// test if variable should be added
		//XQueryExpression parent0Expr = (XQueryExpression) arg.getParentExpression().get(0);
		if (context != IN_FLWR && !(arg.getQType() != null && arg.getQType().isBoolean()) && ((context == this.IN_RETURN && !(arg instanceof FunctionQNAME) && !(arg instanceof FunctionEXPANDED_QNAME)) || arg instanceof AggregateFunctionCall || arg instanceof InputFunctionCall))
			//		(
			//		!(parent0Expr instanceof Variable) && 
			//		!(arg.getQType() != null && arg.getQType().isBoolean()) && 
			//		(inReturn > 0 || arg instanceof AggregateFunctionCall || arg instanceof InputFunctionCall)
			//		) 
			{
			FLWRExpression lastflwr = (FLWRExpression) parentFLWRExpressionStack.peek();
			Variable addvar = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, arg, Variable.RES_CREATOR);
			if (varDefinitionStack.isEmpty()) {
				lastflwr.addVariable(addvar);
			} else {
				Variable lastvar = (Variable) varDefinitionStack.peek();
				int prevpos = lastflwr.getVarPosition(lastvar);
				if (prevpos != -1)
					lastflwr.addVariable(prevpos, addvar);
				else
					lastflwr.addVariable(addvar);
			}
			resultExpr = addvar;
			return;
		}
		//		if (inReturn > 0 && inWhereExpression.isEmpty() && arg.getQType().isBoolean())
		//			throw new TypeException("Boolean expression is not supported : " + arg);
		contextStack.push(IN_FUNCTION);
		ArrayList arguments = arg.getArguments();
		// rule PRIM-1.
		if (arguments != null) {
			boolean change = false;
			ArrayList newArguments = new ArrayList(arguments.size());
			for (int i = 0; i < arguments.size(); i++) {
				XQueryExpression argi = (XQueryExpression) arguments.get(i);
				argi.accept(this);
				if (resultExpr != null) {
					change = true;
					newArguments.add(resultExpr);
				} else
					newArguments.add(argi);
			}
			if (change) {
				//				arg.setBracket(false);
				arg.setArguments(newArguments);
				//resultExpr = arg;
				ruleNumber++;
			}
			//else
			resultExpr = null;
		}
		contextStack.pop();
	}

	/*
	--> Rules for class QuantifiedExpression
	QUANT-1. if mediator change into count otherwise just restructure inner expressions
	QUANT-2. if mediator dont let any
	*/
	public void visit(QuantifiedExpression arg) throws XQueryException {
		if (isRoot) {
			throw new XQueryException(arg.getClass().getName() + " cannot be root expression of parse tree");
		}
		resultExpr = null;
		// rule QUANT-1.
		if (isMediator) {
			FLWRExpression parentFLWR = (FLWRExpression) this.inWhereExpressionStack.peek();
			ArrayList variables = arg.getVariables();
			for (int i = 0; i < variables.size(); i++) {
				((Variable) variables.get(i)).setBindingType(Constants.FOR_BINDINGTYPE);
				((Variable) variables.get(i)).addSubPath(((Variable) variables.get(i)));
			}
			int kind = arg.getKind();
			ArrayList countArgs = new ArrayList(1);
			XQueryExpression tmpExpr = null;
			Variable tmpvar = null;
			if (kind == Constants.SOME_QUANTIFIER) {
				if (variables.size() > 1)
					tmpExpr = new FLWRExpression(variables, arg.getExpression(), new XQueryExpressionSequence((ArrayList) variables.clone(), arg.getParentModule()), arg.getParentModule());
				else
					tmpExpr = new FLWRExpression(variables, arg.getExpression(), (XQueryExpression) variables.get(0), arg.getParentModule());
				tmpExpr.accept(this);
				if (resultExpr != null)
					tmpExpr = resultExpr;
				if (tmpExpr instanceof Variable)
					tmpvar = (Variable)tmpExpr;
				else {
					tmpvar = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, tmpExpr, Variable.RES_CREATOR);
					parentFLWR.addVariable(tmpvar);
				}					
				countArgs.add(tmpvar);
				tmpExpr = new FunctionCOUNT(countArgs, arg.getParentModule());
				tmpExpr = new ListOpCompExpression(tmpExpr, new ValueInteger("0", arg.getParentModule()), Constants.GT_COMPOP, arg.getParentModule());
			} else { // if (kind == Constants.EVERY_QUANTIFIER) {
				ArrayList args = new ArrayList(1);
				args.add(arg.getExpression());
				if (variables.size() > 1)
					tmpExpr = new FLWRExpression(variables, new FunctionNOT(args, arg.getParentModule()), new XQueryExpressionSequence((ArrayList) variables.clone(), arg.getParentModule()), arg.getParentModule());
				else
					tmpExpr = new FLWRExpression(variables, new FunctionNOT(args, arg.getParentModule()), (XQueryExpression) variables.get(0), arg.getParentModule());
				tmpExpr.accept(this);
				if (resultExpr != null)
					tmpExpr = resultExpr;
				if (tmpExpr instanceof Variable)
					tmpvar = (Variable)tmpExpr;
				else {
					tmpvar = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, tmpExpr, Variable.RES_CREATOR);
					parentFLWR.addVariable(tmpvar);
				}
				countArgs.add(tmpvar);
				tmpExpr = new FunctionCOUNT(countArgs, arg.getParentModule());
				tmpExpr = new ListOpCompExpression(tmpExpr, new ValueInteger("0", arg.getParentModule()), Constants.EQ_COMPOP, arg.getParentModule());
			}
			tmpExpr.accept(this);
			if (resultExpr == null)
				resultExpr = tmpExpr;
		} else {
			for (int i = 0; i < arg.getVariables().size(); i++) {
				Variable var = (Variable) arg.getVariables().get(i);
				XQueryExpression varExpr = var.getExpression();
				varExpr.accept(this);
			}
			XQueryExpression tmpExpr = arg.getExpression();
			tmpExpr.accept(this);
			if (resultExpr != null) {
				arg.setExpression(resultExpr);
			}
		}
	}

	/*
	// to be erased in the latest version of the specification
	--> Rules for class SortedExpression
	SORT-1. sort by only supported with located expression or FLWR
	SORT-2. if qtype is not multiple sort clauses are not useful!
	SORT-3. restructure expression to FLWR
	SORT-4. rewrite sort clauses to order by clauses of FLWR
	SORT-5. verify that resulting sort clauses are not constants
	*/
	public void visit(SortedExpression arg) throws XQueryException {
		if (isRoot) {
			isRoot = false;
		}
		XQueryExpression expr = arg.getExpression();
		// ATTENTION : this must be extended to other possible expressions
		// for now SORTBY can only be used with sequence of single FLWR expression !!!
		// rule SORT-1.
		if (!(expr instanceof FLWRExpression) && !(expr instanceof LocatedExpression))
			throw new RestructureException("SORTBY can only be used with sequence of single FLWR expression or with a LocatedExpression!!!");
		// rule SORT-2.
		if (!expr.getQType().isMultiple()) {
			expr.accept(this);
			if (resultExpr == null)
				resultExpr = expr;
			return;
		}
		// rule SORT-3.
		FLWRExpression tmpFLWR = null;
		if (expr instanceof LocatedExpression) {
			expr.accept(this);
			if (resultExpr == null || !(resultExpr instanceof FLWRExpression))
				throw new RestructureException("SORTBY used with incorrect path expression");
			tmpFLWR = (FLWRExpression) resultExpr;
		} else
			tmpFLWR = (FLWRExpression) expr;
		// restructure flwr first to avoid strange(!!!) expression in order by
		//!!! danger it is assumed that a restructured FLWR is still a FLWR
		tmpFLWR.accept(this);
		if (resultExpr != null)
			tmpFLWR = (FLWRExpression) resultExpr;
		// rule SORT-4.
		ArrayList sortClauses = arg.getSortClauses();
		XQueryExpression returnClause = tmpFLWR.getReturnClause();
		ArrayList newSortClauses = new ArrayList(sortClauses.size());
		SubExpressionVisitor subvisitor = new SubExpressionVisitor(typeVisitor);
		// XQueryExpression tmpExpr = null;
		for (int i = 0; i < sortClauses.size(); i++) {
			XQueryExpression sortiexpr = (XQueryExpression) sortClauses.get(i);
			// W3C WD 16 August 2002
			int[] order = sortiexpr.getOrder();
			sortiexpr.setOrder(initOrder);
			ArrayList resultExprs = null;
			XQueryExpression tmpExpr = null;

			// to handle qname
			if (sortiexpr instanceof QName) {
				ruleNumber++;
				subvisitor.setSubExpression((QName) sortiexpr, true);
				returnClause.accept(subvisitor);
				resultExprs = subvisitor.getResultExprs();
				if (resultExprs.isEmpty() || resultExprs.size() > 1)
					throw new TypeException("Invalid sort clause : " + sortiexpr);
				tmpExpr = (XQueryExpression) resultExprs.get(0);
			} else if (sortiexpr instanceof LocatedExpression) {
				LocatedExpression tmpLoc = (LocatedExpression) sortiexpr;
				ArrayList steps = tmpLoc.getSteps();
				Step tmpStep0 = (Step) steps.get(0);
				if (tmpStep0.hasSeparator())
					throw new TypeException("Invalid sort clause : " + sortiexpr);
				XQueryExpression tmpStep0Expr = tmpStep0.getExpression();
				if (steps.size() == 1) {
					ruleNumber++;
					// to handle . or self::node()
					if (tmpStep0Expr instanceof NodeTest) {
						NodeTest tmpNode = (NodeTest) tmpStep0.getExpression();
						if (tmpNode.getKind() != NodeKind.NODE)
							throw new TypeException("Invalid sort clause : " + sortiexpr);
						if (tmpStep0.getAxis() != Axis.SELF)
							throw new TypeException("Invalid sort clause : " + sortiexpr);
						subvisitor.setSubExpression(null, tmpStep0.getAxis(), 0, true);
						returnClause.accept(subvisitor);
						resultExprs = subvisitor.getResultExprs();
						if (resultExprs.isEmpty() || resultExprs.size() > 1)
							throw new TypeException("Invalid sort clause : " + sortiexpr);
						tmpExpr = (XQueryExpression) resultExprs.get(0);
					}
					// to handle @qname or attribute::qname or child::qname
					else if (tmpStep0Expr instanceof QName) {
						QName tmpQName = (QName) tmpStep0.getExpression();
						if (tmpStep0.getAxis() != Axis.CHILD && tmpStep0.getAxis() != Axis.ATTRIBUTE)
							throw new TypeException("Invalid sort clause : " + sortiexpr);
						subvisitor.setSubExpression(tmpQName, tmpStep0.getAxis(), /*(tmpStep0.getAxis() == Axis.CHILD && i!=0) ? 0 :*/
						1, true);
						returnClause.accept(subvisitor);
						resultExprs = subvisitor.getResultExprs();
						if (resultExprs.isEmpty() || resultExprs.size() > 1)
							throw new TypeException("Invalid sort clause : " + sortiexpr);
						tmpExpr = (XQueryExpression) resultExprs.get(0);
					} else
						throw new TypeException("Invalid sort clause : " + sortiexpr);
				} else {
					// to handle step/step/.......step
					ArrayList tmpExprList = new ArrayList(1);
					tmpExprList.add(returnClause);
					int depth = 1;
					for (int j = 0; j < steps.size(); j++) {
						ruleNumber++;
						//if (depth != -1) if (j != steps.size()-1) depth = 1; else depth = 0;
						Step stepj = (Step) steps.get(j);
						// first step must be without separator!
						if (j == 0 && stepj.hasSeparator())
							throw new TypeException("Invalid sort clause : " + sortiexpr);
						XQueryExpression stepjExpr = stepj.getExpression();
						if (stepjExpr instanceof QName) {
							if (j != 0 && stepj.getAxis() != Axis.CHILD && stepj.getAxis() != Axis.ATTRIBUTE)
								throw new TypeException("Invalid sort clause : " + sortiexpr);
							if (stepj.getAxis() == Axis.ATTRIBUTE && j != steps.size() - 1)
								throw new TypeException("Invalid sort clause : " + sortiexpr);
							if (depth == -1) {
								depth = 0;
								subvisitor.setSubExpression((QName) stepjExpr, stepj.getAxis(), depth, j == steps.size() - 1 ? true : false);
								for (int k = 0; k < tmpExprList.size(); k++) {
									XQueryExpression tmpExprListk = (XQueryExpression) tmpExprList.get(k);
									tmpExprListk.accept(subvisitor);
									if (resultExprs == null)
										resultExprs = (ArrayList) subvisitor.getResultExprs().clone();
									else
										resultExprs.addAll(subvisitor.getResultExprs());
								}
								depth = -1;
							}
							if (depth != -1)
								if (j == 0)
									depth = 1;
								else
									depth = 0;
							subvisitor.setSubExpression((QName) stepjExpr, stepj.getAxis(), depth, j == steps.size() - 1 ? true : false);
							for (int k = 0; k < tmpExprList.size(); k++) {
								XQueryExpression tmpExprListk = (XQueryExpression) tmpExprList.get(k);
								tmpExprListk.accept(subvisitor);
								if (resultExprs == null)
									resultExprs = (ArrayList) subvisitor.getResultExprs().clone();
								else
									resultExprs.addAll(subvisitor.getResultExprs());
							}
							if (depth == -1)
								if (j != steps.size() - 1)
									depth = 1;
								else
									depth = 0;
							//if (depth == -1) depth = 1;
							if (resultExprs.isEmpty() || (j == steps.size() - 1 && resultExprs.size() > 1))
								throw new TypeException("Invalid sort clause : " + sortiexpr);
							tmpExprList = (ArrayList) resultExprs.clone();
							resultExprs = null;
						} else if (stepjExpr instanceof NodeTest) {
							NodeTest tmpNode = (NodeTest) stepj.getExpression();
							if (tmpNode.getKind() != NodeKind.NODE)
								throw new TypeException("Invalid sort clause : " + sortiexpr);
							if (stepj.getAxis() != Axis.DESCENDANT_OR_SELF)
								throw new TypeException("Invalid sort clause : " + sortiexpr);
							if (tmpNode.getKind() == NodeKind.NODE)
								depth = -1;
						} else
							throw new TypeException("Invalid sort clause : " + sortiexpr);
					}
					tmpExpr = (XQueryExpression) tmpExprList.get(0);
				}
			}
			if (tmpExpr != null) {
				//tmpExpr.setBrace(false);
				tmpExpr.setOrder(order);
				newSortClauses.add(tmpExpr);
			}

		}

		// verifcation that sort clauses have correct typing
		// erase some sort clause that are not interesting -> constants
		// -> either constants, or expression containing only varibales not used in the current flwr expression
		// rule SORT-5.
		if (!newSortClauses.isEmpty()) {
			GetBoundVariablesVisitor gbvv = new GetBoundVariablesVisitor(true, false);
			for (int i = 0; i < newSortClauses.size(); i++) {
				gbvv.reset();
				XQueryExpression sortiexpr = (XQueryExpression) newSortClauses.get(i);

				sortiexpr.accept(gbvv);
				ArrayList tmplist = gbvv.getVariables();
				if (tmplist == null || tmplist.isEmpty()) {
					newSortClauses.remove(i);
					i--;
					continue;
				}
				boolean cont = false;
				for (int j = 0; j < tmplist.size(); j++) {
					if (tmpFLWR.getVariables().contains(tmplist.get(j))) {
						cont = true;
						break;
					}
				}
				if (cont == false) {
					newSortClauses.remove(i);
					i--;
					continue;
				}
				// verify sort clause typing
				QType qtypei = sortiexpr.getQType();
				if (qtypei == null || !qtypei.isAtomic() || qtypei.isMultiple())
					throw new TypeException("Invalid sort clause : " + sortiexpr);
			}
		}
		if (!newSortClauses.isEmpty())
			tmpFLWR.setOrderBy(newSortClauses);
		resultExpr = tmpFLWR;
	}

	public void visit(UnOpMinusExpression arg) throws XQueryException {
		if (isRoot) {
			makeDummyOneFlwr(arg);
			return;
		}
		// test if variable should be added
		XQueryExpression parent0Expr = (XQueryExpression) arg.getParentExpression().get(0);
		if (!(parent0Expr instanceof Variable)) {
			FLWRExpression lastflwr = (FLWRExpression) parentFLWRExpressionStack.peek();
			Variable addvar = typeVisitor.getStaticContext().createVariable(Constants.LET_BINDINGTYPE, arg, Variable.RES_CREATOR);
			if (varDefinitionStack.isEmpty()) {
				lastflwr.addVariable(addvar);
			} else {
				Variable lastvar = (Variable) varDefinitionStack.peek();
				int prevpos = lastflwr.getVarPosition(lastvar);
				if (prevpos != -1)
					lastflwr.addVariable(prevpos, addvar);
				else
					lastflwr.addVariable(addvar);
			}
			resultExpr = addvar;
			return;
		}
		contextStack.push(IN_FUNCTION);
		XQueryExpression tmpExpr = arg.getExpression();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression(resultExpr);
		}
		resultExpr = null;
		contextStack.pop();
	}

	public void visit(ValidateExpression arg) throws XQueryException {
		XQueryExpression tmpExpr = arg.getExpression();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression(resultExpr);
		}
		resultExpr = null;
	}

	public void visit(Value arg) throws XQueryException {
		if (isRoot) {
			makeDummyFlwr(arg);
			return;
		}
		resultExpr = null;
	}

	/*
	--> Rules for class Variable
	VAR-1. if variable definition then restructure associated expression
	VAR-2. if in return and qtype is multiple then create FLWR to iterate on it
	*/
	public void visit(Variable arg) throws XQueryException {
		if (isRoot) {
			throw new XQueryException(arg.getClass().getName() + " cannot be root expression of parse tree");
		}
		resultExpr = null;
		Integer context = null;
		if (!this.contextStack.isEmpty())
			context = (Integer) this.contextStack.peek();
		// rule VAR-1.
		if (context != null && context.equals(this.IN_FLWR)) {
			varDefinitionStack.add(arg);
			arg.getExpression().accept(this);
			varDefinitionStack.pop();
			if (resultExpr != null)
				arg.setExpression(resultExpr,true);
			resultExpr = arg;
			ruleNumber++;
		} else if (context != null && (context.equals(this.IN_WHERE))) {
			//			QType qtype = arg.getQType();
			//			XQueryExpression tmpExpr = (XQueryExpression) inWhereExpression.peek();
			//			if (!qtype.isBoolean() && (tmpExpr instanceof BinOpANDExpression || tmpExpr instanceof BinOpORExpression)) {
			//				ArrayList tmpList = new ArrayList(1);
			//				tmpList.add(arg);
			//				resultExpr = new FunctionEXISTS(tmpList, typeVisitor);
			//				ruleNumber++;
			//			}
		} else {
			// rule VAR-2.
			if (arg.getExpression() == null)
				return;
			QType qtype = arg.getQType();
			// verification that type is not boolean (not supported yet)
			//			if (qtype.isBoolean())
			//				throw new TypeException("Boolean expression  is not supported : " + arg.getExpression());
			if (context != null && context.equals(this.IN_RETURN) && qtype.isMultiple()) {
				// create new variable to iterate on
				Variable tmpVar = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, arg, Variable.RES_CREATOR);
				ArrayList vars = new ArrayList(1);
				vars.add(tmpVar);
				FLWRExpression tmpFLWR = new FLWRExpression(vars, null, tmpVar, arg.getParentModule());
				tmpFLWR.setParentModule(arg.getParentModule());
				resultExpr = tmpFLWR;
				ruleNumber++;
			}
		}
	}

	/*
	--> Rules for class XQueryBinaryOperatorExpression
	BINOP-1. restructure both operands and replace them if necessary
	*/
	public void visit(XQueryBinaryOperatorExpression arg) throws XQueryException {
		if (isRoot) {
			throw new XQueryException(arg.getClass().getName() + " cannot be root expression of parse tree");
		}
		XQueryExpression tmpExpr = arg.getExpression1();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression1(resultExpr);
		}
		tmpExpr = arg.getExpression2();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression2(resultExpr);
		}
		resultExpr = null;
	}

	//	public void visit(XQueryBooleanOperatorExpression arg) throws XQueryException;

	public void visit(XQueryExpression arg) throws XQueryException {
		//		if (isRoot) {
		//			makeDummyOneFlwr(arg);
		//			return;
		//		}
		resultExpr = null;
	}

	/*
	--> Rules for class XQueryExpressionSequence
	SEQ-1. restructure all operands and replace them if necessary
	*/
	public void visit(XQueryExpressionSequence arg) throws XQueryException {
		if (isRoot) {
			makeDummyFlwr(arg);
			return;
		}
		resultExpr = null;
		ArrayList subExpressions = arg.getSubExpressions();
		if (subExpressions != null) {
			ArrayList newSubExpressions = new ArrayList(subExpressions.size());
			for (int i = 0; i < subExpressions.size(); i++) {
				XQueryExpression expri = (XQueryExpression) subExpressions.get(i);
				expri.accept(this);
				if (resultExpr != null)
					newSubExpressions.add(resultExpr);
				else
					newSubExpressions.add(expri);
			}
			arg.setSubExpressions(newSubExpressions);
			resultExpr = arg;
		}
	}

	public void visit(XQueryModule arg) throws XQueryException {
		ArrayList expressions = arg.getExpressions();
		if (expressions != null)
			for (int i = 0; i < expressions.size(); i++) {
				((XQueryExpression) expressions.get(i)).accept(this);
				if (resultExpr != null) {
					expressions.set(i, resultExpr);
				}
			}
		resultExpr = null;
	}

	/*
	--> Rules for class XQueryUnaryOperatorExpression
	SEQ-1. restructure operand and replace it if necessary
	*/
	public void visit(XQueryUnaryOperatorExpression arg) throws XQueryException {
		if (isRoot) {
			throw new XQueryException(arg.getClass().getName() + " cannot be root expression of parse tree");
		}
		XQueryExpression tmpExpr = arg.getExpression();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression(resultExpr);
		}
		resultExpr = null;
	}

	public void visit(XMLComment arg) throws XQueryException {
		if (isRoot) {
			makeDummyOneFlwr(arg);
			return;
		}
		XQueryExpression tmpExpr = arg.getExpression();
		tmpExpr.accept(this);
		if (resultExpr != null) {
			arg.setExpression(resultExpr);
		}
		resultExpr = null;
	}

	// private methods

	private void makeDummyFlwr(XQueryExpression arg) throws XQueryException {
		isRoot = false;
		Variable var = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, arg, Variable.RES_CREATOR);
		ArrayList vars = new ArrayList(1);
		vars.add(var);
		arg.setRoot(true);
		FLWRExpression tmpFLWR = new FLWRExpression(vars, null, null, var, arg.getParentModule());
		tmpFLWR.setParentModule(arg.getParentModule());
		parentFLWRExpressionStack.push(tmpFLWR);
		tmpFLWR.accept(this);
		parentFLWRExpressionStack.pop();
		if (resultExpr == null)
			resultExpr = tmpFLWR;
	}

	private void makeDummyOneFlwr(XQueryExpression arg) throws XQueryException {
		isRoot = false;
		Variable var = typeVisitor.getStaticContext().createVariable(Constants.FOR_BINDINGTYPE, new ValueInteger("1", arg.getParentModule()), Variable.RES_CREATOR);
		var.isDummy(true);
		ArrayList vars = new ArrayList(1);
		vars.add(var);
		arg.setRoot(true);
		FLWRExpression tmpFLWR = new FLWRExpression(vars, null, null, arg, arg.getParentModule());
		tmpFLWR.setParentModule(arg.getParentModule());
		parentFLWRExpressionStack.push(tmpFLWR);
		tmpFLWR.accept(this);
		parentFLWRExpressionStack.pop();
		if (resultExpr == null)
			resultExpr = tmpFLWR;
	}

}
