/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.*;

import org.xquark.xquery.typing.TypeException;

public class Element extends XQueryExpression implements Cloneable {

	private static final String RCSRevision = "$Revision: 1.12 $";
	private static final String RCSName = "$Name:  $";

	// cannot be null
	protected XQueryExpression startTag = null;
	
	protected HashMap localPrefixes = null;

	// false by default
	protected boolean computed = false;
	private byte numComputedNS = 0;
	private byte numComputedAtt = 0;
	ArrayList attributes = null; // ArrayList of AttributeValuePair
	ArrayList subExpressions = null; // ArrayList of XQueryExpression

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################

	public Element(XQueryExpression startTag, HashMap localPrefixes, ArrayList attributes, ArrayList subExpressions, XQueryModule parentModule) throws TypeException, XQueryException {
		setStartTag(startTag);
		setLocalPrefixes(localPrefixes);
		setAttributes(attributes);
		setSubExpressions(subExpressions);
		setIsComputed(false);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	public Element(QName startTag, HashMap localPrefixes, ArrayList attributes, ArrayList subExpressions, XQueryModule parentModule) throws TypeException, XQueryException {
		setStartTag(startTag);
		setLocalPrefixes(localPrefixes);
		setAttributes(attributes);
		setSubExpressions(subExpressions);
		setIsComputed(false);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	public XQueryExpression getStartTag() {
		return startTag;
	}
	public void setStartTag(XQueryExpression startTag) throws XQueryException {
		if (startTag == null)
			throw new XQueryException("startTag of Element cannot be null");
		this.startTag = startTag;
		this.startTag.setParentExpression(this);
		this.startTag.setParentModule(parentModule);
	}

	public HashMap getLocalPrefixes() {
		return localPrefixes;
	}
	public void setLocalPrefixes(HashMap localPrefixes) {
		this.localPrefixes = localPrefixes;
	}

	public ArrayList getAttributes() {
		return attributes;
	}
	public void setAttributes(ArrayList attributes) {
		this.attributes = attributes;
		if (this.attributes != null)
			for (int j = 0; j < this.attributes.size(); j++) {
				AttributeValuePair attributeValuePair = (AttributeValuePair) this.attributes.get(j);
				attributeValuePair.setParentExpression(this);
				attributeValuePair.setParentModule(parentModule);
			}
	}

	public ArrayList getSubExpressions() {
		return subExpressions;
	}
	public void setSubExpressions(ArrayList subExpressions) {
		this.subExpressions = subExpressions;
		if (this.subExpressions != null)
			for (int j = 0; j < this.subExpressions.size(); j++) {
				XQueryExpression expression = (XQueryExpression) this.subExpressions.get(j);
				expression.setParentExpression(this);
				expression.setParentModule(parentModule);
			}
	}

	public boolean isComputed() {
		return computed;
	}

	public void setIsComputed(boolean computed) {
		this.computed = computed;
	}

	public byte getComputedAttNum() {
		return numComputedAtt;
	}
	
	public void setComputedAttNum(byte num) {
		numComputedAtt = num;
	}
	
	public byte getComputedNSNum() {
		return numComputedNS;
	}
	
	public void setComputedNSNum(byte num) {
		numComputedNS = num;
	}

	public void addSubExpressions(XQueryExpression subExpression) throws XQueryException {
		subExpressions.add(subExpression);
		subExpression.setParentExpression(this);
		subExpression.setParentModule(parentModule);
	}

	public void addSubExpressions(int index, XQueryExpression subExpression) throws XQueryException {
		subExpressions.add(index, subExpression);
		subExpression.setParentExpression(this);
		subExpression.setParentModule(parentModule);
	}

	public void addAttributes(AttributeValuePair attribute) throws XQueryException {
		if (attributes == null)
			attributes = new ArrayList();
		attributes.add(attribute);
		attribute.setParentExpression(this);
		attribute.setParentModule(parentModule);
	}

	public void addSubExpressions(int index, AttributeValuePair attribute) throws XQueryException {
		attributes.add(index, attribute);
		attribute.setParentExpression(this);
		attribute.setParentModule(parentModule);
	}

	public Set getSourceNames() {
		if (sourceNames != null) return sourceNames;
		if (subExpressions == null) return null;
		Set tmplist = null;
		for (int i = 0; i < subExpressions.size(); i++) {
			XQueryExpression expressioni = (XQueryExpression) subExpressions.get(i);
			if (expressioni.getSourceNames() != null && !expressioni.getSourceNames().isEmpty()) {
				if (tmplist == null)
					tmplist = new HashSet(expressioni.getSourceNames());
				else 
					tmplist.addAll(expressioni.getSourceNames());
			}
		}
		return tmplist;
	}

	public Set getUrls() {
		if (urls != null) return urls;
		if (subExpressions == null) return null;
		Set tmplist = null;
		for (int i = 0; i < subExpressions.size(); i++) {
			XQueryExpression expressioni = (XQueryExpression) subExpressions.get(i);
			if (expressioni.getUrls() != null && !expressioni.getUrls().isEmpty()) {
				if (tmplist == null) 
					tmplist = new HashSet(expressioni.getUrls());
				else 
					tmplist.addAll(expressioni.getUrls());
			}
		}
		return tmplist;
	}

	// #############################################################################
	// CLONE STUFF
	// #############################################################################
	public void addParentExpression(XQueryExpression parentExpression) {
		addParentExpression(parentExpression);

		startTag.addParentExpression(parentExpression);
		if (this.attributes != null)
			for (int j = 0; j < this.attributes.size(); j++)
				 ((AttributeValuePair) this.attributes.get(j)).addParentExpression(parentExpression);
		if (subExpressions != null)
			for (int j = 0; j < subExpressions.size(); j++)
				 ((XQueryExpression) subExpressions.get(j)).addParentExpression(parentExpression);
	}


}
