/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.ArrayList;

import org.xquark.xpath.Axis;
import org.xquark.xpath.PathExpr;
import org.xquark.xpath.StepExpr;
import org.xquark.xquery.normalize.LocatedExpressionReducer;
import org.xquark.xquery.typing.TypeException;

public class LocatedExpression extends XQueryExpression implements PathExpr, Cloneable {

	private static final String RCSRevision = "$Revision: 1.18 $";
	private static final String RCSName = "$Name:  $";

	// cannot be null or empty
	protected ArrayList steps = null;

	protected boolean inPredicate = false;
	// for SQL wrapper specific purposes when last step is NodeTest then set nodeAccessor
	protected int nodeAccessor = Step.NONE;
	
	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################

	public LocatedExpression(ArrayList steps, XQueryModule parentModule) throws TypeException, XQueryException {
		this(steps, false, parentModule);
	}

	public LocatedExpression(ArrayList steps, boolean inPredicate, XQueryModule parentModule) throws TypeException, XQueryException {
		init(steps,inPredicate,parentModule,true);
	}
	public LocatedExpression(ArrayList steps, boolean inPredicate, XQueryModule parentModule, boolean simplify) throws TypeException, XQueryException {
		init(steps,inPredicate,parentModule,simplify);
	}
	
	private void init(ArrayList steps, boolean inPredicate, XQueryModule parentModule, boolean simplify) throws TypeException, XQueryException {
		if (steps == null || steps.isEmpty())
			throw new XQueryException("steps of LocatedExpression cannot be null or empty");
		if (simplify) {
			int retValue = LocatedExpressionReducer.simplifySteps(steps,inPredicate, parentModule) ;
			if (retValue == 0)
				throw new XQueryException("steps of LocatedExpression are not valid");
		}
		setSteps(steps);
		setInPredicate(inPredicate);
		setParentModule(parentModule);
		if (parentModule != null && parentModule.getStaticContext().getTypeVisitor() != null)
			accept(parentModule.getStaticContext().getTypeVisitor());
	}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	public String toString() {
		return toString(false, false, false);
	}

	public String getStringValue() {
		return getStringValue(false, false);
	}

	public StepExpr getStep(int num) {
		if (num >= steps.size())
			return null;
		return (Step) steps.get(num);
	}
	public Step getStepNum(int num) {
		if (num >= steps.size())
			return null;
		return (Step) steps.get(num);
	}
	public XQueryExpression getExpression() {
		return ((Step)steps.get(0)).getExpression();
	}
	public ArrayList getSteps() {
		return steps;
	}
	public void setSteps(ArrayList steps) throws XQueryException {
		// cannot be null or empty
		if (steps == null || steps.isEmpty())
			throw new XQueryException("steps of LocatedExpression cannot be null or empty");
		this.steps = steps;
		for (int i = 0; i < this.steps.size(); i++) {
			Step step = (Step) this.steps.get(i);
			step.setParentExpression(this);
			step.setParentModule(parentModule);
		}
	}

	public void addStep(Step step) throws XQueryException {
		if (step == null)
			return;
		step.setParentModule(parentModule);
		if (steps == null)
			steps = new ArrayList(1);
		steps.add(step);
		setSteps(steps);
	}

	public void addSteps(ArrayList steps) throws XQueryException {
		if (steps == null || steps.isEmpty())
			return;
		for (int i=0;i<steps.size();i++) {
			((Step)steps.get(i)).setParentModule(parentModule);
		}
		if (this.steps == null || this.steps.isEmpty()) {
			setSteps(steps);
			return;
		}
		this.steps.addAll(steps);
		setSteps(this.steps);
	}

	public void addStepAt(int index, Step step) throws XQueryException {
		if (step == null)
			return;
		step.setParentModule(parentModule);
		if (steps == null)
			steps = new ArrayList(1);
		steps.add(index,step);
		setSteps(steps);
	}

	public void addStepsAt(int index, ArrayList steps) throws XQueryException {
		if (steps == null || steps.isEmpty())
			return;
		for (int i=0;i<steps.size();i++) {
			((Step)steps.get(i)).setParentModule(parentModule);
		}
		if (this.steps == null || this.steps.isEmpty()) {
			setSteps(steps);
			return;
		}
		this.steps.addAll(index, steps);
		setSteps(this.steps);
	}

	public void setParentModule(XQueryModule parentUnit) {
		if (parentUnit == null)
			return;
		this.parentModule = parentUnit;
		for (int i = 0; i < steps.size(); i++)
			 ((Step) steps.get(i)).setParentModule(parentUnit);
	}

//	public boolean isRelativeExpression() {
//		Step step0 = (Step)steps.get(0);
//		if (step0.getAxis() == Axis.NONE && !step0.hasSeparator())
//			return false;
//		return true;
//	}

	public boolean isRelative() {
		if (((Step) steps.get(0)).getExpression() instanceof QName || ((Step) steps.get(0)).getExpression() instanceof NodeTest)
			return true;
		return false;
	}

	public void setInPredicate(boolean inPredicate) {
		this.inPredicate = inPredicate;
	}

	public boolean getInPredicate() {
		return inPredicate;
	}
	
	public void setNodeAccessor(int nodeAccessor) {
		this.nodeAccessor = nodeAccessor;
	}

	public int getNodeAccessor() {
		return nodeAccessor;
	}

	public boolean startsWith(Variable var) {
		return ((Step)steps.get(0)).getExpression() == var;
	}


	// #############################################################################
	// CLONE STUFF
	// #############################################################################
	public void addParentExpression(XQueryExpression parentExpression) {
		addParentExpression(parentExpression);
		for (int i = 0; i < steps.size(); i++)
			 ((Step) steps.get(i)).addParentExpression(parentExpression);
	}
	
	public Object clone() throws CloneNotSupportedException {
		try {	
			CloneVisitor cloneVisitor = new CloneVisitor(parentModule.getStaticContext(),true);
			this.accept(cloneVisitor);
			return cloneVisitor.getClone();
		} catch (XQueryException xqe) {
			throw new CloneNotSupportedException("Could not clone : " + this);
		}
	}


	// #############################################################################
	// XPATH STUFF
	// #############################################################################

	/**
	 * Returns the PathNode list of the last step.
	 *
	 */
	public ArrayList getXTrees() {
		return ((Step) this.steps.get(this.steps.size() - 1)).getXTrees();
	}

	// #############################################################################
	// RESTRUCTURATION STUFF
	// #############################################################################

	// tells whether the located expression starts with a variable
	public boolean startsWithVariable() {
		return ((Step)steps.get(0)).getExpression() instanceof Variable;
	}

	// tells whether the LocatedExpression has a parent step
	public int hasParentStep() {
		if (steps != null)
			for (int i = 0; i < steps.size(); i++)
				if (((Step) steps.get(i)).getAxis() == Axis.PARENT)
					return i;
		return -1;
	}

	// tells whether the LocatedExpression has a parent step
	public int hasUnionStep() {
		if (steps != null)
			for (int i = 0; i < steps.size(); i++)
				if (((Step) steps.get(i)).getExpression() instanceof ListOpUNIONExpression)
					return i;
		return -1;
	}

	// tells whether the LocatedExpression has predicates
	public boolean hasPredicates() {
		if (steps != null)
			for (int i = 0; i < steps.size(); i++)
				if (((Step) steps.get(i)).hasPredicates())
					return true;
		return false;
	}

	// tells whether the LocatedExpression has position predicates
	public boolean hasPositionPredicates() {
		if (steps != null)
			for (int i = 0; i < steps.size(); i++)
				if (((Step) steps.get(i)).hasPositionPredicates())
					return true;
		return false;
	}

}
