/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.typing;

import java.math.BigDecimal;

import org.xquark.schema.SimpleType;
import org.xquark.schema.datatypes.*;
import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.xquery.parser.XQueryException;

/**
 * Implements XQuery serialization for atomic values. Inputs are java classes
 * used for values in the XQuark XML schema implementation (see
 * @link org.xquark.schema.datatypes) and produce the XQuery conformant string
 * output.
 * NOTE: XQuery serialization is not strictly implemented here for performance
 * reasons. Especially, decimal are using Java serialization because it is
 * considered not only more efficient but "smarter"...
 */
public class QAtomicSerializer {

	private static final String NAN = "NaN";
	private static final String INF = "INF";
	private static final String POSITIVE_INF = "+INF";
	private static final String NEGATIVE_INF = "-INF";
	private static final double MAX_DECIMAL = 1E+6;
	private static final double MIN_DECIMAL = 1E-6;
	private static final String ZERO = "0";
	private static final Double NEG_ZERO_D = new Double("-0.0");
	private static final Float NEG_ZERO_F = new Float("-0.0");

	//private DecimalFormat format = (DecimalFormat)DecimalFormat.getInstance(Locale.US);
	private ValidationContextProvider context = null;

	public QAtomicSerializer() {
	}

	public QAtomicSerializer(ValidationContextProvider contextProvider) {
		context = contextProvider;
	}

	//
	// 				SPECIFIC XQUERY rules
	//
	public String serialize(QName sValue) throws XQueryException {
		throw new XQueryException("QName cannot be cast to string.");
	}
	public String serialize(URI sValue) throws XQueryException {
		if (sValue == null)
			return null;
		return sValue.toString();
	}

	public String serialize(SimpleType sType, String sValue) {
		if (sValue == null)
			return null;
		return sType.normalizedValue(sValue);
	}

	//
	// 				TIME TYPES
	//
	public String serialize(SimpleType sType, DateTime sValue) throws XQueryException {
		if (sValue == null)
			return null;
		String ret = sType.toXMLString(sValue, context);
		switch (sType.getPrimitive().getType()) {
			case PrimitiveType.DATE_TIME :
			case PrimitiveType.TIME :
			case PrimitiveType.DATE :
				if (sValue.hasTimeZone() && sValue.getTimeZone() == 0) // remove 'z'
					ret = ret.substring(0, ret.length() - 1);
				break;
			default :
				}
		return ret;
	}

	//
	// 				NUMERIC TYPES:
	// NOTE: looks like schema processing but slightly different, but piping to
	// schema conversion as specified would be too slow...

	public String serialize(Long sValue) throws XQueryException {
		if (sValue == null)
			return null;
		return sValue.toString();
	}

	public String serialize(BigDecimal sValue) throws XQueryException {
		if (sValue == null)
			return null;
		// check if there are significant decimal digits
		return removeNotSignificantDigits(sValue.toString());
	}

	public String removeNotSignificantDigits(String decimal) {
		int index = decimal.length(), lastZero = index;
		char wChar;
		do {
			if (index == 0)
				return decimal;
			wChar = decimal.charAt(--index);
			if (wChar == '0' && lastZero == index + 1)
				lastZero = index;
			if (wChar == 'E')
				return decimal; // scientific notation used by Double.toString()
		} while (wChar != '.');
		if (lastZero == index + 1)
			return decimal.substring(0, index);
		else
			return decimal.substring(0, lastZero);
	}

	public String serialize(Double sValue) throws XQueryException {
		if (sValue == null)
			return null;
		if (sValue.isNaN())
			return NAN;

		if (sValue.isInfinite()) {
			if (sValue.doubleValue() > 0)
				return INF;
			else
				return NEGATIVE_INF;
		} else {
			if (sValue.equals(NEG_ZERO_D))
				return ZERO;
			else
				return removeNotSignificantDigits(sValue.toString());
		}

		//	SPECIFICATION IMPLEMENTATION (SLOW)		
		//        double value = sValue.doubleValue();
		//        if (value == 0) return ZERO;
		//        
		//        double abs = Math.abs(value);
		//        if (MIN_DECIMAL <= abs && abs <= MAX_DECIMAL)	// TODO: format directly double using decimal rules
		//            return serialize(new BigDecimal(sValue.toString()));	//NOTE: not efficient but a "clean" method to transform floating point into decimals...
		//        else if (sValue.isInfinite()) {
		//            if (sValue.doubleValue() > 0)
		//                return INF;
		//            else
		//                return NEGATIVE_INF;
		//        }
		//        else {
		//            format.applyPattern("0.0##############E0");
		//            return format.format(value);
		//        }
	}

	public String serialize(Float sValue) throws XQueryException {
		if (sValue == null)
			return null;
		if (sValue.isNaN())
			return NAN;

		if (sValue.isInfinite()) {
			if (sValue.floatValue() > 0)
				return INF;
			else
				return NEGATIVE_INF;
		} else {
			if (sValue.equals(NEG_ZERO_F))
				return ZERO;
			else
				return removeNotSignificantDigits(sValue.toString());
		}

		//SPECIFICATION IMPLEMENTATION (SLOW)		
		//		float value = sValue.floatValue();
		//        if (value == 0) return ZERO;
		//        
		//        float abs = Math.abs(value);
		//        if (MIN_DECIMAL <= abs && abs <= MAX_DECIMAL)	// TODO: format directly double using decimal rules
		//            return serialize(new BigDecimal(sValue.toString()));	//NOTE: not efficient but a "clean" method to transform floating point into decimals...
		//        else if (sValue.isInfinite()) {
		//            if (sValue.floatValue() > 0)
		//                return INF;
		//            else
		//                return NEGATIVE_INF;
		//        }
		//        else {
		//            format.applyPattern("0.0######E0");
		//            return format.format(value);
		//        }
	}

	//
	// 				SCHEMA RULES
	//
	public String serialize(SimpleType sType, Boolean sValue) throws XQueryException {
		if (sValue == null)
			return null;
		return sType.toXMLString(sValue, context);
	}

	public String serialize(SimpleType sType, ByteArray sValue) throws XQueryException {
		if (sValue == null)
			return null;
		return sType.toXMLString(sValue, context);
	}

	public String serialize(SimpleType sType, byte[] sValue) throws XQueryException {
		if (sValue == null)
			return null;
		return sType.toXMLString(sValue, context);
	}

	public String serialize(SimpleType sType, Duration sValue) throws XQueryException {
		if (sValue == null)
			return null;
		return sType.toXMLString(sValue, context);
	}

	// ADD 29/03/04
	public String serialize(SimpleType sType, Object obj) throws XQueryException {
		// TODO see if sType should not be always not null
		if (obj == null)
			return null;
		if (sType == null) {
			return obj.toString();
		}
		switch (sType.getPrimitive().getType()) {
			case PrimitiveType.DECIMAL :
				// TODO use info from simple type to dispatch cases
				if (obj instanceof BigDecimal)
					return serialize((BigDecimal) obj);
				else
					return serialize((Long) obj);
			case PrimitiveType.DOUBLE :
				return serialize((Double) obj);
			case PrimitiveType.FLOAT :
				return serialize((Float) obj);
			case PrimitiveType.QNAME :
				return serialize((QName) obj);
			case PrimitiveType.DATE_TIME :
				return serialize(sType, (DateTime) obj);
			case PrimitiveType.ANY_URI :
				return serialize((URI) obj);
			case PrimitiveType.STRING :
				return serialize(sType, (String) obj);

			case PrimitiveType.DATE :
			case PrimitiveType.TIME :
			case PrimitiveType.BOOLEAN :
			case PrimitiveType.BASE64_BINARY :
			case PrimitiveType.HEX_BINARY :
			case PrimitiveType.DURATION :
			default :
				return sType.toXMLString(obj, context);
		}
	}
}
