/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.xdbc;

import java.io.StringWriter;
import java.io.Writer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.*;
import org.xml.sax.ext.LexicalHandler;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.DOMBuilder;
import org.xquark.util.SAXConstants;
import org.xquark.xml.xdbc.*;


/**
 * This class is an abstract class implementing XMLDocument.
 * It describes a document object which provides access to XML data
 * in several ways. When the underlying object is a stored document, it also
 * supports a number of document management methods also available in the
 * XMLCollection interface.<BR><BR>
 */
public abstract class AbstractXMLDocument implements XMLDocument
{
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName     = "$Name:  $";
    
    private static DocumentBuilder DOMCreator;
    
    /**
     * The XML collection.
     */
    private XMLCollection xmlCollection  = null ;
    
    
    /**
     * The document identifier in the XML collection.
     */
    private String identifier = null;
    private boolean readOnly = false;
    
    protected ContentHandler contentHandler ;
    protected ErrorHandler errorHandler ;
    protected LexicalHandler lexicalHandler ;
    
    protected AbstractXMLDocument()
    {
    }
    
    protected AbstractXMLDocument(boolean readOnly)
    {
        setReadOnly(readOnly);
    }
    
    protected AbstractXMLDocument(String Id)
    {
        this(Id, null, false);
    }
    
    protected AbstractXMLDocument(String Id, boolean readOnly)
    {
        this(Id, null, readOnly);
    }
    
    protected AbstractXMLDocument(String Id, XMLCollection collection)
    {
        this(Id, collection, false);
    }
    
    protected AbstractXMLDocument(String Id, XMLCollection collection, boolean readOnly)
    {
        identifier = Id;
        xmlCollection = collection;
        setReadOnly(readOnly);
    }
    
    protected void setReadOnly(boolean readOnly)
    {
        this.readOnly = readOnly;
    }
    
    /**
     * Returns the document identifier in the XML collection.
     * return the document identifier as a String, or null if the document is not a stored
     * object.
     */
    public String getIdentifier()
    {
        return identifier ;
    }
    
    /**
     * Changes the document identifier in the XML collection.
     * This change is persistent in the data source.
     * Note: Depending on the auto-commit mode of the underlying data
     * source connection, the change may only become definitive after a
     * commit operation.
     * @param the new document identifier
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the renaming operation is not supported.
     */
    public void setIdentifier(String identifier) throws XMLDBCException, XMLDBCNotSupportedException
    {
        if (readOnly)
            throw new XMLDBCNotSupportedException("This document is read-only.");
        if (xmlCollection != null)
            xmlCollection.renameDocument(this.identifier, identifier);
        this.identifier = identifier;
    }
    
    /**
     * Returns the XML collection to which this document belongs. This may be null if
     * the document is not a stored object.
     * @return an XMLCollection object, or null if the document is not a stored object.
     */
    public XMLCollection getCollection()
    {
        return xmlCollection ;
    }
    
    
    /**
     * Removes the document from the collection to which it belongs. The object should
     * not be accessed any more after this method is called.
     * Note: Depending on the auto-commit mode of the underlying data source connection, the
     * change may only become definitive after a commit operation.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the remove operation is not supported.
     */
    public void remove() throws XMLDBCException, XMLDBCNotSupportedException
    {
        if (readOnly)
            throw new XMLDBCNotSupportedException("This document is read-only.");
        if (xmlCollection != null)
        {
            xmlCollection.removeDocument(getIdentifier());
            xmlCollection = null;
        }
        else
            throw new XMLDBCNotSupportedException("Document is not a stored object");
    }
    
    /**
     * Sets a (SAX2) content handler to intercept events produced when retrieving
     * the document content.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the content handler implementation (see org.xml.sax.ContentHandler)
     */
    public void setContentHandler(ContentHandler handler)
    {
        this.contentHandler = handler;
    }
    
    /**
     * Sets a (SAX2) lexical handler to intercept events produced when retrieving
     * the document content.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the lexical handler implementation (see org.xml.sax.ext.LexicalHandler)
     */
    public void setLexicalHandler(LexicalHandler handler)
    {
        this.lexicalHandler = handler ;
    }
    
    /**
     * Sets a (SAX2) error handler to intercept error events produced when retrieving
     * the document content.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @param handler the error handler implementation (see org.xml.sax.ErrorHandler)
     */
    public void setErrorHandler(ErrorHandler handler)
    {
        this.errorHandler = handler ;
    }
    
    /**
     * Retrieves the current (SAX2) lexical handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @return the lexical handler implementation (see org.xml.sax.ext.LexicalHandler)
     */
    public LexicalHandler getLexicalHandler()
    {
        return lexicalHandler;
    }
    
    /**
     * Retrieves the current (SAX2) content handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @return the content handler implementation (see org.xml.sax.ContentHandler)
     */
    public ContentHandler getContentHandler()
    {
        return contentHandler;
    }
    
    /**
     * SRetrieves the current (SAX2) error handler.<BR>
     * <P>*** ONLY FOR SAX2 ***</P>
     * @return the error handler implementation (see org.xml.sax.ErrorHandler)
     */
    public ErrorHandler getErrorHandler()
    {
        return errorHandler;
    }
    
    /**
     * Returns the document content as a DOM2 Document.
     * @return the DOM2 Document containing the document data.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public Document getAsDocument() throws XMLDBCException
    {
        createDOMBuilder();
        Document doc = DOMCreator.newDocument();
        DOMBuilder builder = new DOMBuilder(doc);
        getDocument(builder, builder);
        return doc;
    }
    
    public Document getAsDOM() throws XMLDBCException
    {
        return getAsDocument();
    }
    
    public void getAsDOM(Element parent) throws XMLDBCException
    {
        DOMBuilder builder = new DOMBuilder(parent);
        getDocument(builder, builder);
    }
    
    /**
     * Returns the document content as an XML string.
     * @return the XML string containing the document data.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String getAsString() throws XMLDBCException
    {
        StringWriter strWriter = new StringWriter() ;
        getAsStream(strWriter);
        return strWriter.toString() ;
    }
    
    public void getAsStream(Writer out) throws XMLDBCException
    {
        XMLSerializer serializer = new XMLSerializer(out);
        getDocument(serializer, serializer);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                          PRIVATE UTILITIES
    ///////////////////////////////////////////////////////////////////////////
    private void getDocument(ContentHandler contentHandler, LexicalHandler lexicalHandler)
    throws XMLDBCException
    {
        // save handlers plugged by user
        ContentHandler initialContentHandler = getContentHandler();
        LexicalHandler initialLexicalHandler = getLexicalHandler();
        
        setContentHandler(contentHandler);
        setLexicalHandler(lexicalHandler);
        try
        {
            getAsSAX();
        }
        catch (SAXException e)
        {
            throw new XMLDBCException(e.getMessage(), e.getException());
        }
        finally
        {
            // restore handlers plugged by user
            if (initialContentHandler != null)
                setContentHandler(initialContentHandler);
            if (initialLexicalHandler != null)
                setLexicalHandler(initialLexicalHandler);
        }
    }
    
    protected void plugHandlers(XMLReader reader)
    throws XMLDBCException
    {
        try
        {
            if (contentHandler != null)
                reader.setContentHandler(contentHandler);
            if (errorHandler != null)
                reader.setErrorHandler(errorHandler);
            if (lexicalHandler != null)
                reader.setProperty(SAXConstants.SAX_LEXICAL_PROPERTY, lexicalHandler);
        }
        catch (SAXException e)
        {
            throw new XMLDBCException("Could not plug handlers on repository XMLReader.", e);
        }
    }
    
    private synchronized void createDOMBuilder()
    {
        if (DOMCreator == null)
        {
            DocumentBuilderFactory DOMFactory = DocumentBuilderFactory.newInstance();
            DOMFactory.setNamespaceAware(true);
            try
            {
                DOMCreator = DOMFactory.newDocumentBuilder();
            }
            catch (ParserConfigurationException pce)
            {
                throw new RuntimeException("Could not create an empty DOM level 2 document using JAXP. Check a JAXP DOM implementation is in your classpath.") ;
            }
        }
    }
}
